// -*- mode: C++; tab-width: 4; indent-tabs-mode: nil -*- (for GNU Emacs)
//
// Copyright (c) Microsoft Corporation
//
// This file is part of the Microsoft Research Mesh Connectivity Layer.
// You should have received a copy of the Microsoft Research Shared Source
// license agreement (MSR-SSLA) for this software; see the file "license.txt".
// If not, please see http://research.microsoft.com/mesh/license.htm,
// or write to Microsoft Research, One Microsoft Way, Redmond, WA 98052-6399.
//
// Abstract:
//
// Device control (enable/disable) code.  Loosely based on the
// "devcon" sample in the DDK.
//

#include <windows.h>
#include <tchar.h>
#include <stdio.h>
#include <setupapi.h>
#include <devguid.h>

int
ControlDevice(
    HDEVINFO Devs,             // Device set.
    PSP_DEVINFO_DATA DevInfo,  // Device in set.
    DWORD Action)              // Action to perform.
{
    SP_PROPCHANGE_PARAMS Params;

    //
    // Operate on config-specific profile.
    //
    Params.ClassInstallHeader.cbSize = sizeof(SP_CLASSINSTALL_HEADER);
    Params.ClassInstallHeader.InstallFunction = DIF_PROPERTYCHANGE;
    Params.StateChange = Action;
    Params.Scope = DICS_FLAG_CONFIGSPECIFIC;
    Params.HwProfile = 0;

    if (SetupDiSetClassInstallParams(Devs, DevInfo, &Params.ClassInstallHeader,
                                     sizeof(Params)) &&
        SetupDiCallClassInstaller(DIF_PROPERTYCHANGE, Devs, DevInfo)) {
        //
        // Success!
        //
        return 1;
    } else {
        return 0;
    }
}

LPTSTR
GetDeviceRegistryProperty(
    HDEVINFO Devs,             // Device set.
    PSP_DEVINFO_DATA DevInfo,  // Device in set.
    DWORD Property)            // SPDRP code of property to retreive.
{
    LPTSTR Buffer;
    DWORD Attempt;
    DWORD Size;
    DWORD ActualSize;
    DWORD DataType;

    Size = 64;  // Educated guess.

    for (Attempt = 0; Attempt < 2; Attempt++) {

        Buffer = new TCHAR[(Size / sizeof(TCHAR)) + sizeof(TCHAR)];
        if (Buffer == NULL) {
            return NULL;
        }

        if (!SetupDiGetDeviceRegistryProperty(Devs, DevInfo, Property,
                                              &DataType, (LPBYTE)Buffer, Size,
                                              &ActualSize)) {
            if (GetLastError() != ERROR_INSUFFICIENT_BUFFER) {
                delete [] Buffer;
                return NULL;
            }

            //
            // Our guess at a buffer size was too small.  Try again.
            //
            Size = ActualSize;
            delete [] Buffer;
            continue;
        }

        //
        // Make sure the registry entry we found is of the expected type.
        //
        if (DataType != REG_SZ) {
            delete [] Buffer;
            return NULL;
        }

        Size = ActualSize / sizeof(TCHAR);
        Buffer[Size] = TEXT('\0');

        return Buffer;
    }

    return NULL;
}


EXTERN_C int
ControlDeviceClass(
    LPTSTR Query,    // Prefix of device description string to match.
    int QueryIndex,  // Index of device instance in registry's class entry.
    DWORD Action)    // Action to take on found device(s).
{
    HDEVINFO Devs = INVALID_HANDLE_VALUE;
    SP_DEVINFO_DATA DevInfo;
    DWORD DevIndex;
    int Count = 0;

    //
    // Validate QueryIndex is within range.
    //
    if (QueryIndex > 9999)
        return 0;

    //
    // Create a device information set, consisting of our desired devices.
    //
    Devs = SetupDiGetClassDevs(&GUID_DEVCLASS_NET,  // Only "Net" devices.
                               NULL,
                               NULL,
                               DIGCF_PRESENT);      // And currently present.
    if (Devs == INVALID_HANDLE_VALUE) {
        return 0;
    }

    //
    // Run through the set, looking for devices for which the prefix of their
    // description string matches our query.
    //
    DevInfo.cbSize = sizeof(DevInfo);
    for (DevIndex = 0; SetupDiEnumDeviceInfo(Devs, DevIndex, &DevInfo);
         DevIndex++) {

        LPTSTR DevDesc;

        DevDesc = GetDeviceRegistryProperty(Devs, &DevInfo, SPDRP_DEVICEDESC);
        if (DevDesc != NULL) {
            if (_tcsnicmp(DevDesc, Query, _tcslen(Query)) == 0) {
                //
                // Found device with matching description.
                //

                if (QueryIndex >= 0) {
                    //
                    // User wants to act upon only a specific device instance.
                    // Check to see if this is the desired instance.
                    // Note that we should only have devices of class Net.
                    //
                    LPTSTR Driver;
                    TCHAR QueryInstance[45];

                    _stprintf(QueryInstance,
                              "{4D36E972-E325-11CE-BFC1-08002BE10318}\\%04u",
                              QueryIndex);

                    Driver = GetDeviceRegistryProperty(Devs, &DevInfo,
                                                       SPDRP_DRIVER);
                    if (Driver != NULL) {
                        if (_tcsnicmp(Driver, QueryInstance,
                                      _tcslen(QueryInstance)) != 0) {
                            //
                            // Not the correct instance.
                            //
                            delete [] Driver;
                            delete [] DevDesc;
                            continue;
                        }
                        delete [] Driver;
                    } else {
                        //
                        // User specified a specific device instance, but we
                        // can't tell whether this is the right instance or
                        // not.  Be conservative and don't operate on it.
                        //
                        delete [] DevDesc;
                        continue;
                    }
                }

                //
                // Attempt to perform the requested action.
                //
                Count += ControlDevice(Devs, &DevInfo, Action);
            }

            delete [] DevDesc;
        }
    }

    SetupDiDestroyDeviceInfoList(Devs);

    return Count;
}
