// *- mode: C++; tab-width: 4; indent-tabs-mode: nil -*- (for GNU Emacs)
//
// Copyright (c) Microsoft Corporation
//
// This file is part of the Microsoft Research Mesh Connectivity Layer.
// You should have received a copy of the Microsoft Research Shared Source
// license agreement (MSR-SSLA) for this software; see the file "license.txt".
// If not, please see http://research.microsoft.com/mesh/license.htm,
// or write to Microsoft Research, One Microsoft Way, Redmond, WA 98052-6399.
//
// File created at University of Texas-Austin research.
//


#include "headers.h"



//This method is called anytime we have to send an ACK.  It will search through
//the AckCache for the flow that we are dealing with and then fill in the
//cumulative seqnumber and appropiate bitmap.  It will create an entry if necessary.
//
// AckCache *ac - the current Ack Cache for which we get our state from
// LQSRAckId *fillSeqNum - returned parameter, we put in the cumulative ACK seq. # here
// AckBitMap *fillBitmap - returned parameter, the bitmap for the cumulative ACK
// LQSRAckId rcvdSeqNum - The sequence number of the packet that was just received
//
// HAVE TO DEAL WITH WRAP AROUNDS
//
void AckCacheFillAck(AckCache *ac, LQSRAckId *fillSeqNum, AckBitMap *fillBitmap, LQSRAckId rcvdSeqNum) {

	//If we received some duplicate data. If we are close to wrapping around,
	//then we will proceed.
	if (rcvdSeqNum <= ac->AckNum) {
		*fillSeqNum = ac->AckNum ;
		*fillBitmap = ac->bitmap ; 				
		return; 
	}

	//we have a fresh rcvdSeqNum, need to update our data structures and fill in ACK
	// 1. update ac->AckNum
	// 2. update ac->bitmap
	// 3. fill in packet
	//
	// NEED TO DEAL WITH WRAP AROUNDS
	//

	//if our received sequence number is outside the bounds of our 
	//bitmap, then we need to update all our state (ac->AckNum and ac->bitmap); 
	if (rcvdSeqNum > ac->AckNum + BITMAPLENGTH) {
		ac->AckNum = AckCacheAdjustBitMap(&ac->bitmap, ac->AckNum, rcvdSeqNum) ;		
		*fillSeqNum = ac->AckNum ;
		*fillBitmap = ac->bitmap ; 				
		return; 					
	}
	
	//update bitmap
	AckCacheSetBit(&ac->bitmap, (rcvdSeqNum - ac->AckNum) - 1); 
	//see if we need to update the AckNum
	while ((ac->bitmap.bm[0] & (uint) 0x80000000u) != 0){
		ac->AckNum++; 
		//shift the bitmap to the right
		AckCacheShiftLeft(&ac->bitmap, 1); 
	}
			
	//update packet and return
	*fillSeqNum = ac->AckNum ;
	*fillBitmap = ac->bitmap; 
	return; 

}

//Determines whether or not we should delete the packet represented by storeSeqNum given
// that we have received "Ack."
//
// HAVE TO DEAL WITH WRAP AROUND
//
boolint AckCacheDeletePacket(InternalAcknowledgement *Ack, LQSRAckId storeSeqNum) {
	//cumulative ACK clobbers lesser stored ACKs
	if (storeSeqNum <= Ack->opt.identification)
		return 1; 

	//make sure we don't overflow
	if (storeSeqNum > Ack->opt.identification + BITMAPLENGTH) {	
		return 0; 					
	}

	//check bitmap
	if (AckCacheCheckBit(&Ack->opt.bitmap, (storeSeqNum - Ack->opt.identification) - 1))
		return 1;
	return 0 ; 

}


//
// Checks if the given seqNum is in the AckCache already
//
boolint AckCacheCheckSeqNum(AckCache *AC, LQSRAckId seqNum) {
	//cumulative ACK clobbers lesser stored ACKs
	if (seqNum <= AC->AckNum)
		return 1; 

	//make sure we don't overflow
	if (seqNum > AC->AckNum + BITMAPLENGTH) {	
		return 0; 					
	}

	//check bitmap
	if (AckCacheCheckBit(&AC->bitmap, (seqNum - AC->AckNum) - 1))
		return 1;
	return 0 ; 

}


//
// Updates our AckCache given we've received seqNum
//
void AckCacheUpdateSeqNum(AckCache *AC, LQSRAckId seqNum) {
	//If we received some duplicate data. 
	if (seqNum <= AC->AckNum) {			
		return; 
	}

	//if our received sequence number is outside the bounds of our 
	//bitmap, then we need to update all our state (ac->AckNum and ac->bitmap); 
	if (seqNum > AC->AckNum + BITMAPLENGTH) {
		AC->AckNum = AckCacheAdjustBitMap(&AC->bitmap, AC->AckNum, seqNum) ;		
		return; 					
	}
	
	//update bitmap
	AckCacheSetBit(&AC->bitmap, (seqNum - AC->AckNum) - 1); 
	//see if we need to update the AckNum
	while ((AC->bitmap.bm[0] & (uint) 0x80000000u) != 0){
		AC->AckNum++; 
		//shift the bitmap to the right
		AckCacheShiftLeft(&AC->bitmap, 1); 
	}
}



//
//  Some operations on the bitmap
//

//Shift the bitmap to the right, shift amount of times
void AckCacheShiftRight(AckBitMap *abm, int shift) {
	int i = 0 ; 
	int j = 0; 

	for (j = 0; j < shift ; j++) {
		
		abm->bm[NUMACKBITMAP-1] = abm->bm[NUMACKBITMAP-1] >> 1 ;
		for (i = NUMACKBITMAP-2 ; i >= 0; i--) {
			if ((uint) 1 & abm->bm[i]) {
				abm->bm[i+1] = (uint)0x80000000u | abm->bm[i+1]; 
			}
			abm->bm[i] = abm->bm[i] >> 1 ;
		}//end for i

	}//end for j
}//end shiftRight

//Shift the bitmap to the left, shift amount of times
void AckCacheShiftLeft(AckBitMap *abm, int shift) {
	int i = 0 ; 
	int j = 0; 

	for (j = 0; j < shift ; j++) {
		
		abm->bm[0] = abm->bm[0] << 1 ;
		for (i = 1 ; i < NUMACKBITMAP; i++) {
			if ((uint)0x80000000u & abm->bm[i]) {
				abm->bm[i-1] = (uint)1 | abm->bm[i-1]; 
			}
			abm->bm[i] = abm->bm[i] << 1 ;
		}//end for i

	}//end for j
}//end shiftRight

//This function sets the bit denoted by 'index'
// in the bitmap abm.  We'll start the bitmap index at 0.
void AckCacheSetBit(AckBitMap *abm, int index) {
	uint array_index = 0;

	ASSERT (!(index < 0 || index >= BITMAPLENGTH));

	array_index = index / (sizeof(uint)*8); 
	abm->bm[array_index] = ( (uint)0x80000000u >> (index - array_index*sizeof(uint)*8) ) | abm->bm[array_index]; 

}

//This functions checks to see if the bit denoted by 'index'
//is set in the bitmap abm.  We start the bitmap index at 0.
boolint AckCacheCheckBit(AckBitMap *abm, int index) {
	uint array_index = 0;

	ASSERT(!(index < 0 || index >= BITMAPLENGTH));

	array_index = index / (sizeof(uint)*8); 
	if ((((uint)0x80000000u >> (index - array_index*sizeof(uint)*8) ) & abm->bm[array_index]) != (uint) 0) {
		return 1; 
	} 
    
	return 0; 
}


//If we get a sequence number that is outside of our array, then we
//have to move the range of our window.  
//
// AckBitMap *abm - holds the initial bitmap, updated to the new bitmap and returned.
// AckNum oldAck - old AckNum for the initial bitmap
// AckNum newAck - seq # of pkt we just received that causes us to move to window
// @returns - the new value for our cumulative ACK (ac->AckNum)--for the new bitmap
//
// HAVE TO DEAL WITH WRAP-AROUND
//
LQSRAckId AckCacheAdjustBitMap(AckBitMap *abm, LQSRAckId oldAck, LQSRAckId newAck) {
	AckBitMap temp ;  
	uint newbits; //how many bits of new bitmap will be new
	uint j ;
	LQSRAckId retAck = (LQSRAckId) ((newAck - BITMAPLENGTH)); 
    
	ASSERT(newAck > oldAck+BITMAPLENGTH); 

	//don't need to copy any entries over to the new bitmap if this is true
	if (newAck - (oldAck+BITMAPLENGTH) >= BITMAPLENGTH) {
		int i = 0 ;
		for (i = 0 ; i < NUMACKBITMAP-1; i++) {
			abm->bm[i] = (uint) 0 ;
		}//end for
		abm->bm[NUMACKBITMAP-1] = (uint) 0x1;
		return retAck ;
	}//end if
	
	temp = *abm ;
	newbits = (uint) (newAck - (oldAck+BITMAPLENGTH)) ;

	//clear abm, we will now use abm as the new array and temp as the old array
	for (j = 0 ; j < NUMACKBITMAP-1 ; j++) {
		abm->bm[j] = (uint) 0; 
	}
	abm->bm[NUMACKBITMAP-1] = (uint) 0x1;

	//we need to copy the last (bitmaplength - newbits) bits from the old
	//bitmap to the first (bitmaplength - newbits) bits of our new bitmap
	for (j = 0; j < BITMAPLENGTH - newbits ; j++) {
		if (AckCacheCheckBit(&temp, newbits + j)) {
			AckCacheSetBit(abm, j); 
		}
	}//end for j
	
	//now see if we have to collapse our bitmap anymore
	//see if we need to update the AckNum
	while ((abm->bm[0] & (uint) 0x80000000u) != 0){
		retAck++; 
		//shift the bitmap to the right
		AckCacheShiftLeft(abm, 1); 
	}
	return retAck; 
}


// We call thism method when we receive an ACK from a 
// higher priority node. When this happens, we want to update
// all of our ACK state to be consistent with the packets recevied by 
// a higher priority node.
//
// AC - my ack cache for this flow
// Ack - the Ack from the higher priority node
void AckCacheUpdateAckState(AckCache *AC, InternalAcknowledgement *Ack)
{
	uint diff = 0; 
	AckBitMap mask ; 
	uint i ; 


	if (Ack->opt.identification == AC->AckNum) {
		//just want to OR our bitmap
		for (i = 0 ; i < NUMACKBITMAP; i++) {
			AC->bitmap.bm[i] = AC->bitmap.bm[i] | Ack->opt.bitmap.bm[i]; 
		}
	} else if (Ack->opt.identification < AC->AckNum) {
		//shift Ack->id and then OR
		diff = AC->AckNum - Ack->opt.identification ;
		RtlCopyMemory(&mask, &(Ack->opt.bitmap), sizeof(mask));
		AckCacheShiftLeft(&(mask), diff); 
		for (i = 0 ; i < NUMACKBITMAP; i++) {
			AC->bitmap.bm[i] = AC->bitmap.bm[i] | mask.bm[i]; 
		}
	} else {
		//Ack->opt.identification > AC->AckNum
		//adjust our bitmap to Ack->id and then 
		// OR the bitmaps
		diff = Ack->opt.identification - AC->AckNum ;
		AckCacheShiftLeft(&(AC->bitmap), diff);
		for (i = 0 ; i < NUMACKBITMAP; i++) {
			AC->bitmap.bm[i] = AC->bitmap.bm[i] | Ack->opt.bitmap.bm[i]; 
		}
		AC->AckNum = Ack->opt.identification; 

	}

	//shift bitmap and update seqNum as appropiate
	while ((AC->bitmap.bm[0] & (uint) 0x80000000u) != 0){
		AC->AckNum++; 
		//shift the bitmap to the right
		AckCacheShiftLeft(&AC->bitmap, 1); 
	}

}