// -*- mode: C++; tab-width: 4; indent-tabs-mode: nil -*- (for GNU Emacs)
//
// Copyright (c) Microsoft Corporation
//
// This file is part of the Microsoft Research Mesh Connectivity Layer.
// You should have received a copy of the Microsoft Research Shared Source
// license agreement (MSR-SSLA) for this software; see the file "license.txt".
// If not, please see http://research.microsoft.com/mesh/license.htm,
// or write to Microsoft Research, One Microsoft Way, Redmond, WA 98052-6399.
//

#include "headers.h"

#define CACHE_TIMEOUT   (1 * SECOND)

//maximum caching entries for arb_dijkstra
#define MAX_ADC_ENTRIES 200

#define PRINT_VA(f, s) KdPrint(("%s: from %02x-%02x-%02x-%02x-%02x-%02x\n", \
	f, (s)[0], (s)[1], (s)[2], (s)[3], (s)[4], (s)[5])) 


#define PRINT_LINK(s, d, t) KdPrint(("\t\t\t\tLink from %02x-%02x-%02x-%02x-%02x-%02x\
	to %02x-%02x-%02x-%02x-%02x-%02x Metric %u\n", (s)[0], (s)[1], (s)[2], (s)[3], (s)[4], (s)[5], \
	(d)[0], (d)[1], (d)[2], (d)[3], (d)[4], (d)[5], t)) 


#define PRINT_LINK_STR(str, s, d) KdPrint(("%s: Link from %02x-%02x-%02x-%02x-%02x-%02x\
	to %02x-%02x-%02x-%02x-%02x-%02x\n", str, (s)[0], (s)[1], (s)[2], (s)[3], (s)[4], (s)[5], \
	(d)[0], (d)[1], (d)[2], (d)[3], (d)[4], (d)[5])) 


static int dijkstra(MiniportAdapter *VA, Time Now);

//* LinkCacheLinkChange
//
//  Records a link status change.
//
//  Called while holding LC->Lock.
//  
#if !CHANGELOGS
__inline
#endif
static void
LinkCacheLinkChange(
    LinkCache *LC,
    const VirtualAddress From,
    const VirtualAddress To,
    LQSRIf InIf,
    LQSRIf OutIf,
    uint Metric,
    uint Reason)
{
#if CHANGELOGS
    struct LinkChange *Record = &LC->LinkChangeLog[LC->NextLinkRecord];

    KeQuerySystemTime((LARGE_INTEGER *)&Record->TimeStamp);
    RtlCopyMemory(Record->From, From, sizeof(VirtualAddress));
    RtlCopyMemory(Record->To, To, sizeof(VirtualAddress));
    Record->InIf = InIf;
    Record->OutIf = OutIf;
    Record->Metric = Metric;
    Record->Reason = Reason;

    LC->NextLinkRecord = (LC->NextLinkRecord + 1) % NUM_LINKCHANGE_RECORDS;
#else
    UNREFERENCED_PARAMETER(LC);
    UNREFERENCED_PARAMETER(From);
    UNREFERENCED_PARAMETER(To);
    UNREFERENCED_PARAMETER(InIf);
    UNREFERENCED_PARAMETER(OutIf);
    UNREFERENCED_PARAMETER(Metric);
    UNREFERENCED_PARAMETER(Reason);
#endif
}

//* LinkCacheRouteChange
//
//  Records a route change.
//
//  Called while holding LC->Lock.
//  
#if !CHANGELOGS
__inline
#endif
static void
LinkCacheRouteChange(
    LinkCache *LC,
    const VirtualAddress Dest,
    uint PrevMetric,
    uint Metric,
    SourceRoute *SR,
    Link **Hops)
{
#if CHANGELOGS
    struct RouteChange *Record = &LC->RouteChangeLog[LC->NextRouteRecord];

    KeQuerySystemTime((LARGE_INTEGER *)&Record->TimeStamp);
    RtlCopyMemory(Record->Dest, Dest, sizeof(VirtualAddress));
    Record->Metric = Metric;
    Record->PrevMetric = PrevMetric;
    if (SR != NULL) {
        RtlCopyMemory(Record->Buffer, SR, sizeof(LQSROption) + SR->optDataLen);

        //
        // Copy the metrics too.
        //
        if (Hops != NULL) {
            SourceRoute UNALIGNED *Save =
                (SourceRoute UNALIGNED *) Record->Buffer;
            uint NumHops = SOURCE_ROUTE_HOPS(SR->optDataLen);
            uint hop;

            for (hop = 1; hop < NumHops; hop++)
                Save->hopList[hop].Metric = Hops[NumHops - hop - 1]->Metric;
        }

    }
    else {
        //
        // Use an optionType of zero to indicate the absence of a route.
        //
        ((LQSROption UNALIGNED *)Record->Buffer)->optionType = 0;
    }

    LC->NextRouteRecord = (LC->NextRouteRecord + 1) % NUM_ROUTECHANGE_RECORDS;
#else
    UNREFERENCED_PARAMETER(LC);
    UNREFERENCED_PARAMETER(Dest);
    UNREFERENCED_PARAMETER(Metric);
    UNREFERENCED_PARAMETER(PrevMetric);
    UNREFERENCED_PARAMETER(SR);
    UNREFERENCED_PARAMETER(Hops);
#endif
}

//* LinkCacheRouteUsage
//
//  Records the usage of a route for a packet to the destination.
//
//  Called with the link cache locked.
//
#if !CHANGELOGS
__inline
#endif
static void
LinkCacheRouteUsage(
    LinkCache *LC,
    int dnode,
    InternalSourceRoute *SR)    // May be NULL, meaning no route.
{
#if CHANGELOGS
    CacheNode *Dest = &LC->nodes[dnode];
    RouteUsage *RU;
    uint NumHops;
    uint hop;

    //
    // Do we have a current RouteUsage record?
    //
    RU = Dest->CurrentRU;
    if (RU == NULL) {
        //
        // We need to find or create a RouteUsage record.
        //

        if (SR == NULL)
            NumHops = 0; // NB: We can use SR->opt.hopList.
        else
            NumHops = SOURCE_ROUTE_HOPS(SR->opt.optDataLen);

        //
        // Look for an existing RouteUsage record.
        //
        for (RU = Dest->History; ; RU = RU->Next) {
            if (RU == NULL) {
                //
                // Create a new RouteUsage record.
                //
                RU = ExAllocatePool(NonPagedPool,
                                sizeof *RU + NumHops * sizeof(SRAddr));
                if (RU == NULL)
                    return;

                RU->Next = Dest->History;
                Dest->History = RU;

                RU->Usage = 0;
                RU->Hops = NumHops;
                RtlCopyMemory(RU->Route, SR->opt.hopList,
                              NumHops * sizeof(SRAddr));

                //
                // Copy the current metrics for this route.
                // It is possible for Dest->Hops to be NULL
                // (if the allocation in LinkCacheFillSR failed)
                // but otherwise it should match SR.
                //
                if (Dest->Hops != NULL) {
                    ASSERT(SR == NULL ||
                           SR->opt.optDataLen == Dest->SR->optDataLen);

                    for (hop = 1; hop < NumHops; hop++)
                        RU->Route[hop].Metric =
                            Dest->Hops[NumHops - hop - 1]->Metric;
                }
                break;
            }

            //
            // Do we have an existing RouteUsage record?
            // NB: The RouteUsage record has non-zero metrics,
            // while the source route has zero metrics,
            // so RtlEqualMemory does not work.
            //
            if (RU->Hops == NumHops) {
                boolint HaveExisting = TRUE;

                for (hop = 0; hop < NumHops; hop++) {
                    if (! VirtualAddressEqual(RU->Route[hop].addr,
                                              SR->opt.hopList[hop].addr) ||
                        (RU->Route[hop].inif != SR->opt.hopList[hop].inif) ||
                        (RU->Route[hop].outif != SR->opt.hopList[hop].outif)) {
                        HaveExisting = FALSE;
                        break;
                    }
                }

                if (HaveExisting)
                    break;
            }
        }

        //
        // Remember this record for next time.
        //
        Dest->CurrentRU = RU;
    }

    //
    // Record the usage of this route.
    //
    RU->Usage++;

#else
    UNREFERENCED_PARAMETER(LC);
    UNREFERENCED_PARAMETER(dnode);
    UNREFERENCED_PARAMETER(SR);
#endif
}

//* LinkCacheAllocate
//
//  Allocates a new link cache data structure.
//
LinkCache *
LinkCacheAllocate(
    uint size,
    const uchar myaddr[SR_ADDR_LEN])
{
    LinkCache *ret;
    Time Now;

    ret = ExAllocatePool(NonPagedPool, sizeof(LinkCache));
    if (ret == NULL)
        goto fail;

    KeInitializeSpinLock(&ret->Lock);

    ASSERT(size > 1);

    ret->nodeCount = 1;
    ret->maxSize = size;

    ret->metric = ExAllocatePool(NonPagedPool, sizeof(uint)*size);
    ret->hops = ExAllocatePool(NonPagedPool, sizeof(uint)*size);
    ret->link = ExAllocatePool(NonPagedPool, sizeof(Link*)*size);
    ret->prev = ExAllocatePool(NonPagedPool, sizeof(uint)*size);
    ret->nodes = ExAllocatePool(NonPagedPool, sizeof(CacheNode)*size);

    if ((ret->metric == NULL) ||
        (ret->hops == NULL) ||
        (ret->link == NULL) ||
        (ret->prev == NULL) ||
        (ret->nodes == NULL))
        goto fail;

    ret->SmallestMetric = 0;
    ret->LargestMetric = (uint)-1;
    Now = KeQueryInterruptTime();
    ret->timeout = Now;

    // INVARIANT: nodes[0] MUST always refer to me.
    RtlCopyMemory(ret->nodes[0].address, myaddr, SR_ADDR_LEN);
    ret->nodes[0].AdjOut = NULL;
    ret->nodes[0].AdjIn = NULL;
    ret->nodes[0].SR = NULL;
    ret->nodes[0].Hops = NULL;
    ret->nodes[0].Metric = 0;
    ret->nodes[0].RouteChanges = 0;
#if CHANGELOGS
    ret->nodes[0].CurrentRU = NULL;
    ret->nodes[0].History = NULL;
#endif
    ret->metric[0] = 0;
    ret->hops[0] = 0;
    ret->prev[0] = MAXULONG;

    ret->CountAddLinkInvalidate = 0;
    ret->CountAddLinkInsignificant = 0;
    ret->CountRouteFlapDamp = 0;
    ret->CountRouteFlap = 0;

#if CHANGELOGS
    ret->NextLinkRecord = 0;
    RtlZeroMemory(ret->LinkChangeLog, sizeof ret->LinkChangeLog);

    ret->NextRouteRecord = 0;
    RtlZeroMemory(ret->RouteChangeLog, sizeof ret->RouteChangeLog);
#endif

    return ret;

fail:
    if (ret != NULL) {
        if (ret->metric != NULL)
            ExFreePool(ret->metric);
        if (ret->hops != NULL)
            ExFreePool(ret->hops);
        if (ret->prev != NULL)
            ExFreePool(ret->prev);
        if (ret->link != NULL)
            ExFreePool(ret->link);
        if (ret->nodes != NULL)
            ExFreePool(ret->nodes);
        ExFreePool(ret);
    }
    return NULL;
}

//* LinkCacheFree
//
//  Frees the link cache data structure for unload.
//
void
LinkCacheFree(MiniportAdapter *VA)
{
    LinkCache *LC = VA->LC;
    Link *tmp;
#if CHANGELOGS
    RouteUsage *RU;
#endif
    uint i;

    //
    // Free all the links and cached routes.
    //
    for (i = 0; i < LC->nodeCount; i++) {
        while (LC->nodes[i].AdjOut != NULL) {
            tmp = LC->nodes[i].AdjOut;
            LC->nodes[i].AdjOut = LC->nodes[i].AdjOut->NextOut;
            ExFreePool(tmp);
        }

        if (LC->nodes[i].SR != NULL) {
            // This also frees LC->nodes[i].Hops.
            ExFreePool(LC->nodes[i].SR);
        }

#if CHANGELOGS
        while ((RU = LC->nodes[i].History) != NULL) {
            LC->nodes[i].History = RU->Next;
            ExFreePool(RU);
        }
#endif
    }

    ExFreePool(LC->metric);
    ExFreePool(LC->hops);
    ExFreePool(LC->prev);
    ExFreePool(LC->link);
    ExFreePool(LC->nodes);
    ExFreePool(LC);
}

//* LinkCacheResetStatistics
//
//  Resets all counters and statistics gathering for the link cache.
//
void
LinkCacheResetStatistics(MiniportAdapter *VA)
{
    LinkCache *LC = VA->LC;
    Link *Adj;
#if CHANGELOGS
    RouteUsage *RU;
    uint i;
#endif
    KIRQL OldIrql;

    KeAcquireSpinLock(&LC->Lock, &OldIrql);
    LC->CountRouteFlapDamp = 0;
    LC->CountRouteFlap = 0;
    LC->CountAddLinkInvalidate = 0;
    LC->CountAddLinkInsignificant = 0;
    for (Adj = LC->nodes[0].AdjOut; Adj != NULL; Adj = Adj->NextOut)
        Adj->Failures = 0;
    for (i = 0; i < LC->nodeCount; i++) {
        LC->nodes[i].RouteChanges = 0;
#if CHANGELOGS
        while ((RU = LC->nodes[i].History) != NULL) {
            LC->nodes[i].History = RU->Next;
            ExFreePool(RU);
        }
        LC->nodes[i].CurrentRU = NULL;
#endif
    }
    KeReleaseSpinLock(&LC->Lock, OldIrql);
}

//* LinkCacheFlush
//
//  Deletes all cached link information.
//
void
LinkCacheFlush(MiniportAdapter *VA)
{
    LinkCache *LC = VA->LC;
    Link *tmp;
#if CHANGELOGS
    RouteUsage *RU;
#endif
    uint i;
    KIRQL OldIrql;

    KeAcquireSpinLock(&LC->Lock, &OldIrql);

    //
    // Free all the links and cached routes.
    //
    for (i = 0; i < LC->nodeCount; i++) {
        while (LC->nodes[i].AdjOut != NULL) {
            tmp = LC->nodes[i].AdjOut;
            LinkCacheLinkChange(LC,
                                LC->nodes[i].address,
                                LC->nodes[tmp->targetIndex].address,
                                tmp->inif, tmp->outif,
                                (uint)-1,
                                LINK_STATE_CHANGE_DELETE_MANUAL);
            LC->nodes[i].AdjOut = LC->nodes[i].AdjOut->NextOut;
            ExFreePool(tmp);
        }

        if (LC->nodes[i].SR != NULL) {
            // This also frees LC->nodes[i].Hops.
            ExFreePool(LC->nodes[i].SR);
            LinkCacheRouteChange(LC, LC->nodes[i].address,
                                 LC->nodes[i].Metric, (uint)-1,
                                 NULL, NULL);
        }

#if CHANGELOGS
        while ((RU = LC->nodes[i].History) != NULL) {
            LC->nodes[i].History = RU->Next;
            ExFreePool(RU);
        }
#endif
    }

    //
    // Reset to just me.
    //
    LC->nodes[0].AdjIn = NULL;
    LC->timeout = KeQueryInterruptTime();
	delete_arb_dijkstra_caches(VA);
    LC->nodeCount = 1;
    LC->SmallestMetric = 0;
    LC->LargestMetric = (uint)-1;

    KeReleaseSpinLock(&LC->Lock, OldIrql);
}

//* LinkCacheMyDegree
//
//  Returns our own link degree. That is, the number of links
//  from node 0 to other nodes.
//
uint
LinkCacheMyDegree(MiniportAdapter *VA)
{
    LinkCache *LC = VA->LC;
    Link *link;
    uint Degree = 0;
    KIRQL OldIrql;

    KeAcquireSpinLock(&LC->Lock, &OldIrql);
    for (link = LC->nodes[0].AdjOut; link != NULL; link = link->NextOut)
        Degree++;
    KeReleaseSpinLock(&LC->Lock, OldIrql);

    return Degree;
}

//* LinkCacheUseSR
//
//  Called when sending/forwarding a packet with a source route.
//  Updates our usage counter for the next link AND
//  updates the Metric in the source route for the next link.
//
//  Also checks if sending this packet would overflow
//  the transmit queue of the outgoing interface.
//  Returns FALSE to indicate the packet should be dropped.
//
boolint
LinkCacheUseSR(
    MiniportAdapter *VA,
    SRPacket *SRP)
{
    LinkCache *LC = VA->LC;
    InternalSourceRoute *SR = SRP->sr;
    Link *Link;
    int DNode;
    KIRQL OldIrql;
    uint NumHops;
    uint Hop;
    boolint QueueFull;

    NumHops = SOURCE_ROUTE_HOPS(SR->opt.optDataLen);
    ASSERT(NumHops <= MAX_SR_LEN);

    Hop = NumHops - SR->opt.segmentsLeft;
    ASSERT((Hop < NumHops) && (Hop != 0));

    ASSERT(VirtualAddressEqual(SR->opt.hopList[Hop - 1].addr,
                               VA->LC->nodes[0].address));

    //
    // We only want to check for packets that are carrying real payload.
    //
    if (SRP->PacketLength != SRP->PayloadOffset)
        QueueFull = ProtocolQueueFull(VA, SR->opt.hopList[Hop - 1].outif);
    else
        QueueFull = FALSE;

    KeAcquireSpinLock(&LC->Lock, &OldIrql);
    DNode = findNode(LC, SR->opt.hopList[Hop].addr, FALSE);
    if (DNode > 0) {

        //
        // Search our link cache for this link.
        //
        for (Link = LC->nodes[0].AdjOut; Link != NULL; Link = Link->NextOut) {
            if ((Link->targetIndex == DNode) &&
                (Link->outif == SR->opt.hopList[Hop - 1].outif) &&
                (Link->inif == SR->opt.hopList[Hop].inif)) {
                //
                // Found link (should match at most one).
                //

                if (! SR->opt.staticRoute)
                    Link->Usage++;

                if (QueueFull)
                    Link->QueueDrops++;

                //
                // We'll update the option with the metric of this link.
                // Unless we think the link is broken, which is stale info.
                //
                if (! (*VA->IsInfinite)(Link->Metric))
                    SR->opt.hopList[Hop].Metric = Link->Metric;
                break;
            }
        }
    }
    KeReleaseSpinLock(&LC->Lock, OldIrql);

    return !QueueFull;
}

//* LinkCacheRememberLinks
//
//  Stores references to the links in a route to the destination.
//
//  NB: The links are stored backwards, from last to first.
//
//  Called with the link cache locked.
//
void
LinkCacheRememberLinks(
    LinkCache *LC,
    int dnode,
    Link **Hops)
{
    Link *Hop;

    while (dnode != 0) {
        Hop = LC->link[dnode];
        Hop->RefCnt++;
        *Hops++ = Hop;
        dnode = LC->prev[dnode];
        ASSERT(dnode != (uint)-1);
    }
}

//* LinkCacheReleaseLinks
//
//  Releases references to the links in the array.
//
//  Called with the link cache locked.
//
void
LinkCacheReleaseLinks(
    Link **Hops,
    uint NumHops)
{
    uint i;

    for (i = 0; i < NumHops; i++)
        Hops[i]->RefCnt--;
}

//* LinkCacheFlapDamp
//
//  Returns TRUE if the old route is good enough.
//
boolint
LinkCacheFlapDamp(
    uint OldMetric,     // Current metric of the old route.
    uint NewMetric,     // Metric of the current best route.
    Time Age,           // Lifetime of the old route.
    uint DampingFactor) // Damping Factor. 
{
    uint FudgeFactor;
    uint FudgedMetric;

    //
    // If the old route is broken, we should not use it.
    //
    if (OldMetric == (uint)-1)
        return FALSE;

    //
    // The old route can actually look better, if its link metrics
    // have improved since dijkstra was run.
    //
    if (OldMetric <= NewMetric)
        return TRUE;

    //
    // Calculate a "fudge factor". Note that if DampingFactor
    // is 0, we do not do any damping.
    // NB: We check for overflow below.
    //
    if (Age < 100 * MILLISECOND)
        FudgeFactor = (DampingFactor * NewMetric) / 8;
    else if (Age < SECOND)
        FudgeFactor = (DampingFactor * NewMetric) / 16;
    else if (Age < 10 * SECOND)
        FudgeFactor = (DampingFactor * NewMetric) / 32;
    else if (Age < 100 * SECOND)
        FudgeFactor = (DampingFactor * NewMetric) / 64;
    else
        FudgeFactor = 0; 

    FudgedMetric = NewMetric + FudgeFactor; 
    
    //
    // Switch to the new route if it is sufficiently better.
    // The first condition checks for overflow. 
    //
    if ((NewMetric <= FudgedMetric) && (FudgedMetric <= OldMetric))
        return FALSE;

    //
    // The new route is better, but not enough better
    // to warrant switching to it.
    //
    if(debug) KdPrint(("MCL!LinkCacheFlapDamp: old %u new %u\n",
             OldMetric, NewMetric));
    return TRUE;
}



//* inRange
//
// Checks whether two nodes are in range
// of one another.
//
// Looks at the link from src to dest. 
//
// Assumes LC lock is held.
//
boolint
inRange(MiniportAdapter *VA, SRAddr *src, SRAddr *dest, uint ETX_FORWARD_SELECTION_THRESH) 
{
	//LinkCache *LC = VA->LC;
	//uint i ; 
	uint curMetric = LinkCacheLookupMetric(VA, src->addr, dest->addr, dest->inif, src->outif);

	if (curMetric != 0 && !((*VA->IsInfinite)(curMetric)) && 
		(*VA->ConvMetric)(curMetric) < ETX_FORWARD_SELECTION_THRESH) {
		return 1;
	} else {
		return 0 ;
	}

}


//* isForwListConnected
//
// Checks whether 'hopper' is in range of every node in tmps.
// Returns true if so, false otherwise.
//
// Assumes the LC Lock is held.
//
boolint 
isForwListConnected(MiniportAdapter *VA, 
					SourceRoute* tmps, 
					uint tmps_cnt, 
					SRAddr *hopper, 
					uint ETX_FORWARD_SELECTION_THRESH)
{
	uint i = 0 ; 
	
	//if tmps is empty, then list is connected
	if (tmps_cnt == 0 || tmps == NULL) {
		return 1; 
	}
	
	//check hopper against all in tmps
	for (i = 0 ; i < tmps_cnt; i++) {
		if (!inRange(VA, hopper, &(tmps->hopList[i]), ETX_FORWARD_SELECTION_THRESH)) {
			return 0 ; 
		}
	}
	return 1; 
}

//* forwThreshSatisfied
//
// When picking forwarders, we want to ensure that
// the probability of a sucessful transmission is above some
// threshold.
//
// We check all nodes in tmps.
//
// Brute force for now, need to optimize later.
//
// Returns true if threshold is satisified.
//
// retloss is returned
//
boolint
forwThreshSatisfied(MiniportAdapter *VA, SourceRoute* tmps, uint tmps_cnt, uint* retloss) 
{
	uint loss = 0 ; 
	uint i = 0 ; 

	for (i = 0 ; i < tmps_cnt ; i++) {
		uint curLoss = 0 ; 
		//assuming only 1 interface for now!
		uint curMetric = LinkCacheLookupMetric(VA, VA->Address, tmps->hopList[i].addr, 
												tmps->hopList[i].inif, 
												tmps->hopList[i].inif);

		ASSERT(curMetric);
		curMetric = (*VA->ConvMetric) (curMetric); 
		curLoss = LOSS_SCALE - (LOSS_SCALE*4096)/curMetric  ;
		if (i == 0) 
			loss = curLoss;
		else
			loss = (loss*curLoss)/LOSS_SCALE; 
	}
	
	*retloss = loss; 
	if (loss > FORW_LOSS_THRESH)
		return 0;
	else
		return 1;
	
}

// Checks to see if the node 'hop' is already on the
// source route 'sr.' If so, returns true and updates
// the value of the index into 'index'.
//
// Does not need to check against dp or fwding lists.
//
boolint nodeOnSourceRoute(SourceRoute* r, uint sr_cnt, SRAddr *hop, uint* index)
{
	uint i ; 
	for (i = 0 ; i < sr_cnt; i++) {
		if (VirtualAddressEqual(hop->addr, r->hopList[i].addr)) {
			*index = i ;
			return TRUE ;
		}
	}
	return FALSE ; 
}




//* onPath
//
// Checks to see if 'addr' is on the default path
// on srp->sr. Needs to check bitmap.
//
boolint 
onPath(SRPacket* srp, VirtualAddress* addr)
{
	uint i = 0 ;

	for (i = 0 ; i < SOURCE_ROUTE_HOPS(srp->sr->opt.optDataLen); i++) {
		if (VirtualAddressEqual(srp->sr->opt.hopList[i].addr, *addr)) {
			if (checkSRPbitmap(srp->sr->opt.bitmap, i))
				return 1; 
			else 
				return 0;
		}
	}
	return 0; 
} 

//* LinkCacheInsertForwardingList
//
// Injects the *final* forwarding list into the packet.
// This is where we want our forwarding algorithms.
// The condition where non-default path nodes
// need to find the closest default path node to take
// their forwarding list has already been taken care of.
//
// Note that CanForw1 has not been pruned, so we need to do that.
// This method will deallocate the CanForw1 memory.
//
// Bitmap must be set on srp before calling. The callers of this method should 
// have their srp->sr set up, in case this fails.
//
// We assume that the max and min bounds on forwarding list
// have been updated on the srp packet. 
// This method will update these values upon creating the new
// forwarding list.
//
// This method will take care of case that CanForw1 is null.
//
//  Input:
//		CanForw1- the list generated by LinkCacheFindForwarders
//		srp- the packet we are going to send
//		ETX_FORWARD_SELECTION_THRESH- value used in find forwarders
//	Return:
//		status of operations
//
NDIS_STATUS
LinkCacheInsertForwardingList(MiniportAdapter *VA,
							InternalSourceRoute *CanForw1,
							SRPacket* srp,
							uint ETX_FORWARD_SELECTION_THRESH)
{
	//LinkCache *LC = VA->LC;
    //KIRQL OldIrql;
	SourceRoute* CanForw;
	SourceRoute* tmps;//temp array to hold forwarding nodes that will go in srp->sr
	ushort tmps_cnt = 0 ; //index into tmps array
	ushort new_sr_cnt = 0 ; //index into new_sr array
	InternalSourceRoute *new_sr1; 
	SourceRoute* new_sr;//the *whole* source route we'll attach to srp
	uint lowerBound ; //gives an index into srp->sr array (who is sender)
	uint i = 0; 
	uint numForw = 0 ; //number of forwarding nodes we stick in the packet
	boolint forwSat = FALSE;
	uint max_forwarders ; //maximum forwarders = min(MAX_NUM_FORWARDERS, spots left in pkt); 
	int si ; //signed i	

	uint retloss = (uint)-1; //returned loss for rate-limiting

	SourceRoute* cliq_tmps;//temp array to hold forwarding nodes that will go in the clique
	ushort cliq_tmps_cnt = 0 ; //index into clique array

	boolint destOnTmps = FALSE; 
	//uint bestCindex ; //for adding nodes from clique array
	//uint bestCvalue; 
	
	//need to ensure that Dest is still in correct spot in our final fwd list
	VirtualAddress Dest ;
	
	ASSERT(srp->sr != NULL); 
	ASSERT(srp->sr->opt.bitmap); //bitmap should be set
	
	RtlCopyMemory(&Dest, &(srp->sr->opt.hopList[SOURCE_ROUTE_HOPS(srp->sr->opt.optDataLen)-1].addr), sizeof(VirtualAddress));

	if (CanForw1 == NULL) {
		//technically, we finished just fine
		//
		// We may want to change the behavior... if we have no forwarders, then just
		// drop the packet.
		//
		return NDIS_STATUS_SUCCESS; 
	}

	CanForw = &(CanForw1->opt) ;

	cliq_tmps = ExAllocatePool(NonPagedPool, sizeof *cliq_tmps + sizeof(SRAddr)*MAX_SR_LEN); 
	if (cliq_tmps == NULL) {
		if(debug) KdPrint(("LinkCacheInsertForwardingList: Can't alloc cliq_tmps\n"));
		ExFreePool(CanForw1);
		return NDIS_STATUS_RESOURCES ;
    }	
	RtlZeroMemory(cliq_tmps, sizeof *cliq_tmps + sizeof(SRAddr)*MAX_SR_LEN); 

	//have to ensure we don't go over MAX_SR_LEN in packet.
	max_forwarders =  MAX_SR_LEN - SOURCE_ROUTE_HOPS(srp->sr->opt.optDataLen) ;
	if (max_forwarders > MAX_NUM_FORWARDERS)
		max_forwarders = MAX_NUM_FORWARDERS ;

	//grab lock
	//KeAcquireSpinLock(&VA->LC->Lock, &OldIrql);		

	//starting with highest priority node, go thru 
	//forwarding list and add node if 1) graph is connected
	for (i = 0 ; i < SOURCE_ROUTE_HOPS(CanForw->optDataLen) ; i++) {
		//we know highest priority default path node is the first node in CanForw
		if (isForwListConnected(VA, cliq_tmps, cliq_tmps_cnt, &(CanForw->hopList[i]), ETX_FORWARD_SELECTION_THRESH)) {
			PRINT_VA("\t\t\tAdding node to clique: ", CanForw->hopList[i].addr);
			RtlCopyMemory(&(cliq_tmps->hopList[cliq_tmps_cnt]), &(CanForw->hopList[i]), sizeof(SRAddr)); 
			cliq_tmps_cnt++; 
		}
	}
	ASSERT(cliq_tmps_cnt);

	//so now want to add forwarding list in tmps 
	tmps = ExAllocatePool(NonPagedPool, sizeof *tmps + sizeof(SRAddr)*cliq_tmps_cnt); 
	if (tmps == NULL) {
		//KeReleaseSpinLock(&VA->LC->Lock, OldIrql);	
		if(debug) KdPrint(("LinkCacheInsertForwardingList: Can't alloc tmps\n"));
		ExFreePool(cliq_tmps);	
		ExFreePool(CanForw1);
		return NDIS_STATUS_RESOURCES ;
    }	
	RtlZeroMemory(tmps, sizeof *tmps + sizeof(SRAddr)*cliq_tmps_cnt); 

	//Now, add the highest priority node and keep adding nodes until the virtual loss is satisfied
	// or max num  forwarders - 1 is reached.
	for (i = 0 ; i < cliq_tmps_cnt ; i++) { //if we uncomment the above, need to change to "i = 1"
		//find best node
		uint maxProgress = (uint) -1 ;
		uint prev_maxp = maxProgress; 
		uint maxIndex = (uint) -1 ;
		uint bestFromMe = (uint) -1 ;//if we have a tie, then want forwarder closest to me to win
		uint w ;
		//get node with best progress
		for (w = 0 ; w < cliq_tmps_cnt; w++) {
			if (cliq_tmps->hopList[w].Metric != 0 && cliq_tmps->hopList[w].Metric <= maxProgress) {
				uint curMetricm;
				//need to reset bestFromMe if necessary.
				if (cliq_tmps->hopList[w].Metric != prev_maxp) {
					bestFromMe = (uint) -1 ;
					prev_maxp = cliq_tmps->hopList[w].Metric ;
				}
				//let's also pick on best metric from me
				//assuming only one interface for now!!
				curMetricm = LinkCacheLookupMetric(VA, VA->Address,cliq_tmps->hopList[w].addr,
												cliq_tmps->hopList[w].outif, cliq_tmps->hopList[w].outif); 
		
				ASSERT(curMetricm);
				curMetricm = (*VA->ConvMetric)(curMetricm);
				if (bestFromMe == (uint) -1  || curMetricm < bestFromMe) {
					maxProgress = cliq_tmps->hopList[w].Metric;
					maxIndex = w ;
					bestFromMe = curMetricm; 
				}
			}
		}
		ASSERT(maxProgress != (uint) -1);

		//add the node
		RtlCopyMemory(&(tmps->hopList[tmps_cnt]), &(cliq_tmps->hopList[maxIndex]), sizeof(SRAddr)); 
		numForw++ ;
		tmps_cnt++; 
		cliq_tmps->hopList[maxIndex].Metric = 0 ; 

		//check virtual loss and max_forwarders

		//should have this method cache the link metric in the 'tmps' array
		//then be sure to clear them later. Or could have it return an int
		// with the loss percentage and just repass it in everytime along with
		// new fwding node's loss
		if (forwThreshSatisfied(VA, tmps, tmps_cnt, &retloss)) {
			forwSat = TRUE ; 
			break; 
		} else if (tmps_cnt == max_forwarders - 1) {
			break ;
		}
	}
			

	// It is possible that we don't have enough forwarders in CanForw
	// to satisfy the threshold. We don't want to enter the if statement
	// if that occurs.	
	if (!forwSat && (tmps_cnt) != cliq_tmps_cnt ) {
		uint bestMetric = (uint) -1;
		uint bestIndex = 0 ;
		ASSERT(tmps_cnt == max_forwarders - 1); 
		
		//let's just go thru rest of CanForw and find the one
		// that I have best link to.
		for (i = 0 ; i < cliq_tmps_cnt; i++) {
			//assuming only 1 interface for now!
			uint curMetric ;
			if (cliq_tmps->hopList[i].Metric == 0)
					continue; 

			curMetric = LinkCacheLookupMetric(VA, VA->Address, cliq_tmps->hopList[i].addr, 
												cliq_tmps->hopList[i].inif, 
												cliq_tmps->hopList[i].inif);

			ASSERT(curMetric);
			curMetric = (*VA->ConvMetric) (curMetric); 
			if (curMetric < bestMetric) {
				bestMetric = curMetric ;
				bestIndex = i ; 
			}
		}

		ASSERT(bestMetric != (uint) -1); 
					
		//remember to update numForw
		RtlCopyMemory(&(tmps->hopList[tmps_cnt]), &(cliq_tmps->hopList[bestIndex]), sizeof(SRAddr)); 
		tmps_cnt++; 
		numForw++; 
#if (SLIDING_RATE_LIMIT_ENABLED	|| AIMD_RATE_LIMIT_ENABLED)
		//rate-limit
		//we need to get the final loss rate
		forwThreshSatisfied(VA, tmps, tmps_cnt, &retloss);
#endif
	}

	//free cliq_tmps
	ExFreePool(cliq_tmps);	
	cliq_tmps = NULL;


#if (SLIDING_RATE_LIMIT_ENABLED	|| AIMD_RATE_LIMIT_ENABLED)
	//rate-limit
	//if I am the source, then we want to save the virtual loss rate
	if (VirtualAddressEqual(VA->Address, srp->sr->opt.hopList[0].addr)) {
		AckCache *AC; 
		//assuming only one interface for now!!
		AC = MaintBufFindAckFlow(VA, &(VA->Address), &(srp->Dest), 
			srp->sr->opt.hopList[1].inif, srp->sr->opt.hopList[1].inif, &VA->Address); 
		if (AC != NULL) {
			ASSERT(retloss != (uint) -1);
			AC->virtualLoss = retloss; 
			if(debug) KdPrint(("Virtual loss: %u\n", retloss)); 
		}
	}
#endif


	//update srp->sr to add in the tmps array.
	new_sr1 = ExAllocatePool(NonPagedPool, sizeof *new_sr1 + sizeof(SRAddr)*MAX_SR_LEN); 
	if (new_sr1 == NULL) {
		//KeReleaseSpinLock(&VA->LC->Lock, OldIrql);	
		if(debug) KdPrint(("LinkCacheInsertForwardingList: Can't alloc new_sr\n"));
		ExFreePool(CanForw1);
		ExFreePool(tmps);	
		return NDIS_STATUS_RESOURCES ;
    }	
	RtlZeroMemory(new_sr1, sizeof *new_sr1 + sizeof(SRAddr)*MAX_SR_LEN); 
	new_sr = &(new_sr1->opt);
	
	// The value 'lowerBound - 1' should point to the original
	// sender of the packet and 'lowerBound' should indicate the original
	// sender's next hop. This is because we assume segmentsLeft has been updated
	// before calling this method.
	lowerBound = SOURCE_ROUTE_HOPS(srp->sr->opt.optDataLen) - srp->sr->opt.segmentsLeft ; 
	ASSERT(VirtualAddressEqual(srp->sr->opt.hopList[lowerBound-1].addr, VA->Address)); 
	
	//add all low priority nodes on default path
	for (i = 0 ; i < lowerBound-1 ; i++) {	
		RtlCopyMemory(&(new_sr->hopList[new_sr_cnt]), &(srp->sr->opt.hopList[i]), sizeof(SRAddr)); 
		ASSERT(checkSRPbitmap(srp->sr->opt.bitmap, i));
		setSRPbitmap(&(new_sr->bitmap), new_sr_cnt); 
		new_sr_cnt++ ; 
	}
	
	ASSERT(new_sr_cnt == lowerBound-1); 
	ASSERT(VirtualAddressEqual(VA->Address, srp->sr->opt.hopList[i].addr)); 
	RtlCopyMemory(&(new_sr->hopList[new_sr_cnt]), &(srp->sr->opt.hopList[i]), sizeof(SRAddr)); 
	//If sending node is a non-dp node, then one of the entries in the srp->sr
	// is not on the default path!!
	if (checkSRPbitmap(srp->sr->opt.bitmap, lowerBound-1))
		setSRPbitmap(&(new_sr->bitmap), new_sr_cnt); 
	
	//do I set myself as a fwding node? If so, makes priority
	// calculation difficult b/c a fwding node can be to the 
	// left of me in the hopList[]. For now, we do not.
	new_sr_cnt++ ; 


	//add all potential forwarders (can be on dp)
	//--tmps is not sorted anymore, so we must sort
	ASSERT(numForw == tmps_cnt); //we can get ride of one of these
	for (si = numForw-1 ; si >= 0 ; si--) {
		uint index = 0 ;
		uint worstNode = 0 ;
		uint worstMetric = 0; 
		uint ii = 0; 

		//find worst node to use
		for (ii = 0 ; ii < numForw ; ii++) {
			if (tmps->hopList[ii].Metric != 0 && tmps->hopList[ii].Metric >= worstMetric) {
				worstNode = ii;
				worstMetric = tmps->hopList[ii].Metric;
			}
		}

		//need to first check if tmps->hopList[i] is already on new_sr
		if (nodeOnSourceRoute(new_sr, new_sr_cnt, &(tmps->hopList[worstNode]), &index)) {
			setSRPbitmap(&(new_sr->fwd_bitmap), index);
			tmps->hopList[worstNode].Metric = 0 ;
			continue ;
		}

		//This can cause a problem if the destination is in tmps[], but the
		// default next hop is not (because a forwarder had higher priority and
		// provides good enough virtual link quality). What would happen is that
		// the destination would get added here and then the default next hop would
		// be the final entry in the new_sr, which breaks our assumption that the dest
		// is always at the end of the source route.
		if (VirtualAddressEqual(Dest, tmps->hopList[worstNode].addr)) {
			destOnTmps = TRUE; 
			tmps->hopList[worstNode].Metric = 0 ;
			continue; 
		}

		RtlCopyMemory(&(new_sr->hopList[new_sr_cnt]), &(tmps->hopList[worstNode]), sizeof(SRAddr)); 
		if (onPath(srp, &(tmps->hopList[worstNode].addr)))
			setSRPbitmap(&(new_sr->bitmap), new_sr_cnt); 
		setSRPbitmap(&(new_sr->fwd_bitmap), new_sr_cnt);
		tmps->hopList[worstNode].Metric = 0 ;
		new_sr_cnt++ ; 
	}


	//add all dp nodes nodes too far downstream
	for (i = lowerBound ; i < SOURCE_ROUTE_HOPS(srp->sr->opt.optDataLen) ; i++) {
		uint index = 0 ; 
		ASSERT(checkSRPbitmap(srp->sr->opt.bitmap, i));
		if (nodeOnSourceRoute(new_sr, new_sr_cnt, &(srp->sr->opt.hopList[i]), &index)) {
			//all bitmaps should already be set.
			continue; 
		}
		RtlCopyMemory(&(new_sr->hopList[new_sr_cnt]), &(srp->sr->opt.hopList[i]), sizeof(SRAddr)); 
		setSRPbitmap(&(new_sr->bitmap), new_sr_cnt); 
		new_sr_cnt++ ; 		
	}

	if (destOnTmps) 
		setSRPbitmap(&(new_sr->fwd_bitmap), new_sr_cnt-1); //new_sr_cnt should be # of hops 


	//update all other variables 
	new_sr->optionType = LQSR_OPTION_TYPE_SOURCERT; 
	new_sr->optDataLen = SOURCE_ROUTE_LEN(new_sr_cnt) ;
	new_sr->reservedField = srp->sr->opt.reservedField;
    new_sr->staticRoute =  srp->sr->opt.staticRoute;
    new_sr->salvageCount = srp->sr->opt.salvageCount;
	// We would want 'Index' to point to the next hop:
	// Index = SOURCE_ROUTE_HOPS(srp->sr->opt.optDataLen) - srp->sr->opt.segmentsLeft;
    new_sr->segmentsLeft = (uchar) (new_sr_cnt - lowerBound);
	ASSERT(VirtualAddressEqual(VA->Address,
		new_sr->hopList[SOURCE_ROUTE_HOPS(new_sr->optDataLen) - new_sr->segmentsLeft - 1].addr)); 
	
	for (i = 0 ; i < SOURCE_ROUTE_HOPS(new_sr->optDataLen) ; i++) {
		//need to figure out what to do-- just zero them out, LinkCacheUseSR
		// should take care of this. I also commented out the code that
		// passively updates link metrics from a received sr packet.
		new_sr->hopList[i].Metric = 0 ; 
	}

	//Have to update the srp->sr in the packet!!!
	//--free old sr
	SRFreeOptionList((InternalOption *)srp->sr);
	srp->sr = new_sr1; 

	ExFreePool(tmps);	
	ExFreePool(CanForw1);

	ASSERT(VirtualAddressEqual(Dest, srp->sr->opt.hopList[SOURCE_ROUTE_HOPS(srp->sr->opt.optDataLen)-1].addr)); 

	return NDIS_STATUS_SUCCESS; 
}




//* closetDefaultPathNode
//
// Returns the node on the default path that the
// current node is closest to.
//
// Return NULL if none found.
//
VirtualAddress*
closestDefaultPathNode(MiniportAdapter *VA, SRPacket* srp) 
{
    LinkCache *LC = VA->LC;
	uint i = 0 ;
	uint bestValue = (uint) -1;
	uint bestIndex = 0;

	for (i = 0  ; i < SOURCE_ROUTE_HOPS(srp->sr->opt.optDataLen) ; i++) {
		if (checkSRPbitmap(srp->sr->opt.bitmap, i)) {
			Link *walk;
			//find the node in my cache
			for (walk = LC->nodes[0].AdjOut ; walk != NULL; walk = walk->NextOut) {
				if (VirtualAddressEqual(LC->nodes[walk->targetIndex].address, srp->sr->opt.hopList[i].addr)) {
					uint curValue2 = (*VA->ConvMetric) (walk->Metric); 
						
					//change to curValue2 <= bestValue since we want higher priority node if possible
					if (walk->Metric != 0 && !(*VA->IsInfinite)(walk->Metric) && curValue2 <= bestValue ) {
						bestValue = curValue2; 
						bestIndex = i ;
					}
					break;
				}//end if
			}//end for walk

		}//end if checkSRPbitmap
	}//end for i

	if (bestValue == (uint) -1)
		return NULL; 
	return &(srp->sr->opt.hopList[bestIndex].addr); 
}

//* numDefaultPathNodes
//
// Returns the number of nodes in the source route
// that are part of the default path.
//
uint
numDefaultPathNodes(SRPacket* srp) {

	uint i = 0 ;
	uint cnt = 0 ;
	for (i = 0 ; i < SOURCE_ROUTE_HOPS(srp->sr->opt.optDataLen) ; i++) {
		ASSERT(i < MAX_SR_LEN); 
		//if(debug) KdPrint(("\t\ti is: %u and srh: %u\n",i,SOURCE_ROUTE_HOPS(srp->sr->opt.optDataLen)));
		if (checkSRPbitmap(srp->sr->opt.bitmap, i)) {
			cnt++;
		}
	}	
	return cnt; 
}


//* get_ETX_fwd_thresh(VA)
//
// we use to determine the ETX_FORWARD_THRESH is:
// max(ETX-to-default-next-hop*ETX_FACTOR, DEFAULT_ETX_FORWARD_SELECTION_THRESH)
// 
// srp- the packet we are getting forwarders for
// sendFrom- the dp path node that is closest (or me if on dp)
//
uint get_ETX_fwd_thresh_SR(MiniportAdapter *VA, 
					  SourceRoute* SR,
					  VirtualAddress* sendFrom)
{
	uint i = 0; 
	uint j = 0; 
	uint metric = 0 ; 
	uint nextHopETX = 0 ; 

	//need to find next dp path node from sendFrom
	for (i = 0 ; i < SOURCE_ROUTE_HOPS(SR->optDataLen) ; i++) {
		if (VirtualAddressEqual(*sendFrom, SR->hopList[i].addr))
			break; 
	}
	for (j = i+1 ; j < SOURCE_ROUTE_HOPS(SR->optDataLen) ; j++) {
		if (checkSRPbitmap(SR->bitmap, j)) {
			//get link info		
			//assume only one interface for now!
			metric = LinkCacheLookupMetric(VA, *sendFrom, SR->hopList[j].addr, 
												SR->hopList[j].inif, 
												SR->hopList[j].inif);
			
			nextHopETX = (*VA->ConvMetric) (metric); 
			break ;
		}	
	}
	if (metric == 0 || j >= SOURCE_ROUTE_HOPS(SR->optDataLen) ||
		(*VA->IsInfinite)(metric))
        return DEFAULT_ETX_FORWARD_SELECTION_THRESH ;

	if (nextHopETX*ETX_FACTOR > DEFAULT_ETX_FORWARD_SELECTION_THRESH)
        return nextHopETX*ETX_FACTOR ;

    return DEFAULT_ETX_FORWARD_SELECTION_THRESH ;

}


uint get_ETX_fwd_thresh(MiniportAdapter *VA, 
					  SRPacket* srp,
					  VirtualAddress* sendFrom)
{
	return get_ETX_fwd_thresh_SR(VA, &(srp->sr->opt), sendFrom);
}



// isLowerPriority
//
// checks to see if Addr is lower priority than me on the source
// route SR. Returns true if so.
//
// Only checks SR, does no checking of etx values to the destination. ie- progress
// is not considered here.
//
// Assumed it is called on srp->old_sr
//
static boolint
isLowerPriorityFwd(MiniportAdapter *VA, VirtualAddress *Addr, SourceRoute *SR) {
	uint i = 0 ; 
	uint hops = SOURCE_ROUTE_HOPS(SR->optDataLen);
	uint Index = hops - SR->segmentsLeft - 1; //sender of packet

	//if node sent the packet, then it is lower priority
	if (VirtualAddressEqual(SR->hopList[Index].addr, *Addr))
			return TRUE; 

	for (i = 0 ; i < hops; i++) {
		if (VirtualAddressEqual(*Addr, SR->hopList[i].addr)) {
			
			if (checkSRPbitmap(SR->fwd_bitmap, i)) {
				//we haven't found ourselves on the fwd list, but
				// Addr has been found at a lower index... it is lower
				// priority.
				return TRUE;
			} else {
				//Addr is not on the fwd list, then return false.
				return FALSE; 
			}
		} else if (VirtualAddressEqual(VA->Address, SR->hopList[i].addr)) {
			//if we find me first, then the ndoe cannot be lower priority
			return FALSE ; 
		}
	}
	//we should never be here b/c we should always find ourselves on the 
	// fwd list.
	ASSERT(0); 	
	return FALSE ; 
}



//
// Returns TRUE if this node satisfies the constraints
// to potentially be a forwarding node:
//		-makes progress from me
//		-is w/in ETX_THRESH of me
//		-is close to default path
// Returns FALSE otherwise
//
//
// srp- the packet we received and need to send
// myETX- the path ETX from me to dest 
// ForwNode - address of node we want to consider for our forwarding list
// sendFrom - is set to VA->Address if I'm on the default path
// forwETX - returned paramemter, path ETX of ForwNode to dest
// rcurValue - returned parameter, value of link metric from me to ForwNode
// dp_sendFrom - nodes on default path that are in range and make progress
//					--NULL if we are trying to find these default path nodes
// num_dp_sendFrom - number of nodes on dp_sendFrom list
//					-- 0 if we are trying to fidn these default path nodes
// DP_NODE -- set to one if the forwarder we are considering is a default path node
//
//
boolint
addPotentialForwarder(MiniportAdapter *VA, SRPacket* srp, uint myETX,
					  VirtualAddress *ForwNode, VirtualAddress *sendFrom, 
					  uint ETX_FORWARD_SELECTION_THRESH, uint *forwETX, 
					  uint *rcurValue, SourceRoute* dp_sendFrom, uint num_dp_sendFrom, 
					  Time Now, boolint DP_NODE) 

{
	
	uint curValue; //my ETX to adjacent node
	boolint onList = 0 ;
	uint ret1;

	SRAddr *Dest = &(srp->sr->opt.hopList[SOURCE_ROUTE_HOPS(srp->sr->opt.optDataLen)-1]); 

	if (num_dp_sendFrom == 0)
		ASSERT(dp_sendFrom == NULL);

	//never care about me.
	if (VirtualAddressEqual(VA->Address, *ForwNode))
		return FALSE; 
	
	PRINT_VA("\tFFFFF looking at node: ", *ForwNode);

	//never want to add the original sender
	if (VirtualAddressEqual(*ForwNode, srp->sr->opt.hopList[0].addr)) {
		PRINT_VA("\t\tnode is flow source, skip: ", *ForwNode);			
		return FALSE; 
	}

	//if the node was on the old source route and was lower priority, then we want to skip
	if (srp->old_sr && isLowerPriorityFwd(VA, ForwNode, &(srp->old_sr->opt))) {
		PRINT_VA("\t\tnode is lower priority, skip: ", *ForwNode);			
		return FALSE; 
	}

	//Get ETX of link to the adjacent node
	//Assuming only 1 interface for now!!!
	curValue = LinkCacheLookupMetric(VA, VA->Address, *ForwNode, Dest->inif, Dest->inif); 
	if (curValue == 0) {
		PRINT_VA("Missing link value in LinkCacheFindForwarders!", *ForwNode);
		return FALSE ;
	}
	if ((*VA->IsInfinite)(curValue) || 
		(*VA->ConvMetric) (curValue)  >= ETX_FORWARD_SELECTION_THRESH) {
		
		if(debug) KdPrint(("\t\tnode too far away, skip: %d\n", (*VA->ConvMetric) (curValue)));
		return FALSE; 
	}
	curValue = (*VA->ConvMetric) (curValue) ;
	*rcurValue = curValue; 
	
	//If the forwarding node is not on the default path, then we must
	// check the additional constraint that the potential forwarder 
	// is close enough to a default path node.
	onList = 0 ;
	if (!(VirtualAddressEqual(VA->Address, *sendFrom))) {
		uint z; 

		//If ForwNode is on the default path, then we 
		// check if it makes progress from me and if it 
		// is within threshold. Otherwise, we check its
		// connectivity from sendFrom. We already check above
		// that the node has okay connectivity to me, so here we
		// just need to check if the node is on the default path.
		// We check for progress last.
		if (onPath(srp, ForwNode) ) {
			goto if_done; 
		}
		
		ASSERT(num_dp_sendFrom != 0);
		//need to see if ForwNode is in range of *any* node on dp_sendFrom
		for (z = 0 ; z < num_dp_sendFrom ; z++) {
			//assuming only one interface for now!!
			uint curSmetric = LinkCacheLookupMetric(VA, VA->Address, dp_sendFrom->hopList[z].addr, Dest->inif, Dest->inif); 
			if (curSmetric == 0) {
				PRINT_VA("Missing link value in LinkCacheFindForwarders!", *ForwNode);
				continue ;
			}
			if ((*VA->IsInfinite)(curSmetric) || 
				(*VA->ConvMetric) (curSmetric)  >= ETX_FORWARD_SELECTION_THRESH) {
				
				if(debug) KdPrint(("\t\tnode too far away, skip: %d\n", (*VA->ConvMetric) (curSmetric)));
				continue; 
			} 

			//this node is okay
			onList = 1;
			break ;
		}
		
		if (!onList) {
			if(debug) KdPrint(("\t\tI'm a node not on dp, and sendFrom too far\n"));
			return FALSE ;
		}
	}
if_done:

	//check to see if it makes progress towards destination-- if the 
	// node is the destination, then this returns 0.
	if (!VirtualAddressEqual(*ForwNode, Dest->addr)) {
	
		ret1 = arb_dijkstra(VA, *ForwNode, Dest->addr, forwETX, Now) ;
		if (ret1 == 0 || *forwETX == (uint) -1) {
			if(debug) KdPrint(("\t\tret1 is bad or forwETX is bad\n"));
			return FALSE ;
		}

	} else {	
		//hackish: we want to sort the tmps array later, but we also assume
		// that the metric field is nonzero. So let's assign it the least nonzero
		// value since the destination always has maximal progress and we want it
		// to be the first entry in our sorted array.
		*forwETX = 1 ;
	}

	//CACHE_ETX_PERCENTAGE
	// ensure that nodes make enough progress, not just some small amount
	//--if this is a default path node, then we'll want to include it

	if (DP_NODE) {
		//for default path nodes, we'll just ignore the cacheing percentage...
		if ((*forwETX) >= myETX) {
			if(debug) KdPrint(("\t\tfinal1: too far away: %d myETX: %d\n", *forwETX, myETX));
			return FALSE; 
		}
	} else {
		if ((*forwETX)*100 >= myETX*CACHE_ETX_PERCENTAGE) {
			if(debug) KdPrint(("\t\tfinal2: too far away: %d myETX: %d\n", *forwETX, myETX));
			return FALSE; 
		}
	}


	//all tests are passed, hooray!
	return TRUE; 
}

//* LinkCacheFindForwarders
//
// Gets the potential forwarding nodes (including default path nodes) for 
// the given packet. We will leave the final construction/pruning 
// to another method.
//
// This method will ensure that non-default path nodes use the 
// potential forwarding list from the closest default path (DP) node.
//
// Bitmap must be set on srp before calling.
//
// Input:
//	srp- the packet that we will send, srp->sr must be initialized already
//
// Returns: 
//  1. A source route that contains the forwarders. 
//		The route is sorted- highest priority node has lowest index in source route
//		The metric values in this sr are to the dest, so we don't want to use them
//           in final packet.
//   NULL upon error or if no forwarders.
// 2. retETX_FORWARD_SELECTION_THRESH: The value we used for ETX_FORWARD_SELECTION_THRESH
//
InternalSourceRoute*
LinkCacheFindForwarders(MiniportAdapter *VA, 
					  SRPacket* srp,
					  uint *retETX_FORWARD_SELECTION_THRESH) 
{
    LinkCache *LC = VA->LC;
    //KIRQL OldIrql;
	uint numNodes = 0;  //how many forwarding nodes that qualify.
	uint myETX;			//my ETX to the destination.
	uint forwETX;			//forwader's ETX to the destination
	uint i = 0 ; 
	uint qNodes = 0 ; //number of nodes we actually stick in the returned source route
	uint MAX_TMPS_SIZE = MAX_SR_LEN*2 ; //max size of tmps array
	SRAddr Dest; 
	InternalSourceRoute* real_tmps ; //tmps will be a pointer from this
	SourceRoute* tmps; //let's not put this on the kernel stack... tmp array to hold forwarders
	InternalSourceRoute* real_dp_sendFrom;
	SourceRoute* dp_sendFrom; //holds default path nodes that make progress and are in range
	uint num_dp_sendFrom = 0; //number of nodes in the dp_sendFrom list
	InternalSourceRoute* real_sret ; //sret will be a pointer from this
	SourceRoute* sret ; //the source route we'll return
	VirtualAddress *sendFrom ; //which node we base the forwarding list from
	uint ret1; //for checking return of arb_dijkstra
	Time Now = KeQueryInterruptTime();
	uint ETX_FORWARD_SELECTION_THRESH ; 
	
	//used to sort forwarders in returned array
	uint bestNode = 0 ;
	uint bestValue = 0; 
	uint dpIndex ;

	uint Index ;

	ASSERT(srp->sr != NULL); 
	ASSERT(srp->sr->opt.bitmap); 

	

	//get number of potential forwarders-- if the current source route
	// takes up all the space, then we're done
	if (MAX_SR_LEN == SOURCE_ROUTE_HOPS(srp->sr->opt.optDataLen)) {
		return NULL; 
	}

    Dest = srp->sr->opt.hopList[SOURCE_ROUTE_HOPS(srp->sr->opt.optDataLen)-1]; 
	
	real_dp_sendFrom = ExAllocatePool(NonPagedPool, 
		sizeof *real_dp_sendFrom + sizeof(SRAddr)*SOURCE_ROUTE_HOPS(srp->sr->opt.optDataLen)); 
	if (real_dp_sendFrom == NULL) {
		if(debug) KdPrint(("LinkCacheFindForwarders: Can't alloc real_dp_sendFrom\n"));
		return NULL ;
    }
	RtlZeroMemory(real_dp_sendFrom, sizeof *real_dp_sendFrom + sizeof(SRAddr)*SOURCE_ROUTE_HOPS(srp->sr->opt.optDataLen)); 
	dp_sendFrom = &(real_dp_sendFrom->opt);

	real_tmps = ExAllocatePool(NonPagedPool, sizeof *real_tmps + sizeof(SRAddr)*MAX_TMPS_SIZE); 
	if (real_tmps == NULL) {
		if(debug) KdPrint(("LinkCacheFindForwarders: Can't alloc real_tmps\n"));
		ExFreePool(real_dp_sendFrom);
		return NULL ;
    }	
	RtlZeroMemory(real_tmps, sizeof *real_tmps + sizeof(SRAddr)*MAX_TMPS_SIZE); 
	tmps = &(real_tmps->opt);

	//find all incoming ETX values to the destination, and try to select
	//the best ones.
	//KeAcquireSpinLock(&LC->Lock, &OldIrql);	

	
	//Find who we should be creating the forwarding
	// list with respect to. Conditions:
	// 1. Original sender node (bitmap is zero)
	// 2. Forwarder node
	//	   a. not on default path 
	//     b. on default path
	//
	if (onPath(srp, &(VA->Address))) {
		sendFrom = &(VA->Address); 
	} else {
		sendFrom = closestDefaultPathNode(VA, srp); 
	}
	if (sendFrom == NULL) {
		if(debug) KdPrint(("FFFFFF sendFrom null in linkcachefindforwarders FFFFFF\n")); 
		ExFreePool(real_tmps);
		ExFreePool(real_dp_sendFrom);
		//KeReleaseSpinLock(&LC->Lock, OldIrql);	
		return NULL; 
	}

	PRINT_VA("send from is: ", *sendFrom); 

	//assuming only one Physical adapter for now!
	ret1 = arb_dijkstra(VA, VA->Address, Dest.addr, &myETX, Now) ;
	if (ret1 == 0 ||  myETX == (uint) -1) {
		if(debug) KdPrint(("\nERROR LinkCacheFindForwarders, ETX = 0!\n")); 
		ExFreePool(real_tmps);
		ExFreePool(real_dp_sendFrom);
		//KeReleaseSpinLock(&LC->Lock, OldIrql);	
		return NULL;  
	}

	//need to get ETX_FORWARD_SELECTION_THRESH
	ETX_FORWARD_SELECTION_THRESH = get_ETX_fwd_thresh(VA, srp, sendFrom); 
	if(debug) KdPrint(("ETX_FORWARD_SELECTION_THRESH returned is: %u\n", ETX_FORWARD_SELECTION_THRESH)); 
	*retETX_FORWARD_SELECTION_THRESH = ETX_FORWARD_SELECTION_THRESH ;
	

	//look at all default path nodes first
	// if we aren't on the dp, then we want to put the possible forwarders
	// in the dp_sendFrom list
	Index = SOURCE_ROUTE_HOPS(srp->sr->opt.optDataLen) - srp->sr->opt.segmentsLeft; 
	ASSERT(VirtualAddressEqual(VA->Address, srp->sr->opt.hopList[Index-1].addr));
	for (i = Index ; i < SOURCE_ROUTE_HOPS(srp->sr->opt.optDataLen); i++) {
		uint curValue;
		//only want dp nodes
		if (!onPath(srp, &(srp->sr->opt.hopList[i].addr)))
			continue; 

		//only want nodes we can add to fwd list
		if (!addPotentialForwarder(VA, srp, myETX, &(srp->sr->opt.hopList[i].addr), sendFrom, 
			ETX_FORWARD_SELECTION_THRESH, &forwETX, &curValue, NULL, 0, Now, TRUE)) {
		
				continue; 
		}

		PRINT_VA("\t\tFFFFF adding dp node to fwd list: ", srp->sr->opt.hopList[i].addr);
		if(debug) KdPrint(("\t\t\twith forwetx to dest: %d, myetx to dest: %d and link from me: %d\n", forwETX, myETX, curValue));

		//add the node to our list	
		RtlCopyMemory(&(tmps->hopList[numNodes].addr), &(srp->sr->opt.hopList[i].addr), sizeof(VirtualAddress));
		tmps->hopList[numNodes].inif = Dest.inif;  //assume only 1 interface for now!!
		tmps->hopList[numNodes].outif = Dest.inif; 
		tmps->hopList[numNodes].Metric = forwETX;//not used as final metric in path, just so we can sort later

		
		//add the node to our dp_sendFrom list	
		RtlCopyMemory(&(dp_sendFrom->hopList[num_dp_sendFrom].addr), 
						&(srp->sr->opt.hopList[i].addr), sizeof(VirtualAddress));
		dp_sendFrom->hopList[num_dp_sendFrom].inif = Dest.inif;  //assume only 1 interface for now!!
		dp_sendFrom->hopList[num_dp_sendFrom].outif = Dest.inif; 
		dp_sendFrom->hopList[num_dp_sendFrom].Metric = forwETX;//not used as final metric in path, just so we can sort later
		num_dp_sendFrom++; 

		ASSERT(tmps->hopList[numNodes].Metric != 0); 
		numNodes++; 
		if (numNodes >= MAX_TMPS_SIZE)
			break ;
	}

	//if dp_sendFrom is empty, then we just give up
	// -ie: just try to  send to next hop
	if (num_dp_sendFrom == 0) {
		if(debug) KdPrint(("\nERROR LinkCacheFindForwarders, num_dp_sendFrom == 0\n")); 
		ExFreePool(real_tmps);
		ExFreePool(real_dp_sendFrom);
		return NULL;  
	}


	//Need to go thru all nodes in the topology. First try to prune the nodes
	// and then run dijkstra's algorithm if necessary.
	for (i = 0 ; numNodes < MAX_TMPS_SIZE && i < LC->nodeCount ; i++) {
		uint curValue; 

		//dp nodes already considered
		if (onPath(srp, &(LC->nodes[i].address)))
			continue; 

		if (!addPotentialForwarder(VA, srp, myETX, &(LC->nodes[i].address), sendFrom, 
			ETX_FORWARD_SELECTION_THRESH, &forwETX, &curValue, dp_sendFrom, num_dp_sendFrom, Now, FALSE)) {
		
				continue; 
		}

		PRINT_VA("\t\tFFFFF adding node to fwd list: ", LC->nodes[i].address);
		if(debug) KdPrint(("\t\t\twith forwetx to dest: %d, myetx to dest: %d and link from me: %d\n", forwETX, myETX, curValue));

		//add the node to our list	
		RtlCopyMemory(&(tmps->hopList[numNodes].addr), &(LC->nodes[i].address), sizeof(VirtualAddress));
		tmps->hopList[numNodes].inif = Dest.inif;  //assume only 1 interface for now!!
		tmps->hopList[numNodes].outif = Dest.inif; 
		tmps->hopList[numNodes].Metric = forwETX;//not used as final metric in path, just so we can sort later

		ASSERT(tmps->hopList[numNodes].Metric != 0); 
		numNodes++; 
		if (numNodes >= MAX_TMPS_SIZE)
			break ;

	}
	

	// We don't have any new forwarding nodes.  Delete the memory we allocated
	// and return. 
	if (numNodes == 0) {
		if(debug) KdPrint(("FFFFFF NumNodes zero, not forwarders anything FFFFFF\n")); 
		ExFreePool(real_tmps);
		ExFreePool(real_dp_sendFrom);
		//KeReleaseSpinLock(&LC->Lock, OldIrql);	
		return NULL; 
	}

	//sort our nodes
	qNodes = 0 ; //number of nodes we've put in the new 'sret' structure so far.

	//allocate the returned source route
	real_sret = ExAllocatePool(NonPagedPool, sizeof *real_sret + sizeof(SRAddr)*numNodes); 
	
	if (real_sret == NULL) {
		if(debug) KdPrint(("LinkCacheFindForwarders: Can't alloc sret\n"));
		ExFreePool(real_tmps);
		ExFreePool(real_dp_sendFrom);
		//KeReleaseSpinLock(&LC->Lock, OldIrql);	
		return NULL ;
    }	
	RtlZeroMemory(real_sret, sizeof *real_sret + sizeof(SRAddr)*numNodes); 
	sret = &(real_sret->opt); 

	//put the highest priority default path node into the array
	// first. Then add nodes by progress.
	bestNode = 0 ;
	bestValue = myETX;//all nodes in tmps have better ETX than me
	for (dpIndex = 0 ; dpIndex < numNodes ; dpIndex++) {
		if (tmps->hopList[dpIndex].Metric <= bestValue && onPath(srp, &(tmps->hopList[dpIndex].addr))) {
			bestNode = dpIndex;
			bestValue = tmps->hopList[dpIndex].Metric ;
		}
	}
	
	if (bestValue != myETX) {
		PRINT_VA("\t\tFFFFF first forwarder (highest dp): ", tmps->hopList[bestNode].addr);
		RtlCopyMemory(&(sret->hopList[qNodes]), &(tmps->hopList[bestNode]), sizeof(tmps->hopList[bestNode])); 
		qNodes++;		
		tmps->hopList[bestNode].Metric = 0 ; 
	}
	

	//we sort by progress... higher priority nodes go into array first.
	while (qNodes < numNodes) {
		//find best node in list
		uint k = 0 ;
		uint nullcnt = 0 ; 
		bestNode = 0 ;
		bestValue = myETX; //all nodes in tmps have better ETX than me
		for (k = 0 ; k < numNodes; k++) {
			//try to find best node
			if (tmps->hopList[k].Metric != 0 && tmps->hopList[k].Metric <= bestValue) {
				bestNode = k; 
				bestValue = tmps->hopList[k].Metric;
			} else if (tmps->hopList[k].Metric == 0) {
				nullcnt++; 
			}
		}
		//if(debug) KdPrint(("nullcnt: %d numNodes: %d \n", nullcnt, numNodes)); 
		if (nullcnt == numNodes)
			break; 
		ASSERT(bestValue != myETX); 
		//if(debug) KdPrint(("FFFFFF Copying new sr list entry FFFFFF\n")); 
		//add the node to our list
		RtlCopyMemory(&(sret->hopList[qNodes]), &(tmps->hopList[bestNode]), sizeof(tmps->hopList[bestNode])); 
		qNodes++;		
		tmps->hopList[bestNode].Metric = 0 ; 
	}
	ASSERT(qNodes == numNodes); //can get ride nullcnt above.
	//KeReleaseSpinLock(&LC->Lock, OldIrql);	

	sret->optDataLen = SOURCE_ROUTE_LEN(qNodes); 
	
	//if(debug) KdPrint(("Adding %d nodes to local recovery list\n",qNodes)); 
	ExFreePool(real_dp_sendFrom);
	ExFreePool(real_tmps);
	return real_sret; 
}





//* LinkCacheFillSR
//
//  Initializes SR with a route to the destination.
//  Always initializes SR->free to NULL.
//  The remaining SR fields are only modified if we succeed.
//
//  Return codes:
//      NDIS_STATUS_NO_ROUTE_TO_DESTINATION     No usable route.
//      NDIS_STATUS_RESOURCES                   Resource shortage.
//      NDIS_STATUS_SUCCESS
//
NDIS_STATUS
LinkCacheFillSR(
    MiniportAdapter *VA,
    const VirtualAddress Dest,
    InternalSourceRoute *SR)
{
    LinkCache *LC = VA->LC;
    SourceRoute *CachedSR;
    int dnode;
    uint NumHops;
    int curr;
    KIRQL OldIrql;
    Time Now;
    NDIS_STATUS ret;
	

	//ejr
    Link *Walk;
    int DNode;
	boolint isInf = 0 ;
	uint ETX_FORWARD_SELECTION_THRESH ;
	uint i ;
	uint fwd_max;

    //
    // We must update SR->next even when we fail, for SRPacketFree.
    //
    SR->next = NULL;

    KeAcquireSpinLock(&LC->Lock, &OldIrql);
    Now = KeQueryInterruptTime();

    dnode = findNode(LC, Dest, FALSE);

    //
    // We do not route to ourself.
    // Loopback is handled by layer 3.
    //
    if (dnode <= 0) {
        ret = NDIS_STATUS_NO_ROUTE_TO_DESTINATION;
        goto Return; // No route change log: CachedSR is not initialized.
    }

    //
    // Ensure that our route computation is current.
    // NB: This must be <= not < because
    // KeQueryInterruptTime might return the same value twice.
    //
    if (LC->timeout <= Now) {
        //
        // Recompute shortest path routes.
        //
        if (! dijkstra(VA, Now)) {
            ret = NDIS_STATUS_RESOURCES;
            goto Return; // No route change log: CachedSR is not initialized.
        }
    }

    //
    // Check for a cached route.
    //
    CachedSR = LC->nodes[dnode].SR;
    if (CachedSR != NULL) {
        //
        // We can use the cached route if it is static
        // or if its path metric is current.
        //
        if (CachedSR->staticRoute ||
            (LC->nodes[dnode].Metric != 0)) {

        UseCachedRoute:
            RtlCopyMemory(&SR->opt, CachedSR,
                          sizeof(LQSROption) + CachedSR->optDataLen);
            LinkCacheRouteUsage(LC, dnode, SR);
            ret = NDIS_STATUS_SUCCESS;
            goto Return; // No route change log in this case.
        }

        //
        // Calculate a current path metric
        // and check if we should still use this route.
        //
        LC->nodes[dnode].Metric =
            (*VA->PathMetric)(VA, LC->nodes[dnode].Hops,
                              SOURCE_ROUTE_HOPS(CachedSR->optDataLen) - 1);

        if (LinkCacheFlapDamp(LC->nodes[dnode].Metric,
                              LC->metric[dnode],
                              Now - LC->nodes[dnode].FirstUsage,
                              VA->RouteFlapDampingFactor)) {
            LC->CountRouteFlapDamp++;
            goto UseCachedRoute;
        }
        LC->CountRouteFlap++;
    }

    //
    // NB: The hop count will be (uint)-1 if there is no route.
    // We also reject routes that are too long.
    //
    if (LC->hops[dnode] >= MAX_SR_LEN) {
        //
        // We have no usable route to the destination.
        //
        LinkCacheRouteUsage(LC, dnode, NULL);
        if (CachedSR != NULL) {
            //
            // We have an existing source route.
            // Delete it and log the change.
            //
            LinkCacheReleaseLinks(LC->nodes[dnode].Hops,
                                  SOURCE_ROUTE_HOPS(CachedSR->optDataLen) - 1);
            ExFreePool(CachedSR);
            LC->nodes[dnode].SR = NULL;
            LC->nodes[dnode].Hops = NULL;
            LinkCacheRouteChange(LC, Dest,
                                 LC->nodes[dnode].Metric, (uint)-1,
                                 NULL, NULL);
            LC->nodes[dnode].RouteChanges++;
        }
        LC->nodes[dnode].Metric = (uint)-1;
        ret = NDIS_STATUS_NO_ROUTE_TO_DESTINATION;
        goto Return;
    }
    NumHops = LC->hops[dnode] + 1; // Add one to get number of nodes.

    SR->opt.optionType = LQSR_OPTION_TYPE_SOURCERT;
    SR->opt.optDataLen = SOURCE_ROUTE_LEN(NumHops);
    SR->opt.reservedField = 0;
    SR->opt.staticRoute = 0;
    SR->opt.salvageCount = 0;
    SR->opt.segmentsLeft = (uchar) (NumHops - 1);

    SR->opt.hopList[NumHops-1].outif = 0;
    SR->opt.hopList[0].inif = 0;

    curr = -1;
    do {
        if (curr == -1)
            curr = dnode;
        else
            curr = LC->prev[curr];
        
        ASSERT(curr != (uint)-1); 

        NumHops--;
        RtlCopyMemory(SR->opt.hopList[NumHops].addr, 
                      LC->nodes[curr].address, SR_ADDR_LEN);
        SR->opt.hopList[NumHops].Metric = 0;
        if (NumHops != 0) {
            SR->opt.hopList[NumHops].inif = LC->link[curr]->inif;
            SR->opt.hopList[NumHops-1].outif = LC->link[curr]->outif;
        }
    } while (curr);

    if (CachedSR != NULL) {
        //
        // We have an existing source route.
        // If this one is different, cache it and log it.
        //
        if ((CachedSR->optDataLen != SR->opt.optDataLen) ||
            ! RtlEqualMemory(CachedSR, &SR->opt,
                             sizeof(LQSROption) + SR->opt.optDataLen)) {

            LinkCacheReleaseLinks(LC->nodes[dnode].Hops,
                                  SOURCE_ROUTE_HOPS(CachedSR->optDataLen) - 1);
            // This also frees LC->nodes[dnode].Hops.
            ExFreePool(CachedSR);
            goto CacheAndLogSourceRoute;
        }
    }
    else {
        uint SRSize;
        Link **Hops;

        //
        // We do not have an existing source route.
        // We cache this new source route and log it.
        //
    CacheAndLogSourceRoute:
        NumHops = LC->hops[dnode]; // We want number of links, not nodes.
        SRSize = ROUND_UP_COUNT(sizeof(LQSROption) + SR->opt.optDataLen,
                                sizeof *Hops);
        CachedSR = ExAllocatePool(NonPagedPool,
                                  SRSize + NumHops * sizeof *Hops);
        if (CachedSR != NULL) {
            RtlCopyMemory(CachedSR, &SR->opt,
                          sizeof(LQSROption) + SR->opt.optDataLen);

            Hops = (Link **) ((uchar *)CachedSR + SRSize);
            LinkCacheRememberLinks(LC, dnode, Hops);
        }
        else
            Hops = NULL;

        LC->nodes[dnode].SR = CachedSR;
        LC->nodes[dnode].Hops = Hops;
        LC->nodes[dnode].FirstUsage = Now;
        LinkCacheRouteChange(LC, Dest,
                             LC->nodes[dnode].Metric, LC->metric[dnode],
                             &SR->opt, Hops);
        LC->nodes[dnode].Metric = LC->metric[dnode];
        LC->nodes[dnode].RouteChanges++;
    }

    LinkCacheRouteUsage(LC, dnode, SR);
    ret = NDIS_STATUS_SUCCESS;

Return:
	
	//  ejr
	//Fill in the forwarding list boundaries and ensure SR->opt.sender == 0
	//
	if (ret != NDIS_STATUS_SUCCESS)
		goto realReturn; 


	//set max value of forwarding list, the forwarding list is 
	//bounded by ETX_FORWARD_SELECTION_THRESH (if no such nodes 
	//exist, then we at least include the next hop)
	fwd_max = (ushort) 1 ;

	ETX_FORWARD_SELECTION_THRESH = get_ETX_fwd_thresh_SR(VA, &(SR->opt), &(VA->Address)); 

	//
	// Let's change this loop so it *does not* put in any more metric values.
	//
	while (fwd_max < SOURCE_ROUTE_HOPS(SR->opt.optDataLen)) {
		uint curMetric = 0 ;

		DNode = findNode(LC, SR->opt.hopList[fwd_max].addr, FALSE);
		Walk = NULL; 
		if (DNode > 0) {
			//if(debug) KdPrint(("Dnode nonzero\n")); 	

			//
			// Search our link cache for this link.
			//
			for (Walk = LC->nodes[0].AdjOut; Walk != NULL; Walk = Walk->NextOut) {
				if ((Walk->targetIndex == DNode) &&
				    (Walk->outif == SR->opt.hopList[0].outif) &&
				    (Walk->inif == SR->opt.hopList[fwd_max].inif)) {
				    //
				    // Found link (should match at most one).
				    //
					if ((*VA->IsInfinite)(Walk->Metric))
						isInf = 1 ;
					else
						curMetric = Walk->Metric;
					//if(debug) KdPrint(("\t\t\tMetric found: %d\n", (*VA->ConvMetric)(Walk->Metric))); 
				    break;
				}//end if link equal
			}//end for all links
		} else {
			if(debug) KdPrint(("Dnode zero\n")); 	
		}//end if dnode > 0 


		if (DNode <= 0 || Walk == NULL || isInf ||
			 (*VA->ConvMetric)(curMetric) >= ETX_FORWARD_SELECTION_THRESH) 
			{
				if (isInf)
					if(debug) KdPrint(("Infinite metric\n")); 
				break; 
			}

		fwd_max++; 
	}//end while	

	//in the loop above, we always advance 1 too many.
	fwd_max--;
	if (fwd_max == 0)
		fwd_max++; //need to have max be 1 at very least.
	
	//set all fwd list bitmaps
	SR->opt.fwd_bitmap = 0 ; //reset
	for (i = 1 ; i <= fwd_max ; i++) {
		setSRPbitmap(&(SR->opt.fwd_bitmap), i);		
	}
	//fill dp
	SR->opt.bitmap = (ushort) -1; 
realReturn:

    KeReleaseSpinLock(&LC->Lock, OldIrql);
    return ret;
}




//* LinkCacheFillACKSR
//
//  Initializes SR with a route to the destination.
//	Is at most 1 hop (two nodes): Source and Dest
//  Always initializes SR->free to NULL.
//
//
//  Return codes:
//      NDIS_STATUS_RESOURCES                   Resource shortage.
//      NDIS_STATUS_SUCCESS
//
NDIS_STATUS
LinkCacheFillACKSR(
    MiniportAdapter *VA,
    InternalSourceRoute *SR,
	VirtualAddress *Source,
	VirtualAddress *Dest)
{
	uint NumHops = 2 ;   
	LinkCache *LC = VA->LC;
	Link *Walk ; 
	int DNode ;
    KIRQL OldIrql;
	NDIS_STATUS ret ; 
	uint fwd_max;
	
	KeAcquireSpinLock(&LC->Lock, &OldIrql);
    DNode = findNode(LC, *Dest, FALSE);

    if (DNode <= 0) {
        ret = NDIS_STATUS_NO_ROUTE_TO_DESTINATION;
		KeReleaseSpinLock(&LC->Lock, OldIrql);
        return ret ;; // No route change log: CachedSR is not initialized.
    }


	
    SR->opt.optionType = LQSR_OPTION_TYPE_SOURCERT;
    SR->opt.optDataLen = SOURCE_ROUTE_LEN(NumHops);
    SR->opt.reservedField = 0;
    SR->opt.staticRoute = 1;
    SR->opt.salvageCount = 0;
    SR->opt.segmentsLeft = (uchar) (NumHops - 1);
	fwd_max = (ushort) 1 ;
	
    RtlCopyMemory(SR->opt.hopList[0].addr, *Source, SR_ADDR_LEN);
    RtlCopyMemory(SR->opt.hopList[1].addr, *Dest, SR_ADDR_LEN);
	
    SR->opt.hopList[NumHops-1].outif = 0;
    SR->opt.hopList[0].inif = 0;

	//update the fwd_list bitmap and dp bitmap
	SR->opt.bitmap = (ushort) -1 ;
	setSRPbitmap(&(SR->opt.fwd_bitmap), fwd_max); 

	//we assume only 1 interface for now!!!
	SR->opt.hopList[NumHops-1].inif = 1 ;
	SR->opt.hopList[0].outif = 1 ;

	//
	// Search our link cache for this link.
	//
	for (Walk = LC->nodes[0].AdjOut; Walk != NULL; Walk = Walk->NextOut) {
		if ((Walk->targetIndex == DNode) &&
		    (Walk->outif == SR->opt.hopList[0].outif) &&
		    (Walk->inif == SR->opt.hopList[1].inif)) {
		    //
		    // Found link (should match at most one).
		    //
			if (! (*VA->IsInfinite)(Walk->Metric))
				SR->opt.hopList[1].Metric = Walk->Metric;
		    break;
		}//end if link equal
	}//end for all links

   
    ret = NDIS_STATUS_SUCCESS;
    KeReleaseSpinLock(&LC->Lock, OldIrql);
	return ret; 
}






//* LinkCacheGetSR
//
//  Initializes SR with a route to the destination,
//  including the link metrics. This is for IoQuerySourceRoute only,
//  not for actually sending packets.
//
//  Return codes:
//      NDIS_STATUS_NO_ROUTE_TO_DESTINATION     No usable route.
//      NDIS_STATUS_RESOURCES                   Resource shortage.
//      NDIS_STATUS_SUCCESS
//
NDIS_STATUS
LinkCacheGetSR(
    MiniportAdapter *VA,
    const VirtualAddress Dest,
    InternalSourceRoute *SR)
{
    LinkCache *LC = VA->LC;
    SourceRoute *CachedSR;
    int dnode;
    uint NumHops;
    int curr;
    KIRQL OldIrql;
    Time Now;
    NDIS_STATUS ret;

    KeAcquireSpinLock(&LC->Lock, &OldIrql);
    Now = KeQueryInterruptTime();

    dnode = findNode(LC, Dest, FALSE);

    //
    // We do not route to ourself.
    // Loopback is handled by layer 3.
    //
    if (dnode <= 0) {
        ret = NDIS_STATUS_NO_ROUTE_TO_DESTINATION;
        goto Return;
    }

    //
    // Ensure that our route computation is current.
    // NB: This must be <= not < because
    // KeQueryInterruptTime might return the same value twice.
    //
    if (LC->timeout <= Now) {
        //
        // Recompute shortest path routes.
        //
        if (! dijkstra(VA, Now)) {
            ret = NDIS_STATUS_RESOURCES;
            goto Return;
        }
    }

    //
    // Check for a cached route.
    //
    CachedSR = LC->nodes[dnode].SR;
    if (CachedSR != NULL) {
        //
        // We can use the cached route if it is static
        // or if its path metric is current.
        //
        if (CachedSR->staticRoute ||
            (LC->nodes[dnode].Metric != 0)) {

            RtlCopyMemory(&SR->opt, CachedSR,
                          sizeof(LQSROption) + CachedSR->optDataLen);

            if (LC->nodes[dnode].Hops != NULL) {
                //
                // Get the metric values for the cached route.
                //
                NumHops = SOURCE_ROUTE_HOPS(CachedSR->optDataLen);
                for (curr = 1; curr < (int)NumHops; curr++)
                    SR->opt.hopList[curr].Metric =
                        LC->nodes[dnode].Hops[NumHops - curr - 1]->Metric;
            }

            ret = NDIS_STATUS_SUCCESS;
            goto Return;
        }
    }

    //
    // NB: The hop count will be (uint)-1 if there is no route.
    // We also reject routes that are too long.
    //
    if (LC->hops[dnode] >= MAX_SR_LEN) {
        //
        // We have no usable route to the destination.
        //
        ret = NDIS_STATUS_NO_ROUTE_TO_DESTINATION;
        goto Return;
    }

    NumHops = LC->hops[dnode] + 1; // Add one to get number of nodes.

    SR->opt.optionType = LQSR_OPTION_TYPE_SOURCERT;
    SR->opt.optDataLen = SOURCE_ROUTE_LEN(NumHops);
    SR->opt.reservedField = 0;
    SR->opt.staticRoute = 0;
    SR->opt.salvageCount = 0;
    SR->opt.segmentsLeft = (uchar) (NumHops - 1);

    SR->opt.hopList[NumHops-1].outif = 0;
    SR->opt.hopList[0].inif = 0;

    curr = -1;
    do {
        if (curr == -1)
            curr = dnode;
        else
            curr = LC->prev[curr];

        NumHops--;
        RtlCopyMemory(SR->opt.hopList[NumHops].addr, 
                      LC->nodes[curr].address, SR_ADDR_LEN);
        if (NumHops != 0) {
            SR->opt.hopList[NumHops].Metric = LC->link[curr]->Metric;
            SR->opt.hopList[NumHops].inif = LC->link[curr]->inif;
            SR->opt.hopList[NumHops-1].outif = LC->link[curr]->outif;
        }
    } while (curr);
    ret = NDIS_STATUS_SUCCESS;

Return:
    KeReleaseSpinLock(&LC->Lock, OldIrql);
    return ret;
}

//* LinkCacheAddRoute
//
//  Adds a static route to the destination.
//
NDIS_STATUS
LinkCacheAddRoute(
    MiniportAdapter *VA,
    const VirtualAddress Dest,
    SourceRoute *SR)
{
    LinkCache *LC = VA->LC;
    KIRQL OldIrql;
    NDIS_STATUS Status;
    int dnode;

    ASSERT(SR->optionType == LQSR_OPTION_TYPE_SOURCERT);
    ASSERT(SR->staticRoute);
    ASSERT(SOURCE_ROUTE_HOPS(SR->optDataLen) <= MAX_SR_LEN);

    KeAcquireSpinLock(&LC->Lock, &OldIrql);
    dnode = findNode(LC, Dest, TRUE);
    if (dnode < 0) {
        //
        // Could not expand the link cache.
        //
        Status = NDIS_STATUS_RESOURCES;
    }
    else if (dnode == 0) {
        //
        // Can not specify a route to ourself.
        //
        Status = NDIS_STATUS_INVALID_ADDRESS;
    }
    else {
        CacheNode *node = &LC->nodes[dnode];

        //
        // Insert this source route.
        //
        if (node->SR != NULL) {
            if (node->Hops != NULL) {
                LinkCacheReleaseLinks(node->Hops,
                                      SOURCE_ROUTE_HOPS(node->SR->optDataLen) - 1);
                node->Hops = NULL;
            }
            // This also frees the old node->Hops.
            ExFreePool(node->SR);
        }
        node->SR = SR;
#if CHANGELOGS
        node->CurrentRU = NULL;
#endif
        Status = NDIS_STATUS_SUCCESS;

        //
        // Log the change.
        //
        LinkCacheRouteChange(LC, Dest, node->Metric, 0, SR, NULL);
        node->Metric = 0;
        node->RouteChanges++;
    }
    KeReleaseSpinLock(&LC->Lock, OldIrql);

    return Status;
}

//* LinkCacheCreateLI
//
//  Creates a LinkInfo option that represents the links from this node.
//  The option size can not exceed the specified MaxSize.
//
InternalLinkInfo *
LinkCacheCreateLI(
    MiniportAdapter *VA,
    uint MaxSize)
{
    LinkCache *LC = VA->LC;
    InternalLinkInfo *LI = NULL;
    Link *Link;
    uint NumLinks;
    KIRQL OldIrql;

    ASSERT(VA->MetricType != METRIC_TYPE_HOP);

    KeAcquireSpinLock(&LC->Lock, &OldIrql);

    //
    // Count the number of links from this node.
    // Stop counting if we'd exceed the amount of data that can be
    // represented in a linkinfo option.
    //
    NumLinks = 0;
    for (Link = LC->nodes[0].AdjOut; Link != NULL; Link = Link->NextOut) {
        if (LINKINFO_HOPS(LINKINFO_LEN(NumLinks + 1)) != NumLinks + 1) {
            //
            // We can not return all the links in one option.
            //
            InterlockedIncrement((PLONG)&VA->CountLinkInfoTooManyLinks);
            break;
        }
        NumLinks++;
    }

    //
    // Don't create a linkinfo option if we don't have any links.
    //
    if (NumLinks == 0)
        goto ErrorReturn;

    //
    // Is there room for all of these links in the packet?
    //
    if (sizeof(LQSROption) + LINKINFO_LEN(NumLinks) > MaxSize) {
        //
        // No, we can not return a complete option.
        //
        InterlockedIncrement((PLONG)&VA->CountLinkInfoTooLarge);

        //
        // Give up now if there isn't room for even one link.
        //
        if (MaxSize < sizeof(LQSROption) + LINKINFO_LEN(1))
            goto ErrorReturn;

        //
        // How many links will fit?
        //
        NumLinks = LINKINFO_HOPS(MaxSize - sizeof(LQSROption));
    }

    //
    // Allocate the Link Info option.
    //
    LI = ExAllocatePool(NonPagedPool, sizeof *LI + NumLinks * sizeof(SRAddr));
    if (LI != NULL) {
        SRAddr *SRA = LI->Opt.Links;

        //
        // Initialize the Link Info option.
        //
        LI->Next = NULL;
        LI->Opt.optionType = LQSR_OPTION_TYPE_LINKINFO;
        LI->Opt.optDataLen = LINKINFO_LEN(NumLinks);
        RtlCopyMemory(LI->Opt.From, LC->nodes[0].address, SR_ADDR_LEN);

        //
        // Traverse the links again, adding them to the option.
        // If they don't all fit, we just stop early.
        //
        for (Link = LC->nodes[0].AdjOut; NumLinks != 0; Link = Link->NextOut) {
            RtlCopyMemory(SRA->addr,
                      LC->nodes[Link->targetIndex].address, SR_ADDR_LEN);
            SRA->inif = Link->inif;
            SRA->outif = Link->outif;
            SRA->Metric = Link->Metric;

            NumLinks--;
            SRA++;
        }
    }

  ErrorReturn:
    KeReleaseSpinLock(&LC->Lock, OldIrql);
    return LI;
}

//* LinkCacheDeleteInterface
//
//  Deletes all links that use the interface.
//
void
LinkCacheDeleteInterface(
    MiniportAdapter *VA,
    LQSRIf IF)
{
    LinkCache *LC = VA->LC;
    Link **Prev, *This;
    uint i;
    KIRQL OldIrql;

    KeAcquireSpinLock(&LC->Lock, &OldIrql);
    LC->timeout = KeQueryInterruptTime(); // Invalidate dijkstra.
	delete_arb_dijkstra_caches(VA);
	

    //
    // Delete links from this node.
    // Also resets the links' reference count.
    //
    Prev = &LC->nodes[0].AdjOut;
    while ((This = *Prev) != NULL) {
        //
        // Delete this link if originates from the interface.
        //
        if (This->outif == IF) {
            LinkCacheLinkChange(LC,
                                LC->nodes[0].address,
                                LC->nodes[This->targetIndex].address,
                                This->inif, This->outif,
                                (uint)-1,
                                LINK_STATE_CHANGE_DELETE_INTERFACE);

            *Prev = This->NextOut;
            if (This->NextOut != NULL)
                This->NextOut->PrevOut = Prev;

            *This->PrevIn = This->NextIn;
            if (This->NextIn != NULL)
                This->NextIn->PrevIn = This->PrevIn;

            ExFreePool(This);
        }
        else {
            This->RefCnt = 0;
            Prev = &This->NextOut;
        }
    }

    //
    // Delete links to this node.
    // Also resets the links' reference count.
    //
    for (i = 1; i < LC->nodeCount; i++) {
        CacheNode *node = &LC->nodes[i];

        Prev = &node->AdjOut;
        while ((This = *Prev) != NULL) {
            //
            // Delete this link if terminates at the interface.
            //
            if (This->inif == IF) {
                LinkCacheLinkChange(LC,
                                    node->address,
                                    LC->nodes[This->targetIndex].address,
                                    This->inif, This->outif,
                                    (uint)-1,
                                    LINK_STATE_CHANGE_DELETE_INTERFACE);

                *Prev = This->NextOut;
                if (This->NextOut != NULL)
                    This->NextOut->PrevOut = Prev;

                *This->PrevIn = This->NextIn;
                if (This->NextIn != NULL)
                    This->NextIn->PrevIn = This->PrevIn;

                ExFreePool(This);
            }
            else {
                This->RefCnt = 0;
                Prev = &This->NextOut;
            }
        }
    }

    //
    // Delete all cached routes.
    // NB: We already reset the link reference counts,
    // so LinkCacheReleaseLinks is not needed.
    //
    for (i = 1; i < LC->nodeCount; i++) {
        CacheNode *node = &LC->nodes[i];

        if (node->Hops != NULL) {
            ASSERT((node->SR != NULL) && ! node->SR->staticRoute);
            ExFreePool(node->SR); // This also frees node->Hops.
            node->Hops = NULL;
            node->SR = NULL;
        }
    }

    KeReleaseSpinLock(&LC->Lock, OldIrql);
}

//* LinkCacheAddLink
//
//  Add a link to the link cache, or if the link already exists,
//  updates the metric.
//
//  An infinite Metric value means the link is actually non-functional.
//  We keep state for non-working links, if they were recently working.
//
void
LinkCacheAddLink(
    MiniportAdapter *VA,
    const VirtualAddress From,
    const VirtualAddress To,
    LQSRIf InIf,
    LQSRIf OutIf,
    uint Metric,
    uint Reason)
{
    LinkCache *LC = VA->LC;
    uint OldMetric;
    uint NewMetric;
    uint MetricDiff;
    int SNode, DNode;
    Link *Link;
    Time Now;
    KIRQL OldIrql;

    if (! VA->Snooping &&
        (Reason != LINK_STATE_CHANGE_ADD_REPLY))
        return;

	Now = KeQueryInterruptTime();
#if STOP_LINK_METRIC
			//ejr- want to stop updating the link weights
			// so do nothing here.
			if (Now > VA->start_link_time + (Time) (LINK_MEASURE_TIME * SECOND))
			{
				return; 
			}	
#endif
    //
    // First find the source and destination nodes in the cache.
    //
    KeAcquireSpinLock(&LC->Lock, &OldIrql);
    if (((SNode = findNode(LC, From, TRUE)) >= 0) &&
        ((DNode = findNode(LC, To, TRUE)) >= 0)) {
        Now = KeQueryInterruptTime();

        //
        // Search our link cache to see if we already have this link.
        //
        for (Link = LC->nodes[SNode].AdjOut;
             Link != NULL;
             Link = Link->NextOut) {
            if ((Link->targetIndex == DNode) &&
                (Link->outif == OutIf) &&
                (Link->inif == InIf)) {

                OldMetric = (*VA->ConvMetric)(Link->Metric);
                goto UpdateExisting;
			} else if (Link->targetIndex == DNode) {
				if(debug) KdPrint(("EEE Added redundant link in LinkCacheAddLink b/c inifs are wrong!!!!!!!!!!!!!!!!!!!!!!!\n"));
			}
        }

        //
        // Create a cache entry for this new link.
        //

        Link = ExAllocatePool(NonPagedPool, sizeof *Link);
        if (Link != NULL) {

            Link->RefCnt = 0;

            Link->sourceIndex = SNode;
            Link->targetIndex = DNode;
            Link->outif = OutIf;
            Link->inif = InIf;

            Link->Usage = 0;
            Link->Failures = 0;
            Link->DropRatio = 0;
            Link->ArtificialDrops = 0;
            Link->QueueDrops = 0;
			//if(debug) KdPrint(("EEE Adding from add\n"));
            (*VA->InitLinkMetric)(VA, SNode, Link, Now);
            Link->TimeStamp = Now;

            Link->PrevOut = &LC->nodes[SNode].AdjOut;
            Link->NextOut = LC->nodes[SNode].AdjOut;
            if (Link->NextOut != NULL)
                Link->NextOut->PrevOut = &Link->NextOut;
            LC->nodes[SNode].AdjOut = Link;

            Link->PrevIn = &LC->nodes[DNode].AdjIn;
            Link->NextIn = LC->nodes[DNode].AdjIn;
            if (Link->NextIn != NULL)
                Link->NextIn->PrevIn = &Link->NextIn;
            LC->nodes[DNode].AdjIn = Link;

            //
            // The link previously did not exist.
            //
            OldMetric = (uint)-1;

        UpdateExisting:
            if (Metric != 0) {
                if ((SNode == 0) && (VA->MetricType != METRIC_TYPE_HOP)) {
                    //
                    // This link is directly adjacent. We are directly
                    // monitoring the quality of this link, via probing
                    // and LinkCachePenalizeLink. Any other information
                    // will be stale so ignore it.
                    //
                }
                else {
                    //
                    // We have metric information, so update the metric.
                    //
                    Link->Metric = Metric;
                }
            }
            else {
                //
                // We do not have a metric, but we know the link exists.
                // If we previously believed the link did not exist,
                // then reinitialize to a default metric.
                //
				if ((*VA->IsInfinite)(Link->Metric)) {
					//if(debug) KdPrint(("EEE Adding from infinite\n"));
					(*VA->InitLinkMetric)(VA, SNode, Link, Now);
				}
            }

            //
            // This is a quick heuristic check of whether this
            // change is likely to affect dijkstra.
            // We assume the metric change affects dijskstra
            // if the change is large enough to be meaningful AND
            // this link is likely to be part of either the old
            // or new shortest-path tree.
            // NB: It is critical that for the HOP metric,
            // a change from -1 to 1 or vice-versa invalidates dijkstra.
            //
            NewMetric = (*VA->ConvMetric)(Link->Metric);
			if (OldMetric < NewMetric) {
                MetricDiff = NewMetric - OldMetric;
			} else {
                MetricDiff = OldMetric - NewMetric;
				/*
				if (MetricDiff > 4096) {
					if(debug) KdPrint(("EEE Major link difference in updateLinks. Link from %02x-%02x-%02x-%02x-%02x-%02x to %02x-%02x-%02x-%02x-%02x-%02x NewMetric: %d OldMetric: %d\n", (From)[0], 
							(From)[1], (From)[2], (From)[3], (From)[4], (From)[5], 
							(To)[0], (To)[1], (To)[2], (To)[3], (To)[4], (To)[5], NewMetric, OldMetric));

				}
				*/ 
			}

            if ((MetricDiff > LC->SmallestMetric) &&
                ((NewMetric <= LC->LargestMetric) ||
                 (OldMetric <= LC->LargestMetric))) {

                LC->CountAddLinkInvalidate++;

                //
                // Invalidate the dijkstra computation, because
                // we've significantly changed the metric of a link.
                //
                LC->timeout = Now;
				//delete_arb_dijkstra_caches(VA);

                //
                // Record the change.
                //
                LinkCacheLinkChange(LC,
                                    From, To,
                                    InIf, OutIf,
                                    Metric, Reason);
            }
            else {
                //
                // This was not a significant change.
                // However the dijkstra has a max timeout
                // of CACHE_TIMEOUT (one second) so
                // it will run again fairly soon anyway.
                //
                LC->CountAddLinkInsignificant++;
            }

            //
            // If Metric is not infinite (so the link exists)
            // update the TimeStamp to keep this link structure in existence.
            //
            if (! (*VA->IsInfinite)(Link->Metric))
                Link->TimeStamp = Now;
        }
    }
    KeReleaseSpinLock(&LC->Lock, OldIrql);
}


//* LinkCacheControlLink
//
//  Adjust link attributes.
//
NDIS_STATUS
LinkCacheControlLink(
    MiniportAdapter *VA,
    const VirtualAddress From,
    const VirtualAddress To,
    LQSRIf InIf,
    LQSRIf OutIf,
    uint DropRatio)
{
    LinkCache *LC = VA->LC;
    int SNode, DNode;
    Link *Link;
    NDIS_STATUS Status = STATUS_INVALID_PARAMETER;
    KIRQL OldIrql;

    //
    // First find the source and destination nodes in the cache.
    //
    KeAcquireSpinLock(&LC->Lock, &OldIrql);
    if (((SNode = findNode(LC, From, TRUE)) >= 0) &&
        ((DNode = findNode(LC, To, TRUE)) >= 0)) {

        //
        // Search our link cache for this link.
        //
        for (Link = LC->nodes[SNode].AdjOut;
             Link != NULL;
             Link = Link->NextOut) {
            if ((Link->targetIndex == DNode) &&
                (Link->outif == OutIf) &&
                (Link->inif == InIf)) {

                //
                // Found the link.
                //
                Link->DropRatio = DropRatio;
                Status = STATUS_SUCCESS;
                break;
            }
        }
    }
    KeReleaseSpinLock(&LC->Lock, OldIrql);
    return Status;
}

//* LinkCachePenalizeLink
//
//  Penalizes a link's metric because the link failed to send a packet.
//
void
LinkCachePenalizeLink(
    MiniportAdapter *VA,
    const VirtualAddress To,
    LQSRIf InIf,
    LQSRIf OutIf)
{
    LinkCache *LC = VA->LC;
    int DNode;
    Link *Adj;
    KIRQL OldIrql;
	

#if STOP_LINK_METRIC
	Time Now = KeQueryInterruptTime();

	if (Now > VA->start_link_time + (Time) (LINK_MEASURE_TIME * SECOND))
	{
		return;
	}
#endif


    //
    // First find the destination node in the cache.
    //
    KeAcquireSpinLock(&LC->Lock, &OldIrql);
    if ((DNode = findNode(LC, To, FALSE)) >= 0) {

        //
        // Next find the link from us to the destination.
        // It might not exist.
        //
        for (Adj = LC->nodes[0].AdjOut; Adj != NULL; Adj = Adj->NextOut) {
            if ((Adj->targetIndex == DNode) &&
                (Adj->outif == OutIf) &&
                (Adj->inif == InIf)) {

                //
                // Penalize the link.
                //
                if (VA->Penalize != NULL)
                    (*VA->Penalize)(VA, Adj);

                //
                // Log the penalty to this link.
                //
                Adj->Failures++;
                LinkCacheLinkChange(LC,
                                    VA->Address, To,
                                    InIf, OutIf,
                                    (uint)-1,
                                    LINK_STATE_CHANGE_PENALIZED);

                //
                // Ensure that dijkstra runs again immediately
                // to recalculate routes.
                //
                LC->timeout = KeQueryInterruptTime();
				//delete_arb_dijkstra_caches(VA);
                break;
            }
        }
    }
    KeReleaseSpinLock(&LC->Lock, OldIrql);
}


//* LinkCacheCountLinkUse
//
//  Increments the Usage value on a given link.
//
void
LinkCacheCountLinkUse(
    MiniportAdapter *VA,
    const VirtualAddress From,
    const VirtualAddress To,
    LQSRIf InIf,
    LQSRIf OutIf)
{
    LinkCache *LC = VA->LC;
    int SNode, DNode;
    Link *Link;
    KIRQL OldIrql;

    //
    // Find indices of source and destination 
    // in our link cache. 
    //
    KeAcquireSpinLock(&LC->Lock, &OldIrql);
    if (((SNode = findNode(LC, From, FALSE)) >= 0) &&
        ((DNode = findNode(LC, To, FALSE)) >= 0)) {

        //
        // Search our link cache for this link.
        //
        for (Link = LC->nodes[SNode].AdjOut; Link != NULL; Link = Link->NextOut) {
            if ((Link->targetIndex == DNode) &&
                (Link->outif == OutIf) &&
                (Link->inif == InIf)) {

                //
                // Found link (should match at most one).
                // Increment usage count.
                //
                Link->Usage++;
                break;
			} else if (Link->targetIndex == DNode) {
				if(debug) KdPrint(("EEE Missed LinkCacheLinkUse b/c inif's are wrong!!!!!!!!!!\n")); 
			}
        }
    }

    KeReleaseSpinLock(&LC->Lock, OldIrql);
}


//* LinkCacheCheckForDrop
//
//  See whether we should artificially drop a packet traversing this link.
//
boolint
LinkCacheCheckForDrop(
    MiniportAdapter *VA,
    const VirtualAddress From,
    const VirtualAddress To,
    LQSRIf InIf,
    LQSRIf OutIf)
{
    LinkCache *LC = VA->LC;
    int SNode, DNode;
    Link *Link;
    boolint Drop = FALSE;
    KIRQL OldIrql;

    //
    // Find indices of source and destination 
    // in our link cache. 
    //
    KeAcquireSpinLock(&LC->Lock, &OldIrql);
    if (((SNode = findNode(LC, From, FALSE)) >= 0) &&
        ((DNode = findNode(LC, To, FALSE)) >= 0)) {

        //
        // Search our link cache for this link.
        //
        for (Link = LC->nodes[SNode].AdjOut; Link != NULL; Link = Link->NextOut) {
            if ((Link->targetIndex == DNode) &&
                (Link->outif == OutIf) &&
                (Link->inif == InIf)) {
                uint Random;

                //
                // Found link (should match at most one).
                // Check against random drop ratio.
                //
                if (Link->DropRatio) {
					//if(debug) KdPrint(("drop ratio: %u\n", Link->DropRatio));
                    GetRandom((uchar *)&Random, sizeof Random);
                    if (Link->DropRatio > Random) {
						Drop = TRUE;
                        Link->ArtificialDrops++;
                    }
                }
                break;
            }
        }
    }

    KeReleaseSpinLock(&LC->Lock, OldIrql);

    return Drop;
}





//* LinkCacheCheckForDrop
//
//  See whether we should artificially drop a packet traversing this link.
//
boolint
LinkCacheCheckProbeForDrop(
    MiniportAdapter *VA,
    const VirtualAddress From,
    const VirtualAddress To,
    LQSRIf InIf,
    LQSRIf OutIf,
	uint seqNum)
{
    LinkCache *LC = VA->LC;
    int SNode, DNode;
    Link *Link;
    boolint Drop = FALSE;
    KIRQL OldIrql;

    UNREFERENCED_PARAMETER(seqNum);


    //
    // Find indices of source and destination 
    // in our link cache. 
    //
    KeAcquireSpinLock(&LC->Lock, &OldIrql);
    if (((SNode = findNode(LC, From, FALSE)) >= 0) &&
        ((DNode = findNode(LC, To, FALSE)) >= 0)) {

        //
        // Search our link cache for this link.
        //
        for (Link = LC->nodes[SNode].AdjOut; Link != NULL; Link = Link->NextOut) {
            if ((Link->targetIndex == DNode) &&
                (Link->outif == OutIf) &&
                (Link->inif == InIf)) {
                uint Random;
                Random = 0 ;
				
				//
                // Found link (should match at most one).
                // Check against random drop ratio.
                //
				
                if (Link->DropRatio) {
					//This value can be changed
					if (seqNum % 3 == 0) {
						Drop = TRUE;
						Link->ArtificialDrops++;
					}
				
					break;
				}
			}	
		}
	}

    KeReleaseSpinLock(&LC->Lock, OldIrql);

    return Drop;
}




//* LinkCacheLookupMetric
//
//  Returns the current metric value for the specified link,
//  or zero if we do not have that information.
//
uint
LinkCacheLookupMetric(
    MiniportAdapter *VA,
    const VirtualAddress From,
    const VirtualAddress To,
    LQSRIf InIf,
    LQSRIf OutIf)
{
    LinkCache *LC = VA->LC;
    int SNode, DNode;
    Link *Link;
    KIRQL OldIrql;
    uint Metric = 0;

    //
    // Find indices of source and destination 
    // in our link cache. 
    //
    KeAcquireSpinLock(&LC->Lock, &OldIrql);
    if (((SNode = findNode(LC, From, FALSE)) >= 0) &&
        ((DNode = findNode(LC, To, FALSE)) >= 0)) {

        //
        // Search our link cache for this link.
        //
        for (Link = LC->nodes[SNode].AdjOut; Link != NULL; Link = Link->NextOut) {
            if ((Link->targetIndex == DNode) &&
                (Link->outif == OutIf) &&
                (Link->inif == InIf)) {

                //
                // Found link (should match at most one).
                //
                Metric = Link->Metric;
                break;
				} else if (Link->targetIndex == DNode) {
					if(debug) KdPrint(("EEE Missed link in LinkCacheLookupMetric b/c inif's wrong!!!!!!!!!!!!!!!!!!11\n"));
				}
        }
    }

    KeReleaseSpinLock(&LC->Lock, OldIrql);
    return Metric;
}


//////////////////////////////////////////////////////////////////////////////
//                 HELPER FUNCTIONS FOR LINKCACHE FUNCTIONS                 //
//////////////////////////////////////////////////////////////////////////////

//* findNode
//
//  Find a node index in the link cache.
//  Returns the index if found, or -1 if not.
//  If Force, attempts to create a new node.
//
//  Must be called while holding LC->Lock.
//
int
findNode(
    LinkCache *LC,
    const VirtualAddress addr,
    boolint Force)
{
    uint i;

    for (i = 0; i < LC->nodeCount; i++)
        if (VirtualAddressEqual(LC->nodes[i].address, addr))
            return i;

    if (! Force)
        return -1;

    if (i == LC->maxSize) {
        uint maxSize;
        uint *metric;
        uint *hops;
        Link **link;
        uint *prev;
        CacheNode *nodes;
        uint i;

        //
        // Calculate a new size for the link cache.
        // Growing by a fixed percentage means we can
        // reach steady-state in log time instead of linear time.
        //
        maxSize = (LC->maxSize * 5) / 4;
        maxSize = max(maxSize, LC->maxSize + 1);

        //
        // Allocate larger arrays.
        //
        metric = ExAllocatePool(NonPagedPool, sizeof *metric * maxSize);
        if (metric == NULL)
            return -1;
        hops = ExAllocatePool(NonPagedPool, sizeof *hops * maxSize);
        if (hops == NULL)
            return -1;
        link = ExAllocatePool(NonPagedPool, sizeof *link * maxSize);
        if (link == NULL)
            return -1;
        prev = ExAllocatePool(NonPagedPool, sizeof *prev * maxSize);
        if (prev == NULL)
            return -1;
        nodes = ExAllocatePool(NonPagedPool, sizeof *nodes * maxSize);
        if (nodes == NULL)
            return -1;

        //
        // Copy to the new arrays.
        //
        RtlCopyMemory(metric, LC->metric, sizeof *metric * LC->maxSize);
        RtlCopyMemory(hops, LC->hops, sizeof *hops * LC->maxSize);
        RtlCopyMemory(link, LC->link, sizeof *link * LC->maxSize);
        RtlCopyMemory(prev, LC->prev, sizeof *prev * LC->maxSize);
        RtlCopyMemory(nodes, LC->nodes, sizeof *nodes * LC->maxSize);

        //
        // Fix the PrevOut/PrevIn links.
        //
        for (i = 0; i < LC->nodeCount; i++) {
            if (nodes[i].AdjOut != NULL)
                nodes[i].AdjOut->PrevOut = &nodes[i].AdjOut;
            if (nodes[i].AdjIn != NULL)
                nodes[i].AdjIn->PrevIn = &nodes[i].AdjIn;
        }

        //
        // Delete the old arrays.
        //
        ExFreePool(LC->metric);
        ExFreePool(LC->hops);
        ExFreePool(LC->link);
        ExFreePool(LC->prev);
        ExFreePool(LC->nodes);

        //
        // And update the link cache.
        //
        LC->metric = metric;
        LC->hops = hops;
        LC->link = link;
        LC->prev = prev;
        LC->nodes = nodes;
        LC->maxSize = maxSize;
    }

    //
    // Add a new node.
    //
    LC->nodeCount++;
    LC->nodes[i].AdjOut = NULL;
    LC->nodes[i].AdjIn = NULL;
    LC->nodes[i].SR = NULL;
    LC->nodes[i].Hops = NULL;
    LC->nodes[i].Metric = 0;
    LC->nodes[i].RouteChanges = 0;
#if CHANGELOGS
    LC->nodes[i].CurrentRU = NULL;
    LC->nodes[i].History = NULL;
#endif
    LC->metric[i] = MAXULONG;
    LC->hops[i] = MAXULONG;
    LC->prev[i] = MAXULONG;

    RtlCopyMemory(LC->nodes[i].address, addr, SR_ADDR_LEN);
    return i;
}

//////////////////////////////////////////////////////////////////////////////
//                DIJKSTRA'S ALGORITHM AND HELPER FUNCTIONS                 //
//////////////////////////////////////////////////////////////////////////////

#define HEAP_LEFT(x)   (2*(x)+1)
#define HEAP_RIGHT(x)  (2*(x)+2)
#define HEAP_PARENT(x) (((x)-1)>>1)

//* LinkCacheClean
//
//  Ensure efficient memory usage by freeing Links that have expired.
//  For debugging/testing purposes, if LinkTimeout is configured to zero
//  then we never remove stale links.
//
//  Called while holding LC->Lock.
//
static void
LinkCacheClean(MiniportAdapter *VA, Time Now)
{
#if STOP_LINK_METRIC
			//ejr- want to stop updating the link weights
			// so do nothing here.
			if (Now > VA->start_link_time + (Time) (LINK_MEASURE_TIME * SECOND))
			{
				//KdPrint(("Link clean metric skipped because LINK_MEASURE_TIME is up\n"));
				return; 
			}	
#endif
    if (VA->LinkTimeout != 0) {
        LinkCache *LC = VA->LC;
        Link **Prev;
        Link *This;
        uint Node;

        for (Node = 0; Node < LC->nodeCount; Node++) {
            Prev = &LC->nodes[Node].AdjOut;
            while ((This = *Prev) != NULL) {
                if ((This->TimeStamp + VA->LinkTimeout < Now) &&
                    (This->DropRatio == 0) &&
                    (This->RefCnt == 0)) {
                  
						
					//ejr- do not delete links we are directly monitoring
						if (VirtualAddressEqual(VA->Address, LC->nodes[This->sourceIndex].address) ||
							VirtualAddressEqual(VA->Address, LC->nodes[This->targetIndex].address)) {
								Prev = &This->NextOut;
								continue; 
							}

					PRINT_LINK_STR("EEE Cleaning link: ", LC->nodes[This->sourceIndex].address, 
													LC->nodes[This->targetIndex].address);


					//
                    // This link has not been updated for a long time,
                    // and it's not set for artificial drop, so delete it.
                    //
                    LinkCacheLinkChange(LC,
                                        LC->nodes[Node].address,
                                        LC->nodes[This->targetIndex].address,
                                        This->inif, This->outif,
                                        (uint)-1,
                                        LINK_STATE_CHANGE_DELETE_TIMEOUT);

                    *Prev = This->NextOut;
                    if (This->NextOut != NULL)
                        This->NextOut->PrevOut = Prev;

                    *This->PrevIn = This->NextIn;
                    if (This->NextIn != NULL)
                        This->NextIn->PrevIn = This->PrevIn;

                    ExFreePool(This);
                }
                else {
                    Prev = &This->NextOut;
                }
            }
        }
    }
}

//* relax
//
//  See if the shortest distance to v can be improved by going through u.
//  Takes the cache LC, node indices for v and u,
//  and the link that allows you to go from u to v.
//
static boolint
relax(MiniportAdapter *VA, int v, int u, Link *uv)
{
    LinkCache *LC = VA->LC;
    Link **Hops, **Hop;
    int node;
    uint Metric;

    //
    // Make sure that we did not call relax
    // with a link from a node to which we have 
    // no route.
    //
    ASSERT(LC->hops[u] != (uint)-1);

    //
    // Two quick sanity checks.
    //
	// If the Metric value is zero, then we do not want to consider the link.
	// This is true at least for ETX. But since we use that, I will add it here.
	//
	if ((*VA->IsInfinite)(uv->Metric) || (LC->hops[u] + 1 >= MAX_SR_LEN))
        return FALSE; // Can not use the link.

    //
    // Create an array of links for the path to v via u.
    //
    Hop = Hops = alloca(sizeof *Hops * (LC->hops[u] + 1));
    *Hop++ = uv;
    node = u;
    while (node != 0) {
        *Hop++ = LC->link[node];
        node = LC->prev[node];
    }

    //
    // Calculate the metric of the new path.
    // This will return (uint)-1 in case of overflow,
    // which will never be an improvement.
    //
    Metric = (*VA->PathMetric)(VA, Hops, LC->hops[u] + 1);
    
    //
    // Verify that this path metric is non-decreasing.
    // Note that this ASSERT is different from the
    // "if" statement below - here, we check against
    // node u, while in the "if" statement, we check
    // against node v. 
    //
    ASSERT(Metric >= LC->metric[u]);

	//PRINT_LINK(LC->nodes[u].address, LC->nodes[v].address, Metric);  
    if (Metric < LC->metric[v]) {
        LC->metric[v] = Metric;
        LC->hops[v] = LC->hops[u] + 1;

        //
        // Will we end up creating a 2-node loop?
        //
        ASSERT(LC->prev[u] != (uint)v);

        LC->link[v] = uv;
        LC->prev[v] = u;
        return TRUE;
    }

    return FALSE;
}

//* node_trich
//
//  Trichotomy function for nodes. Useful for bubble_up.
//  Returns +1 if a is better, -1 if b is better, 0 if equal.
//
static int
node_trich(int a, int b, const uint *metric)
{
    if (metric[a] < metric[b])
        return 1; // a is better

    if (metric[a] > metric[b])
        return -1; // b is betetr

    return 0;
}

//* swap
//
//  Swap the heap positions.
//
static void
swap(int a, int b, uint *myHeap, uint *heapLoc)
{
    uint tmp = myHeap[a];
    heapLoc[myHeap[a]] = b;
    heapLoc[myHeap[b]] = a;
    myHeap[a]          = myHeap[b];
    myHeap[b]          = tmp;
}

//* bubble_down
//
// Helper for delete_min.
//
static void
bubble_down(uint *myHeap, uint *heapLoc, int heapsz, uint *metric)
{
    int pos = 0;
    int bestkid;

    while (pos < heapsz) {
        if (HEAP_RIGHT(pos) > heapsz)
            return;
        if (HEAP_RIGHT(pos) == heapsz) {
            // only have a left child. swap if the left child is better
            if (node_trich(myHeap[HEAP_LEFT(pos)], myHeap[pos],
                           metric) > 0)
                swap(pos, HEAP_LEFT(pos), myHeap, heapLoc);
            return;
        }
        // have two kids. figure out which one is better.
        if (node_trich(myHeap[HEAP_LEFT(pos)], myHeap[HEAP_RIGHT(pos)],
                       metric) > 0)
            bestkid = HEAP_LEFT(pos);
        else
            bestkid = HEAP_RIGHT(pos);
        
        if (node_trich(myHeap[bestkid], myHeap[pos],
                       metric) > 0) {
            swap(pos, bestkid, myHeap, heapLoc);
            pos = bestkid;
        } else {
            return;
        }
    }
}

static void
bubble_up(int pos, uint *myHeap, uint *heapLoc, int heapsz,
          uint *metric)
{
    UNREFERENCED_PARAMETER(heapsz);

    while (pos) {
        if (node_trich(myHeap[pos], myHeap[HEAP_PARENT(pos)], 
                       metric) > 0) {
            swap(pos, HEAP_PARENT(pos), myHeap, heapLoc);
            pos = HEAP_PARENT(pos);
        } else {
            return;
        }
    }
}

static boolint
check_heap(uint nodeCount,
           const uint *myHeap, const uint *heapLoc, uint heapsz,
           const uint *metric)
{
    uint i;

    //
    // Check the consistency of myHeap/heapLoc.
    //
    for (i = 0; i < nodeCount; i++)
        if (myHeap[heapLoc[i]] != i)
            return FALSE;

    //
    // Check the consistency of the heap.
    //
    for (i = 0; i < heapsz; i++) {
        //
        // The left child, if present, must be >= the parent.
        //
        if ((HEAP_LEFT(i) < heapsz) &&
            (node_trich(myHeap[i], myHeap[HEAP_LEFT(i)],
                        metric) < 0)) {
            if(debug) KdPrint(("MCL!check_heap: The left child, if present, must be >= the parent.\n"));
            return FALSE;
        }

        //
        // The right child, if present, must be >= the parent.
        //
        if ((HEAP_RIGHT(i) < heapsz) &&
            (node_trich(myHeap[i], myHeap[HEAP_RIGHT(i)],
                        metric) < 0)) {
            if(debug) KdPrint(("MCL!check_heap: The right child, if present, must be >= the parent."));
            return FALSE;
        }
    }

    if (heapsz < nodeCount) {
        //
        // The smallest node in the heap is not smaller than the removed nodes.
        //
        if (node_trich(myHeap[heapsz], myHeap[0],
                       metric) < 0) {
            if(debug) KdPrint(("MCL!check_heap: The smallest node in the heap is not smaller than the removed nodes.\n"));
            return FALSE;
        }

        //
        // The nodes that were removed have non-decreasing metric,
        // which since they are in reverse order, means non-increasing metric.
        //
        for (i = heapsz; i < nodeCount - 1; i++) {
            if (node_trich(myHeap[i + 1], myHeap[i],
                           metric) < 0) {
                if(debug) KdPrint(("MCL!check_heap: The nodes that were removed have non-decreasing metric.\n"));
                return FALSE;
            }
        }
    }

    return TRUE;
}

#if DBG
//* LinkCacheCheckLinks
//
//  Verifies that the link reference counts are correct.
//
void
LinkCacheCheckLinks(LinkCache *LC)
{
    uint i;
    uint hop;
    Link *walk;
    Link **prev;
    
    //
    // Check the consistency of the AdjOut/AdjIn lists.
    //
    for (i = 0; i < LC->nodeCount; i++) {

        for (prev = &LC->nodes[i].AdjOut;
             (walk = *prev) != NULL;
             prev = &walk->NextOut) {

            ASSERT(walk->PrevOut == prev);
            ASSERT(walk->sourceIndex == (int)i);
        }

        for (prev = &LC->nodes[i].AdjIn;
             (walk = *prev) != NULL;
             prev = &walk->NextIn) {

            ASSERT(walk->PrevIn == prev);
            ASSERT(walk->targetIndex == (int)i);
        }
    }

    //
    // Reset the check counts.
    //
    for (i = 0; i < LC->nodeCount; i++) {
        for (walk = LC->nodes[i].AdjOut; walk != NULL; walk = walk->NextOut)
            walk->CheckRefCnt = 0;
    }

    //
    // Calculate the check counts.
    //
    for (i = 0; i < LC->nodeCount; i++) {
        if (LC->nodes[i].Hops != NULL) {
            for (hop = 0; hop < SOURCE_ROUTE_HOPS(LC->nodes[i].SR->optDataLen) - 1; hop++)
                LC->nodes[i].Hops[hop]->CheckRefCnt++;
        }
    }

    //
    // Check against the real counts.
    //
    for (i = 0; i < LC->nodeCount; i++) {
        for (walk = LC->nodes[i].AdjOut; walk != NULL; walk = walk->NextOut)
            ASSERT(walk->RefCnt == walk->CheckRefCnt);
    }
}
#endif // DBG




//* dijkstra
//
//  Implements the Dijkstra algorithm to compute shortest path routes.
//  Returns FALSE to indicate failure (resource shortage).
//
static int
dijkstra(MiniportAdapter *VA, Time Now)
{
    LinkCache *LC = VA->LC;
    uint *myHeap;
    uint *heapLoc;
    uint heapsz;
    uint i;
    uint curr;
    Link *walk;

	//ejr
	ASSERT(LC->timeout <= Now);


    //
    // Remove expired links.
    //
    LinkCacheClean(VA, Now);

	//trash the arb_dijkstra state
	delete_arb_dijkstra_caches(VA);

#if DBG
    LinkCacheCheckLinks(LC);
#endif

    //
    // Calculate a new value of SmallestMetric.
    // If there are no links, then leave SmallestMetric as zero.
    //
    LC->SmallestMetric = (uint)-1;
    for (i = 0; i < LC->nodeCount; i++) {
        for (walk = LC->nodes[i].AdjOut; walk != NULL; walk = walk->NextOut) {
            uint Metric = (*VA->ConvMetric)(walk->Metric);
            if (Metric < LC->SmallestMetric)
                LC->SmallestMetric = Metric;
        }

        //
        // Also reset the path metrics of any cached routes.
        //
        LC->nodes[i].Metric = 0;

#if CHANGELOGS
        //
        // Along the way, clear the cached RouteUsage records.
        // We want to do this whether dijkstra succeeds or fails.
        //
        LC->nodes[i].CurrentRU = NULL;
#endif
    }
    if (LC->SmallestMetric == (uint)-1)
        LC->SmallestMetric = 0;

    //
    // myHeap is the heap data structure thingy that we de-q from.
    // heapLoc is the inverse table, so in theory myHeap[heapLoc[i]] == i,
    // when i is still in the heap.
    //
    heapsz  = LC->nodeCount;

    myHeap = ExAllocatePool(NonPagedPool, sizeof(int) * heapsz);
    heapLoc = ExAllocatePool(NonPagedPool, sizeof(int) * heapsz);

    if ((myHeap == NULL) || (heapLoc == NULL))
        goto lowmem;

    //
    // Initialize the heap and the metric arrays.
    // Initially all nodes have infinite metric values,
    // except node 0 has metric zero.
    //
    for (i = 0; i < LC->nodeCount; i++) {
        myHeap[i] = i;
        heapLoc[i] = i;
        LC->metric[i] = MAXULONG;
        LC->hops[i] = MAXULONG;
    }
    LC->metric[0] = 0;
    LC->hops[0] = 0;


    while (heapsz && LC->hops[myHeap[0]] < MAX_SR_LEN) {
        ASSERT(check_heap(LC->nodeCount, myHeap, heapLoc, heapsz, LC->metric));

        //
        // Remove the smallest node from the heap.
        //
        curr = myHeap[0];
        swap(0, heapsz-1, myHeap, heapLoc);
        heapsz--;
        bubble_down(myHeap, heapLoc, heapsz, LC->metric);

        //
        // Update the value of other nodes,
        // using the links from the smallest node.
        //
        for (walk = LC->nodes[curr].AdjOut; walk; walk = walk->NextOut) {
            if (relax(VA, walk->targetIndex, curr, walk))
                bubble_up(heapLoc[walk->targetIndex],
                          myHeap, heapLoc, heapsz, LC->metric);
        }
    }

    //
    // So we don't return unitialized memory in IoQueryCacheNode.
    //
    LC->prev[0] = MAXULONG;

    //
    // Calculate a new value of LargestMetric.
    // This is the maximum of the metrics of the links
    // that are actually used in the shortest-path tree.
    // But leave it at infinite if there are no links.
    //
    ASSERT(myHeap[LC->nodeCount - 1] == 0);
    LC->LargestMetric = 0;
    for (i = heapsz; i < LC->nodeCount - 1; i++) {
        Link *link = LC->link[myHeap[i]];
        uint Metric = (*VA->ConvMetric)(link->Metric);
        if (Metric > LC->LargestMetric)
            LC->LargestMetric = Metric;
    }
    if (LC->LargestMetric == 0)
        LC->LargestMetric = (uint)-1;

    //
    // Don't wait very long to re-run dijkstra
    // because link metric values are always changing.
    //
    LC->timeout = Now + CACHE_TIMEOUT;

    ExFreePool(myHeap);
    ExFreePool(heapLoc);

    return TRUE;

lowmem:
    if (myHeap)
        ExFreePool(myHeap);
    if (heapLoc)
        ExFreePool(heapLoc);

    return FALSE;
}



//* extract_min
//
//  Finds vertex u with least d[u] value in setQ.
//  Needs to remove the element u from the set, so
//  let's just assign it to (uint) -1 for now.
//
// setQ_size is size of setQ, not the number of elements
// in it.
//
uint 
extract_min(uint *setQ, uint setQ_size, uint *d_vals)
{
	uint i ; 
	uint bestIndex = (uint) -1;
	uint bestValue = (uint) -1 ;

	for (i = 0 ; i < setQ_size ; i++) {
		if (setQ[i] == (uint) -1)
			continue ;
		if (d_vals[i] < bestValue) {
			bestIndex = i;
			bestValue = d_vals[i];
		}
	}

	
	//return best value
	if (bestIndex != (uint) -1)
		setQ[bestIndex] = (uint) -1 ;
    return bestIndex; 
}


//* arb_dijkstra
//
//  Implements the Dijkstra algorithm to compute shortest path route.
//  We take as input the source and destination.
//  Returns FALSE to indicate failure (resource shortage).
//
//  Uses pseudo-code from wikipedia. (not optimized)
//
// Returns true upon successful completion.
// Sticks the value of the path metric in 'retMetric'
//  The returned path metric is already converted!
//
//  Must be called with lock held.
//
int
arb_dijkstra(MiniportAdapter *VA, VirtualAddress Source, VirtualAddress Dest, uint *retMetric, Time Now)
{
    LinkCache *LC = VA->LC;
	uint i ;
	uint *setQ;
	uint numQ = 0; //number of elements in setQ
	uint curr ; //used to dequeue from setQ and enqueue to setS
	uint *d_vals ; //to store cost of shortest path to node i
    Link *walk;
	uint sourceIndex ; //index of source in LC
	uint destIndex ; //index of destination in LN
	//Time Now = KeQueryInterruptTime();
	ArbDijkstraCache *tmpadc ; 
	ArbDijkstraCache *curadc ; 

	sourceIndex = findNode(LC, Source, 0);
	destIndex = findNode(LC, Dest, 0);

	//if(debug) KdPrint(("Starting dijkstra from %d to %d\n", sourceIndex, destIndex)); 

	if (sourceIndex == -1 || destIndex == -1) {
		if(debug) KdPrint(("Unable to find source/dest in arb_dijkstra\n"));
		return FALSE ;
	}

	//check for stale state
	if (LC->timeout <= Now) {		
		//since we are going to update our state, we 
		// also need to update dijkstra's state.
		if (!dijkstra(VA, Now)) {
			if(debug) KdPrint(("ERROR: arb_dijkstra... Dijkstra did not finish!\n")); 
		}		
		//dijkstra will update LC->timeout
		// and also delete state of arb_dijkstra
	}



	curadc = VA->adc ;
	while (curadc != NULL) {
		tmpadc = curadc->next; 
		//PRINT_VA("Examing source: ", curadc->source);
		//PRINT_VA("Examing source: ", Source);
		//PRINT_VA("Examing dest: ", curadc->dest);
		//if(debug) KdPrint(("Now: %u\n", Now));
		//if(debug) KdPrint(("Examing metric: %u\n", curadc->Metric));
		//if(debug) KdPrint(("Examing timeout: %u\n", curadc->timeout));
		
		
		//found the caching entry
		if (VirtualAddressEqual(curadc->source, Source) && VirtualAddressEqual(curadc->dest, Dest)) {
			//if(debug) KdPrint(("match, returning %u\n", curadc->Metric)); 
			*retMetric = curadc->Metric ;
			return TRUE; 			
		}
		curadc = tmpadc ;
	}

	//No caching entries, just do dijkstra's algorithm
	numQ = LC->nodeCount ;

    setQ = ExAllocatePool(NonPagedPool, sizeof(int) * numQ);
    d_vals = ExAllocatePool(NonPagedPool, sizeof(int) * numQ);

    if ((setQ == NULL) || (d_vals == NULL))//| (setS == NULL) | (prev == NULL)
        goto lowmem;

    //
    // Initialize the heap and the metric arrays.
    // Initially all nodes have infinite metric values,
    // except the source node has metric zero.
    //
    for (i = 0; i < LC->nodeCount; i++) {
        d_vals[i] = (uint) -1; //represent infinity
		setQ[i] = i ;
    }
    d_vals[sourceIndex] = 0;

    while (numQ) {
        //
        // Remove the smallest node from the setQ.
        //
        curr = extract_min(setQ, LC->nodeCount, d_vals);
		if (curr == (uint) -1)
			break ;

		numQ--;

	
        //
        // Update the value of other nodes,
        // using the links from the smallest node.
        //
		//
		// Need to check for overflow.
		//
        for (walk = LC->nodes[curr].AdjOut; walk; walk = walk->NextOut) {
			uint curMetric = (*VA->ConvMetric)(walk->Metric);
			//PRINT_LINK(LC->nodes[curr].address, LC->nodes[walk->targetIndex].address, curMetric);
			if (walk->Metric == 0 || (*VA->IsInfinite)(walk->Metric) || d_vals[curr] + curMetric < d_vals[curr]) {
				//overflow
				continue; 
			}

			if (d_vals[curr] + curMetric < d_vals[walk->targetIndex]) {
				/*	
				if(debug) KdPrint(("d_vals[%d]: %u looking at sum: %u + %u = %u\n",walk->targetIndex, 
						d_vals[walk->targetIndex], d_vals[curr], curMetric, curMetric + d_vals[curr]));
				*/
				d_vals[walk->targetIndex] = d_vals[curr] + curMetric ;
			}
        }
	}

	//if(debug) KdPrint(("Dijkstra's returning: %u for index: %d\n", d_vals[destIndex], destIndex));
	*retMetric = d_vals[destIndex]; 	

	//cache this entry
	//if(debug) KdPrint(("num_adcs %u\n", VA->num_adcs)); 
	if (VA->num_adcs < MAX_ADC_ENTRIES) {	
		tmpadc = ExAllocatePool(NonPagedPool, sizeof(*tmpadc));
		if (tmpadc != NULL) {
			//if we have memory, set up the cache
			RtlZeroMemory(tmpadc, sizeof(*tmpadc));
			RtlCopyMemory(tmpadc->source, Source, sizeof(tmpadc->source));
			/*
			if(debug) KdPrint(("Size of source: %d and tmpsrc: %d\n", sizeof(Source), sizeof(tmpadc->source))); 
			if(debug) KdPrint(("Now: %u\n", Now)); 
			PRINT_VA("tmpsrc ", tmpadc->source);
			PRINT_VA("realsrc ", Source);
			*/
			RtlCopyMemory(tmpadc->dest, Dest, sizeof(tmpadc->dest));
			//PRINT_VA("tmpdest ", tmpadc->dest);

			if (VA->num_adcs == 0)
				ASSERT(VA->adc == NULL); 
			tmpadc->next = VA->adc;
			tmpadc->prev = NULL; 
			if (VA->adc != NULL)
				VA->adc->prev = tmpadc ;
			VA->adc = tmpadc ;
			tmpadc->Metric = d_vals[destIndex]; 
			
			//if(debug) KdPrint(("Metric: %u and timeout: %u\n", tmpadc->Metric, tmpadc->timeout));
			VA->num_adcs++; 
		}
	}

    ExFreePool(setQ);
    ExFreePool(d_vals);

    return TRUE;

lowmem:
    if (setQ)
        ExFreePool(setQ);
	if (d_vals)
        ExFreePool(d_vals);
    return FALSE;
}

//
// delete_arb_dijkstra_caches(MiniportAdapter *VA)
//
// In some instances, drastic changes in link values
// should cause the cache to be flushed. We make this consistent
// with LQSR- only do it when they do.
//
// Instead of resetting the timeouts (like they do), we'll just delete
// all the entries. Can be optimized.
//
// Must be called with lock held.
//
// This method is also called by MiniportHalt.
//
void
delete_arb_dijkstra_caches(MiniportAdapter *VA)
{
	ArbDijkstraCache *tmpadc ; 
	ArbDijkstraCache *curadc ; 

	curadc = VA->adc ;
	while (curadc != NULL) {
		tmpadc = curadc->next; 
		ExFreePool(curadc); 
		curadc = tmpadc ;
		VA->num_adcs--; 
	}
	VA->adc = NULL; 
	ASSERT(VA->num_adcs == 0); 
}