// -*- mode: C++; tab-width: 4; indent-tabs-mode: nil -*- (for GNU Emacs)
//
// Copyright (c) Microsoft Corporation
//
// This file is part of the Microsoft Research Mesh Connectivity Layer.
// You should have received a copy of the Microsoft Research Shared Source
// license agreement (MSR-SSLA) for this software; see the file "license.txt".
// If not, please see http://research.microsoft.com/mesh/license.htm,
// or write to Microsoft Research, One Microsoft Way, Redmond, WA 98052-6399.
//

#include "headers.h"

#define PRINT_ADDRESS(f, s, d, t) KdPrint(("%s: from %02x-%02x-%02x-%02x-%02x-%02x\
	to %02x-%02x-%02x-%02x-%02x-%02x type %i\n", f, (s)[0], (s)[1], (s)[2], (s)[3], (s)[4], (s)[5], \
	(d)[0], (d)[1], (d)[2], (d)[3], (d)[4], (d)[5], t)) 


#define PRINT_VA(f, s) KdPrint(("%s: from %02x-%02x-%02x-%02x-%02x-%02x\n", \
	f, (s)[0], (s)[1], (s)[2], (s)[3], (s)[4], (s)[5])) 


//* MaintBufAckExpected
//
//  Returns TRUE if we expect to receive an ack, that is,
//  we have any unacknowledged ack requests outstanding.
// ejr- disabling this
__inline boolint
MaintBufAckExpected(MaintBufNode *MBN)
{
	
    UNREFERENCED_PARAMETER(MBN);
	return FALSE; 
	/*
    return (LQSRAckId)(MBN->LastAckNum + 1) != MBN->NextAckNum;
	*/
}

//* MaintBufValidAck
//
//  Returns TRUE if AckNum is a valid sequence number, that is,
//  corresponds to an unacknowledged ack request.
//
__inline boolint
MaintBufValidAck(
    MaintBufNode *MBN,
    LQSRAckId AckNum)
{
    //
    // LastAckNum is the last acknowledged sequence number,
    // and NextAckNum is the next sequence number that will be used.
    // NB: The sequence numbers can wrap.
    //
    return ((0 < (LQSRAckId)(AckNum - MBN->LastAckNum)) &&
            ((LQSRAckId)(AckNum - MBN->LastAckNum) <
             (LQSRAckId)(MBN->NextAckNum - MBN->LastAckNum)));
}

//* MaintBufAckNum
//
//  Returns the most-recently used sequence number.
//
__inline LQSRAckId
MaintBufAckNum(MaintBufNode *MBN)
{
    return (LQSRAckId)(MBN->NextAckNum - 1);
}

//* MaintBufNew
//
//  Allocates a new maintenance buffer.
//
MaintBuf *
MaintBufNew(void)
{
    MaintBuf *MB;

    MB = ExAllocatePool(NonPagedPool, sizeof *MB);
    if (MB == NULL)
        return NULL;

    KeInitializeSpinLock(&MB->Lock);
    MB->MBN = NULL;

    MB->NumPackets = 0;
    MB->HighWater = 0;

    return MB;
}

//* MaintBufFree
//
//  Frees an existing maintenance buffer.
//
void
MaintBufFree(MiniportAdapter *VA)
{
    MaintBuf *MB = VA->MaintBuf;
    MaintBufNode *MBN;
    MaintBufPacket *MBP;

    //
    // Free the node structures.
    // There can be internally-generated packets waiting to be sent.
    //
    while ((MBN = MB->MBN) != NULL) {
        MB->MBN = MBN->Next;

        while ((MBP = MBN->MBP) != NULL) {
            SRPacket *SRP = MBP->srp;

            MBN->MBP = MBP->Next;

            (*SRP->TransmitComplete)(VA, SRP, NDIS_STATUS_SUCCESS);
            ExFreePool(MBP);
        }
		
        ExFreePool(MBN);
    }

    ExFreePool(MB);
}

//* MaintBufFindNode
//
//  Find a MaintBufNode with specified address and interfaces.
//  If one doesn't exist, creates it.
//
//  Called with the MaintBuf locked.
//
//ejr- this used to be static, and was defined only here.
MaintBufNode *
MaintBufFindNode(MiniportAdapter *VA, MaintBuf *MB,
                 const VirtualAddress Address,
                 LQSRIf InIf, LQSRIf OutIf)
{
    MaintBufNode *MBN;
	
    UNREFERENCED_PARAMETER(VA);
    //
    // Search for an existing MaintBufNode.
    //
    for (MBN = MB->MBN; MBN != NULL; MBN = MBN->Next) {
        if (VirtualAddressEqual(MBN->Address, Address) &&
            (MBN->InIf == InIf) &&
			(MBN->OutIf == OutIf)) {
            return MBN;
			} else if (VirtualAddressEqual(MBN->Address, Address)) {
				if(debug) KdPrint(("EEE Created new maintbuf node b/c inif's are wrong!!!!!!!!!!\n"));
			}
    }

	PRINT_VA("Creating new maintbufnode ", Address);

    //
    // Create a new MaintBufNode for this node.
    //
    MBN = ExAllocatePool(NonPagedPool, sizeof *MBN);
    if (MBN == NULL)
        return NULL;

    RtlCopyMemory(MBN->Address, Address, SR_ADDR_LEN);
    MBN->OutIf = OutIf;
    MBN->InIf = InIf;
    MBN->NextAckNum = 0;
    MBN->LastAckNum = (LQSRAckId)-1;
    MBN->LastAckRcv = 0;
    MBN->FirstAckReq = 0;
    MBN->LastAckReq = 0;
    MBN->MBP = NULL;

    MBN->NumPackets = 0;
    MBN->HighWater = 0;

    MBN->NumAckReqs = 0;
    MBN->NumFastReqs = 0;
    MBN->NumValidAcks = 0;
    MBN->NumInvalidAcks = 0;

    //
    // Add the new node to the Maintenance Buffer.
    //
    MBN->Next = MB->MBN;
    MB->MBN = MBN;

    return MBN;
}

//* MaintBufPacketRelease
//
//  Releases a reference for a MaintBufPacket.
//
void
MaintBufPacketRelease(
    MiniportAdapter *VA,
    MaintBufPacket *MBP)
{
    if (InterlockedDecrement((PLONG)&MBP->RefCnt) == 0) {
        SRPacket *srp = MBP->srp;
        //
        // There is no outstanding use of this MaintBufPacket,
        // so deallocate it and complete the packet.
        //
        (*srp->TransmitComplete)(VA, srp, NDIS_STATUS_SUCCESS);
        ExFreePool(MBP);
    }
}

//* MaintBufTransmitComplete
//
//  Called when ProtocolTransmit finishes transmitting a packet that is
//  in the maintenance buffer. Releases a reference to the MBP.
//
static void
MaintBufTransmitComplete(
    MiniportAdapter *VA,
    NDIS_PACKET *Packet,
    NDIS_STATUS Status)
{
    MaintBufPacket *MBP = PC(Packet)->MBP;

    UNREFERENCED_PARAMETER(Status);

    NdisFreePacketClone(Packet);

    MaintBufPacketRelease(VA, MBP);
}

//* MaintBufAddPacket
//
//  Adds a MaintBufPacket to a MaintBufNode,
//  returning an NDIS_PACKET to transmit.
//  Adds a reference to the MaintBufPacket
//  when returning non-NULL.
//
//  Our caller must update MBN->FirstAckReq, MBN->LackAckReq,
//  and MBN->NumAckReqs.
//
//  Called with the MaintBuf locked.
//
static NDIS_PACKET *
MaintBufAddPacket(
    MiniportAdapter *VA,
    MaintBufNode *MBN,
    MaintBufPacket *MBP)
{
    MaintBuf *MB = VA->MaintBuf;
    SRPacket *srp = MBP->srp;
    NDIS_PACKET *Packet;
    NDIS_STATUS Status;

    //
    // Get the Ack identifier for this packet.
    //
    MBP->AckNum = srp->ackreq->opt.identification = MBN->NextAckNum++;

    //
    // Add the MaintBufPacket to the MaintBufNode.
    //
    MBP->Next = MBN->MBP;
    MBN->MBP = MBP;

    if (++MBN->NumPackets > MBN->HighWater)
        MBN->HighWater = MBN->NumPackets;
    if (++MB->NumPackets > MB->HighWater)
        MB->HighWater = MB->NumPackets;

    //
    // Check for options that can be piggy-backed on this packet.
    //
    PbackSendPacket(VA, srp);

    Status = SRPacketToPkt(VA, srp, &Packet);

    //
    // We only want to send Acks once.
    //
    if (srp->ack != NULL) {
        SRFreeOptionList((InternalOption *) srp->ack);
        srp->ack = NULL;
    }

    if (Status != NDIS_STATUS_SUCCESS) {
        //
        // We do not add a reference to MBP.
        //
        return NULL;
    }
    else {
        //
        // Add a reference to MBP for MaintBufTransmitComplete.
        //
        InterlockedIncrement((PLONG)&MBP->RefCnt);
        return Packet;
    }
}

//* MaintBufTransmit
//
//  If Packet is non-NULL, our caller provides a reference
//  for the MaintBufPacket that we pass through
//  to MaintBufTransmitComplete.
//
//  Called with the MaintBuf locked.
//
static void
MaintBufTransmit(
    MiniportAdapter *VA,
    MaintBufPacket *MBP,
    NDIS_PACKET *Packet)
{
    NDIS_BUFFER *Buffer;
    void *Data;
    uint BufferLength;
    EtherHeader UNALIGNED *Ether;
	SRPacket *srpkt  ;
	SourceRoute *sroute  ;
	uint i ; 


    if (Packet == NULL) {
        //
        // MaintBufAddPacket did not give us a ref for MBP.
        // MaintBufTimer will retransmit.
        //
        return;
    }

    //
    // MaintBufTransmitComplete needs these fields.
    // We have a reference for MBP from MaintBufAddPacket
    // that we pass on to MaintBufTransmitComplete.
    //
    PC(Packet)->TransmitComplete = MaintBufTransmitComplete;
    PC(Packet)->MBP = MBP;


    if (PC(Packet)->PA == NULL) {
        //
        // This means the source route is trying to use a physical adapter
        // that no longer exists. LinkCacheDeleteInterface has been called.
        // REVIEW - We know this packet will never transmit successfully,
        // so we could immediately try salvaging instead of letting
        // MaintBufTimer get there eventually. But this is a rare situation.
        //
        MiniportSendRouteError(VA, MBP->srp);
        MaintBufTransmitComplete(VA, Packet, NDIS_STATUS_FAILURE);
        return;
    }
	
	
    Buffer = NdisFirstBuffer(Packet);
	NdisQueryBuffer(Buffer, &Data, &BufferLength);
	Ether = (EtherHeader UNALIGNED *) Data;
	if(debug) KdPrint((" call PT from MBT\n"));
	if (MBP->srp->ackreq != NULL) 
		if(debug) KdPrint(("DEBUG Send_Packet %d\n", MBP->srp->ackreq->opt.identification)); 
	else
		if(debug) KdPrint(("DEBUG Send_Packet %d\n", -1)); 
			
	srpkt = MBP->srp ;
	sroute = &((srpkt->sr)->opt) ;
	for (i = 0; i < SOURCE_ROUTE_HOPS(sroute->optDataLen); i++) {
		VirtualAddress *Address;
		Address = &(sroute->hopList[i].addr) ;
		if(debug) KdPrint(("hop %d is: %02x-%02x-%02x-%02x-%02x-%02x\n", i,
            (*Address)[0], (*Address)[1], (*Address)[2],
            (*Address)[3], (*Address)[4], (*Address)[5]));
	}

    //PRINT_ADDRESS("pt->mbt1", Ether->Source,Ether->Dest, Ether->Type); 
	IEEE_802_ADDR_SET_BROADCAST(&(Ether->Dest)) ;
    //PRINT_ADDRESS("pt->mbt2", Ether->Source,Ether->Dest, Ether->Type); 
	ProtocolTransmit(PC(Packet)->PA, Packet);
}

//* MaintBufSalvage
//
//  If possible, salvages the packet by finding a new route
//  to the destination. Our caller gives us a reference for the MBP
//  that we must dispose of.
//
Time
MaintBufSalvage(
    MiniportAdapter *VA,
    MaintBufPacket *MBP,
    Time Now)
{
    MaintBuf *MB = VA->MaintBuf;
    SRPacket *srp = MBP->srp;
    VirtualAddress NextHopAddr;
    LQSRIf NextHopOutIf, NextHopInIf;
    VirtualAddress Dest;
    uint Index;
    MaintBufNode *MBN;
    NDIS_PACKET *Packet = NULL;
    NDIS_STATUS Status;
    Time Timeout = MAXTIME;
    uint SalvageCount;
    KIRQL OldIrql;

    //
    // We do not salvage statically routed packets. 
    //
    if (srp->sr->opt.staticRoute) {
        InterlockedIncrement((PLONG)&VA->CountSalvageStatic);
        MaintBufPacketRelease(VA, MBP);
        return Timeout;
    }
    
    InterlockedIncrement((PLONG)&VA->CountSalvageAttempt);

    SalvageCount = ++srp->sr->opt.salvageCount;
    if (SalvageCount == 0) {
        //
        // We can not salvage because the count overflowed.
        //
        InterlockedIncrement((PLONG)&VA->CountSalvageOverflow);
        MaintBufPacketRelease(VA, MBP);
        return Timeout;
    }

    //
    // Get the current next-hop for this packet.
    // We failed to send to this next-hop.
    //
    Index = SOURCE_ROUTE_HOPS(srp->sr->opt.optDataLen) -
        srp->sr->opt.segmentsLeft - 1;
    ASSERT(VirtualAddressEqual(srp->sr->opt.hopList[Index].addr,
                               VA->Address));
    RtlCopyMemory(NextHopAddr, srp->sr->opt.hopList[Index + 1].addr,
                  sizeof(VirtualAddress));
    NextHopInIf = srp->sr->opt.hopList[Index + 1].inif;
    NextHopOutIf = srp->sr->opt.hopList[Index].outif;

    //
    // Get the final destination of this packet.
    //
    Index = SOURCE_ROUTE_HOPS(srp->sr->opt.optDataLen) - 1;
    RtlCopyMemory(Dest, srp->sr->opt.hopList[Index].addr,
                  sizeof(VirtualAddress));

    //
    // The DSR spec says that when salvaging, you leave the originator
    // in hopList[0] and insert the new route starting with hopList[1].
    // But this seems like an unnecessary complication. Besides complicating
    // this code, it would require a change in CacheSRPacket.
    //

    Status = LinkCacheFillSR(VA, Dest, srp->sr);
    if (Status != NDIS_STATUS_SUCCESS) {
        //
        // We can not salvage because we do not have a route.
        //
        InterlockedIncrement((PLONG)&VA->CountSalvageNoRoute);
        MaintBufPacketRelease(VA, MBP);
        return Timeout;
    }

    Index = SOURCE_ROUTE_HOPS(srp->sr->opt.optDataLen) -
        srp->sr->opt.segmentsLeft - 1;
    ASSERT(Index == 0); // This is a new source route.

    //
    // Check that we are not trying the same bad next-hop again.
    //
    if (VirtualAddressEqual(NextHopAddr,
                            srp->sr->opt.hopList[Index + 1].addr) &&
        (NextHopInIf == srp->sr->opt.hopList[Index + 1].inif) &&
        (NextHopOutIf == srp->sr->opt.hopList[Index].outif)) {
        //
        // We can not salvage because the route tries the same next-hop.
        //
        InterlockedIncrement((PLONG)&VA->CountSalvageSameNextHop);
        MaintBufPacketRelease(VA, MBP);
        return Timeout;
    }

    if (! LinkCacheUseSR(VA, srp)) {
        //
        // We can not salvage because the transmit queue is full.
        //
        InterlockedIncrement((PLONG)&VA->CountSalvageQueueFull);
        MaintBufPacketRelease(VA, MBP);
        return Timeout;
    }

    // LinkCacheFillSR zeroed salvageCount, so initialize it.
    srp->sr->opt.salvageCount = (ushort)SalvageCount;

    //
    // Find the MaintBufNode for the packet.
    //
    KeAcquireSpinLock(&MB->Lock, &OldIrql);
    MBN = MaintBufFindNode(VA, MB, srp->sr->opt.hopList[Index + 1].addr,
                           srp->sr->opt.hopList[Index + 1].inif,
                           srp->sr->opt.hopList[Index].outif);
    if (MBN != NULL) {
        //
        // Are we requesting the first ack?
        //
        if (! MaintBufAckExpected(MBN))
            MBN->FirstAckReq = Now;

        MBN->LastAckReq = Now;
        Timeout = MBN->LastAckReq + MAINTBUF_REXMIT_TIMEOUT;

        if ((MBN->LastAckRcv + MAINTBUF_HOLDOFF_TIME > Now) ||
            (MBN->NumPackets >= MAINTBUF_MAX_QUEUE)) {

            //
            // We have recent confirmation that this link is working.
            // Or we are already holding many packets for this destination.
            // In any case, just send this packet without holding it.
            // We request an ack but we do not add the MBP to the MBN.
            // So we dispose of our reference for the MBP.
            //
            MBN->NumFastReqs++;
            srp->ackreq->opt.identification = MBN->NextAckNum++;

            //
            // Check for options that can be piggy-backed on this packet.
            //
            PbackSendPacket(VA, srp);

            Status = SRPacketToPkt(VA, srp, &Packet);

            if (Status != NDIS_STATUS_SUCCESS) {
                //
                // Release our reference to MBP below.
                //
                MBN = NULL;
            }
            else {
                //
                // Donate our reference to MBP to MaintBufTransmitComplete.
                //
            }
        }
        else {
            //
            // We are sending an Ack Request to this node.
            //
            MBN->NumAckReqs++;

            //
            // Our reference for MBP goes to keep the MBP on the MBN.
            // MaintBufAddPacket returns an additional reference
            // if Packet is non-NULL.
            //
            Packet = MaintBufAddPacket(VA, MBN, MBP);
        }
    }
    KeReleaseSpinLock(&MB->Lock, OldIrql);

    if (MBN == NULL) {
        //
        // We could not salvage.
        //
        MaintBufPacketRelease(VA, MBP);
        return Timeout;
    }

    InterlockedIncrement((PLONG)&VA->CountSalvageTransmit);
    MaintBufTransmit(VA, MBP, Packet);
    return Timeout;
}

//* MaintBufCreateAckRequest
//
//  Creates an ack request packet.
//
//  Called with the MaintBuf locked.
//
static SRPacket *
MaintBufCreateAckRequest(
    MiniportAdapter *VA,
    MaintBufNode *MBN)
{
    SRPacket *SRP;
    InternalSourceRoute *SR;
    InternalAcknowledgementRequest *AR;
	uint fwd_max;
		
    //
    // Allocate a packet with which to send the ack request.
    //

    SRP = ExAllocatePool(NonPagedPool, sizeof *SRP);
    if (SRP == NULL)
        return NULL;

    SR = ExAllocatePool(NonPagedPool,
                        sizeof *SR + sizeof(SRAddr)*MAX_SR_LEN);
    if (SR == NULL) {
        ExFreePool(SRP);
        return NULL;
    }

    AR = ExAllocatePool(NonPagedPool, sizeof *AR);
    if (AR == NULL) {
        ExFreePool(SR);
        ExFreePool(SRP);
        return NULL;
    }

    //
    // Initialize the packet.
    //
    RtlZeroMemory(SRP, sizeof *SRP);
	//SRP->MAC=NULL; 
    RtlCopyMemory(SRP->Dest, MBN->Address, sizeof(VirtualAddress));
    RtlCopyMemory(SRP->Source, VA->Address, sizeof(VirtualAddress));
    SRP->sr = SR;
    SRP->ackreq = AR;

    //
    // Initialize the source route.
    //
    SR->next = NULL;
    SR->opt.optionType = LQSR_OPTION_TYPE_SOURCERT;
	
	fwd_max = (ushort) 1 ; //ejr- just sending to one hop 
	setSRPbitmap(&(SR->opt.fwd_bitmap), fwd_max); 
	SR->opt.bitmap = (ushort) - 1; 

    SR->opt.optDataLen = SOURCE_ROUTE_LEN(2);
    SR->opt.reservedField = 0;
    SR->opt.staticRoute = 0;
    SR->opt.salvageCount = 0;
    SR->opt.segmentsLeft = (uchar) 1;
    SR->opt.hopList[0].inif = 0;
    SR->opt.hopList[0].outif = MBN->OutIf;
    RtlCopyMemory(SR->opt.hopList[0].addr, SRP->Source,
                  sizeof(VirtualAddress));
    SR->opt.hopList[0].Metric = 0;
    SR->opt.hopList[1].inif = MBN->InIf;
    SR->opt.hopList[1].outif = 0;
    RtlCopyMemory(SR->opt.hopList[1].addr, SRP->Dest,
                  sizeof(VirtualAddress));
    SR->opt.hopList[1].Metric = 0;

    //
    // Initialize the Acknowledgement Request.
    //
    AR->next = NULL;
    AR->opt.optionType = LQSR_OPTION_TYPE_ACKREQ;
    AR->opt.optDataLen = ACK_REQUEST_LEN;
    AR->opt.identification = MaintBufAckNum(MBN);

    return SRP;
}

//* MaintBufAckRequestSendComplete
//
//  Called to complete the transmission of a packet
//  from MaintBufCreateAckRequest.
//
static void
MaintBufAckRequestSendComplete(
    MiniportAdapter *VA,
    NDIS_PACKET *Packet,
    NDIS_STATUS Status)
{
    SRPacket *SRP = PC(Packet)->srp;

    UNREFERENCED_PARAMETER(VA);
    UNREFERENCED_PARAMETER(Status);

    NdisFreePacketClone(Packet);
    SRPacketFree(SRP);
}

//* MaintBufTimer
//
//  Called periodically to retransmit packets in the Maintenance Buffer.
//
Time
MaintBufTimer(MiniportAdapter *VA, Time Now)
{
    MaintBuf *MB = VA->MaintBuf;
    MaintBufNode **PrevMBN, *MBN;
    MaintBufPacket *MBP;
    MaintBufPacket *Salvage = NULL;
    NDIS_PACKET *Packet;
    NDIS_PACKET *RexmitPackets = NULL;
    //SRPacket *SRP;
    Time Timeout = MAXTIME;
    Time Deadline;
    KIRQL OldIrql;
    //NDIS_STATUS Status;
	EtherHeader UNALIGNED *Ether; 
	void *Data;
    NDIS_BUFFER *Buffer;
    uint BufferLength;


    KeAcquireSpinLock(&MB->Lock, &OldIrql);

    //
    // Inspect each MaintBufNode in the Maintenance Buffer.
    // We are only interested in MBNs that have unacknowledged ack requests.
    // While doing this, we update Timeout to track
    // the next time by which we want to run.
    //
    PrevMBN = &MB->MBN;
    while ((MBN = *PrevMBN) != NULL) {
        if (MaintBufAckExpected(MBN)) {

            //
            // If it's been too long since we received an ack,
            // we conclude that the link has failed.
            //
            Deadline = max(MBN->FirstAckReq, MBN->LastAckRcv) +
                                                        MAINTBUF_LINK_TIMEOUT;
            if (Deadline <= Now) {

                //
                // Penalize the link for the failure.
                // We must do this before sending the Route Error,
                // which should carry the updated metric,
                // or attempting to salvage the packets.
                //
				if(debug) KdPrint(("\nEEE Penalizing link\n")); 
                LinkCachePenalizeLink(VA, MBN->Address, MBN->InIf, MBN->OutIf);

                //
                // Move any waiting packets to the salvage list.
                //
                while ((MBP = MBN->MBP) != NULL) {
                    MBN->MBP = MBP->Next;
                    MBP->Next = Salvage;
                    Salvage = MBP;
                }
                MB->NumPackets -= MBN->NumPackets;
                MBN->NumPackets = 0;

                //
                // We no longer expect an ack.
                //
                MBN->LastAckNum = MaintBufAckNum(MBN);

                // Next loop iteration will inspect the same MBN again.
                continue;
            }

            if (Deadline < Timeout)
                Timeout = Deadline;

		}
        else {
            //
            // There are no outstanding ack requests.
            // Hence we must have no waiting packets.
            //
			uint okay = 1;
#if STOP_LINK_METRIC
			if (Now > VA->start_link_time + (Time) (LINK_MEASURE_TIME * SECOND))
			{
				okay = 0;
			}
#endif
            ASSERT(MBN->MBP == NULL);
            Deadline = MBN->LastAckReq + MAINTBUF_IDLE_TIMEOUT;
            if (okay && Deadline <= Now) {
                //
                // This Maintenance Buffer Node has been idle for a long time.
                // If we delete it, we lose the AckNum state for this neighbor.
                // But this is OK after sufficient time.
                //
                *PrevMBN = MBN->Next;

                ExFreePool(MBN);
                // Next loop iteration looks at the next MBN.
                continue;
            }
        }

        if (Deadline < Timeout)
            Timeout = Deadline;

        // Move to the next MBN.
        PrevMBN = &MBN->Next;
    }
    KeReleaseSpinLock(&MB->Lock, OldIrql);

    //
    // Send any queued retransmissions.
    //
    while ((Packet = RexmitPackets) != NULL) {
        RexmitPackets = PC(Packet)->OrigPacket;

        if (PC(Packet)->PA == NULL) {
            //
            // This means the source route is trying to use a physical adapter
            // that no longer exists. LinkCacheDeleteInterface has been called.
            //
            if (PC(Packet)->TransmitComplete == MaintBufTransmitComplete)
                MiniportSendRouteError(VA, PC(Packet)->MBP->srp);
            (*PC(Packet)->TransmitComplete)(VA, Packet, NDIS_STATUS_FAILURE);
        }
        else {
            //
            // Transmit the packet.
            //
            
			if(debug) KdPrint(("EJR call PT from MBtimer\n")); 
			
			/*
			if (PC(Packet)->srp != NULL && PC(Packet)->srp->ackreq != NULL)
				if(debug) KdPrint(("DEBUG Resend_Packet %d\n", PC(Packet)->srp->ackreq->opt.identification)); 
			else
				if(debug) KdPrint(("DEBUG Resend_Packet %d\n", -1)); 
			*/

			Buffer = NdisFirstBuffer(Packet);
			NdisQueryBuffer(Buffer, &Data, &BufferLength);
			Ether = (EtherHeader UNALIGNED *) Data;
			
			IEEE_802_ADDR_SET_BROADCAST(&(Ether->Dest)) ;
			
			ProtocolTransmit(PC(Packet)->PA, Packet);
        }
    }

    //
    // If there were any waiting packets,
    // send errors and salvage them.
    // NB: These packets may have outstanding references.
    //
    while ((MBP = Salvage) != NULL) {
        Salvage = MBP->Next;

        MiniportSendRouteError(VA, MBP->srp);

        //
        // Try to salvage the packet.
        // This returns a new timeout.
        //
        Deadline = MaintBufSalvage(VA, MBP, Now);
        if (Deadline < Timeout)
            Timeout = Deadline;
    }

    //
    // Return the time of our next call.
    //
    ASSERT(Now < Timeout);
    return Timeout;
}

//* MaintBufRecvAck
//
//  Handles a received acknowledgement.
//
void
MaintBufRecvAck(
    MiniportAdapter *VA,
    const VirtualAddress Address,
    LQSRIf InIf,
    LQSRIf OutIf,
    LQSRAckId AckNum)
{
    MaintBuf *MB = VA->MaintBuf;
    MaintBufNode *MBN;
    MaintBufPacket **PrevMBP;
    MaintBufPacket *MBP = NULL;
    MaintBufPacket *NextMBP;
    uint NumPackets;
    KIRQL OldIrql;

    //
    // Find the appropriate MaintBufNode.
    //
    KeAcquireSpinLock(&MB->Lock, &OldIrql);
    MBN = MaintBufFindNode(VA, MB, Address, InIf, OutIf);
    if (MBN != NULL) {

        //
        // Is this a valid ack?
        // That is, between LastAckNum and NextAckNum.
        //
		//ejr: now all ACKs coming directly to us are valid
		//
        if ( 1 /*MaintBufValidAck(MBN, AckNum)*/) {
			
            //
            // We have received a valid ack, confirming the link.
            //
            MBN->NumValidAcks++;
            MBN->LastAckNum = AckNum;
            MBN->LastAckRcv = KeQueryInterruptTime();

            //
            // Remove any acknowledged packets.
            // This maintains the invariant that waiting packets
            // have a sequence number between LastAckNum and NextAckNum.
            //
            NumPackets = 0;
            for (PrevMBP = &MBN->MBP;
                 (MBP = *PrevMBP) != NULL;
                 PrevMBP = &MBP->Next) {
					 
                //
                // If we receive MBP->AckNum, will it be valid?
                //
				//ejr-- all acks go here
                if (1 /*! MaintBufValidAck(MBN, MBP->AckNum)*/) {
                    //
                    // Remove this packet (and any older ones).
                    //
                    *PrevMBP = NULL;
                    MB->NumPackets -= MBN->NumPackets - NumPackets;
                    MBN->NumPackets = NumPackets;
                    break;
                }
                NumPackets++;
            }
        }
        else {
            //
            // This is an invalid ack, because the ack sequence number
            // is not in the proper range.
            //
            MBN->NumInvalidAcks++;
        }
    }
    KeReleaseSpinLock(&MB->Lock, OldIrql);

    //
    // Did we find any packets to complete?
    //
    while (MBP != NULL) {
        NextMBP = MBP->Next;
        MaintBufPacketRelease(VA, MBP);

        MBP = NextMBP;
    }
}




//* MaintBufRetransmitSendComplete
//
// ejr- used for packets that should be retransmitted.
//
void
MaintBufRetransmitSendComplete(
    MiniportAdapter *VA,
    NDIS_PACKET *Packet,
    NDIS_STATUS Status)
{
    SRPacket *SRP ;
	
	KIRQL OldIrql;
	
	ASSERT(Packet != NULL && PC(Packet) != NULL); 

    UNREFERENCED_PARAMETER(Status);

	SRP = PC(Packet)->srp;
	//check
	ASSERT(SRP != NULL);
	ASSERT(SRP->sendCompleteDone != 1); 

    NdisFreePacketClone(Packet);
	
	//make sure no race conditions are causing errors.
	KeAcquireSpinLock(&VA->RecoveryLock, &OldIrql);

	SRP->sendCompleteDone = 1; 

	if (!OLD_NDIS_M_COMPLETE) {
	
	if (SRP->TransmitComplete == MiniportRetransmitSendComplete && !SRP->ndisMSendCompleteDone) {
			ASSERT(SRP->OriginalPacket != NULL); 	
			NdisMSendComplete(VA->MiniportHandle,
				              SRP->OriginalPacket,
					          SRP->ForwardStatus);
			SRP->ndisMSendCompleteDone = (boolint) 1; 
			SRP->OriginalPacket = NULL;

	} 

	}

	
	if (SRP->sendCompleteShouldDelete) {
		
		if (OLD_NDIS_M_COMPLETE) {
		
		if (SRP->TransmitComplete == MiniportRetransmitSendComplete) {
			//we only call ndissendcomplete after all retransmissions are complete			
			NdisMSendComplete(VA->MiniportHandle,
				              PC(SRP->Packet)->OrigPacket,
					          SRP->ForwardStatus);
			
		}

		}

		//we don't delete SRPacketHolder in deletePacket(), so do it here.
		if (SRP->RecoveryHolder)
			ExFreePool(SRP->RecoveryHolder); 
		SRP->RecoveryHolder = NULL; 
		
		//delete packet whether or not we forward data or retransmit.
		SRPacketFree(SRP);	
	} else {
		ASSERT(SRP->RecoveryHolder != NULL);
		SRP->RecoveryHolder->Packet = NULL; 
	}

	KeReleaseSpinLock(&VA->RecoveryLock, OldIrql);
}


//* MaintBufStaticSendComplete
//
//  Completion handler for statically source-routed packets.
//
void
MaintBufStaticSendComplete(
    MiniportAdapter *VA,
    NDIS_PACKET *Packet,
    NDIS_STATUS Status)
{
    SRPacket *SRP ;
	ASSERT(Packet != NULL && PC(Packet) != NULL); 

	SRP = PC(Packet)->srp;

    NdisFreePacketClone(Packet);

	//check
	ASSERT(SRP != NULL);
	

	(*SRP->TransmitComplete)(VA, SRP, Status);
}

//* MaintBufSendPacket
//
//  Sends a Source-Routed packet using Route Maintenance.
//  That is, requests an Ack and retransmits as necessary.
//
//  LinkCacheUseSR has already been called, but if we retransmit
//  then we should call it again.
//
//  NB: We must create a new NDIS_PACKET for every (re)transmission.
//  We can not allow a single NDIS_PACKET to be sent twice simultaneously.
//
void
MaintBufSendPacket(
    MiniportAdapter *VA,
    SRPacket *srp,
    void (*Complete)(MiniportAdapter *VA, SRPacket *srp, NDIS_STATUS Status))
{
    MaintBuf *MB = VA->MaintBuf;
    InternalAcknowledgementRequest *AR = NULL;
    MaintBufNode *MBNnexthop = NULL ;
    //MaintBufNode *MBNdest = NULL ;
	AckCache *ACdest = NULL ; 
    MaintBufPacket *MBP = NULL;
    NDIS_PACKET *Packet = NULL;
	Time Now = KeQueryInterruptTime();
    Time Timeout = 0;
    uint Index;
    KIRQL OldIrql;
    NDIS_STATUS Status;  
	EtherHeader UNALIGNED *Ether; 
	void *Data;
    NDIS_BUFFER *Buffer;
    uint BufferLength;
	
	SRPacket *srpkt  ;
	SourceRoute *sroute  ;
	uint i ; 
	uint DestIndex; 

    ASSERT(srp->sr != NULL);

	if (Complete != MiniportForwardComplete)
		ASSERT(srp->ackreq == NULL);
	
    //
    // We have to initialize this before unlocking MaintBuf.
    //
    srp->TransmitComplete = Complete;

    //
    // If the packet is using a static route,
    // then we do not use Route Maintenance.
    //
    if (! srp->sr->opt.staticRoute) {
        LQSRAckId saveack = 0 ; 
		//
        // Use the Maintenance Buffer to send the packet,
        // requesting an acknowledgement.
        //
        InterlockedIncrement((PLONG)&VA->CountXmitMaintBuf);

        //
        // SegmentsLeft has already been decremented.
        //
        Index = SOURCE_ROUTE_HOPS(srp->sr->opt.optDataLen) -
            srp->sr->opt.segmentsLeft - 1;
		DestIndex = SOURCE_ROUTE_HOPS(srp->sr->opt.optDataLen) - 1 ;
        ASSERT(VirtualAddressEqual(srp->sr->opt.hopList[Index].addr, 
                                   VA->Address));


		//we need to keep the same sequence number, so we keep the ackreq
		//packet expects ackreq to be null, so we'd better remove it.
		if (srp->TransmitComplete == MiniportForwardComplete && srp->ackreq != NULL) {
			saveack = srp->ackreq->opt.identification; 
			ExFreePool(srp->ackreq);
			srp->ackreq = NULL;
		}		

		//only want an ackreq if we are sending data packets
		if (srp->TransmitComplete != MiniportReceiveComplete) {
			ASSERT(srp->ackreq == NULL);
			AR = ExAllocatePool(NonPagedPool, sizeof *AR);
			if (AR == NULL) {
				if (srp->TransmitComplete == MiniportRetransmitSendComplete)
					srp->TransmitComplete = MiniportSendComplete ;
				else if (srp->TransmitComplete == MiniportForwardComplete) {
					srp->TransmitComplete = MiniportReceiveComplete; 
				} 		
				(srp->TransmitComplete)(VA, srp, NDIS_STATUS_RESOURCES);
				return;
			}

			srp->ackreq = AR;
			RtlZeroMemory(AR, sizeof *AR);
			AR->opt.optionType = LQSR_OPTION_TYPE_ACKREQ;
			AR->opt.optDataLen = ACK_REQUEST_LEN;
			// AR->opt.identification initialized in MaintBufAddPacket.
		}

        //
        // Find the MaintBufNode for the packet.
        //
        KeAcquireSpinLock(&MB->Lock, &OldIrql);
        Now = KeQueryInterruptTime();

        MBNnexthop = MaintBufFindNode(VA, MB, srp->sr->opt.hopList[Index + 1].addr,
                               srp->sr->opt.hopList[Index + 1].inif,
                               srp->sr->opt.hopList[Index].outif);
		srp->MBNnexthop = (char*) MBNnexthop ;
	

		if (MBNnexthop != NULL /*&& MBNdest != NULL*/) {

            //
            // Check if we will hold onto this packet.
            //
			if (1) { 
			/*
            if ((MBN->LastAckRcv + MAINTBUF_HOLDOFF_TIME > Now) ||
                (MBN->NumPackets >= MAINTBUF_MAX_QUEUE)) {
			*/
                //
                // We have recent confirmation that this link is working.
                // Or we are already holding many packets for this destination.
                // In any case, just send this packet without holding it.
                // We still request an ack.
                //
                if (! MaintBufAckExpected(MBNnexthop))
                    MBNnexthop->FirstAckReq = Now;

				//we only want to change sequence numbers for new flows
				if (srp->TransmitComplete != MiniportReceiveComplete && srp->TransmitComplete != MiniportForwardComplete) {
			
					//we only need ACdest if we are the original sender for this flow
					ACdest = MaintBufFindAckFlow(VA, &VA->Address, &srp->sr->opt.hopList[DestIndex].addr, 
						srp->sr->opt.hopList[Index + 1].inif, srp->sr->opt.hopList[Index].outif, &VA->Address); 

#if SLIDING_RATE_LIMIT_ENABLED					
					//rate-limit
					//we are the original source of the flow, let's ensure rate-limiting will let us send
					// the packet.
					ASSERT(srp->sr);
					ASSERT(VirtualAddressEqual(srp->sr->opt.hopList[0].addr, VA->Address));
					ASSERT(ACdest); 
					//
					// should we update our state here? ie- prune list b/c sliding window
					//
					if (srp->ackreq && shouldRateLimitDrop(VA, ACdest)) {
						//drop the packet
						if (srp->TransmitComplete == MiniportRetransmitSendComplete)
							srp->TransmitComplete = MiniportSendComplete ;
						else if (srp->TransmitComplete == MiniportForwardComplete) {
							srp->TransmitComplete = MiniportReceiveComplete; 
						} 		
						(srp->TransmitComplete)(VA, srp, NDIS_STATUS_RESOURCES);
						KeReleaseSpinLock(&MB->Lock, OldIrql);
						return; 
					} 
					//add the transmission to our sending times--done in ProtocolTransmit, ProtocolRetransmit and ProtocolCheckPacket
						
#endif				

					
#if AIMD_RATE_LIMIT_ENABLED
					//we are the original source of the flow, let's ensure rate-limiting will let us send
					// the packet.
					ASSERT(srp->sr);
					ASSERT(VirtualAddressEqual(srp->sr->opt.hopList[0].addr, VA->Address));
					ASSERT(ACdest); 
					
					//
					// check to see if interval has passed
					//
					//only for flows i am the original sender on
					if (Now >= ACdest->intervalDeadline) {
						updateAIMDdeadline(ACdest);
					}
					if (srp->ackreq && shouldAIMDRateLimitDrop(VA, ACdest)) {
						//drop the packet
						if (srp->TransmitComplete == MiniportRetransmitSendComplete)
							srp->TransmitComplete = MiniportSendComplete ;
						else if (srp->TransmitComplete == MiniportForwardComplete) {
							srp->TransmitComplete = MiniportReceiveComplete; 
						} 		
						(srp->TransmitComplete)(VA, srp, NDIS_STATUS_RESOURCES);
						KeReleaseSpinLock(&MB->Lock, OldIrql);
						return; 
					} 
#endif
					//
					//we are now going to assign sequence numbers in protocol transmit and
					// protocolcheckqueue (or at least were it updates the state), instead of
					// here. This makes rate-limit dropping a bit better.
					//
					//srp->ackreq->opt.identification = ACdest->nextSeqNumToSend++; 
					
					//If I assign an initial sequence number of zero, then the packet
					// will get deleted in the preIFQ when an ACK arrives. So let's assign
					// a special value.
					srp->ackreq->opt.identification = INITIAL_SEQ_NO ; //(uint) -1; 
				} else if (srp->TransmitComplete == MiniportForwardComplete) {
					srp->ackreq->opt.identification = saveack;
				}

                MBNnexthop->NumFastReqs++;
                MBNnexthop->LastAckReq = Now;
                Timeout = MBNnexthop->LastAckReq + MAINTBUF_REXMIT_TIMEOUT;
                KeReleaseSpinLock(&MB->Lock, OldIrql);

                MiniportRescheduleTimeout(VA, Now, Timeout);
              				
				goto SendDirectly;
            }
			
		}
        KeReleaseSpinLock(&MB->Lock, OldIrql);

        if (Timeout != 0) {
            //
            // Reschedule the next MaintBufTimer call.
            //
            MiniportRescheduleTimeout(VA, Now, Timeout);
        }

        if (MBP == NULL) {
			if (srp->TransmitComplete == MiniportRetransmitSendComplete)
				srp->TransmitComplete = MiniportSendComplete ;
			else if (srp->TransmitComplete == MiniportForwardComplete) {
				srp->TransmitComplete = MiniportReceiveComplete; 
			} 		
            (srp->TransmitComplete)(VA, srp, NDIS_STATUS_RESOURCES);
            return;
        }


		//
		// We need to be careful here. If we end up doing this, we don't use any
		// of the retransmission functions.  Therefore, we need to ensure we update
		// the complete functions.
		if (srp->TransmitComplete == MiniportRetransmitSendComplete)
			srp->TransmitComplete = MiniportSendComplete ;
		else if (srp->TransmitComplete == MiniportForwardComplete) {
			srp->TransmitComplete = MiniportReceiveComplete; 
		} 							
	
        MaintBufTransmit(VA, MBP, Packet);
    }
    else {
        //
        // Just send the packet directly.
        //
    SendDirectly:
		//don't want to do this if forwarding or sending an E2E ack....
#if E2E_RATE_LIMIT_ENABLED	
		if (!(srp && srp->ack && srp->ack->opt.isE2E)) 
#endif
			PbackSendPacket(VA, srp);


        Status = SRPacketToPkt(VA, srp, &Packet);
        if (Status != NDIS_STATUS_SUCCESS) {
			if (srp->TransmitComplete == MiniportRetransmitSendComplete)
				srp->TransmitComplete = MiniportSendComplete ;
			else if (srp->TransmitComplete == MiniportForwardComplete) {
				srp->TransmitComplete = MiniportReceiveComplete; 
			} 		
            (srp->TransmitComplete)(VA, srp, NDIS_STATUS_RESOURCES);
            return;
        }

        PC(Packet)->srp = srp;
        PC(Packet)->TransmitComplete = MaintBufStaticSendComplete;

        if (PC(Packet)->PA == NULL) {
            //
            // This means the packet is trying to use a physical adapter that
            // no longer exists.
            //
			
			//
			// We need to be careful here. If we end up doing this, we don't use any
			// of the retransmission functions.  Therefore, we need to ensure we update
			// the complete functions.
			if (srp->TransmitComplete == MiniportRetransmitSendComplete)
				srp->TransmitComplete = MiniportSendComplete ;
			else if (srp->TransmitComplete == MiniportForwardComplete) {
				srp->TransmitComplete = MiniportReceiveComplete; 
			} 
            MaintBufStaticSendComplete(VA, Packet, NDIS_STATUS_FAILURE);
        }
        else {
			Buffer = NdisFirstBuffer(Packet);
			NdisQueryBuffer(Buffer, &Data, &BufferLength);
			Ether = (EtherHeader UNALIGNED *) Data;
			

			srpkt = srp ;
			if (srp->ackreq != NULL)
					if(debug) KdPrint(("DEBUG Send_Packet %d\n", srp->ackreq->opt.identification)); 

			sroute = &((srpkt->sr)->opt) ;
			for (i = 0; i < SOURCE_ROUTE_HOPS(sroute->optDataLen); i++) {
				VirtualAddress *Address;
				Address = &(sroute->hopList[i].addr) ;
				if(debug) KdPrint(("hop %d is: %02x-%02x-%02x-%02x-%02x-%02x\n", i,
					(*Address)[0], (*Address)[1], (*Address)[2],
					(*Address)[3], (*Address)[4], (*Address)[5]));
			}

#if E2E_RATE_LIMIT_ENABLED	
			if (!(srp->ack && srp->ack->opt.isE2E))
#endif
				IEEE_802_ADDR_SET_BROADCAST(&(Ether->Dest)) ;


			//PRINT_VA("Source address: ", PC(Packet)->PA->Address); 
            ProtocolTransmit(PC(Packet)->PA, Packet);
        }
    }
}


#if NEW_E2E_RATE_CONTROL
//ejr
//* MaintbufE2EAckTimeout
//
//  Called when e2eACK thinning results in an e2eACK that needs to be sent
// because of a timeout.  
//
void MaintbufE2EAckTimeout(
    PKDPC Dpc,
    void *Context,
    void *Unused1,
    void *Unused2)
{
	AckCache *AC = (AckCache*) Context; 

    UNREFERENCED_PARAMETER(Dpc);
    UNREFERENCED_PARAMETER(Unused1);
    UNREFERENCED_PARAMETER(Unused2);
	
	AC->E2ENumAckSkipped = 0; 	
	ProtocolSendE2EAck(AC, AC->E2ENumNewPktAck); 
}
#endif


//ejr
//* MaintbufAckTimeout
//
//  Called when ACK thinning results in a ACK that needs to be sent
// because of a timeout.  Since we have cumulative ACKs, we just send 
// the most recent ACK state.
//
void MaintbufAckTimeout(
    PKDPC Dpc,
    void *Context,
    void *Unused1,
    void *Unused2)
{
	
    //MaintBufNode *MBN = (MaintBufNode *) Context;
	AckCache *AC = (AckCache*) Context; 
	MiniportAdapter *VA = AC->VA ; 
	InternalAcknowledgement *Ack;
	PbackCache *PCache ;
	SRPacket *srpack ;
    InternalSourceRoute *SR;
    boolint OKToSend;
    uint Size;
    NDIS_STATUS Status;

	EtherHeader UNALIGNED *Ether; 
	void *Data;
    NDIS_BUFFER *Buffer;
    uint BufferLength;
	NDIS_PACKET *Packet = NULL;
    KIRQL OldIrql;
#if E2E_DEBUG
	InternalAcknowledgement *curAck; 
#endif


	ASSERT(VA != NULL); 

    UNREFERENCED_PARAMETER(Dpc);
    UNREFERENCED_PARAMETER(Unused1);
    UNREFERENCED_PARAMETER(Unused2);

	//
    // Allocate the Acknowledgement option.
    //
    Ack = ExAllocatePool(NonPagedPool, sizeof *Ack);
    if (Ack == NULL)
        return;
	
	KeAcquireSpinLock(&VA->MaintBuf->Lock, &OldIrql);
    RtlZeroMemory(Ack, sizeof *Ack);
    RtlCopyMemory(Ack->opt.from, VA->Address, SR_ADDR_LEN);
	ASSERT(AC->to != NULL); 

	//need to fill in source and dest. endpoints
	RtlCopyMemory(Ack->opt.flowDest, AC->dest, SR_ADDR_LEN);
	RtlCopyMemory(Ack->opt.flowSource, AC->source, SR_ADDR_LEN);


	Ack->opt.inif = AC->InIf;
	Ack->opt.outif = AC->OutIf;

    Ack->opt.optionType = LQSR_OPTION_TYPE_ACK;
    Ack->opt.optDataLen = ACKNOWLEDGEMENT_LEN;
	//fill in opt.identification and opt.bitmap--if we use MBN->ac.AckNum, no state will be updated
	AckCacheFillAck(AC ,&(Ack->opt.identification), &(Ack->opt.bitmap), AC->AckNum); 
	KeReleaseSpinLock(&VA->MaintBuf->Lock, OldIrql);

	//Increment counter variables
	PCache = &VA->PCache;
	InterlockedIncrement((PLONG)&PCache->CountAloneAck);
	
	//Allocate a new packet to hold our ACK
	srpack = ExAllocatePool(NonPagedPool, sizeof *srpack);
	if (srpack == NULL)
          goto MainFirstDie;
    RtlZeroMemory(srpack, sizeof *srpack);
	
    //
    // Initialize source & destination of this packet.
    //
	RtlCopyMemory(srpack->Dest, AC->to, SR_ADDR_LEN);
	RtlCopyMemory(srpack->Source, VA->Address, SR_ADDR_LEN);
	//
    // Initialize a source route to this destination.
    //
    SR = ExAllocatePool(NonPagedPool, sizeof *SR + sizeof(SRAddr)*MAX_SR_LEN);
    if (SR == NULL)
		goto MainACK_Die;
    RtlZeroMemory(SR, sizeof *SR);
	
	//
	// For E2E ACKs, can we just use the regular FillSR method?
	// -be sure to manually check the fwding bitmap
	//
	Status = LinkCacheFillACKSR(VA, SR, &VA->Address, &srpack->Dest);
    if (Status != NDIS_STATUS_SUCCESS) {
		ExFreePool(SR);
		SR = NULL; 
        if (Status == NDIS_STATUS_NO_ROUTE_TO_DESTINATION) {

                //
                // Didn't find a source route. Send via route request.
                // Note that MiniportSendViaRouteRequest will add
                // a RouteRequest option, so we must leave room for it.
                //
                Size = sizeof(LQSRHeader) + sizeof(LQSROption) + ROUTE_REQUEST_LEN(MAX_SR_LEN);
                ASSERT(Size <= PROTOCOL_MIN_FRAME_SIZE);
				//
				//ejr- Add option to the packet
				//
				srpack->ack = Ack;
				
				AC->NumAckSkipped = 0 ; 
                MiniportSendViaRouteRequest(VA, srpack, ACKSendComplete);
                return ;
		}
        MainACK_Die:
            SRPacketFree(srpack);
            //
            // Drop this option in the bit bucket.
            //
		MainFirstDie:		
			ExFreePool(Ack);
            return;
	}

	
	// Send the packet using the source route.
    srpack->sr = SR;

    Size = sizeof(LQSRHeader) + sizeof(LQSROption) + SOURCE_ROUTE_LEN(MAX_SR_LEN);
    ASSERT(Size == PbackPacketSize(srpack));
    ASSERT(Size <= PROTOCOL_MIN_FRAME_SIZE);

	//Add option to the packet
	srpack->ack = Ack;

	//
    // This packet only contains acknowledgements.
    // Requesting an ack for an ack is a bad idea,
    // so suppress acknowledgements.
    //
    SR->opt.staticRoute = TRUE;
		
    
    // Send this packet using the maintenance buffer.
    OKToSend = LinkCacheUseSR(VA, srpack);
    ASSERT(OKToSend);

	//add cross flow acks
	Size = PbackPacketSize(srpack);
	if (Size < PROTOCOL_MIN_FRAME_SIZE) {
		Size = PROTOCOL_MIN_FRAME_SIZE - Size; //amount left to use
		//if(debug) KdPrint(("Piggyback size is: %u\n", Size)); 
		srpack->ack->next = MaintBufGetCrossFlowAcks(VA, AC, Size, 1);
	}
 
#if E2E_DEBUG
	for (curAck = srpack->ack ; curAck != NULL ; curAck = curAck->next) {
		ASSERT(curAck->opt.isE2E == 0);
	}
#endif

	//we'll just send the ACK immediately
    srpack->TransmitComplete = ACKSendComplete;	

    Status = SRPacketToPkt(VA, srpack, &Packet);
    if (Status != NDIS_STATUS_SUCCESS) {
		ACKSendComplete(VA, srpack, Status);
        return;
    }

    PC(Packet)->srp = srpack;
    PC(Packet)->TransmitComplete = MaintBufStaticSendComplete;

    if (PC(Packet)->PA == NULL) {
		    //
            // This means the packet is trying to use a physical adapter that
            // no longer exists.
            //
            MaintBufStaticSendComplete(VA, Packet, NDIS_STATUS_FAILURE);
    }
    else {
			
			Buffer = NdisFirstBuffer(Packet);
			NdisQueryBuffer(Buffer, &Data, &BufferLength);
			Ether = (EtherHeader UNALIGNED *) Data;
			IEEE_802_ADDR_SET_BROADCAST(&(Ether->Dest)) ;

			
			AC->NumAckSkipped = 0 ; 

			PRINT_ADDRESS("sending tack: ", srpack->Source, srpack->Dest, -1); 
			if(debug) KdPrint(("DEBUG tSending_ACK %d BitMap: %x %x %x %x %x %x %x %x\n", srpack->ack->opt.identification,
				srpack->ack->opt.bitmap.bm[0], srpack->ack->opt.bitmap.bm[1], srpack->ack->opt.bitmap.bm[2], 
				srpack->ack->opt.bitmap.bm[3], srpack->ack->opt.bitmap.bm[4], srpack->ack->opt.bitmap.bm[5], 
				srpack->ack->opt.bitmap.bm[6], srpack->ack->opt.bitmap.bm[7])); 		

			//if(debug) KdPrint(("DEBUG Sending_ACK %d\n", srpack->ack->opt.identification)); 
            ProtocolTransmit(PC(Packet)->PA, Packet);
	}

}



//* MaintBufFindAckFlow
//
// Find our AckCache for the flow of (src,dest).
// Creates a new AckCache if necessary.
//
// "to" is the address of the node that the ACK should be addressed to
//
//
AckCache* 
MaintBufFindAckFlow(MiniportAdapter *VA, VirtualAddress *Source, VirtualAddress *Dest, LQSRIf InIf, LQSRIf OutIf,
					VirtualAddress *to) {
	AckCache *tmp = VA->acl.head; 
	uint i ; 
	while (tmp != NULL) {
		if (VirtualAddressEqual(tmp->dest, *Dest) && VirtualAddressEqual(tmp->source, *Source)) {
			if (tmp->to == NULL && to != NULL)
				RtlCopyMemory(&tmp->to, to, SR_ADDR_LEN);
			return tmp; 
		}
		tmp = tmp->next; 
	}
	
	PRINT_ADDRESS("Creating new ackcache. ", *Source, *Dest, -1); 
	//if(debug) KdPrint(("Creating new ackcache.\n")); 

	//we have to create a new AckCache
	ASSERT (tmp == NULL); 
	tmp = ExAllocatePool(NonPagedPool, sizeof *tmp);
    if (tmp == NULL)
        return NULL ;	
    RtlZeroMemory(tmp, sizeof *tmp);
	tmp->AckNum =  (LQSRAckId) 0 ; 
	for (i = 0 ; i < NUMACKBITMAP ; i++) {
		tmp->bitmap.bm[i] = 0 ; 
	}
	tmp->NumAckSkipped = 0 ; 
	tmp->VA = VA; 
	RtlCopyMemory(&tmp->dest, Dest, SR_ADDR_LEN);
	if (to != NULL)
		RtlCopyMemory(&tmp->to, to, SR_ADDR_LEN);
	RtlCopyMemory(&tmp->source, Source, SR_ADDR_LEN);

	tmp->InIf = InIf; 
	tmp->OutIf = OutIf; 

	//rto calculation stuff
	tmp->firstRTT = TRUE ;

	tmp->nextSeqNumToSend = 1 ;

#if SLIDING_RATE_LIMIT_ENABLED	
	//rate-limit at source (per flow)
	tmp->SendTimes = NULL;	//linked list of timestamps for send packets
	tmp->lastSendTimes= NULL;	//pointer to last item in linked list
	tmp->numSendTimes= 0;		//size of SendTimes
	tmp->AckTimes= NULL;	//linked list of timestamps for acked packets
	tmp->lastAckTimes= NULL;	//pointer to last item in linked list
	tmp->numAckTimes= 0;		//size of AckTimes
	tmp->numForwarders= 0;		//number of forwarders included in last packet 
	tmp->virtualLoss= 0;		//virtual loss to next hop
#endif

#if AIMD_RATE_LIMIT_ENABLED
	tmp->curNumSent = 0;			//number of packet sent in the current interval
	tmp->curNumAck = 0;				//number of packets ACKed in the current interval--done on cum. ack state
	tmp->intervalDeadline = 0;		//when the current interval ends
	tmp->aimd_prob = AIMD_PROB_SCALE;	//probability we use to to determine if we should send the packet
	tmp->prev_aimd_prob = AIMD_PROB_SCALE;	
	tmp->firstInterval = TRUE;		//whether or not it is the first interval for this flow --need to initialize in ackcache init!
	tmp->virtualLoss= 0;		//virtual loss to next hop
#endif


#if E2E_RATE_LIMIT_ENABLED
	tmp->firstInterval = TRUE;		//whether or not it is the first interval for this flow --need to initialize in ackcache init!
#endif
	KeInitializeDpc(&tmp->AckTimeoutDpc, MaintbufAckTimeout, tmp);
	KeInitializeTimer(&tmp->AckTimer);

#if NEW_E2E_RATE_CONTROL
	KeInitializeDpc(&tmp->E2EAckTimeoutDpc, MaintbufE2EAckTimeout, tmp);
	KeInitializeTimer(&tmp->E2EAckTimer);
#endif

	//always add to head of list
	tmp->next = VA->acl.head; 
	VA->acl.head = tmp  ;	

	return tmp ; 

}

//* AckIncluded
// 
// Checks to see if curAck is in the link list
// pointed to by retAck.
//
// Returns true if curAck is included in retAck, false otherwise
boolint AckIncluded(InternalAcknowledgement *retAck, 
					AckCache *curAck) 

{
	InternalAcknowledgement *tmpAck; 

	for (tmpAck = retAck ; tmpAck != NULL ; tmpAck = tmpAck->next) {
		if (VirtualAddressEqual(tmpAck->opt.flowDest, curAck->dest) &&
			VirtualAddressEqual(tmpAck->opt.flowSource, curAck->source)) 
			return TRUE; 
	}

	return FALSE; 

}



//* getCrossFlowAcks
//
// AC - the AckCache from ACK we already added in pkt (can be NULL)
// Size - the amount of space left in the packet					
// numAcks - the number of ACKs already in the packet
//
InternalAcknowledgement* MaintBufGetCrossFlowAcks(
	MiniportAdapter *VA, 
	AckCache *AC, 
	uint Size, 
	uint numAcks) 
{
	
    KIRQL OldIrql;
    MaintBuf *MB = VA->MaintBuf;
	AckCache *ACcross ; 
	InternalAcknowledgement *retAck = NULL;
	InternalAcknowledgement *curAck = NULL;//to have linked list
	uint OptionSize = sizeof(LQSROption) + ACKNOWLEDGEMENT_LEN; 

	ASSERT(numAcks == 0 || numAcks == 1); 

	//
	// For cross-flow ACKs, we want to go through all the Ack caches and add them
	//
	ACcross = VA->acl.head; 
	while (TRUE) {
		InternalAcknowledgement *tAck;

		AckCache *curAckCache ;
		uint highestAckOut = 0 ;
		AckCache *highestAckCache = NULL; 

		if (Size < OptionSize) {
			//if(debug) KdPrint(("Size limit reached in piggyback ACK, numAcks: %u\n", numAcks));
			return retAck; 
		}
	
		//we want to limit the number of acks in a packet
		if (numAcks >= NUM_ACK_THRESH) {
			//if(debug) KdPrint(("Ack limit reached in piggyback ACK, numAcks: %u\n", numAcks));
			return retAck; 
		}


		//find best ACK
		for (curAckCache = VA->acl.head ; curAckCache != NULL ; curAckCache = curAckCache->next) {
			//don't want to add flows we already added
			if (curAckCache == AC || AckIncluded(retAck, curAckCache)) {
				/*
				PRINT_ADDRESS("already have Ack flow: ", curAckCache->source, curAckCache->dest, 
					curAckCache->AckNum);
				*/
				continue;
			}
			
			//if there are no outstanding ack requests, then skip
			if (curAckCache->numOutAckRequest == 0) {
				/*
				PRINT_ADDRESS("zero requests Ack flow: ", curAckCache->source, curAckCache->dest, 
					curAckCache->AckNum);
				*/
				continue; 
			}
			/*
			PRINT_ADDRESS("\tcur Ack flow: ", curAckCache->source, curAckCache->dest, 
				curAckCache->numOutAckRequest);
			*/
			if (curAckCache->numOutAckRequest >= highestAckOut) {
				highestAckOut = curAckCache->numOutAckRequest ;
				highestAckCache = curAckCache; 
			}
		}

		//want to set ACcross to the best ACK we choose
		if (highestAckCache == NULL) {
			//if(debug) KdPrint(("No more ACKs to add in MaintBufGetCrossFlowAcks\n"));
			break; 
		}
		ACcross = highestAckCache; 

		
		PRINT_ADDRESS("including Ack flow: ", ACcross->source, ACcross->dest, 
			ACcross->AckNum);

		//
		// Allocate the Acknowledgement option.
		//
		tAck = ExAllocatePool(NonPagedPool, sizeof *tAck);
		if (tAck == NULL)
			break;
		
		RtlZeroMemory(tAck, sizeof *tAck);
		RtlCopyMemory(tAck->opt.from, VA->Address, SR_ADDR_LEN);
		//RtlCopyMemory(tAck->opt.to, ACcross->to, SR_ADDR_LEN);

		//we need to set the flow information in the ACK
		RtlCopyMemory(tAck->opt.flowSource, ACcross->source, SR_ADDR_LEN);
		RtlCopyMemory(tAck->opt.flowDest, ACcross->dest, SR_ADDR_LEN);

		//assuming only 1 interface for now
		tAck->opt.inif = ACcross->InIf; 
		tAck->opt.outif = ACcross->OutIf; 

		KeAcquireSpinLock(&MB->Lock, &OldIrql);

		//fill in opt.identification and opt.bitmap
		AckCacheFillAck(ACcross, &(tAck->opt.identification), &(tAck->opt.bitmap), 0); 
		
		KeReleaseSpinLock(&MB->Lock, OldIrql);

		//
		// Initialize the Acknowledgement option.
		//
		tAck->opt.optionType = LQSR_OPTION_TYPE_ACK;
		tAck->opt.optDataLen = ACKNOWLEDGEMENT_LEN;

		//update Size-- amount left to use
		Size = Size - OptionSize; 
		numAcks++; 

		if (retAck == NULL) {
			retAck = tAck ;
		} else { 
			curAck->next = tAck;
		}
		curAck = tAck ;

		//ACcross = ACcross->next;
	}	


	return retAck; 
}




//* MaintBufFillInPiggyBackACk
//
//  Fills in the ack data structure when we piggy-back
//  ACKs onto forwarded data packets.
//
InternalAcknowledgement* MaintBufFillInPiggyBackAck(
	MiniportAdapter *VA, 
	SRPacket *SRP) 
{
	InternalAcknowledgement *Ack = NULL;
    uint Index;
    uint myIndex;
	AckCache *AC = NULL; 
    KIRQL OldIrql;
    MaintBuf *MB = VA->MaintBuf;
	
	uint numAcks = 0; 
	//AckCache *ACcross ; 
	//InternalAcknowledgement *prevAck;
 
	//only piggyback if packet isn't too big. (taken from pback.c)
	uint OptionSize ; 
	uint Size = PbackPacketSize(SRP);
	if (Size >= PROTOCOL_MIN_FRAME_SIZE) {
		if(debug) KdPrint(("Cannot piggy back ACK b/c pkt size\n"));
		return NULL; 
	}

	//'Index - 1' points to node that sent the packet-- not necessarily due
	// to new packet headers. The fact that this is inaccurate is probably
	// ok, since we only use Ack.from for determining priorities.
    Index = SOURCE_ROUTE_HOPS(SRP->sr->opt.optDataLen) - SRP->sr->opt.segmentsLeft - 1;

	//
    // Allocate the Acknowledgement option.
    //
    Ack = ExAllocatePool(NonPagedPool, sizeof *Ack);
    if (Ack == NULL)
        return NULL;

	
    RtlZeroMemory(Ack, sizeof *Ack);
    RtlCopyMemory(Ack->opt.from, VA->Address, SR_ADDR_LEN);
    //RtlCopyMemory(Ack->opt.to, SRP->sr->opt.hopList[Index - 1].addr, SR_ADDR_LEN);

	//we need to set the flow information in the ACK
	RtlCopyMemory(Ack->opt.flowSource, SRP->Source, SR_ADDR_LEN);
	RtlCopyMemory(Ack->opt.flowDest, SRP->Dest, SR_ADDR_LEN);
	
	myIndex = 0; 
	while (!VirtualAddressEqual(VA->Address, SRP->sr->opt.hopList[myIndex].addr)) {
		myIndex++;
		ASSERT(myIndex < SOURCE_ROUTE_HOPS(SRP->sr->opt.optDataLen)) ;
	}

	//assuming only 1 interface for now, need to check this for correctness
    Ack->opt.inif = SRP->sr->opt.hopList[myIndex].inif;
    Ack->opt.outif = SRP->sr->opt.hopList[Index - 1].outif;

    KeAcquireSpinLock(&MB->Lock, &OldIrql);

	AC = MaintBufFindAckFlow(VA, &SRP->sr->opt.hopList[0].addr, 
							&SRP->sr->opt.hopList[SOURCE_ROUTE_HOPS(SRP->sr->opt.optDataLen)-1].addr,
							Ack->opt.inif, Ack->opt.outif, &(SRP->sr->opt.hopList[Index - 1].addr)); 
	if (AC == NULL) {
		KeReleaseSpinLock(&MB->Lock, OldIrql);
		ExFreePool(Ack);	
		return NULL; 
	}

	//fill in opt.identification and opt.bitmap
	AckCacheFillAck(AC ,&(Ack->opt.identification), &(Ack->opt.bitmap), SRP->ackreq->opt.identification); 
	
	KeReleaseSpinLock(&MB->Lock, OldIrql);

    //
    // Initialize the Acknowledgement option.
    //
    Ack->opt.optionType = LQSR_OPTION_TYPE_ACK;
    Ack->opt.optDataLen = ACKNOWLEDGEMENT_LEN;


	//check size again	
	Size = PROTOCOL_MIN_FRAME_SIZE - Size; //amount left to use
	//if(debug) KdPrint(("Piggyback size is: %u\n", Size)); 
	OptionSize = sizeof(LQSROption) + ACKNOWLEDGEMENT_LEN; 
	//if(debug) KdPrint(("Piggyback option size is: %u\n", OptionSize)); 
	if (OptionSize > Size) {
		//need to delete the ACK
		if(debug) KdPrint(("Freeing piggyback ack b/c not enough room in pkt\n"));
			ExFreePool(Ack);	
		return NULL; 
	}
 
	//update Size-- amount left to use
	Size = Size - OptionSize; 
	numAcks++; 

	Ack->next = MaintBufGetCrossFlowAcks(VA, AC, Size, numAcks);

	return Ack ;
}


//* MaintBufSendAck
//
//  Requests immediate(ejr) transmission of an Ack.
//
// returns true if packet has already been received.
//
boolint
MaintBufSendAck(
    MiniportAdapter *VA,
    SRPacket *SRP)
{
	boolint retBool = 0 ; //equals 1 if the seqNum was already set
	AckCache *AC ; 
    InternalAcknowledgement *Ack;
    uint Index;
	//ejr
	PbackCache *PCache ;
	SRPacket *srpack ;
    InternalSourceRoute *SR;
    boolint OKToSend;
    uint Size;
    NDIS_STATUS Status;

	EtherHeader UNALIGNED *Ether; 
	void *Data;
    NDIS_BUFFER *Buffer;
    uint BufferLength;
	NDIS_PACKET *Packet = NULL;

    MaintBuf *MB = VA->MaintBuf;
    KIRQL OldIrql;
#if E2E_DEBUG
	InternalAcknowledgement *curAck; 
#endif



    ASSERT((SRP->sr != NULL) && (SRP->ackreq != NULL));
    Index = SOURCE_ROUTE_HOPS(SRP->sr->opt.optDataLen) - SRP->sr->opt.segmentsLeft;

    //
    // Allocate the Acknowledgement option.
    //
    Ack = ExAllocatePool(NonPagedPool, sizeof *Ack);
    if (Ack == NULL)
        return retBool;

	
    RtlZeroMemory(Ack, sizeof *Ack);
    RtlCopyMemory(Ack->opt.from, VA->Address, SR_ADDR_LEN);
    //RtlCopyMemory(Ack->opt.to, SRP->sr->opt.hopList[Index - 1].addr, SR_ADDR_LEN);

	//we need to set the flow information in the ACK
	RtlCopyMemory(Ack->opt.flowSource, SRP->Source, SR_ADDR_LEN);
	RtlCopyMemory(Ack->opt.flowDest, SRP->Dest, SR_ADDR_LEN);

    Ack->opt.inif = SRP->sr->opt.hopList[Index].inif;
    Ack->opt.outif = SRP->sr->opt.hopList[Index - 1].outif;

    KeAcquireSpinLock(&MB->Lock, &OldIrql);

	AC = MaintBufFindAckFlow(VA, &SRP->sr->opt.hopList[0].addr, 
							&SRP->sr->opt.hopList[SOURCE_ROUTE_HOPS(SRP->sr->opt.optDataLen)-1].addr,
							Ack->opt.inif, Ack->opt.outif, &(SRP->sr->opt.hopList[Index - 1].addr)); 
	if (AC == NULL) {
		KeReleaseSpinLock(&MB->Lock, OldIrql);
		goto MainFirstDie; 
	}

	retBool = AckCacheCheckSeqNum(AC, SRP->ackreq->opt.identification) ;

	//fill in opt.identification and opt.bitmap
	AckCacheFillAck(AC ,&(Ack->opt.identification), &(Ack->opt.bitmap), SRP->ackreq->opt.identification); 
	

	//We just received a new packet which we need to send an ACK for.
	AC->numOutAckRequest++ ;



#if NEW_E2E_RATE_CONTROL
	//only for flows we are the dest
	if (VirtualAddressEqual(AC->dest, VA->Address)) {
		//update num pkts received only if this newly recvd pkt is not retx'd
		if (!retBool) {
			AC->E2ENumNewPktAck++; 
		} else {
			AC->totalReTxRcv++; 
		}

		//E2E ACK thinning, basically the same as per hop ACK thinning below
		if (AC->E2ENumAckSkipped < E2E_NUM_ACKS_SKIPPED) {
			if (AC->E2ENumAckSkipped == 0) {
				LARGE_INTEGER Delay;  
				Delay.QuadPart = - (LONGLONG) (E2E_ACK_THIN_TIMEOUT); // Negative so its relative.
				KeSetTimer(&AC->E2EAckTimer, Delay, &AC->E2EAckTimeoutDpc);
				if(debug) KdPrint(("Setting E2E ACK timer.\n")); 
			}
			//we can skip this ACK
			AC->E2ENumAckSkipped++;
		} else {
			//reset, we'll send this time
			AC->E2ENumAckSkipped = 0 ; 
			KeCancelTimer(&AC->E2EAckTimer);
			KeReleaseSpinLock(&MB->Lock, OldIrql);
			ProtocolSendE2EAck(AC, AC->E2ENumNewPktAck); 
			KeAcquireSpinLock(&MB->Lock, &OldIrql);
		}
	}
#endif


	//
	// We want to have some ACK thinning to help the throughput of the protocol.  For now
	// we will send an ACK for every NUM_ACKS_SKIPPED packets, or after some timeout elipses.  Also,
	// we need state from MBN, which prevents us from doing this earlier.  Regardless,
	// we want the state for cumulative ACKs to be updated after each packet, so this is
	// probably an okay place to add the functionality we are looking for.  
	//	
	if (AC->NumAckSkipped < NUM_ACKS_SKIPPED) {
		if (AC->NumAckSkipped == 0) {
			LARGE_INTEGER Delay;  
			Delay.QuadPart = - (LONGLONG) (ACK_THIN_TIMEOUT); // Negative so its relative.
			KeSetTimer(&AC->AckTimer, Delay, &AC->AckTimeoutDpc);
			if(debug) KdPrint(("Setting ACK timer.\n")); 
		}
		//update state for our AC
		AC->InIf = Ack->opt.inif;
		AC->OutIf = Ack->opt.outif ;
		RtlCopyMemory(AC->to, SRP->sr->opt.hopList[Index - 1].addr, SR_ADDR_LEN);
		
		//we can skip this ACK
		AC->NumAckSkipped++;
		KeReleaseSpinLock(&MB->Lock, OldIrql);
		goto MainFirstDie;
	} else {
		//reset, we'll send this time
		AC->NumAckSkipped = 0 ; 
		KeCancelTimer(&AC->AckTimer);
	}
	
	KeReleaseSpinLock(&MB->Lock, OldIrql);

    //
    // Initialize the Acknowledgement option.
    //
    Ack->opt.optionType = LQSR_OPTION_TYPE_ACK;
    Ack->opt.optDataLen = ACKNOWLEDGEMENT_LEN;
    //Ack->opt.identification = SRP->ackreq->opt.identification;

    //
    // Send the Acknowledgement option immediately.
 	//
	//
	//ejr- One thing to be concerned about here:
	// In the original implementation of LQSR, the Route Reply
	// is usually sent along with the ACK.  Now, the route reply
	// will more then likely be send in a seperate packet. We can
	// possibly change the code to check if a route reply needs to be
	// sent, and if so, piggy-back with the ACK here.
    //
	PCache = &VA->PCache;
	InterlockedIncrement((PLONG)&PCache->CountAloneAck);
	
	//Allocate a new packet to hold our ACK
	srpack = ExAllocatePool(NonPagedPool, sizeof *srpack);
	if (srpack == NULL)
          goto MainFirstDie;
    RtlZeroMemory(srpack, sizeof *srpack);
	
    //
    // Initialize source & destination of this packet.
    //
    RtlCopyMemory(srpack->Dest, SRP->sr->opt.hopList[Index - 1].addr, SR_ADDR_LEN);
    RtlCopyMemory(srpack->Source, VA->Address, SR_ADDR_LEN);
	//
    // Initialize a source route to this destination.
    //
    SR = ExAllocatePool(NonPagedPool, sizeof *SR + sizeof(SRAddr)*MAX_SR_LEN);
    if (SR == NULL)
		goto MainACK_Die;
    RtlZeroMemory(SR, sizeof *SR);
	
	// For E2E ACKs, can we just reuse the LinkCacheFillSR() function and
	// then manually set the fwd bitmap to next node?
	//
	Status = LinkCacheFillACKSR(VA, SR, &VA->Address, &srpack->Dest);
    if (Status != NDIS_STATUS_SUCCESS) {
		ExFreePool(SR);
		SR = NULL; 
        if (Status == NDIS_STATUS_NO_ROUTE_TO_DESTINATION) {

                //
                // Didn't find a source route. Send via route request.
                // Note that MiniportSendViaRouteRequest will add
                // a RouteRequest option, so we must leave room for it.
                //
                Size = sizeof(LQSRHeader) + sizeof(LQSROption) + ROUTE_REQUEST_LEN(MAX_SR_LEN);
                ASSERT(Size <= PROTOCOL_MIN_FRAME_SIZE);
				//
				//ejr- Add option to the packet
				//
				srpack->ack = Ack;

                MiniportSendViaRouteRequest(VA, srpack, ACKSendComplete);
                return retBool;
		}
        MainACK_Die:
            SRPacketFree(srpack);
            //
            // Drop this option in the bit bucket.
            //
		MainFirstDie:		
			ExFreePool(Ack);
            return retBool;
	}

	//
	// Send the packet using the source route.
    //
    srpack->sr = SR;

    Size = sizeof(LQSRHeader) + sizeof(LQSROption) + SOURCE_ROUTE_LEN(MAX_SR_LEN);
    ASSERT(Size == PbackPacketSize(srpack));
    ASSERT(Size <= PROTOCOL_MIN_FRAME_SIZE);

	//
	//ejr- Add option to the packet
	//
	srpack->ack = Ack;

	//
    // This packet only contains acknowledgements.
    // Requesting an ack for an ack is a bad idea,
    // so suppress acknowledgements.
    //
    SR->opt.staticRoute = TRUE;
		
    //
    // Send this packet using the maintenance buffer.
    //
	OKToSend = LinkCacheUseSR(VA, srpack);
    ASSERT(OKToSend);

	//add cross flow acks
	Size = PbackPacketSize(srpack);
	if (Size < PROTOCOL_MIN_FRAME_SIZE) {
		Size = PROTOCOL_MIN_FRAME_SIZE - Size; //amount left to use
		//if(debug) KdPrint(("Piggyback size is: %u\n", Size)); 
		srpack->ack->next = MaintBufGetCrossFlowAcks(VA, AC, Size, 1);
	}
#if E2E_DEBUG
	for (curAck = srpack->ack ; curAck != NULL ; curAck = curAck->next) {
		ASSERT(curAck->opt.isE2E == 0);
	}
#endif

    srpack->TransmitComplete = ACKSendComplete;

    Status = SRPacketToPkt(VA, srpack, &Packet);
    if (Status != NDIS_STATUS_SUCCESS) {
		ACKSendComplete(VA, srpack, Status);
        return retBool;
    }

    PC(Packet)->srp = srpack;
    PC(Packet)->TransmitComplete = MaintBufStaticSendComplete;

    if (PC(Packet)->PA == NULL) {
		    //
            // This means the packet is trying to use a physical adapter that
            // no longer exists.
            //
            MaintBufStaticSendComplete(VA, Packet, NDIS_STATUS_FAILURE);
    }
    else {
			
			Buffer = NdisFirstBuffer(Packet);
			NdisQueryBuffer(Buffer, &Data, &BufferLength);
			Ether = (EtherHeader UNALIGNED *) Data;
			IEEE_802_ADDR_SET_BROADCAST(&(Ether->Dest)) ;

			PRINT_ADDRESS("sending ack: ", srpack->Source, srpack->Dest, -1); 
			if(debug) KdPrint(("DEBUG Sending_ACK %d (%d) BitMap: %x %x %x %x %x %x %x %x\n", srpack->ack->opt.identification,
				AC->AckNum,
				srpack->ack->opt.bitmap.bm[0], srpack->ack->opt.bitmap.bm[1], srpack->ack->opt.bitmap.bm[2], 
				srpack->ack->opt.bitmap.bm[3], srpack->ack->opt.bitmap.bm[4], srpack->ack->opt.bitmap.bm[5], 
				srpack->ack->opt.bitmap.bm[6], srpack->ack->opt.bitmap.bm[7])); 		

			//if(debug) KdPrint(("DEBUG Sending_ACK %d\n", srpack->ack->opt.identification)); 
            ProtocolTransmit(PC(Packet)->PA, Packet);
	}
	return retBool; 

}



//* ACKSendComplete
//
//  Completes the transmission of a control packet,
//  which is carrying options that could not be piggy-backed.
//
void
ACKSendComplete(
    MiniportAdapter *VA,
    SRPacket *SRP,
    NDIS_STATUS Status)
{
    UNREFERENCED_PARAMETER(VA);
    UNREFERENCED_PARAMETER(Status);

    SRPacketFree(SRP);
	VA->junk++; 
}



//* MaintBufResetStatistics
//
//  Resets all counters and statistics gathering for the maintenance buffer.
//
void
MaintBufResetStatistics(MiniportAdapter *VA)
{
    MaintBuf *MB = VA->MaintBuf;
    MaintBufNode *MBN;
    KIRQL OldIrql;

    KeAcquireSpinLock(&MB->Lock, &OldIrql);
    MB->HighWater = MB->NumPackets;
    for (MBN = MB->MBN; MBN != NULL; MBN = MBN->Next) {
        MBN->HighWater = MBN->NumPackets;
        MBN->NumAckReqs = 0;
        MBN->NumFastReqs = 0;
        MBN->NumValidAcks = 0;
        MBN->NumInvalidAcks = 0;
    }
    KeReleaseSpinLock(&MB->Lock, OldIrql);
}

