// -*- mode: C++; tab-width: 4; indent-tabs-mode: nil -*- (for GNU Emacs)
//
// Copyright (c) Microsoft Corporation
//
// This file is part of the Microsoft Research Mesh Connectivity Layer.
// You should have received a copy of the Microsoft Research Shared Source
// license agreement (MSR-SSLA) for this software; see the file "license.txt".
// If not, please see http://research.microsoft.com/mesh/license.htm,
// or write to Microsoft Research, One Microsoft Way, Redmond, WA 98052-6399.
//

#include "headers.h"

#define debug_enabled 1

#define PRINT_ADDRESS(f, s, d, t) KdPrint(("%s: from %02x-%02x-%02x-%02x-%02x-%02x\
	to %02x-%02x-%02x-%02x-%02x-%02x seqnum %i\n", f, (s)[0], (s)[1], (s)[2], (s)[3], (s)[4], (s)[5], \
	(d)[0], (d)[1], (d)[2], (d)[3], (d)[4], (d)[5], t)) 

#define PRINT_VA(f, s)  KdPrint(("%s: from %02x-%02x-%02x-%02x-%02x-%02x\n", \
	f, (s)[0], (s)[1], (s)[2], (s)[3], (s)[4], (s)[5])) 


static void
MiniportSendInfo(MiniportAdapter *VA, ProtocolAdapter *PA,
                 const SRPacket *srp,
                 const InternalInfoRequest *InfoReq);

static void
MiniportRescheduleTimeoutHelper(MiniportAdapter *VA, Time Now, Time Timeout);

static void
MiniportRescheduleRecoveryTimeoutHelper(MiniportAdapter *VA, Time Now, Time Timeout);


uint NextVirtualAdapterIndex;
NDIS_HANDLE MiniportHandle;
NDIS_HANDLE OurDeviceHandle;


struct MiniportAdapters MiniportAdapters;

#define MINIPORT_MAX_MULTICAST_ADDRESS  16
#define MINIPORT_MAX_FRAME_SIZE         1280
#define MINIPORT_HEADER_SIZE            sizeof(EtherHeader)
#define MINIPORT_MAX_LOOKAHEAD          0xffff

#define MAX_RECV_QUEUE_SIZE 16

#define MINIPORT_TIMEOUT        100     // Milliseconds.

//ejr- may need to change this depending on how we use local recovery timers, etc
#define MINIPORT_RECOVERY_TIMEOUT 100 //Milliseconds

#define DEFAULT_METRIC WCETT 

static void
MiniportTimeout(PKDPC Dpc, void *Context, void *Unused1, void *Unused2);

static void
MiniportRecoveryTimeout(PKDPC Dpc, void *Context, void *Unused1, void *Unused2);

//* FindVirtualAdapterFromIndex
//
//  Given an index, returns the virtual adapter.
//
MiniportAdapter *
FindVirtualAdapterFromIndex(uint Index)
{
    MiniportAdapter *VA;
    KIRQL OldIrql;

    KeAcquireSpinLock(&MiniportAdapters.Lock, &OldIrql);
    for (VA = MiniportAdapters.VirtualAdapters;
         VA != NULL;
         VA = VA->Next) {
        //
        // Is this the desired virtual adapter?
        //
        if (VA->Index == Index)
            break;
    }
    KeReleaseSpinLock(&MiniportAdapters.Lock, OldIrql);

    return VA;
}

//* FindVirtualAdapterFromGuid
//
//  Given a guid, returns the virtual adapter.
//
MiniportAdapter *
FindVirtualAdapterFromGuid(const GUID *Guid)
{
    MiniportAdapter *VA;
    KIRQL OldIrql;

    KeAcquireSpinLock(&MiniportAdapters.Lock, &OldIrql);
    for (VA = MiniportAdapters.VirtualAdapters;
         VA != NULL;
         VA = VA->Next) {
        //
        // Is this the desired virtual adapter?
        //
        if (RtlEqualMemory(Guid, &VA->Guid, sizeof(GUID)))
            break;
    }
    KeReleaseSpinLock(&MiniportAdapters.Lock, OldIrql);

    return VA;
}

//* MiniportReadOrCreateAdapterAddress
//
//  Reads a virtual address from the registry,
//  or if the address is not present, creates a new address
//  and persists it in the registry.
//
NDIS_STATUS
MiniportReadOrCreateAdapterAddress(
    IN HANDLE KeyHandle,
    OUT VirtualAddress Address)
{
    NTSTATUS Status;

    Status = GetRegNetworkAddress(KeyHandle, L"NetworkAddress", Address);
    if (! NT_SUCCESS(Status)) {

        //
        // Create a pseudo-random address.
        //
        GetRandom(Address, VIRTUAL_ADDR_LENGTH);

        //
        // Clear the individual/group bit and
        // the local/universal bit:
        // We want a universal individual address.
        //
        IEEE_802_ADDR_CLEAR_GROUP(Address);
        IEEE_802_ADDR_CLEAR_LOCAL(Address);

        //
        // Write the new address to the registry.
        //
        Status = SetRegNetworkAddress(KeyHandle, L"NetworkAddress", Address);
        if (! NT_SUCCESS(Status)) {
            if(debug) KdPrint(("MCL!SetRegNetworkAddress -> %x\n", Status));
            return NDIS_STATUS_FAILURE;
        }
    }

    return NDIS_STATUS_SUCCESS;
}

//* MiniportIsInfinite
//
//  Returns TRUE if the link metric indicates
//  that we do not know about the link.
//
boolint
MiniportIsInfinite(uint Metric)
{
    return Metric == (uint)-1;
}

//* MiniportIsBad
//
//  Returns TRUE if the link metric indicates
//  that the link is bad and should never be used.
//
boolint
MiniportIsBad(uint Metric)
{
    UNREFERENCED_PARAMETER(Metric);

    return FALSE;
}

//* MiniportConvMetric
//
//  Converts a link metric to a path metric.
//
uint
MiniportConvMetric(uint LinkMetric)
{
    return LinkMetric;
}

//* MiniportPathMetric
//
//  Calculates the path metric of an array of links,
//  by summing the metrics of the links.
//
//  Called with the link cache locked.
//
uint
MiniportPathMetric(
    MiniportAdapter *VA,
    Link **Hops,
    uint NumHops)
{
    uint Metric;
    uint New;
    uint i;

    Metric = 0;
    for (i = 0; i < NumHops; i++) {
        //
        // Check for a broken link.
        //
        if ((*VA->IsInfinite)(Hops[i]->Metric)) {
            Metric = (uint)-1;
            break;
        }

        //
        // Check for overflow.
        //
        New = Metric + (*VA->ConvMetric)(Hops[i]->Metric);
        if (New < Metric) {
            Metric = (uint)-1;
            break;
        }
        Metric = New;
    }
    return Metric;
}

//* MiniportPenalize
//
//  Penalize the link metric for a transmission failure.
//
void
MiniportPenalize(
    MiniportAdapter *VA,
    Link *Link)
{
    UNREFERENCED_PARAMETER(VA);

    Link->Metric = (uint)-1;
}

//* MiniportInitLinkMetric
//
//  Initialize metric information for a new link.
//
void 
MiniportInitLinkMetric(
    MiniportAdapter *VA,
    int SNode,
    Link *Link, 
    Time Now)
{
    UNREFERENCED_PARAMETER(VA);
    UNREFERENCED_PARAMETER(SNode);
    UNREFERENCED_PARAMETER(Now);

    Link->Metric = 1;
}

//* MiniportProbeComplete
//
//  Completes an individual packet transmission for MiniportForwardProbe.
//
void
MiniportProbeComplete(
    MiniportAdapter *VA,
    NDIS_PACKET *Packet,
    NDIS_STATUS Status)
{
    SRPacket *srp = PC(Packet)->srp;

    //
    // Free the packet structure.
    //
    NdisFreePacketClone(Packet);

    //
    // Update the current cumulative status.
    // If the current status is NDIS_STATUS_SUCCESS,
    // we replace it with our new status.
    //
    InterlockedCompareExchange((PLONG)&srp->ForwardStatus,
                               Status,
                               NDIS_STATUS_SUCCESS);

    if (InterlockedDecrement((PLONG)&srp->ForwardCount) == 0) {
        //
        // Complete the ProtocolForwardRequest operation.
        //
        (*srp->TransmitComplete)(VA, srp, srp->ForwardStatus);
    }
}

//* MiniportForwardProbe
//
//  Sends a probe packet via every physical adapter simultaneously.
//  The operation completes when the last underlying transmit completes.
//  The operation completes successfully only if every
//  underlying transmit is successful.
//
//  Our caller supplies srp->TransmitComplete, which always called
//  to consume the SRPacket.
//
void
MiniportForwardProbe(
    MiniportAdapter *VA,
    SRPacket *srp)
{
    ProtocolAdapter *PA;
    NDIS_PACKET *PacketList = NULL;
    NDIS_PACKET *Packet;
    NDIS_STATUS Status;
    KIRQL OldIrql;

    ASSERT(srp->Probe != NULL);

    //
    // We start with one reference (our own) for the SRPacket.
    //
    srp->ForwardCount = 1;
    srp->ForwardStatus = NDIS_STATUS_SUCCESS;

    //
    // First we build a temporary list of packet structures
    // and corresponding physical adapters. While doing this,
    // we also initialize the transmission count.
    // It starts with one, for our own reference for OrigPacket.
    //
    KeAcquireSpinLock(&VA->Lock, &OldIrql);
    for (PA = VA->PhysicalAdapters;
         PA != NULL;
         PA = PA->Next) {

        if (!PA->Use)
          continue;

        srp->Probe->Opt.OutIf = (LQSRIf) PA->Index;

        Status = SRPacketToPkt(VA, srp, &Packet);
        if (Status != NDIS_STATUS_SUCCESS) {
            srp->ForwardStatus = Status;
            break;
        }

        //
        // Remember the packet and corresponding physical adapter.
        // We temporarily use the OrigPacket field to build our list.
        //
        PC(Packet)->PA = PA;
        PC(Packet)->OrigPacket = PacketList;
        PacketList = Packet;
        srp->ForwardCount++;
    }
    KeReleaseSpinLock(&VA->Lock, OldIrql);

    //
    // Now we can transmit the packet via each physical adapter,
    // using the new packet structures.
    //
    while ((Packet = PacketList) != NULL) {
        PacketList = PC(Packet)->OrigPacket;
        PA = PC(Packet)->PA;

        //
        // Send the packet via a physical adapter.
        //
        PC(Packet)->srp = srp;
        PC(Packet)->TransmitComplete = MiniportProbeComplete;
		//if(debug) KdPrint(("\tSending Probe Packet.\n")); 
        ProtocolTransmit(PA, Packet);
    }

    //
    // Release our reference for the original packet.
    //
    if (InterlockedDecrement((PLONG)&srp->ForwardCount) == 0) {
        //
        // Complete the ProtocolForwardRequest operation.
        //
        (*srp->TransmitComplete)(VA, srp, srp->ForwardStatus);
    }
}

//* MiniportPersistControl
//
//  Writes configuration parameters to the registry.
//
NTSTATUS
MiniportPersistControl(
    MiniportAdapter *VA,
    MCL_CONTROL_VIRTUAL_ADAPTER *Control)
{
    NTSTATUS Status;
    HANDLE DeviceKey;
    HANDLE Key;

    //
    // We do not need to validate the configuration parameters.
    // They will be validated after being read from the registry.
    //

    Status = IoOpenDeviceRegistryKey(VA->PhysicalDeviceObject,
                                     PLUGPLAY_REGKEY_DRIVER,
                                     GENERIC_READ,
                                     &DeviceKey);
    if (! NT_SUCCESS(Status))
        return Status;

    Status = OpenRegKey(&Key, DeviceKey, L"Parameters", OpenRegKeyCreate);
    ZwClose(DeviceKey);
    if (! NT_SUCCESS(Status))
        return Status;

    if (Control->Snooping != (uint)-1) {
        Status = SetRegDWORDValue(Key, L"Snooping", Control->Snooping);
        if (! NT_SUCCESS(Status))
            goto CloseReturn;
    }

    if (Control->ArtificialDrop != (uint)-1) {
        Status = SetRegDWORDValue(Key, L"ArtificialDrop",
                                  Control->ArtificialDrop);
        if (! NT_SUCCESS(Status))
            goto CloseReturn;
    }

    if (Control->RouteFlapDampingFactor != (uint)-1) {
        Status = SetRegDWORDValue(Key, L"RouteFlapDampingFactor",
                                  Control->RouteFlapDampingFactor);
        if (! NT_SUCCESS(Status))
            goto CloseReturn;
    }

    if (Control->Crypto != (uint)-1) {
        Status = SetRegDWORDValue(Key, L"Crypto", Control->Crypto);
        if (! NT_SUCCESS(Status))
            goto CloseReturn;
    }

    if (LQSR_KEY_PRESENT(Control->CryptoKeyMAC)) {
        Status = SetRegBinaryValue(Key, L"CryptoKeyMAC",
                        Control->CryptoKeyMAC, sizeof Control->CryptoKeyMAC);
        if (! NT_SUCCESS(Status))
            goto CloseReturn;
    }

    if (LQSR_KEY_PRESENT(Control->CryptoKeyAES)) {
        Status = SetRegBinaryValue(Key, L"CryptoKeyAES",
                        Control->CryptoKeyAES, sizeof Control->CryptoKeyAES);
        if (! NT_SUCCESS(Status))
            goto CloseReturn;
    }

    if (Control->LinkTimeout != (uint)-1) {
        Status = SetRegDWORDValue(Key, L"LinkTimeout", Control->LinkTimeout);
        if (! NT_SUCCESS(Status))
            goto CloseReturn;
    }

    if (Control->MetricType != (uint)-1) {
        Status = SetRegDWORDValue(Key, L"MetricType", Control->MetricType);
        if (! NT_SUCCESS(Status))
            goto CloseReturn;
    }

    switch (Control->MetricType) {
    case METRIC_TYPE_RTT:
        if (Control->MetricParams.Rtt.Alpha != (uint)-1) {
            Status = SetRegDWORDValue(Key, L"Alpha",
                                      Control->MetricParams.Rtt.Alpha);
            if (! NT_SUCCESS(Status))
                goto CloseReturn;
        }
        if (Control->MetricParams.Rtt.ProbePeriod != (uint)-1) {
            Status = SetRegDWORDValue(Key, L"ProbePeriod",
                                      Control->MetricParams.Rtt.ProbePeriod);
            if (! NT_SUCCESS(Status))
                goto CloseReturn;
        }
        if (Control->MetricParams.Rtt.PenaltyFactor != (uint)-1) {
            Status = SetRegDWORDValue(Key, L"PenaltyFactor",
                                      Control->MetricParams.Rtt.PenaltyFactor);
            if (! NT_SUCCESS(Status))
                goto CloseReturn;
        }
        if (Control->MetricParams.Rtt.HysteresisPeriod != (uint)-1) {
            Status = SetRegDWORDValue(Key, L"HysteresisPeriod",
                                      Control->MetricParams.Rtt.HysteresisPeriod);
            if (! NT_SUCCESS(Status))
                goto CloseReturn;
        }
        if (Control->MetricParams.Rtt.SweepPeriod != (uint)-1) {
            Status = SetRegDWORDValue(Key, L"SweepPeriod",
                                      Control->MetricParams.Rtt.SweepPeriod);
            if (! NT_SUCCESS(Status))
                goto CloseReturn;
        }
        if (Control->MetricParams.Rtt.Random != (uint)-1) {
            Status = SetRegDWORDValue(Key, L"Random",
                                      Control->MetricParams.Rtt.Random);
            if (! NT_SUCCESS(Status))
                goto CloseReturn;
        }
        if (Control->MetricParams.Rtt.OutIfOverride != (uint)-1) {
            Status = SetRegDWORDValue(Key, L"OutIfOverride",
                                      Control->MetricParams.Rtt.OutIfOverride);
            if (! NT_SUCCESS(Status))
                goto CloseReturn;
        }
        break;

    case METRIC_TYPE_PKTPAIR:
        if (Control->MetricParams.PktPair.Alpha != (uint)-1) {
            Status = SetRegDWORDValue(Key, L"Alpha",
                                    Control->MetricParams.PktPair.Alpha);
            if (! NT_SUCCESS(Status))
                goto CloseReturn;
        }
        if (Control->MetricParams.PktPair.ProbePeriod != (uint)-1) {
            Status = SetRegDWORDValue(Key, L"ProbePeriod",
                                    Control->MetricParams.PktPair.ProbePeriod);
            if (! NT_SUCCESS(Status))
                goto CloseReturn;
        }
        if (Control->MetricParams.PktPair.PenaltyFactor != (uint)-1) {
            Status = SetRegDWORDValue(Key, L"PenaltyFactor",
                                    Control->MetricParams.PktPair.PenaltyFactor);
            if (! NT_SUCCESS(Status))
                goto CloseReturn;
        }
        break;

    case METRIC_TYPE_ETX:
        if (Control->MetricParams.Etx.Alpha != (uint)-1) {
            Status = SetRegDWORDValue(Key, L"Alpha",
                                    Control->MetricParams.Etx.Alpha);
            if (! NT_SUCCESS(Status))
                goto CloseReturn;
        }
        if (Control->MetricParams.Etx.ProbePeriod != (uint)-1) {
            Status = SetRegDWORDValue(Key, L"ProbePeriod",
                                    Control->MetricParams.Etx.ProbePeriod);
            if (! NT_SUCCESS(Status))
                goto CloseReturn;
        }
        if (Control->MetricParams.Etx.PenaltyFactor != (uint)-1) {
            Status = SetRegDWORDValue(Key, L"PenaltyFactor",
                                    Control->MetricParams.Etx.PenaltyFactor);
            if (! NT_SUCCESS(Status))
                goto CloseReturn;
        }
        if (Control->MetricParams.Etx.LossInterval != (uint)-1) {
            Status = SetRegDWORDValue(Key, L"LossInterval",
                                    Control->MetricParams.Etx.LossInterval);
            if (! NT_SUCCESS(Status))
                goto CloseReturn;
        }
        break;

    case METRIC_TYPE_WCETT:
        if (Control->MetricParams.Wcett.Alpha != (uint)-1) {
            Status = SetRegDWORDValue(Key, L"Alpha",
                                    Control->MetricParams.Wcett.Alpha);
            if (! NT_SUCCESS(Status))
                goto CloseReturn;
        }
        if (Control->MetricParams.Wcett.ProbePeriod != (uint)-1) {
            Status = SetRegDWORDValue(Key, L"ProbePeriod",
                                    Control->MetricParams.Wcett.ProbePeriod);
            if (! NT_SUCCESS(Status))
                goto CloseReturn;
        }
        if (Control->MetricParams.Wcett.PenaltyFactor != (uint)-1) {
            Status = SetRegDWORDValue(Key, L"PenaltyFactor",
                                    Control->MetricParams.Wcett.PenaltyFactor);
            if (! NT_SUCCESS(Status))
                goto CloseReturn;
        }
        if (Control->MetricParams.Wcett.LossInterval != (uint)-1) {
            Status = SetRegDWORDValue(Key, L"LossInterval",
                                    Control->MetricParams.Wcett.LossInterval);
            if (! NT_SUCCESS(Status))
                goto CloseReturn;
        }
        if (Control->MetricParams.Wcett.Beta != (uint)-1) {
            Status = SetRegDWORDValue(Key, L"Beta",
                                    Control->MetricParams.Wcett.Beta);
            if (! NT_SUCCESS(Status))
                goto CloseReturn;
        }
        if (Control->MetricParams.Wcett.PktPairProbePeriod != (uint)-1) {
            Status = SetRegDWORDValue(Key, L"PktPairProbePeriod",
                                    Control->MetricParams.Wcett.PktPairProbePeriod);
            if (! NT_SUCCESS(Status))
                goto CloseReturn;
        }
        if (Control->MetricParams.Wcett.PktPairMinOverProbes != (uint)-1) {
            Status = SetRegDWORDValue(Key, L"PktPairMinOverProbes",
                                    Control->MetricParams.Wcett.PktPairMinOverProbes);
            if (! NT_SUCCESS(Status))
                goto CloseReturn;
        }
        break;

    case METRIC_TYPE_ULT:
        if (Control->MetricParams.Ult.UcastProbePeriod != (uint)-1) {
            Status = SetRegDWORDValue(Key, L"UcastProbePeriod",
                                Control->MetricParams.Ult.UcastProbePeriod);
            if (! NT_SUCCESS(Status))
                goto CloseReturn;
        }
        if (Control->MetricParams.Ult.UcastLowThreshold != (uint)-1) {
            Status = SetRegDWORDValue(Key, L"UcastLowThreshold",
                                Control->MetricParams.Ult.UcastLowThreshold);
            if (! NT_SUCCESS(Status))
                goto CloseReturn;
        }
        if (Control->MetricParams.Ult.UcastHighThreshold != (uint)-1) {
            Status = SetRegDWORDValue(Key, L"UcastHighThreshold",
                                Control->MetricParams.Ult.UcastHighThreshold);
            if (! NT_SUCCESS(Status))
                goto CloseReturn;
        }
        if (Control->MetricParams.Ult.UcastLossPenalty != (uint)-1) {
            Status = SetRegDWORDValue(Key, L"UcastLossPenalty",
                                Control->MetricParams.Ult.UcastLossPenalty);
            if (! NT_SUCCESS(Status))
                goto CloseReturn;
        }
        if (Control->MetricParams.Ult.BcastProbePeriod != (uint)-1) {
            Status = SetRegDWORDValue(Key, L"BcastProbePeriod",
                                Control->MetricParams.Ult.BcastProbePeriod);
            if (! NT_SUCCESS(Status))
                goto CloseReturn;
        }
        if (Control->MetricParams.Ult.BcastLossThreshold != (uint)-1) {
            Status = SetRegDWORDValue(Key, L"BcastLossThreshold",
                                Control->MetricParams.Ult.BcastLossThreshold);
            if (! NT_SUCCESS(Status))
                goto CloseReturn;
        }
        break;
    }

    Status = STATUS_SUCCESS;
CloseReturn:
    ZwClose(Key);
    return Status;
}

//* MiniportReadControl
//
//  Reads configuration parameters from the registry.
//
void
MiniportReadControl(
    MiniportAdapter *VA,
    MCL_CONTROL_VIRTUAL_ADAPTER *Control)
{
    NTSTATUS Status;
    HANDLE DeviceKey;
    HANDLE Key;

    //
    // We must always return an initialized Control structure.
    //
    MCL_INIT_CONTROL_VIRTUAL_ADAPTER(Control);

    Status = IoOpenDeviceRegistryKey(VA->PhysicalDeviceObject,
                                     PLUGPLAY_REGKEY_DRIVER,
                                     GENERIC_READ,
                                     &DeviceKey);
    if (! NT_SUCCESS(Status))
        return;

    Status = OpenRegKey(&Key, DeviceKey, L"Parameters", OpenRegKeyCreate);
    ZwClose(DeviceKey);
    if (! NT_SUCCESS(Status))
        return;

    //
    // Read configuration parameters.
    //

    (void) GetRegDWORDValue(Key, L"Snooping",
                            (PULONG)&Control->Snooping);
    (void) GetRegDWORDValue(Key, L"ArtificialDrop",
                            (PULONG)&Control->ArtificialDrop);
    (void) GetRegDWORDValue(Key, L"RouteFlapDampingFactor",
                            (PULONG)&Control->RouteFlapDampingFactor);

    (void) GetRegDWORDValue(Key, L"Crypto",
                            (PULONG)&Control->Crypto);
    (void) GetRegBinaryValue(Key, L"CryptoKeyMAC",
                Control->CryptoKeyMAC, sizeof Control->CryptoKeyMAC);
    (void) GetRegBinaryValue(Key, L"CryptoKeyAES",
                Control->CryptoKeyAES, sizeof Control->CryptoKeyAES);

    (void) GetRegDWORDValue(Key, L"LinkTimeout",
                            (PULONG)&Control->LinkTimeout);

    (void) GetRegDWORDValue(Key, L"MetricType",
                            (PULONG)&Control->MetricType);
    switch (Control->MetricType) {
    case METRIC_TYPE_RTT:
        MCL_INIT_CONTROL_RTT(Control);
        (void) GetRegDWORDValue(Key, L"Alpha",
                                (PULONG)&Control->MetricParams.Rtt.Alpha);
        (void) GetRegDWORDValue(Key, L"ProbePeriod",
                                (PULONG)&Control->MetricParams.Rtt.ProbePeriod);
        (void) GetRegDWORDValue(Key, L"PenaltyFactor",
                                (PULONG)&Control->MetricParams.Rtt.PenaltyFactor);
        (void) GetRegDWORDValue(Key, L"HysteresisPeriod",
                                (PULONG)&Control->MetricParams.Rtt.HysteresisPeriod);
        (void) GetRegDWORDValue(Key, L"SweepPeriod",
                                (PULONG)&Control->MetricParams.Rtt.SweepPeriod);
        (void) GetRegDWORDValue(Key, L"Random",
                                (PULONG)&Control->MetricParams.Rtt.Random);
        (void) GetRegDWORDValue(Key, L"OutIfOverride",
                                (PULONG)&Control->MetricParams.Rtt.OutIfOverride);
        break;

    case METRIC_TYPE_PKTPAIR:
        MCL_INIT_CONTROL_PKTPAIR(Control);
        (void) GetRegDWORDValue(Key, L"Alpha",
                                (PULONG)&Control->MetricParams.PktPair.Alpha);
        (void) GetRegDWORDValue(Key, L"ProbePeriod",
                                (PULONG)&Control->MetricParams.PktPair.ProbePeriod);
        (void) GetRegDWORDValue(Key, L"PenaltyFactor",
                                (PULONG)&Control->MetricParams.PktPair.PenaltyFactor);
        break;

    case METRIC_TYPE_ETX:
        MCL_INIT_CONTROL_ETX(Control);
        (void) GetRegDWORDValue(Key, L"Alpha",
                                (PULONG)&Control->MetricParams.Etx.Alpha);
        (void) GetRegDWORDValue(Key, L"ProbePeriod",
                                (PULONG)&Control->MetricParams.Etx.ProbePeriod);
        (void) GetRegDWORDValue(Key, L"PenaltyFactor",
                                (PULONG)&Control->MetricParams.Etx.PenaltyFactor);
        (void) GetRegDWORDValue(Key, L"LossInterval",
                                (PULONG)&Control->MetricParams.Etx.LossInterval);
        break;

    case METRIC_TYPE_WCETT:
        MCL_INIT_CONTROL_WCETT(Control);
        (void) GetRegDWORDValue(Key, L"Alpha",
                                (PULONG)&Control->MetricParams.Wcett.Alpha);
        (void) GetRegDWORDValue(Key, L"ProbePeriod",
                                (PULONG)&Control->MetricParams.Wcett.ProbePeriod);
        (void) GetRegDWORDValue(Key, L"PenaltyFactor",
                                (PULONG)&Control->MetricParams.Wcett.PenaltyFactor);
        (void) GetRegDWORDValue(Key, L"LossInterval",
                                (PULONG)&Control->MetricParams.Wcett.LossInterval);
        (void) GetRegDWORDValue(Key, L"Beta",
                                (PULONG)&Control->MetricParams.Wcett.Beta);
        (void) GetRegDWORDValue(Key, L"PktPairProbePeriod",
                                (PULONG)&Control->MetricParams.Wcett.PktPairProbePeriod);
        (void) GetRegDWORDValue(Key, L"PktPairMinOverProbes",
                                (PULONG)&Control->MetricParams.Wcett.PktPairMinOverProbes);
        break;

    case METRIC_TYPE_ULT:
        MCL_INIT_CONTROL_ULT(Control);
        (void) GetRegDWORDValue(Key, L"UcastProbePeriod",
                        (PULONG)&Control->MetricParams.Ult.UcastProbePeriod);
        (void) GetRegDWORDValue(Key, L"UcastLowThreshold",
                        (PULONG)&Control->MetricParams.Ult.UcastLowThreshold);
        (void) GetRegDWORDValue(Key, L"UcastHighThreshold",
                        (PULONG)&Control->MetricParams.Ult.UcastHighThreshold);
        (void) GetRegDWORDValue(Key, L"UcastLossPenalty",
                        (PULONG)&Control->MetricParams.Ult.UcastLossPenalty);
        (void) GetRegDWORDValue(Key, L"BcastProbePeriod",
                        (PULONG)&Control->MetricParams.Ult.BcastProbePeriod);
        (void) GetRegDWORDValue(Key, L"BcastLossThreshold",
                        (PULONG)&Control->MetricParams.Ult.BcastLossThreshold);
        break;
    }

    ZwClose(Key);
}

//* MiniportControl
//
//  Sets configuration parameters from the structure.
//  Note that CryptoKeyMAC, CryptoKeyAES, and MetricType
//  can not be modified.
//
NTSTATUS
MiniportControl(
    MiniportAdapter *VA,
    MCL_CONTROL_VIRTUAL_ADAPTER *Control)
{
    // 
    // Check each parameter first.
    //

    if ((Control->MetricType != (uint)-1) &&
        (Control->MetricType != VA->MetricType)) {
        //
        // We do not allow the MetricType to change.
        //
        return STATUS_INVALID_PARAMETER_2;
    }

    switch (Control->MetricType) {
    case METRIC_TYPE_RTT:
        if (Control->MetricParams.Rtt.Alpha != (uint)-1) {
            //
            // Alpha is interpreted by rtt.c as Alpha/MAXALPHA.
            // Since the result of this division should
            // not exceed 1, the maximum value of Alpha is MAXALPHA.
            //   
            if (Control->MetricParams.Rtt.Alpha > MAXALPHA)
                return STATUS_INVALID_PARAMETER_3;
        }

        if (Control->MetricParams.Rtt.ProbePeriod != (uint)-1) {
            //
            // The probe period should not be less than 100ms to
            // prevent overload.
            // 
            if (Control->MetricParams.Rtt.ProbePeriod < 100 * MILLISECOND)
                return STATUS_INVALID_PARAMETER_4;
        }

        if (Control->MetricParams.Rtt.HysteresisPeriod != (uint)-1) {
            //
            // The hysteresis period should not be less than 100ms to
            // prevent overload.
            // 
            if (Control->MetricParams.Rtt.HysteresisPeriod < 100 * MILLISECOND)
                return STATUS_INVALID_PARAMETER_5;
        }

        if (Control->MetricParams.Rtt.PenaltyFactor != (uint)-1) {
            //
            // The penalty factor should be at most 32 to avoid
            // overflow problems.
            // 
            if ((Control->MetricParams.Rtt.PenaltyFactor < 1) ||
                (Control->MetricParams.Rtt.PenaltyFactor > 32))
                return STATUS_INVALID_PARAMETER_6;
        }

        if (Control->MetricParams.Rtt.SweepPeriod != (uint)-1) {
            //
            // The sweep period should not be less than 1ms 
            // to prevent system overload. 
            // 
            if (Control->MetricParams.Rtt.SweepPeriod < 1 * MILLISECOND)
                return STATUS_INVALID_PARAMETER_7;
        }
        break;

    case METRIC_TYPE_PKTPAIR:
        if (Control->MetricParams.PktPair.Alpha != (uint)-1) {
            //
            // Alpha is interpreted as Alpha/MAXALPHA.
            // Since the result of this division should
            // not exceed 1, the maximum value of Alpha is MAXALPHA.
            //   
            if (Control->MetricParams.PktPair.Alpha > MAXALPHA)
                return STATUS_INVALID_PARAMETER_3;
        }

        if (Control->MetricParams.PktPair.ProbePeriod != (uint)-1) {
            //
            // The probe period should not be less than 100ms to
            // prevent system overload.
            // 
            if (Control->MetricParams.PktPair.ProbePeriod < 100 * MILLISECOND)
                return STATUS_INVALID_PARAMETER_4;
        }

        if (Control->MetricParams.PktPair.PenaltyFactor != (uint)-1) {
            //
            // The penalty factor should be at most 32 to avoid
            // overflow problems.
            // 
            if ((Control->MetricParams.PktPair.PenaltyFactor < 1) ||
                (Control->MetricParams.PktPair.PenaltyFactor > 32))
                return STATUS_INVALID_PARAMETER_5;
        }
        break;

    case METRIC_TYPE_ETX:
        if (Control->MetricParams.Etx.ProbePeriod != (uint)-1) {
            //
            // The probe period should not be less than 100ms. 
            // to prevent system overload.
            // 
            if (Control->MetricParams.Etx.ProbePeriod < 100 * MILLISECOND)
                return STATUS_INVALID_PARAMETER_3;
        }

        if (Control->MetricParams.Etx.LossInterval != (uint)-1) {
            //
            // The loss interval should not be less than 100ms. 
            // to prevent system overload. The upper bound is 1 minute
            // to prevent too much memory consumption.
            // 
            if ((Control->MetricParams.Etx.LossInterval < 100 * MILLISECOND) || 
                (Control->MetricParams.Etx.LossInterval > 1 * MINUTE))
                return STATUS_INVALID_PARAMETER_4;
        }

        if (Control->MetricParams.Etx.Alpha != (uint)-1) {
            //
            // Alpha is interpreted as Alpha/MAXALPHA.
            // Since the result of this division should
            // not exceed 1, the maximum value of Alpha is MAXALPHA.
            //   
            if (Control->MetricParams.Etx.Alpha > MAXALPHA)
                return STATUS_INVALID_PARAMETER_5;
        }

        if (Control->MetricParams.Etx.PenaltyFactor != (uint)-1) {
            //
            // The penalty factor should be at most 32 to avoid
            // overflow problems. 
            // 
            if ((Control->MetricParams.Etx.PenaltyFactor < 1) ||
                (Control->MetricParams.Etx.PenaltyFactor > 32))
                return STATUS_INVALID_PARAMETER_6;
        }
        break;

    case METRIC_TYPE_WCETT:
        if (Control->MetricParams.Wcett.ProbePeriod != (uint)-1) {
            //
            // The probe period should not be less than 100ms
            // to prevent system overload.
            // 
            if (Control->MetricParams.Wcett.ProbePeriod < 100 * MILLISECOND)
                return STATUS_INVALID_PARAMETER_3;
        }

        if (Control->MetricParams.Wcett.LossInterval != (uint)-1) {
            //
            // The loss interval should not be less than 100ms
            // to prevent system overload. The upper bound is 1 minute
            // to prevent too much memory consumption.
            // 
            if ((Control->MetricParams.Wcett.LossInterval < 100 * MILLISECOND) || 
                (Control->MetricParams.Wcett.LossInterval > 1 * MINUTE))
                return STATUS_INVALID_PARAMETER_4;
        }

        if (Control->MetricParams.Wcett.Alpha != (uint)-1) {
            //
            // Alpha is interpreted as Alpha/MAXALPHA.
            // Since the result of this division should
            // not exceed 1, the maximum value of Alpha is MAXALPHA.
            //   
            if (Control->MetricParams.Wcett.Alpha > MAXALPHA)
                return STATUS_INVALID_PARAMETER_5;
        }

        if (Control->MetricParams.Wcett.PenaltyFactor != (uint)-1) {
            //
            // The penalty factor should be at most 32 to avoid
            // overflow problems. 
            // 
            if ((Control->MetricParams.Wcett.PenaltyFactor < 1) ||
                (Control->MetricParams.Wcett.PenaltyFactor > 32))
                return STATUS_INVALID_PARAMETER_6;
        }

        if (Control->MetricParams.Wcett.Beta != (uint)-1) {
            //
            // Beta is interpreted as Beta/MAXALPHA.
            // Since the result of this division should
            // not exceed 1, the maximum value of Weight is MAXALPHA.
            //   
            if (Control->MetricParams.Wcett.Beta > MAXALPHA)
                return STATUS_INVALID_PARAMETER_7;
        }

        if (Control->MetricParams.Wcett.PktPairProbePeriod != (uint)-1) {
            //
            // The probe period should not be less than 100ms to
            // prevent system overload.
            // 
            if (Control->MetricParams.Wcett.PktPairProbePeriod < 100 * MILLISECOND)
                return STATUS_INVALID_PARAMETER_8;
        }

        if (Control->MetricParams.Wcett.PktPairMinOverProbes != (uint)-1) {
            //
            // The number of probes should be at least one.
            //
            if (Control->MetricParams.Wcett.PktPairMinOverProbes < 1)
                return STATUS_INVALID_PARAMETER_9;
        }
        break;

    case METRIC_TYPE_ULT:
        if (Control->MetricParams.Ult.UcastProbePeriod != (uint)-1) {
            //
            // The probe period should not be less than 100ms. 
            // to prevent system overload.
            // 
            if (Control->MetricParams.Ult.UcastProbePeriod < 100 * MILLISECOND)
                return STATUS_INVALID_PARAMETER_3;
        }

        if (Control->MetricParams.Ult.UcastLowThreshold != (uint)-1) {
            //
            // LowThreshold is interpreted as LowThreshold/MAXPERCENT.
            // Since the result of this division should
            // not exceed 1, the maximum value of LowThreshold is MAXPERCENT.
            //   
            if (Control->MetricParams.Ult.UcastLowThreshold > MAXPERCENT)
                return STATUS_INVALID_PARAMETER_5;
        }

        if (Control->MetricParams.Ult.UcastHighThreshold != (uint)-1) {
            //
            // HighThreshold is interpreted as HighThreshold/MAXPERCENT.
            // Since the result of this division should
            // not exceed 1, the maximum value of HighThreshold is MAXPERCENT.
            //   
            if (Control->MetricParams.Ult.UcastHighThreshold > MAXPERCENT)
                return STATUS_INVALID_PARAMETER_6;
        }

        if (Control->MetricParams.Ult.UcastLossPenalty != (uint)-1) {
            //
            // LossPenalty is interpreted as LossPenalty/MAXPERCENT.
            // Since the result of this division should
            // not exceed 1, the maximum value of LossPenalty is MAXPERCENT.
            //   
            if (Control->MetricParams.Ult.UcastLossPenalty > MAXPERCENT)
                return STATUS_INVALID_PARAMETER_7;
        }

        if (Control->MetricParams.Ult.BcastProbePeriod != (uint)-1) {
            //
            // The probe period should not be less than 100ms. 
            // to prevent system overload.
            // 
            if (Control->MetricParams.Ult.BcastProbePeriod < 100 * MILLISECOND)
                return STATUS_INVALID_PARAMETER_8;
        }

        if (Control->MetricParams.Ult.BcastLossThreshold != (uint)-1) {
            //
            // LossThreshold must be non-zero.
            //   
            if (Control->MetricParams.Ult.BcastLossThreshold == 0)
                return STATUS_INVALID_PARAMETER_9;
        }
        break;
    }

    //
    // Now, set the ones that we need to set.
    //

    if (Control->Snooping != (boolint) -1)
        VA->Snooping = Control->Snooping;

    if (Control->ArtificialDrop != (boolint) -1)
        VA->ArtificialDrop = Control->ArtificialDrop;

    if (Control->RouteFlapDampingFactor != (uint) -1)
        VA->RouteFlapDampingFactor = Control->RouteFlapDampingFactor;

#ifdef DDKBUILD
    //
    // No crypto support in DDK builds.
    //
    VA->Crypto = FALSE;
#else
    if (Control->Crypto != (boolint) -1)
        VA->Crypto = Control->Crypto;
#endif

    if (Control->LinkTimeout != (uint)-1)
        VA->LinkTimeout = (Time)Control->LinkTimeout * SECOND;

    switch (Control->MetricType) {
    case METRIC_TYPE_RTT:
        if (Control->MetricParams.Rtt.Random != (boolint) -1)
            VA->MetricParams.Rtt.Random = Control->MetricParams.Rtt.Random;
        if (Control->MetricParams.Rtt.OutIfOverride != (boolint) -1)
            VA->MetricParams.Rtt.OutIfOverride = Control->MetricParams.Rtt.OutIfOverride;
        if (Control->MetricParams.Rtt.Alpha != (uint)-1)
            VA->MetricParams.Rtt.Alpha = Control->MetricParams.Rtt.Alpha;
        if (Control->MetricParams.Rtt.ProbePeriod != (uint)-1)
            VA->MetricParams.Rtt.ProbePeriod = Control->MetricParams.Rtt.ProbePeriod;
        if (Control->MetricParams.Rtt.HysteresisPeriod != (uint)-1)
            VA->MetricParams.Rtt.HysteresisPeriod = Control->MetricParams.Rtt.HysteresisPeriod;
        if (Control->MetricParams.Rtt.PenaltyFactor != (uint)-1)
            VA->MetricParams.Rtt.PenaltyFactor = Control->MetricParams.Rtt.PenaltyFactor;
        if (Control->MetricParams.Rtt.SweepPeriod != (uint)-1)
            VA->MetricParams.Rtt.SweepPeriod = Control->MetricParams.Rtt.SweepPeriod;
        break;

    case METRIC_TYPE_PKTPAIR:
        if (Control->MetricParams.PktPair.Alpha != (uint)-1)
            VA->MetricParams.PktPair.Alpha = Control->MetricParams.PktPair.Alpha;
        if (Control->MetricParams.PktPair.ProbePeriod != (uint)-1)
            VA->MetricParams.PktPair.ProbePeriod = Control->MetricParams.PktPair.ProbePeriod;
        if (Control->MetricParams.PktPair.PenaltyFactor != (uint)-1)
            VA->MetricParams.PktPair.PenaltyFactor = Control->MetricParams.PktPair.PenaltyFactor;
        break;

    case METRIC_TYPE_ETX:
        if (Control->MetricParams.Etx.ProbePeriod != (uint)-1)
            VA->MetricParams.Etx.ProbePeriod = Control->MetricParams.Etx.ProbePeriod;
        if (Control->MetricParams.Etx.LossInterval != (uint)-1)
            VA->MetricParams.Etx.LossInterval = Control->MetricParams.Etx.LossInterval;
        if (Control->MetricParams.Etx.Alpha != (uint)-1)
            VA->MetricParams.Etx.Alpha = Control->MetricParams.Etx.Alpha;
        if (Control->MetricParams.Etx.PenaltyFactor != (uint)-1)
            VA->MetricParams.Etx.PenaltyFactor = Control->MetricParams.Etx.PenaltyFactor;
        break;

    case METRIC_TYPE_WCETT:
        if (Control->MetricParams.Wcett.ProbePeriod != (uint)-1)
            VA->MetricParams.Wcett.ProbePeriod = Control->MetricParams.Wcett.ProbePeriod;
        if (Control->MetricParams.Wcett.LossInterval != (uint)-1)
            VA->MetricParams.Wcett.LossInterval = Control->MetricParams.Wcett.LossInterval;
        if (Control->MetricParams.Wcett.Alpha != (uint)-1)
            VA->MetricParams.Wcett.Alpha = Control->MetricParams.Wcett.Alpha;
        if (Control->MetricParams.Wcett.PenaltyFactor != (uint)-1)
            VA->MetricParams.Wcett.PenaltyFactor = Control->MetricParams.Wcett.PenaltyFactor;
        if (Control->MetricParams.Wcett.Beta != (uint)-1)
            VA->MetricParams.Wcett.Beta = Control->MetricParams.Wcett.Beta;
        if (Control->MetricParams.Wcett.PktPairProbePeriod != (uint)-1)
            VA->MetricParams.Wcett.PktPairProbePeriod = Control->MetricParams.Wcett.PktPairProbePeriod;
        if (Control->MetricParams.Wcett.PktPairMinOverProbes != (uint)-1)
            VA->MetricParams.Wcett.PktPairMinOverProbes = Control->MetricParams.Wcett.PktPairMinOverProbes;
        break;

    case METRIC_TYPE_ULT:
        if (Control->MetricParams.Ult.UcastProbePeriod != (uint)-1)
            VA->MetricParams.Ult.UcastProbePeriod = Control->MetricParams.Ult.UcastProbePeriod;
        if (Control->MetricParams.Ult.UcastLowThreshold != (uint)-1)
            VA->MetricParams.Ult.UcastLowThreshold = Control->MetricParams.Ult.UcastLowThreshold;
        if (Control->MetricParams.Ult.UcastHighThreshold != (uint)-1)
            VA->MetricParams.Ult.UcastHighThreshold = Control->MetricParams.Ult.UcastHighThreshold;
        if (Control->MetricParams.Ult.UcastLossPenalty != (uint)-1)
            VA->MetricParams.Ult.UcastLossPenalty = Control->MetricParams.Ult.UcastLossPenalty;
        if (Control->MetricParams.Ult.BcastProbePeriod != (uint)-1)
            VA->MetricParams.Ult.BcastProbePeriod = Control->MetricParams.Ult.BcastProbePeriod;
        if (Control->MetricParams.Ult.BcastLossThreshold != (uint)-1)
            VA->MetricParams.Ult.BcastLossThreshold = Control->MetricParams.Ult.BcastLossThreshold;
        break;
    }

    return STATUS_SUCCESS;
}

//* MiniportInitialize
//
//  Called by NDIS to initialize an adapter.
//
NDIS_STATUS
MiniportInitialize(
    OUT PNDIS_STATUS OpenErrorStatus,
    OUT PUINT SelectedMediumIndex,
    IN PNDIS_MEDIUM MediumArray,
    IN UINT MediumArraySize,
    IN NDIS_HANDLE MiniportAdapterContext,
    IN NDIS_HANDLE WrapperConfigurationContext)
{
    MCL_CONTROL_VIRTUAL_ADAPTER Control;
    MiniportAdapter *VA;
    HANDLE KeyHandle;
    LARGE_INTEGER Timeout;
    KIRQL OldIrql;
    NDIS_STATUS Status;
    NTSTATUS NtStatus;
    uint i;

    LARGE_INTEGER RecoveryTimeout;
	
	Time now;

    UNREFERENCED_PARAMETER(OpenErrorStatus);
    UNREFERENCED_PARAMETER(WrapperConfigurationContext);

    //
    // We emulate 802.3. Figure out which index that is.
    //
    for (i = 0; ; i++) {
        if (i == MediumArraySize) {
            Status = NDIS_STATUS_UNSUPPORTED_MEDIA;
            goto ErrorReturn;
        }

        if (MediumArray[i] == NdisMedium802_3) {
            *SelectedMediumIndex = i;
            break;
        }
    }

    //
    // Allocate our virtual adapter structure.
    //
    VA = (MiniportAdapter *) ExAllocatePool(NonPagedPool, sizeof *VA);
    if (VA == NULL) {
        Status = NDIS_STATUS_RESOURCES;
        goto ErrorReturn;
    }

    if(debug) KdPrint(("MCL!MiniportInitialize -> VA %p\n", VA));

    RtlZeroMemory(VA, sizeof *VA);

	//ejr
	VA->acl.head = NULL ; 
	VA->adc = NULL; 
	VA->num_adcs = 0 ; 

    VA->Index = InterlockedIncrement((PLONG)&NextVirtualAdapterIndex);
    KeInitializeSpinLock(&VA->Lock);
	//ejr- for local recovery
	KeInitializeSpinLock(&VA->RecoveryLock);

    NeighborCacheInit(&VA->NC);
    PbackInit(VA);
    VA->MiniportHandle = MiniportAdapterContext;

    VA->Snooping = TRUE;
    VA->ArtificialDrop = FALSE;
    VA->RouteFlapDampingFactor = DEFAULT_ROUTE_FLAP_DAMPING_FACTOR;
    
	VA->LinkTimeout = (Time)1 * MINUTE;


    VA->LookAhead = 0;
    VA->MediumLinkSpeed = 60000; // ejr: 6 Mbps, was 10 Mbps.
    VA->MediumMinPacketLen = MINIPORT_HEADER_SIZE;
    VA->MediumMaxPacketLen = MINIPORT_HEADER_SIZE + MINIPORT_MAX_FRAME_SIZE;
    VA->MediumMacHeaderLen = MINIPORT_HEADER_SIZE;
    VA->MediumMaxFrameLen = MINIPORT_MAX_FRAME_SIZE;

    VA->MinVersionSeen = Version;
    VA->MaxVersionSeen = Version;

	//ejr
	VA->mlq.count = 0;
	VA->mlq.firstHolder = NULL; 
	VA->mlq.lastHolder = NULL; 

	VA->queueCounter = 1 ;


    //
    // Get our device objects.
    //
    NdisMGetDeviceProperty(MiniportAdapterContext,
                           &VA->PhysicalDeviceObject,
                           &VA->DeviceObject,
                           NULL, NULL, NULL);

    //
    // Allocate a packet pool.
    //
    NdisAllocatePacketPool(&Status, &VA->PacketPool,
                           PACKET_POOL_SZ, sizeof(PacketContext));
    if (Status != NDIS_STATUS_SUCCESS) {
        if(debug) KdPrint(("MCL!NdisAllocatePacketPool -> %x\n", Status));
        goto ErrorFreePool;
    }

    //
    // Allocate a buffer pool.
    //
    NdisAllocateBufferPool(&Status, &VA->BufferPool, PACKET_POOL_SZ);
    if (Status != NDIS_STATUS_SUCCESS) {
        if(debug) KdPrint(("MCL!NdisAllocateBufferPool -> %x\n", Status));
        goto ErrorFreePacketPool;
    }

    //
    // Give NDIS some information about our virtual adapter.
    //
    NdisMSetAttributesEx(MiniportAdapterContext, VA, 0,
                         NDIS_ATTRIBUTE_IGNORE_TOKEN_RING_ERRORS |
                         NDIS_ATTRIBUTE_IGNORE_PACKET_TIMEOUT |
                         NDIS_ATTRIBUTE_IGNORE_REQUEST_TIMEOUT |
                         NDIS_ATTRIBUTE_NO_HALT_ON_SUSPEND |
                         NDIS_ATTRIBUTE_SURPRISE_REMOVE_OK |
                         NDIS_ATTRIBUTE_USES_SAFE_BUFFER_APIS |
                         NDIS_ATTRIBUTE_DESERIALIZE,
                         0);

    //
    // Open our configuration information in the registry.
    //
    NtStatus = IoOpenDeviceRegistryKey(VA->PhysicalDeviceObject,
                                       PLUGPLAY_REGKEY_DRIVER,
                                       GENERIC_READ,
                                       &KeyHandle);
    if (! NT_SUCCESS(NtStatus)) {
        if(debug) KdPrint(("MCL!IoOpenDeviceRegistryKey -> %x\n", NtStatus));
        Status = NDIS_STATUS_FAILURE;
        goto ErrorFreeBufferPool;
    }

    //
    // Read the adapter's virtual link-layer address.
    //
    Status = MiniportReadOrCreateAdapterAddress(KeyHandle, VA->Address);
    if (Status != NDIS_STATUS_SUCCESS)
        goto ErrorCloseRegistry;

    if(debug) KdPrint(("MCL!MiniportInitialize(VA %p): "
             "addr %02x-%02x-%02x-%02x-%02x-%02x\n", VA,
             VA->Address[0], VA->Address[1], VA->Address[2],
             VA->Address[3], VA->Address[4], VA->Address[5]));

    //
    // Read the adapter's guid.
    //
    NtStatus = GetRegGuid(KeyHandle, L"NetCfgInstanceId", &VA->Guid);
    if (! NT_SUCCESS(NtStatus)) {
        if(debug) KdPrint(("MCL!GetRegGuid -> %x\n", NtStatus));
        Status = NDIS_STATUS_FAILURE;
        goto ErrorCloseRegistry;
    }

    ZwClose(KeyHandle);

    //
    // Read configuration parameters from the registry.
    //
    MiniportReadControl(VA, &Control);

    // 
    // Initialize Metric parameters.
    // The default MetricType is WCETT.
    //
    switch (Control.MetricType) {
    case METRIC_TYPE_HOP:
        VA->MetricType = METRIC_TYPE_HOP;
        VA->IsInfinite = MiniportIsInfinite;
        VA->IsBad = MiniportIsBad;
        VA->ConvMetric = MiniportConvMetric;
        VA->PathMetric = MiniportPathMetric;
        VA->InitLinkMetric = MiniportInitLinkMetric;
        VA->Penalize = MiniportPenalize;
        break;
    case METRIC_TYPE_RTT:
        VA->MetricType = METRIC_TYPE_RTT;
        RttInit(VA);
        break;
    case METRIC_TYPE_PKTPAIR:
        VA->MetricType = METRIC_TYPE_PKTPAIR;
        PktPairInit(VA);
        break;
    case METRIC_TYPE_ETX:
        VA->MetricType = METRIC_TYPE_ETX;
        EtxInit(VA);
        break;
    case METRIC_TYPE_ULT:
        VA->MetricType = METRIC_TYPE_ULT;
        UltInit(VA);
        break;
    case METRIC_TYPE_WCETT:
    default:
        VA->MetricType = METRIC_TYPE_WCETT;
        WcettInit(VA);
        break;
    }

    if (LQSR_KEY_PRESENT(Control.CryptoKeyMAC) &&
        LQSR_KEY_PRESENT(Control.CryptoKeyAES)) {
        //
        // Initialize the crypto parameters.
        // Crypto will be enabled, unless MiniportControl overrides.
        //
        VA->Crypto = TRUE;
        RtlCopyMemory(VA->CryptoKeyMAC, Control.CryptoKeyMAC,
                      sizeof VA->CryptoKeyMAC);
        RtlCopyMemory(VA->CryptoKeyAES, Control.CryptoKeyAES,
                      sizeof VA->CryptoKeyAES);
    }

    //
    // Modify our MAC key, using Version and MetricType.
    // If they disagree then nodes should fail to communicate.
    //
    CryptoKeyMACModify(VA, VA->CryptoKeyMAC);

    //
    // Update the virtual adapter using the parameters.
    // This can modify VA->Crypto but not MetricType,
    // CryptoKeyMAC, or CryptoKeyAES.
    //
    (void) MiniportControl(VA, &Control);

    //
    // Now we have our address, so we can set up our link cache.
    //
    VA->LC = LinkCacheAllocate(4, VA->Address);
    if (VA->LC == NULL)
        goto ErrorFreeBufferPool;

    //
    // Set up our request table.
    //
    Status = ReqTableNew(VA, 16);
    if (Status != NDIS_STATUS_SUCCESS)
        goto ErrorFreeLinkCache;

    //
    // Set up our send buffer.
    //
    Status = SendBufNew(VA, SEND_BUF_SZ);
    if (Status != NDIS_STATUS_SUCCESS)
        goto ErrorFreeRequestTable;

    //
    // Set up our maintenance buffer.
    //
    VA->MaintBuf = MaintBufNew();
    if (VA->MaintBuf == NULL)
        goto ErrorFreeSendBuffer;

    //
    // Set up our forwarding queue.
    //
    VA->ForwardLast = &VA->ForwardList;
    ASSERT(VA->ForwardList == NULL);
    ASSERT(VA->ForwardTime == 0);
	

	//
	// Set up the local recovery cache
	//
	LocalRecoveryInitialize(VA);

	//junk, used to enforce SendACKComplete != PbackSendComplete
	VA->junk = 0 ;
	VA->etx_seqno = 0 ;
	ASSERT(ACKSendComplete != PbackSendComplete);

	//
    // Initialize with NDIS as a protocol. This will bind
    // our virtual adapter to physical adapters,
    // based on the protocol name that we supply here.
    //
    Status = ProtocolInit(L"msmcltp", VA);
    if (Status != NDIS_STATUS_SUCCESS)
        goto ErrorFreeMaintBuffer;

    //
    // We have an aperiodic timeout function,
    // with a maximum timeout of MINIPORT_TIMEOUT.
    //
    VA->TimeoutMinLoops = MAXULONG;
    KeInitializeDpc(&VA->TimeoutDpc, MiniportTimeout, VA);
    KeInitializeTimer(&VA->Timer);
    Timeout.QuadPart = - (LONGLONG) (MINIPORT_TIMEOUT * 10000);
    VA->Timeout = KeQueryInterruptTime() - Timeout.QuadPart;
    KeSetTimer(&VA->Timer, Timeout, &VA->TimeoutDpc);


	
	//ejr
	// We will use this timer function to deal with all of our 
	// local recover packets.
	//
	KeInitializeDpc(&VA->RecoveryTimeoutDpc, MiniportRecoveryTimeout, VA);
    KeInitializeTimer(&VA->RecoveryTimer);
    RecoveryTimeout.QuadPart = - (LONGLONG) (MINIPORT_RECOVERY_TIMEOUT * 10000);
	now = KeQueryInterruptTime();
    VA->RecoveryTimeout = now - RecoveryTimeout.QuadPart;
    KeSetTimer(&VA->RecoveryTimer, RecoveryTimeout, &VA->RecoveryTimeoutDpc);


    //
    // Add the virtual adapter to our list.
    //
    KeAcquireSpinLock(&MiniportAdapters.Lock, &OldIrql);
    VA->Prev = &MiniportAdapters.VirtualAdapters;
    VA->Next = MiniportAdapters.VirtualAdapters;
    if (MiniportAdapters.VirtualAdapters != NULL)
        MiniportAdapters.VirtualAdapters->Prev = &VA->Next;
    MiniportAdapters.VirtualAdapters = VA;
    KeReleaseSpinLock(&MiniportAdapters.Lock, OldIrql);
   

#if RATE_LIMIT_FILE
	ProtocolCreateRateFile(MAX_DATA_PKTS_PER_INTERVAL); 
#endif

#if STOP_LINK_METRIC
	VA->start_link_time = now;
#endif


    return NDIS_STATUS_SUCCESS;

ErrorCloseRegistry:
    ZwClose(KeyHandle);
    goto ErrorFreeBufferPool;

ErrorFreeMaintBuffer:
    MaintBufFree(VA);
ErrorFreeSendBuffer:
    SendBufFree(VA);
ErrorFreeRequestTable:
    ReqTableFree(VA);
ErrorFreeLinkCache:
    LinkCacheFree(VA);
ErrorFreeBufferPool:
    NdisFreeBufferPool(VA->BufferPool);
ErrorFreePacketPool:
    NdisFreePacketPool(VA->PacketPool);
ErrorFreePool:
    ExFreePool(VA);
ErrorReturn:
    if(debug) KdPrint(("MCL!MiniportInitialize -> %x\n", Status));
    return Status;
}

NDIS_OID MiniportSupportedOidArray[] =
{
    OID_GEN_SUPPORTED_LIST,
    OID_GEN_HARDWARE_STATUS,
    OID_GEN_MEDIA_SUPPORTED,
    OID_GEN_MEDIA_IN_USE,
    OID_GEN_MAXIMUM_LOOKAHEAD,
    OID_GEN_CURRENT_LOOKAHEAD,
    OID_GEN_MAXIMUM_FRAME_SIZE,
    OID_GEN_MAC_OPTIONS,
    OID_GEN_LINK_SPEED,
    OID_GEN_TRANSMIT_BUFFER_SPACE,
    OID_GEN_RECEIVE_BUFFER_SPACE,
    OID_GEN_MAXIMUM_TOTAL_SIZE,
    OID_GEN_TRANSMIT_BLOCK_SIZE,
    OID_GEN_RECEIVE_BLOCK_SIZE,
    OID_GEN_VENDOR_ID,
    OID_GEN_VENDOR_DESCRIPTION,
    OID_GEN_CURRENT_PACKET_FILTER,
    OID_GEN_DRIVER_VERSION,
    OID_GEN_MEDIA_CONNECT_STATUS,
    OID_GEN_MAXIMUM_SEND_PACKETS,

    OID_802_3_PERMANENT_ADDRESS,
    OID_802_3_CURRENT_ADDRESS,
    OID_802_3_MULTICAST_LIST,           // Set only?
    OID_802_3_MAXIMUM_LIST_SIZE,

    OID_802_3_RCV_ERROR_ALIGNMENT,
    OID_802_3_XMIT_ONE_COLLISION,
    OID_802_3_XMIT_MORE_COLLISIONS,

    OID_PNP_CAPABILITIES,
    OID_PNP_SET_POWER,                  // Set only.
    OID_PNP_QUERY_POWER,
};

uchar MiniportVendorDescription[] = "MCL Miniport Driver";
uchar MiniportVendorId[3] = {0xFF, 0xFF, 0xFF};

NDIS_PNP_CAPABILITIES MiniportPnpCapabilities = {
    0, 
    NdisDeviceStateUnspecified,
    NdisDeviceStateUnspecified,
    NdisDeviceStateUnspecified
};

//* MiniportQueryInformation
//
//  Called by NDIS to retrieve information about our virtual adapter.
//
NDIS_STATUS
MiniportQueryInformation(
    IN NDIS_HANDLE MiniportAdapterContext,
    IN NDIS_OID Oid,
    IN PVOID InformationBuffer,
    IN ULONG InformationBufferLength,
    OUT PULONG BytesWritten,
    OUT PULONG BytesNeeded)
{
    MiniportAdapter *VA = (MiniportAdapter *) MiniportAdapterContext;
    uint GenericUint;
    ushort GenericUshort;
    void *Data;
    uint DataLength;

    //
    // Prepare to return a uint - this is the common case.
    //
    Data = &GenericUint;
    DataLength = sizeof GenericUint;

    //
    // If you change this switch, be sure to update MiniportSupportedOidArray.
    //
    switch (Oid) {
    case OID_GEN_SUPPORTED_LIST:
        Data = MiniportSupportedOidArray;
        DataLength = sizeof MiniportSupportedOidArray;
        break;

    case OID_GEN_HARDWARE_STATUS:
        GenericUint = NdisHardwareStatusReady;
        break;

    case OID_GEN_MEDIA_SUPPORTED:
    case OID_GEN_MEDIA_IN_USE:
        GenericUint = VA->Medium;
        break;

    case OID_GEN_MAXIMUM_LOOKAHEAD:
        GenericUint = MINIPORT_MAX_LOOKAHEAD;
        break;

    case OID_GEN_CURRENT_LOOKAHEAD:
        GenericUint = VA->LookAhead;
        break;

    case OID_GEN_MAXIMUM_FRAME_SIZE:
        GenericUint = VA->MediumMaxFrameLen;
        break;

    case OID_GEN_MAC_OPTIONS:
        GenericUint = NDIS_MAC_OPTION_NO_LOOPBACK;
        break;

    case OID_GEN_LINK_SPEED:
        GenericUint = VA->MediumLinkSpeed;
        break;

    case OID_GEN_TRANSMIT_BUFFER_SPACE:
    case OID_GEN_RECEIVE_BUFFER_SPACE:
    case OID_GEN_MAXIMUM_TOTAL_SIZE:
        GenericUint = VA->MediumMaxPacketLen;
        break;

    case OID_GEN_TRANSMIT_BLOCK_SIZE:
    case OID_GEN_RECEIVE_BLOCK_SIZE:
        GenericUint = 1;
        break;

    case OID_GEN_VENDOR_ID:
        Data = MiniportVendorId;
        DataLength = sizeof MiniportVendorId;
        break;

    case OID_GEN_VENDOR_DESCRIPTION:
        Data = MiniportVendorDescription;
        DataLength = sizeof MiniportVendorDescription;
        break;

    case OID_GEN_CURRENT_PACKET_FILTER:
        GenericUint = VA->PacketFilter;
        break;

    case OID_GEN_DRIVER_VERSION:
        GenericUshort = (NDIS_MINIPORT_MAJOR_VERSION << 8) + NDIS_MINIPORT_MINOR_VERSION;
        Data = &GenericUshort;
        DataLength = sizeof GenericUshort;
        break;

    case OID_GEN_MEDIA_CONNECT_STATUS:
        if (VA->PhysicalAdapters != NULL)
            GenericUint = NdisMediaStateConnected;
        else
            GenericUint = NdisMediaStateDisconnected;
        break;

    case OID_GEN_MAXIMUM_SEND_PACKETS:
        GenericUint = MAX_RECV_QUEUE_SIZE;
        break;

    case OID_802_3_PERMANENT_ADDRESS:
    case OID_802_3_CURRENT_ADDRESS:
        Data = VA->Address;
        DataLength = sizeof VA->Address;
        break;

    case OID_802_3_MAXIMUM_LIST_SIZE:
        GenericUint = MINIPORT_MAX_MULTICAST_ADDRESS;
        break;

    case OID_802_3_RCV_ERROR_ALIGNMENT:
    case OID_802_3_XMIT_ONE_COLLISION:
    case OID_802_3_XMIT_MORE_COLLISIONS:
        GenericUint = 0;
        break;

    case OID_PNP_CAPABILITIES:
        Data = &MiniportPnpCapabilities;
        DataLength = sizeof MiniportPnpCapabilities;
        break;

    case OID_PNP_QUERY_POWER:
        DataLength = 0;
        break;

    default:
        if(debug) KdPrint(("MCL!MiniportQueryInformation(VA %p OID %x)\n", VA, Oid));

        //
        // The following OIDs are called in practice,
        // but we do not support them.
        //
    case OID_GEN_VENDOR_DRIVER_VERSION:
    case OID_GEN_SUPPORTED_GUIDS:
    case OID_GEN_PHYSICAL_MEDIUM:
    case OID_GEN_MEDIA_CAPABILITIES:
    case OID_FFP_SUPPORT:
    case OID_TCP_TASK_OFFLOAD:
    case 0xff54554e: // OID_CUSTOM_TUNMP_INSTANCE_ID
    case OID_GEN_RCV_NO_BUFFER:
        return NDIS_STATUS_NOT_SUPPORTED;
    }

    *BytesNeeded = DataLength;
    if (DataLength > InformationBufferLength)
        return NDIS_STATUS_BUFFER_TOO_SHORT;

    RtlCopyMemory(InformationBuffer, Data, DataLength);
    *BytesWritten = DataLength;
    return NDIS_STATUS_SUCCESS;
}

//* MiniportSetInformation
//
//  Called by NDIS to specify information for our virtual adapter.
//
NDIS_STATUS
MiniportSetInformation(
    IN NDIS_HANDLE MiniportAdapterContext,
    IN NDIS_OID Oid,
    IN PVOID InformationBuffer,
    IN ULONG InformationBufferLength,
    OUT PULONG BytesRead,
    OUT PULONG BytesNeeded)
{
    MiniportAdapter *VA = (MiniportAdapter *) MiniportAdapterContext;

    switch (Oid) {
    case OID_GEN_CURRENT_LOOKAHEAD: {
        uint LookAhead;

        if (InformationBufferLength != sizeof LookAhead) {
            *BytesNeeded = sizeof LookAhead;
            return NDIS_STATUS_INVALID_LENGTH;
        }

        LookAhead = * (uint UNALIGNED *) InformationBuffer;
        *BytesRead = sizeof LookAhead;

        if (LookAhead > MINIPORT_MAX_LOOKAHEAD)
            return NDIS_STATUS_INVALID_DATA;

        if (LookAhead > VA->LookAhead)
            VA->LookAhead = LookAhead;

        return NDIS_STATUS_SUCCESS;
    }

    case OID_PNP_SET_POWER: {
        NDIS_DEVICE_POWER_STATE PowerState;

        if (InformationBufferLength != sizeof PowerState) {
            *BytesNeeded = sizeof PowerState;
            return NDIS_STATUS_INVALID_LENGTH;
        }

        PowerState = * (NDIS_DEVICE_POWER_STATE UNALIGNED *) InformationBuffer;
        *BytesRead = sizeof PowerState;

        if(debug) KdPrint(("MCL!MiniportSetInformation(VA %p SetPower %x)\n",
                 VA, PowerState));

        //
        // Should we pend until any outstanding operations are completed,
        // when transitioning to a sleep state?
        //

        if (PowerState == NdisDeviceStateD0) {
            //
            // We are waking up after being asleep.
            // The environment has likely changed, so flush our link cache.
            // The neighbor cache does not need to be flushed.
            //
            LinkCacheFlush(VA);
        }

        return NDIS_STATUS_SUCCESS;
    }

    case OID_802_3_MULTICAST_LIST:
    case OID_GEN_CURRENT_PACKET_FILTER:
        *BytesRead = InformationBufferLength;
        return NDIS_STATUS_SUCCESS;

    default:
        if(debug) KdPrint(("MCL!MiniportSetInformation(VA %p OID %x)\n", VA, Oid));

        //
        // The following OIDs are called in practice,
        // but we do not support them.
        //
    case OID_GEN_TRANSPORT_HEADER_OFFSET:
    case OID_GEN_NETWORK_LAYER_ADDRESSES:
    case OID_PNP_ADD_WAKE_UP_PATTERN:
    case OID_PNP_REMOVE_WAKE_UP_PATTERN:
        return NDIS_STATUS_NOT_SUPPORTED;
    }
}

//* ForwardRequestP
//
//  Should we forward this Route Request?
//  Returns TRUE if we should.
//
static boolint
ForwardRequestP(
    MiniportAdapter *VA,
    SRPacket *srp)
{
    InternalRouteRequest *RR;
    uint Hops;
    uint i;

    RR = srp->req;
    ASSERT(RR != NULL);

    //
    // Is the path too long? If so we can not forward.
    //
    Hops = ROUTE_REQUEST_HOPS(RR->opt.optDataLen);
    ASSERT(Hops != 0);
    if (Hops >= MAX_SR_LEN)
        return FALSE;

    //
    // Make sure we're not on the path yet.
    // If so we can not forward.
    //
    for (i = 0; i < Hops; i++)
        if (VirtualAddressEqual(RR->opt.hopList[i].addr, VA->Address))
            return FALSE;

    //
    // Check if we've forwarded this request id before.
    // If so we should not forward.
    //
    if (ReqTableSuppress(VA,
                         RR->opt.hopList[0].addr,
                         RR->opt.targetAddress,
                         RR->opt.identification))
        return FALSE;

    return TRUE;
}

//* MiniportControlSendComplete
//
//  Completes the transmission of a Source-Routed control packet,
//  such as a Route Request, Route Reply or Route Error.
//
static void
MiniportControlSendComplete(
    MiniportAdapter *VA,
    SRPacket *srp,
    NDIS_STATUS Status)
{
    UNREFERENCED_PARAMETER(VA);
    UNREFERENCED_PARAMETER(Status);

    SRPacketFree(srp);
}

//* MiniportSendRouteReplyFromSR
//
//  Sends a Route Reply in response to a Source Route,
//  to propogate link metric information back to the originator.
//
//  Does not consume srp or srp->Packet.
//
static void
MiniportSendRouteReplyFromSR(
    MiniportAdapter *VA,
    const SRPacket *SRP)
{
    InternalRouteReply *RR;
    uint Hops;

    ASSERT(SRP->sr != NULL);
    Hops = SOURCE_ROUTE_HOPS(SRP->sr->opt.optDataLen);
    ASSERT(VirtualAddressEqual(SRP->sr->opt.hopList[Hops - 1].addr,
                               VA->Address));

    //
    // Allocate the Route Reply option.
    //
    RR = ExAllocatePool(NonPagedPool, sizeof *RR + sizeof(SRAddr)*MAX_SR_LEN);
    if (RR == NULL)
        return;

    //
    // Initialize the Route Reply option.
    //
    RtlZeroMemory(RR, sizeof *RR + sizeof(SRAddr)*MAX_SR_LEN);
    RR->opt.optionType = LQSR_OPTION_TYPE_REPLY;
    RR->opt.optDataLen = ROUTE_REPLY_LEN(Hops);

    RtlCopyMemory(RR->opt.hopList, SRP->sr->opt.hopList,
                  Hops * sizeof(SRAddr));

    //
    // Send the Route Reply option.
    // PbackSendOption will squash duplicate Route Replies.
    //
    PbackSendOption(VA, SRP->sr->opt.hopList[0].addr,
                    (InternalOption *) RR,
                    PBACK_ROUTE_REPLY_SR_TIMEOUT);
}

//* MiniportSendRouteReply
//
//  Sends a Route Reply in response to a Route Request.
//
//  Does not consume srp or srp->Packet.
//
static void
MiniportSendRouteReply(
    MiniportAdapter *VA,
    const SRPacket *SRP)
{
    InternalRouteReply *RR;
    uint Hops;

    InterlockedIncrement((PLONG)&VA->CountXmitRouteReply);

    ASSERT(SRP->req != NULL);
    ASSERT(VirtualAddressEqual(SRP->req->opt.targetAddress, VA->Address));
    Hops = ROUTE_REQUEST_HOPS(SRP->req->opt.optDataLen);
    ASSERT(VirtualAddressEqual(SRP->req->opt.hopList[Hops - 1].addr,
                               VA->Address));
    SRP->req->opt.hopList[Hops - 1].outif = 0;

    //
    // Allocate the Route Reply option.
    //
    RR = ExAllocatePool(NonPagedPool, sizeof *RR + sizeof(SRAddr)*MAX_SR_LEN);
    if (RR == NULL)
        return;

    //
    // Initialize the Route Reply option.
    //
    RtlZeroMemory(RR, sizeof *RR + sizeof(SRAddr)*MAX_SR_LEN);
    RR->opt.optionType = LQSR_OPTION_TYPE_REPLY;
    RR->opt.optDataLen = ROUTE_REPLY_LEN(Hops);

    RtlCopyMemory(RR->opt.hopList, SRP->req->opt.hopList,
                  Hops * sizeof(SRAddr));

    //
    // Send the Route Reply option.
    //
    PbackSendOption(VA, SRP->req->opt.hopList[0].addr,
                    (InternalOption *) RR,
                    PBACK_ROUTE_REPLY_TIMEOUT);
}

//* MiniportForwardCleanup
//
//  Cleans up the forwarding queue.
//
static void
MiniportForwardCleanup(MiniportAdapter *VA)
{
    SRPacket *SRP;

    while ((SRP = VA->ForwardList) != NULL) {
        VA->ForwardList = SRP->Next;

        (*SRP->TransmitComplete)(VA, SRP, NDIS_STATUS_FAILURE);
    }
}

//* MiniportForwardRequest
//
//  Implements a rate-limiting queue in front of ProtocolForwardRequest.
//
static void
MiniportForwardRequest(
    MiniportAdapter *VA,
    SRPacket *srp)
{
    Time Now;
    KIRQL OldIrql;

    KeAcquireSpinLock(&VA->Lock, &OldIrql);
    Now = KeQueryInterruptTime();

    //
    // Can we forward this packet now, or must we queue it?
    //
    if (Now < VA->ForwardTime) {
        if (VA->ForwardNum >= MAX_FORWARD_QUEUE) {
            //
            // The queue is too long, so drop this packet.
            //
            VA->CountForwardDrop++;
            KeReleaseSpinLock(&VA->Lock, OldIrql);

            (*srp->TransmitComplete)(VA, srp, NDIS_STATUS_FAILURE);
        }
        else {
            //
            // Queue the packet.
            //
            VA->CountForwardQueue++;
            *VA->ForwardLast = srp;
            VA->ForwardLast = &srp->Next;
            ASSERT(srp->Next == NULL);

            if (++VA->ForwardNum > VA->ForwardMax)
                VA->ForwardMax = VA->ForwardNum;

            //
            // Possibly reschedule MiniportTimeout to run earlier,
            // so the waiting packet is sent asap.
            //
            if (VA->ForwardTime < VA->Timeout)
                MiniportRescheduleTimeoutHelper(VA, Now, VA->ForwardTime);

            KeReleaseSpinLock(&VA->Lock, OldIrql);
        }
    }
    else {
        //
        // We will forward it now.
        // Calculate time at which we can next forward.
        //
        VA->CountForwardFast++;
        VA->ForwardTime = Now + MIN_BROADCAST_GAP;
        KeReleaseSpinLock(&VA->Lock, OldIrql);

        ProtocolForwardRequest(VA, srp);
    }
}

//* MiniportForwardDelayedRequests
//
//  Checks if we can forward any queued requests.
//
static void
MiniportForwardDelayedRequests(
    MiniportAdapter *VA,
    Time Now)
{
    SRPacket *srp;
    KIRQL OldIrql;

    KeAcquireSpinLock(&VA->Lock, &OldIrql);
    if ((VA->ForwardList != NULL) &&
        (VA->ForwardTime <= Now)) {
        //
        // Dequeue the packet.
        //
        srp = VA->ForwardList;
        VA->ForwardList = srp->Next;
        if (VA->ForwardList == NULL)
            VA->ForwardLast = &VA->ForwardList;
        VA->ForwardNum--;

        //
        // Calculate time at which we can next forward.
        // We can update VA->Timeout directly instead of
        // using MiniportRescheduleTimeoutHelper because we are
        // called from MiniportTimeout.
        //
        VA->ForwardTime = Now + MIN_BROADCAST_GAP;
        if (VA->ForwardTime < VA->Timeout)
            VA->Timeout = VA->ForwardTime;
    }
    else {
        //
        // No packet to forward.
        //
        srp = NULL;
    }
    KeReleaseSpinLock(&VA->Lock, OldIrql);

    if (srp != NULL) {
        //
        // Forward the waiting packet.
        //
        ProtocolForwardRequest(VA, srp);
    }
}

//* MiniportGenerateRouteRequest
//
//  Generates a Route Request.
//
//  Does NOT consume srp or srp->Packet,
//  that is the responsibility of the completion function.
//
static void
MiniportGenerateRouteRequest(
    MiniportAdapter *VA,
    SRPacket *srp,
    const VirtualAddress Target,
    LQSRReqId Identifier,
    void (*Complete)(MiniportAdapter *VA, SRPacket *srp, NDIS_STATUS Status))
{
    InternalRouteRequest *RR;

    ASSERT((srp->req == NULL) && (srp->sr == NULL));

    //
    // Create and initialize a Route Request option.
    //

    RR = ExAllocatePool(NonPagedPool, sizeof *RR + sizeof(SRAddr)*MAX_SR_LEN);
    if (RR == NULL) {
        (*Complete)(VA, srp, NDIS_STATUS_RESOURCES);
        return;
    }

    srp->req = RR;
    RtlZeroMemory(RR, sizeof *RR + sizeof(SRAddr)*MAX_SR_LEN);
    RR->opt.optionType = LQSR_OPTION_TYPE_REQUEST;
    RR->opt.optDataLen = ROUTE_REQUEST_LEN(1);

    RR->opt.identification = Identifier;
    RtlCopyMemory(RR->opt.targetAddress, Target, SR_ADDR_LEN);

    RtlCopyMemory(RR->opt.hopList[0].addr, VA->Address, SR_ADDR_LEN);
    ASSERT(RR->opt.hopList[0].inif == 0);
    // RR->opt.hopList[0].outif initialized in ProtocolForwardRequest.

    if (VA->MetricType != METRIC_TYPE_HOP) {
        uint Size;

        //
        // Add a Link Info option to the packet.
        // We give LinkCacheCreateLI an upper bound for the option size.
        //
        Size = PbackPacketSize(srp);
        ASSERT(Size <= PROTOCOL_MIN_FRAME_SIZE);
        srp->LinkInfo = LinkCacheCreateLI(VA, PROTOCOL_MIN_FRAME_SIZE - Size);
        InterlockedWrite64((LONGLONG *)&VA->LastLinkInfo,
                           KeQueryInterruptTime());
    }

    //
    // Send the packet.
    //
    srp->TransmitComplete = Complete;
    MiniportForwardRequest(VA, srp);
}

//* MiniportSendViaRouteRequest
//
void
MiniportSendViaRouteRequest(
    MiniportAdapter *VA,
    SRPacket *SRP,
    void (*Complete)(MiniportAdapter *VA, SRPacket *SRP, NDIS_STATUS Status))
{
    LQSRReqId Identifier;

    Identifier = ReqTableIdentifier(VA, SRP->Dest);
	MiniportGenerateRouteRequest(VA, SRP,
                                 SRP->Dest, Identifier,
                                 Complete);
}

//* ejr- MiniportRetransmitSendComplete
//
//  Called when a MiniportSendPacket operation has finished
//  and the packet must be returned to NDIS.
//
void
MiniportRetransmitSendComplete(
    MiniportAdapter *VA,
    SRPacket *srp,
    NDIS_STATUS Status)
{
    //recoverycache.c should take care of everything for us.
    UNREFERENCED_PARAMETER(Status);
    UNREFERENCED_PARAMETER(srp);
    UNREFERENCED_PARAMETER(VA);
    
}



//* MiniportSendComplete
//
//  Called when a MiniportSendPacket operation has finished
//  and the packet must be returned to NDIS.
//
void
MiniportSendComplete(
    MiniportAdapter *VA,
    SRPacket *srp,
    NDIS_STATUS Status)
{
    //NDIS_PACKET *Packet = srp->Packet;

    UNREFERENCED_PARAMETER(Status);
    //UNREFERENCED_PARAMETER(Packet);
    
	NdisMSendComplete(VA->MiniportHandle,
					srp->OriginalPacket,
					srp->ForwardStatus);

	srp->OriginalPacket = NULL ;

	SRPacketFree(srp);
}



//* MiniportFreePacket
//
//  Helper function for MiniportTransmitPacket.
//
static void
MiniportFreePacket(
	MiniportAdapter *VA,
    ProtocolAdapter *PA,
    NDIS_PACKET *Packet)
{
	
	//ejr
	// support for packet copying
	//
	NDIS_BUFFER *Buffer;
	
    UNREFERENCED_PARAMETER(VA);
	
    Buffer = NdisFirstBuffer(Packet);
    ASSERT((Buffer != NULL) && (Buffer->Next == NULL));

    NdisFreeBuffer(Buffer);
		

    NdisFreePacket(Packet);


	//old way...
    UNREFERENCED_PARAMETER(PA);
    //NdisFreePacketClone(Packet);
}


//* MiniportTransmitPacket
//
//  Send an LQSR packet via our virtual link.
//
void
MiniportTransmitPacket(
    MiniportAdapter *VA,
    NDIS_PACKET *Packet)
{
    SRPacket *srp;
    EtherHeader UNALIGNED *Ether;
    uint Length;
    NDIS_STATUS Status;
	NDIS_PACKET *tmpPacket;
	uint Size ;
	
	if(debug) KdPrint(("Inside MiniportTransmitPacket\n")); 

    //
    // Check that the packet is not too small or too large.
    // The bridge driver, for example, does not respect our MTU.
    // The most important checks are in ProtocolTransmit and
    // MiniportReceiveLocally (ensuring that we do not send down
    // or pass up a bad packet), but this check here is needed
    // to avoid asserts elsewhere in our code.
    //
    NdisQueryPacketLength(Packet, &Length);
    if (Length < VA->MediumMinPacketLen) {
        InterlockedIncrement((PLONG)&VA->CountXmitSmall);
        Status = NDIS_STATUS_INVALID_LENGTH;
		if(debug) KdPrint(("MiniportTransmit: Packet length too small\n"));
        goto ErrorReturn;
    }
    else if (Length > VA->MediumMaxPacketLen) {
        InterlockedIncrement((PLONG)&VA->CountXmitLarge);
        Status = NDIS_STATUS_INVALID_LENGTH;
		if(debug) KdPrint(("MiniportTransmit: Packet length too large\n"));
        goto ErrorReturn;
    }

	
	InterlockedIncrement((PLONG)&VA->CountXmitLocally);

    //
    // Allocate an SRPacket structure for this packet.
    //
    srp = ExAllocatePool(NonPagedPool, sizeof *srp);
    if (srp == NULL) {
        Status = NDIS_STATUS_RESOURCES;
		if(debug) KdPrint(("MiniportTransmit: Can't alloc from packet pool\n"));
        goto ErrorReturn;
    }

    RtlZeroMemory(srp, sizeof *srp);
	srp->dataPtr = NULL; 
	srp->VA = VA ;

	//ejr- we put a copy of the original packet here, so we can remove it later.
	// This whole process needs to be optimized. Currently, this is a hack.
	Status = CopyWholePacket(VA, Packet, &srp->OriginalPacket, VA->PacketPool, VA->BufferPool, &srp->dataPtr); 
	if (Status != NDIS_STATUS_SUCCESS) {
		//if CopyWholePacket fails, then srp->dataPtr is already taken care of
		if(debug) KdPrint(("Error in MiniportTransmitPacket0\n")); 
		NdisMSendComplete(VA->MiniportHandle, Packet, Status);
		srp->OriginalPacket = NULL ;
		SRPacketFree(srp);
		return ;
	}

	tmpPacket = Packet ;
	Packet = srp->OriginalPacket; 
	srp->OriginalPacket = tmpPacket;

	srp->copiedPacket = Packet; 

    
	//
    // Encrypt the packet payload (including the EType).
    // This generates an IV.
    Status = CryptoEncryptPacket(VA, Packet,
                                 sizeof(EtherHeader) - sizeof(ushort),
                                 srp->IV, &srp->Packet);

	if (Status != NDIS_STATUS_SUCCESS) {
		if(debug) KdPrint(("MiniportTransmit: CryptoEncryptPacket failed\n"));
		//if this fails, we are in a bind: srp->Packet will be 
		// null, so the FreePacket functions won't be called. We
		// need to take care of srp->copiedPacket and srp->dataPtr here.
		srp->Packet = NULL; 
		if (srp->dataPtr != NULL) {
			ExFreePool(srp->dataPtr); 
			srp->dataPtr = NULL; 
		}
		if (srp->copiedPacket != NULL) {
				NDIS_BUFFER *Buffer;
				Buffer = NdisFirstBuffer(srp->copiedPacket);
				ASSERT((Buffer != NULL) && (Buffer->Next == NULL));

				NdisFreeBuffer(Buffer);
				NdisFreePacket(srp->copiedPacket);
				srp->copiedPacket = NULL; 
		}
        goto ErrorFreeSRP;
	}
    //
    // MiniportFreePacket will free the encrypted packet
    // using NdisFreePacketClone.
    //
    srp->PA = NULL;
	srp->FreePacket = NULL; 
	srp->FreePacket2 = MiniportFreePacket;

    //
    // This is a dummy value, so srp->PacketLength vs srp->PayloadOffset
    // comparisons will work properly. We do not need the actual
    // length of srp->Packet.
    //
    ASSERT(srp->PayloadOffset == 0);
    srp->PacketLength = 1;

    //
    // MiniportSendComplete will complete the original packet.
    //
    PC(srp->Packet)->OrigPacket = Packet;

    //
    // Initialize the virtual source & destination addresses
    // in the SRPacket. If our caller supplies a bogus
    // source address, we ignore it. This can happen,
    // for example when the IPv4 stack detects a duplicate address
    // it sends an ARP with the "wrong" source address.
    //
	Ether = NdisFirstBuffer(Packet)->MappedSystemVa;

	//PRINT_ADDRESS("Sending packet",  srp->Source, srp->Dest, Ether->Type) ; 

	RtlCopyMemory(srp->Dest, Ether->Dest, SR_ADDR_LEN);
    RtlCopyMemory(srp->Source, VA->Address, SR_ADDR_LEN);
	
    if (IEEE_802_ADDR_IS_BROADCAST(srp->Dest)) {
        //
        // We piggy-back multicast/broadcast packets on a Route Request.
        //
		if(debug) KdPrint(("Sending out Bcast packet\n")); 
        InterlockedIncrement((PLONG)&VA->CountXmitMulticast);
        MiniportSendViaRouteRequest(VA, srp, MiniportSendComplete);
    }
    else {        
		InternalSourceRoute *SR;
		uint etxThresh ; 
		uint j ;
		InternalSourceRoute *SR_;
		uint Index ;
		AckCache *AC ;
		KIRQL OldIrql;
		uint dcnt = 0 ; 

        SR = ExAllocatePool(NonPagedPool,
                            sizeof *SR + sizeof(SRAddr)*MAX_SR_LEN);
        if (SR == NULL) {
            Status = NDIS_STATUS_RESOURCES;
			if(debug) KdPrint(("MiniportTransmit: Can't alloc SR\n"));
            goto ErrorFreeSRP;
        }
		RtlZeroMemory(SR, sizeof *SR + sizeof(SRAddr)*MAX_SR_LEN);
        srp->sr = SR;

        //
        // See if we can find a route in the link cache. 
        //
        Status = LinkCacheFillSR(VA, srp->Dest, SR);
        if (Status != NDIS_STATUS_SUCCESS) {
            if (Status == NDIS_STATUS_NO_ROUTE_TO_DESTINATION) {
                //
                // Didn't find a route. Stick this packet in the send buffer.
                //
                InterlockedIncrement((PLONG)&VA->CountXmitSendBuf);
                SendBufInsert(VA, srp);
                return;
            }
            //
            // Looks like we didn't have enough memory to run dijkstra().
            // Drop this packet on the floor.
            //
            ASSERT(Status == NDIS_STATUS_RESOURCES);
			if(debug) KdPrint(("MiniportTransmit: No memory for dijkstra\n"));
            goto ErrorFreeSRP;
        }

		//currently, every node on forwarding list is a dp node.
		srp->sr->opt.bitmap = (ushort) -1;
		
		KeAcquireSpinLock(&(VA->LC->Lock), &OldIrql);	
		SR_ = LinkCacheFindForwarders(VA, srp, &etxThresh);
	
		Status = LinkCacheInsertForwardingList(VA, SR_, srp, etxThresh);
		if (Status != NDIS_STATUS_SUCCESS) {
			//no action required b/c we should still have the packet
			// prepared to send anyways.
			if(debug) KdPrint(("Unsuccessful in linkcacheinsertforwardinglist\n")); 
		}
		KeReleaseSpinLock(&(VA->LC->Lock), OldIrql);
		srp->updateTime = KeQueryInterruptTime();

		//
		// DEBUG
		//
		Index = SOURCE_ROUTE_HOPS(srp->sr->opt.optDataLen) - srp->sr->opt.segmentsLeft ;
		if(debug) KdPrint(("%d hops on the fwd list, %d is next hop\n", SOURCE_ROUTE_HOPS(srp->sr->opt.optDataLen), Index));
		for (j = 0 ; j < SOURCE_ROUTE_HOPS(srp->sr->opt.optDataLen); j++) {
				VirtualAddress *Address;
				Address = &(srp->sr->opt.hopList[j].addr) ;

				if (checkSRPbitmap(srp->sr->opt.bitmap, j)) {
					dcnt++; 
					if (checkSRPbitmap(srp->sr->opt.fwd_bitmap, j))
						PRINT_VA("\t\t\tDFDFDFDFDF Forwarding node: ", *Address); 
					else
						PRINT_VA("\t\t\tDNDNDNDNDN Forwarding node: ", *Address); 
				} else {
					if (checkSRPbitmap(srp->sr->opt.fwd_bitmap, j))
						PRINT_VA("\t\t\tNFNFNFNFNF Forwarding node: ", *Address); 
					else
						PRINT_VA("\t\t\tNNNNNNNNNN Forwarding node: ", *Address); 
				}
		} 
		if(debug) KdPrint(("Done with %d default hops on the fwd list.\n", dcnt));
	
		//
		// END DEBUG
		//
		
		
		//assuming 1 interface for now!
		AC = MaintBufFindAckFlow(VA, &(VA->Address), &(srp->Dest), 
			srp->sr->opt.hopList[1].inif, srp->sr->opt.hopList[1].inif, &VA->Address); 
	
#if SLIDING_RATE_LIMIT_ENABLED	
		//rate-limit
		//calculate numForwarders
		if (AC != NULL) {
			uint count = 0 ;
			for (j = 0 ; j < SOURCE_ROUTE_HOPS(srp->sr->opt.optDataLen); j++) {
					if (checkSRPbitmap(srp->sr->opt.fwd_bitmap, j)) 
						count++;
			}
			ASSERT(count);
			AC->numForwarders = count; 
		}
#endif

		//fill in piggy back ACKs
		Size = PbackPacketSize(srp);
		if (Size < PROTOCOL_MIN_FRAME_SIZE) {
			if (AC != NULL) {
				InternalAcknowledgement *curAck; 
				Size = PROTOCOL_MIN_FRAME_SIZE - Size; //amount left to use
				srp->ack = MaintBufGetCrossFlowAcks(VA, AC, Size, 0);
				//debug
				for (curAck = srp->ack ; curAck != NULL ; curAck = curAck->next) {
					PRINT_ADDRESS("transmitPacket Ack flow: ", curAck->opt.flowSource, curAck->opt.flowDest, 
						curAck->opt.identification);
#if E2E_DEBUG
					ASSERT(curAck->opt.isE2E == 0); 
#endif
				}
			}
		}

        if (! LinkCacheUseSR(VA, srp)) {
            InterlockedIncrement((PLONG)&VA->CountXmitQueueFull);
			if(debug) KdPrint(("MiniportTransmit: LinkCacheUseSR failed\n"));
            goto ErrorFreeSRP;
        }
		/*
		if (srp->ackreq != NULL) {
			PRINT_ADDRESS("Sending packet (srp)",  srp->Source, srp->Dest, srp->ackreq->opt.identification) ; 
			//PRINT_ADDRESS("Sending packet (srp-real)",  srp->Source, srp->realDest, Ether->Type) ; 
			PRINT_ADDRESS("Sending packet (eth)",  Ether->Source, Ether->Dest, srp->ackreq->opt.identification) ; 
		} else {
			PRINT_ADDRESS("Sending packet (srp)",  srp->Source, srp->Dest, -1) ; 
			//PRINT_ADDRESS("Sending packet (srp-real)",  srp->Source, srp->realDest, Ether->Type) ; 
			PRINT_ADDRESS("Sending packet (eth)",  Ether->Source, Ether->Dest, -1) ; 

		}
		*/
		MaintBufSendPacket(VA, srp, MiniportRetransmitSendComplete);
		//MaintBufSendPacket(VA, srp, MiniportSendComplete);
	}

    return;

ErrorFreeSRP:
	if(debug) KdPrint(("Error in MiniportTransmitPacket1\n")); 
	//ejr- try to get cloned data
	NdisMSendComplete(VA->MiniportHandle, srp->OriginalPacket, Status);
	srp->OriginalPacket = NULL ;
    SRPacketFree(srp);
	return ;
ErrorReturn:
	if(debug) KdPrint(("Error in MiniportTransmitPacket2\n")); 
    
	NdisMSendComplete(VA->MiniportHandle, Packet, Status);
}

//* MiniportTransmitPackets
//
//  Called by NDIS to send packets via our virtual adapter.
//
void
MiniportTransmitPackets(
    IN NDIS_HANDLE MiniportAdapterContext,
    IN PPNDIS_PACKET PacketArray,
    IN UINT NumberOfPackets)
{
    MiniportAdapter *VA = (MiniportAdapter *) MiniportAdapterContext;
    uint i;
	

	for (i = 0; i < NumberOfPackets; i++) {
		//if(debug) KdPrint(("Sending packet %d of %d in MiniportTransmitPackets\n", i+1, NumberOfPackets)); 
		MiniportTransmitPacket(VA, PacketArray[i]);
	}
}

//* ArtificialDrop
//
//  Checks to see if the link on which this packet arrived has been flagged
//  for extra packet drops.  If so, determines whether or not to drop this
//  packet.  Return TRUE if packet should be dropped, FALSE otherwise.
//
boolint
ArtificialDrop(
    MiniportAdapter *VA,
	ProtocolAdapter *PA,
    const SRPacket *srp)
{
	boolint ret; 

    //
    // Check that ArtificialDrop is enabled.  Also, we only consider
    // (non-static) source routed packets as candidates for dropping.
    //
    if ((VA->ArtificialDrop) &&
        (srp->sr != NULL) && !srp->sr->opt.staticRoute) {
        uint Index;

        Index = SOURCE_ROUTE_HOPS(srp->sr->opt.optDataLen) -
            srp->sr->opt.segmentsLeft;
	
		//
		// we assume only one interface for now!!!!!
        ret = LinkCacheCheckForDrop(VA,
                                     srp->sr->opt.hopList[Index-1].addr,
									 VA->Address,
                                     srp->sr->opt.hopList[Index].inif,
                                     srp->sr->opt.hopList[Index-1].outif);
		if (ret)
			if(debug) KdPrint(("ArtificialDrop is dropping the current data packet.\n")); 
		return ret; 
    } else if ((VA->ArtificialDrop) && (srp->Probe != NULL)) {
		boolint mod_drop = 0 ;

		if (mod_drop) {
			//ejr- let's add support for dropping probe packets.
			//This drops probe packets in a modulo fashion, based
			// on their seqNum
			ret = LinkCacheCheckProbeForDrop(VA,
				srp->Probe->Opt.From,
				VA->Address,
				(LQSRIf) PA->Index,
				srp->Probe->Opt.OutIf,
				srp->Probe->Opt.Seq); 
		} else {
			//ejr- let's add support for dropping probe packets.
			//This is the default MCL way of dropping packets.
			ret = LinkCacheCheckForDrop(VA,
				srp->Probe->Opt.From,
				VA->Address,
				(LQSRIf) PA->Index,
				srp->Probe->Opt.OutIf); 

		}
		return ret; 

	} else if ((srp->req != NULL)) {
		//check if route request????
		RouteRequest *RR = &srp->req->opt;
        uint Hops = ROUTE_REQUEST_HOPS(RR->optDataLen);
		ret =  LinkCacheCheckForDrop(VA,
			RR->hopList[Hops -1].addr,
			VA->Address,
			(LQSRIf) PA->Index,
			RR->hopList[Hops -1].outif); 
		/*
		if (ret)
			if(debug) KdPrint(("ArtificialDrop is dropping the current route request.\n"));
		*/
		return ret; 
	} else if ((srp->ack != NULL)) {

		uint Index = SOURCE_ROUTE_HOPS(srp->sr->opt.optDataLen) - srp->sr->opt.segmentsLeft;

		ASSERT(srp->sr); 

#if E2E_RATE_LIMIT_ENABLED	
		//test: FOR NOW, DO NOT DROP E2E ACKS
		if (srp->ack->opt.isE2E) {
			return FALSE;	
		}
#endif

		ret =  LinkCacheCheckForDrop(VA,
			srp->sr->opt.hopList[Index-1].addr,
			VA->Address,
			(LQSRIf) PA->Index,	
			srp->ack->opt.inif); 
		
		if (ret)
			if(debug) KdPrint(("ArtificialDrop is dropping the current ACK.\n")); 
		return ret; 

	} 
	//should also add support for dropping route replies....

    return FALSE;
}


//* CacheSRPacket
//
//  Updates the link cache and neighbor cache with information
//  from a newly-received packet.
//
//  Also calls SendBufCheck, to send any packets that have been
//  waiting for a route.
//
void
CacheSRPacket(
    MiniportAdapter *VA,
    ProtocolAdapter *PA,
    const SRPacket *srp)
{
    InternalRouteReply *Rep;
    InternalRouteError *Err;
    InternalLinkInfo *LI;
    boolint LinkAdded = FALSE;
    uint Metric;
    uint i;


	//if(debug) KdPrint(("starting CacheSRPacket\n"));

	//
    // We must process Link Info before Route Request options.
    // If the Link Info mistakenly says that a link from A -> B
    // is broken, and B receives the Route Request / Link Info
    // from A, then we need to ensure that after CacheSRPacket
    // B's link cache says the A -> B link exists and that B then
    // puts that metric into the Route Request that it forwards,
    // so A will overhear this and update its link cache.
    //
    for (LI = srp->LinkInfo; LI != NULL; LI = LI->Next) {
        //
        // If this link information originated from ourselves,
        // then don't update our link cache based upon it.
        // It is stale information. Furthermore it would reset
        // the link timestamps, preventing them from ever timing out.
        //
        if (! VirtualAddressEqual(LI->Opt.From, VA->Address)) {
            for (i = 0; i < LINKINFO_HOPS(LI->Opt.optDataLen); i++) {				
				/*	
				if(debug) KdPrint(("EEE Calling LinkCacheAddLink from CacheSRPacket1 from %02x-%02x-%02x-%02x-%02x-%02x to %02x-%02x-%02x-%02x-%02x-%02x with metric: %u\n", 
					LI->Opt.From[0], LI->Opt.From[1], 
					LI->Opt.From[2], LI->Opt.From[3], 
					LI->Opt.From[4], LI->Opt.From[5], 
					LI->Opt.Links[i].addr[0], LI->Opt.Links[i].addr[1], 
					LI->Opt.Links[i].addr[2], LI->Opt.Links[i].addr[3], 
					LI->Opt.Links[i].addr[4], LI->Opt.Links[i].addr[5],
					(*VA->ConvMetric)(LI->Opt.Links[i].Metric)));
				if (LI->Opt.Links[i].Metric == 0)
					if(debug) KdPrint(("Zero metric\n"));
				*/
                LinkCacheAddLink(VA,
                                 LI->Opt.From,
                                 LI->Opt.Links[i].addr,
                                 LI->Opt.Links[i].inif,
                                 LI->Opt.Links[i].outif,
                                 LI->Opt.Links[i].Metric,
                                 LINK_STATE_CHANGE_ADD_LINKINFO);
                LinkAdded = TRUE;
            }
        }
    }

    if (srp->sr != NULL) {
        uint Index;

		Index = SOURCE_ROUTE_HOPS(srp->sr->opt.optDataLen) - srp->sr->opt.segmentsLeft;

		//ejr
		// Let's only call LinkCacheAddLink on the link we know is good.
		//--in future, change above method so we update all links from sender
		//  to its forwarding list nodes.
		//
		Metric = srp->sr->opt.hopList[Index].Metric;
        if ((*VA->IsInfinite)(Metric)) {
			//
            // This shouldn't happen, because we never send infinity
            // in a Source Route.
            //
            if(debug) KdPrint(("MCL!CacheSRPacket: SR infinite metric\n"));
            Metric = 0;
		}

        //
        // We know the forward link exists.
        //
		/*
		if(debug) KdPrint(("EEE Calling LinkCacheAddLink from CacheSRPacket2 \n"));
		if (Metric == 0)
			if(debug) KdPrint(("Zero metric\n"));
        */
		LinkCacheAddLink(VA,
						srp->sr->opt.hopList[Index-1].addr,
                        srp->sr->opt.hopList[Index].addr,
                        srp->sr->opt.hopList[Index].inif,
                        srp->sr->opt.hopList[Index-1].outif,
                        Metric,
                        LINK_STATE_CHANGE_ADD_SNOOP_SR);
		LinkAdded = TRUE;


        NeighborReceivePassive(&VA->NC,
                               srp->sr->opt.hopList[Index-1].addr,
                               srp->sr->opt.hopList[Index-1].outif,
                               srp->EtherSource);

        if (! srp->sr->opt.staticRoute) {
            //
            // Increment the usage count for this link.
            // REVIEW: Count static-routed packets?
            //
            LinkCacheCountLinkUse(VA,
                                  srp->sr->opt.hopList[Index-1].addr,
                                  srp->sr->opt.hopList[Index].addr,
                                  srp->sr->opt.hopList[Index].inif,
                                  srp->sr->opt.hopList[Index-1].outif);
        }
    }

    if (srp->req != NULL) {
        ASSERT(srp->sr == NULL);

        for (i = 1; i < ROUTE_REQUEST_HOPS(srp->req->opt.optDataLen); i++) {
            Metric = srp->req->opt.hopList[i].Metric;
            if ((*VA->IsInfinite)(Metric)) {
                //
                // This shouldn't happen, because we never send infinity
                // in a Route Request.
                //
                if(debug) KdPrint(("MCL!CacheSRPacket: Request infinite metric\n"));
                Metric = 0;
            }
			/*
			if(debug) KdPrint(("EEE Calling LinkCacheAddLink from CacheSRPacket3 from %02x-%02x-%02x-%02x-%02x-%02x to %02x-%02x-%02x-%02x-%02x-%02x with metric: %u\n", 
				srp->req->opt.hopList[i-1].addr[0], srp->req->opt.hopList[i-1].addr[1], srp->req->opt.hopList[i-1].addr[2],
				srp->req->opt.hopList[i-1].addr[3], srp->req->opt.hopList[i-1].addr[4], srp->req->opt.hopList[i-1].addr[5], 
				srp->req->opt.hopList[i].addr[0], srp->req->opt.hopList[i].addr[1], srp->req->opt.hopList[i].addr[2],
				srp->req->opt.hopList[i].addr[3], srp->req->opt.hopList[i].addr[4], srp->req->opt.hopList[i].addr[5],
				(*VA->ConvMetric)(Metric)));
			if (Metric == 0)
				if(debug) KdPrint(("Zero metric\n"));
            */
			LinkCacheAddLink(VA,
                             srp->req->opt.hopList[i-1].addr,
                             srp->req->opt.hopList[i].addr,
                             srp->req->opt.hopList[i].inif,
                             srp->req->opt.hopList[i-1].outif,
                             Metric,
                             LINK_STATE_CHANGE_ADD_SNOOP_REQUEST);
        }
		//if(debug) KdPrint(("EEE Calling LinkCacheAddLink from CacheSRPacket4 \n"));
        LinkCacheAddLink(VA,
                         srp->req->opt.hopList[i-1].addr,
                         VA->Address,
                         (LQSRIf) PA->Index,
                         srp->req->opt.hopList[i-1].outif,
                         0, // Metric.
                         LINK_STATE_CHANGE_ADD_SNOOP_REQUEST);
        LinkAdded = TRUE;

        NeighborReceivePassive(&VA->NC,
                               srp->req->opt.hopList[i-1].addr,
                               srp->req->opt.hopList[i-1].outif,
                               srp->EtherSource);
    }

    for (Rep = srp->rep; Rep != NULL; Rep = Rep->next) {
        uint Reason;

        if (VirtualAddressEqual(Rep->opt.hopList[0].addr, VA->Address))
            Reason = LINK_STATE_CHANGE_ADD_REPLY;
        else
            Reason = LINK_STATE_CHANGE_ADD_SNOOP_REPLY;

        for (i = 1; i < ROUTE_REPLY_HOPS(Rep->opt.optDataLen); i++) {
            Metric = Rep->opt.hopList[i].Metric;
            if ((*VA->IsInfinite)(Metric)) {
                //
                // This shouldn't happen, because we never send infinity
                // in a Route Reply.
                //
                if(debug) KdPrint(("MCL!CacheSRPacket: Reply infinite metric\n"));
                Metric = 0;
            }
			/*
			if(debug) KdPrint(("EEE Calling LinkCacheAddLink from CacheSRPacket5 \n"));
			if (Metric == 0)
				if(debug) KdPrint(("Zero metric\n"));
            */
			LinkCacheAddLink(VA,
                             Rep->opt.hopList[i-1].addr,
                             Rep->opt.hopList[i].addr,
                             Rep->opt.hopList[i].inif,
                             Rep->opt.hopList[i-1].outif,
                             Metric,
                             Reason);
            LinkAdded = TRUE;
        }

        //
        // Reset the backoff parameters for sending a Route Request
        // for the target address in this Route Reply.
        // NB: We do this whether or not we requested the reply.
        //
        ReqTableReceivedReply(VA, Rep->opt.hopList[i-1].addr);
    }

    for (Err = srp->err; Err != NULL; Err = Err->next) {
		/*
		if(debug) KdPrint(("EEE Calling LinkCacheAddLink from CacheSRPacket6 \n"));
		if (Err->opt.Metric == 0)
			if(debug) KdPrint(("Zero metric\n"));
        */
		LinkCacheAddLink(VA,
                         Err->opt.errorSrc,
                         Err->opt.unreachNode,
                         Err->opt.inIf,
                         Err->opt.outIf,
                         Err->opt.Metric,
                         (VirtualAddressEqual(Err->opt.errorDst,
                                              VA->Address) ?
                          LINK_STATE_CHANGE_ERROR :
                          LINK_STATE_CHANGE_SNOOP_ERROR));
        LinkAdded = TRUE;
    }

    //
    // Call SendBufCheck once, after the link cache is fully updated.
    //
    if (LinkAdded)
        SendBufCheck(VA);
}

//* MiniportReceiveLocally
//
//  Receives an SRPacket locally, using NdisMIndicateReceivePacket.
//  Decrypts the packet data for receiving it.
//
//  The ReceiveComplete handler is always called.
//
//  Does NOT directly consume srp or srp->Packet,
//  but the ReceiveComplete handler may do that.
//
void
MiniportReceiveLocally(
    MiniportAdapter *VA,
    SRPacket *srp,
    void (*Complete)(MiniportAdapter *VA, SRPacket *srp, NDIS_STATUS Status))
{
    NDIS_PACKET *Packet;
    NDIS_BUFFER *Buffer;
    void *Data;
    uint BufferLength;
    uint PacketLength;
    EtherHeader UNALIGNED *Ether;
    NDIS_STATUS Status;

    //
    // srp->Packet is NULL only in SRPackets that we generate ourself
    // (for example, for Acks, Route Requests, Route Replies)
    // and we should never try to receive such a packet locally.
    //
    ASSERT(srp->Packet != NULL);



    if (srp->PacketLength == srp->PayloadOffset) {
        //
        // This happens when we receive a packet which only contains
        // options, no payload.
        //
		//if(debug) KdPrint(("MRL: Tossing empty pkt\n")); 
        InterlockedIncrement((PLONG)&VA->CountRecvEmpty);
        (*Complete)(VA, srp, NDIS_STATUS_SUCCESS);
        return;
    }


    //
    // The ethernet Type field is part of the encrypted data.
    //
    Status = CryptoDecryptPacket(VA, srp->Packet,
                                 srp->PayloadOffset,
                                 sizeof *Ether - sizeof Ether->Type,
                                 srp->IV, &Packet); 
    if (Status != NDIS_STATUS_SUCCESS) { 
        if(debug) KdPrint(("MCL!MiniportReceiveLocally: CryptoDecryptPacket: %x\n", Status));
        InterlockedIncrement((PLONG)&VA->CountRecvDecryptFailure);
        (*Complete)(VA, srp, Status);
        return;
    }

    //
    // Get the ethernet header in the packet.
    //
    Buffer = NdisFirstBuffer(Packet);
    NdisQueryBuffer(Buffer, &Data, &BufferLength);
    if (BufferLength < sizeof *Ether) {
        InterlockedIncrement((PLONG)&VA->CountRecvSmall);
        NdisFreePacketClone(Packet);
        (*Complete)(VA, srp, NDIS_STATUS_INVALID_LENGTH);
        return;
    }

    //
    // We must not indicate packets larger than MediumMaxPacketLen.
    //
    NdisQueryPacketLength(Packet, &PacketLength);
    if (PacketLength > VA->MediumMaxPacketLen) {
        InterlockedIncrement((PLONG)&VA->CountRecvLarge);
        NdisFreePacketClone(Packet);
        (*Complete)(VA, srp, NDIS_STATUS_INVALID_LENGTH);
        return;
    }

    //
    // If the Type is ETYPE_MSFT, we do not receive the packet.
    //
    Ether = (EtherHeader UNALIGNED *) Data;
    if (Ether->Type == ETYPE_MSFT) {
        InterlockedIncrement((PLONG)&VA->CountRecvRecursive);
        NdisFreePacketClone(Packet);
        (*Complete)(VA, srp, NDIS_STATUS_SUCCESS);
        return;
    }

    InterlockedIncrement((PLONG)&VA->CountRecvLocally);
	/*
	if (srp->ackreq != NULL) {		
		PRINT_ADDRESS("MRL (srp->dest)", srp->Source, srp->Dest, srp->ackreq->opt.identification);
	} else {
		PRINT_ADDRESS("MRL (srp->dest)", srp->Source, srp->Dest, -1);
	}
	*/
    if (srp->sr != NULL) {
        if (srp->sr->opt.salvageCount != 0)
            InterlockedIncrement((PLONG)&VA->CountRecvLocallySalvaged);

        //
        // On its way to us, the source route was augmented
        // with link metric information. Send this info
        // back to the originator in a route reply.
        //
		if (! srp->sr->opt.staticRoute) {
            MiniportSendRouteReplyFromSR(VA, srp);
		}
    }

    //
    // Initialize the ethernet header.
    //
	RtlCopyMemory(Ether->Dest, srp->Dest, SR_ADDR_LEN);
    RtlCopyMemory(Ether->Source, srp->Source, SR_ADDR_LEN);
	
	

    //
    // Prepare for MiniportReturnPacket.
    //
    PC(Packet)->srp = srp;
    PC(Packet)->ReceiveComplete = Complete;


	if(debug) KdPrint(("MRL: sending to higher layer.\n")); 
    //
    // Indicate the packet up to any transports bound to us.
    // MiniportReturnPacket will be called asynchronously.
    //
    NDIS_SET_PACKET_STATUS(Packet, NDIS_STATUS_SUCCESS);
    NDIS_SET_PACKET_HEADER_SIZE(Packet, sizeof *Ether);
    NdisMIndicateReceivePacket(VA->MiniportHandle, &Packet, 1);
}




//* MiniportReturnPacket
//
//  Called by NDIS to return a packet after the transports have
//  finished their receive processing.
//
void
MiniportReturnPacket(
    IN NDIS_HANDLE MiniportAdapterContext,
    IN PNDIS_PACKET Packet)
{
    MiniportAdapter *VA = (MiniportAdapter *) MiniportAdapterContext;
    SRPacket *srp = PC(Packet)->srp;
    void (*Complete)(MiniportAdapter *VA, SRPacket *srp, NDIS_STATUS Status) =
        PC(Packet)->ReceiveComplete;

    //
    // Free the clear-text packet that was created
    // in MiniportReceiveLocally.
    //
    NdisFreePacketClone(Packet);

    //
    // Indicate the completion of receive processing.
    //
    (*Complete)(VA, srp, NDIS_STATUS_SUCCESS);
}


//* MiniportForwardComplete
//ejr
// This method is intended to be used after we forward a SRPacket
// via ReceiveSRPacket and MaintBufSendPacket.  We need to delete
// the SRPacket iff we don't try to retransmit the packet.
//
void 
MiniportForwardComplete(
    MiniportAdapter *VA,
    SRPacket *srp,
    NDIS_STATUS Status)
{
    UNREFERENCED_PARAMETER(VA);
    UNREFERENCED_PARAMETER(Status);

    //
    // Free the SRPacket. This frees srp->Packet.
    //
	if (srp->ackreq == NULL || srp->PayloadOffset == srp->PacketLength)
		SRPacketFree(srp);
}


//* MiniportReceiveComplete
//
//  Called after the completion of local receive processing
//  of an SRPacket.
//
void
MiniportReceiveComplete(
    MiniportAdapter *VA,
    SRPacket *srp,
    NDIS_STATUS Status)
{
    UNREFERENCED_PARAMETER(VA);
    UNREFERENCED_PARAMETER(Status);

    //
    // Free the SRPacket. This frees srp->Packet.
    //
    SRPacketFree(srp);
}

//* MiniportForwardRequestComplete
//
//  Called after the completion of forwarding of an SRPacket
//  with a Route Request option. The SRPacket still needs
//  to be received locally.
//
static void
MiniportForwardRequestComplete(
    MiniportAdapter *VA,
    SRPacket *srp,
    NDIS_STATUS Status)
{
    UNREFERENCED_PARAMETER(Status);

    //
    // Receive the packet locally too.
    //
	//if(debug) KdPrint(("Calling MRL from MFRC\n")); 
    MiniportReceiveLocally(VA, srp, MiniportReceiveComplete);
}

//* MiniportForwardingStall
//
//  Stalls execution very briefly, to prevent synchronization
//  when forwarding a broadcast route request.
//
void
MiniportForwardingStall(MiniportAdapter *VA)
{
    uint Degree;
    uint Microseconds;

    //
    // How many neighbors do we have?
    // With more neighbors we need to choose from a larger interval.
    //
    Degree = LinkCacheMyDegree(VA);

    //
    // We have a lower bound in case there are neighbors we don't know about.
    //
    Degree = max(Degree, 2);

    //
    // We have an upper bound to prevent the stall from being too large.
    // The DDK doc recommends stalling no longer than 50 microseconds.
    //
    Degree = min(Degree, 10);

    //
    // We spread ourselves across this range. The multiplicative
    // factor decreases collisions with neighbors performing
    // this same calculation.
    //
    Microseconds = GetRandomNumber(Degree * 3);
    InterlockedExchangeAdd((PLONG)&VA->TotalForwardingStall, Microseconds);

    NdisStallExecution(Microseconds);
}



//*		ejr
//	OnForwardingList(srp->sr, Index, srp->sr->opt.max)
//
// Checks to see if 'Addr' is on the forwarding list, specified
// between (Start, End) index into the array (inclusive)
//
boolint
OnForwardingList(MiniportAdapter *VA, VirtualAddress *Addr, SourceRoute *SR) {
	uint i = 0 ; 
    UNREFERENCED_PARAMETER(VA);

	for (i = 0 ; i < SOURCE_ROUTE_HOPS(SR->optDataLen) ; i++) {
		if (VirtualAddressEqual(*Addr, SR->hopList[i].addr)) {
			if (checkSRPbitmap(SR->fwd_bitmap, i))
				return TRUE;
			else
				return FALSE; 
		} 
	}
	return FALSE; 	
}

//*	ejr
//		Checks to see if:
//			0. The node receiving the packet is the destination
//			1. This packet is for the destination and if so:
//			3. If the destination is allowed to touch the packet (ie- is it on the forwarding
//				list?)
//
boolint
destinationShouldReceive(MiniportAdapter *VA, SourceRoute *SR) {
	uint index = SOURCE_ROUTE_HOPS(SR->optDataLen) - 1 ;
	if (VirtualAddressEqual(VA->Address, SR->hopList[index].addr)) {
		if (checkSRPbitmap(SR->fwd_bitmap, index))
			return TRUE;
		else
			return FALSE ;
	}
	return FALSE  ;

}


// getFwdSlots
//
// This method is used when we need to forward a packet. We need to 
// get the priorities of the nodes by realizing what slot they are 
// in the fwding list.
//
// Returns:
//  max_fwd_slot = normalized slot in fwding list of highest priority node
//  my_fwd_slot = normalized slot in the fwding list of me
//     ---above two variables are zero-based
void getFwdSlots(MiniportAdapter *VA, SRPacket *srp, int *max_fwd_slot, int *my_fwd_slot) {
	
	uint cur_cnt = 0 ; //index of each fwding node
	uint i = 0 ; 
		
	*max_fwd_slot = -1;
	*my_fwd_slot = -1 ;

	for (i = 0  ; i < SOURCE_ROUTE_HOPS(srp->sr->opt.optDataLen) ; i++) {
		if (checkSRPbitmap(srp->sr->opt.fwd_bitmap, i)) {
			*max_fwd_slot = cur_cnt ;
			if (VirtualAddressEqual(VA->Address, srp->sr->opt.hopList[i].addr))
				*my_fwd_slot = cur_cnt ;
			
			cur_cnt++; 
		}
	}
}



//*		ejr
//	We call this method when we need to forward a packet. It is responsible
// for setting the new boundaries of the forwarding list (source route) and 
// also updating segmentsLeft to the correct value.
//
// It will first prune off all forwarding nodes that are not on the dp,
// leaving only the sender of the packet (if the sender is not on dp).
//
// Biggest thing will changing this method to prepare for LinkCacheInsertForwarders.
// We need to remove all non-dp nodes (except for me if I am the sender) as well
// as appropiately set the state of the source route (because even if InsertForwarders
// fails, we still want to send the packet using the dp).
//
//
NDIS_STATUS 
updateForwardInfo(MiniportAdapter *VA, SRPacket *srp) {
	
    LinkCache *LC = VA->LC;
	InternalSourceRoute *SR = srp->sr; 
    Link *Walk;
    int DNode;

	//data to create our new fwd list w/ only dp path nodes and me
	InternalSourceRoute *SR_new ; 
	SourceRoute *sr_ ;
	uint sr_index = 0 ; //index into new sr_ hoplist[]
	uint i ; 
	uint Index;
	VirtualAddress Dest  ; //debugging
	uint fwd_max ;

	RtlCopyMemory(&Dest, &(srp->sr->opt.hopList[SOURCE_ROUTE_HOPS(srp->sr->opt.optDataLen)-1].addr), sizeof(VirtualAddress));

	//first get rid of all non-dp nodes that are not me.
	SR_new =  ExAllocatePool(NonPagedPool, sizeof *SR_new + sizeof(SRAddr)*SOURCE_ROUTE_HOPS(SR->opt.optDataLen)); 
	if (SR_new == NULL) {
		return NDIS_STATUS_RESOURCES ;
    }	
	RtlZeroMemory(SR_new, sizeof *SR_new + sizeof(SRAddr)*SOURCE_ROUTE_HOPS(SR->opt.optDataLen)); 
	sr_ = &(SR_new->opt);

	for (i = 0 ; i < SOURCE_ROUTE_HOPS(SR->opt.optDataLen) ; i++) {
		if (checkSRPbitmap(srp->sr->opt.bitmap, i) || VirtualAddressEqual(VA->Address, SR->opt.hopList[i].addr)) {
			RtlCopyMemory(&(sr_->hopList[sr_index]), &(srp->sr->opt.hopList[i]), sizeof(SRAddr)); 
					
			//need to zero out the metric in the new source route b/c LinkCacheUseSR will take
			// care of this for us. Furthermore, since we're changing the SR, the Metric
			// for the link in the new sr can be incorrect (but probably doesn't matter b/c
			// cachesrpacket would not read this value).	   	
			sr_->hopList[sr_index].Metric = 0 ;
			
			//want to set dp bitmap for me if I am a forwarder
			if (checkSRPbitmap(srp->sr->opt.bitmap, i))
				setSRPbitmap(&(sr_->bitmap), sr_index);
			else
				ASSERT(VirtualAddressEqual(VA->Address, SR->opt.hopList[i].addr)); 
			sr_index++; 
		}
	}

	ASSERT(sr_index > 1); 
	sr_->optionType = LQSR_OPTION_TYPE_SOURCERT; 
	sr_->optDataLen = SOURCE_ROUTE_LEN(sr_index) ;
	sr_->reservedField = srp->sr->opt.reservedField;
    sr_->staticRoute =  srp->sr->opt.staticRoute;
    sr_->salvageCount = srp->sr->opt.salvageCount;
	sr_->segmentsLeft = (uchar) (sr_index - 1); //just iterate through all nodes to get correct one

	SRFreeOptionList((InternalOption *)srp->sr);
	srp->sr = SR_new ;
	SR = srp->sr; 

	//let's first update segmentsLeft.  Right now, Index will denote the 'next hop' in
	//the original case and Index-1 is supposed to denote me (the forwarder).  So our basic
	// goal is to have Index-1 denote me.
	srp->sr->opt.segmentsLeft = SOURCE_ROUTE_HOPS(srp->sr->opt.optDataLen) - 1 ;//want equation below to start at 0
    Index = SOURCE_ROUTE_HOPS(srp->sr->opt.optDataLen) - srp->sr->opt.segmentsLeft - 1 ;//want to be me
	ASSERT(Index == 0) ;
	while (!VirtualAddressEqual(VA->Address, srp->sr->opt.hopList[Index].addr)) {
		srp->sr->opt.segmentsLeft--; 
		Index = SOURCE_ROUTE_HOPS(srp->sr->opt.optDataLen) - srp->sr->opt.segmentsLeft - 1 ;//want to be me
		ASSERT(srp->sr->opt.segmentsLeft != 0 ) ;
	}

	//
	// Now we need to update the max variable
	//
	//set max value of forwarding list, the forwarding list is 
	//bounded by ETX_FORWARD_SELECTION_THRESH (if no such nodes 
	//exist, then we at least include the next hop)
	fwd_max = (ushort) Index + 1  ; //at least the next hop--Index is pointing to me
	
	while (fwd_max < SOURCE_ROUTE_HOPS(SR->opt.optDataLen)) {
		uint curMetric = 0; 
		DNode = findNode(LC, SR->opt.hopList[fwd_max].addr, FALSE);
		Walk = NULL; 
		if (DNode > 0) {

			//
			// Search our link cache for this link.
			//
			for (Walk = LC->nodes[0].AdjOut; Walk != NULL; Walk = Walk->NextOut) {
				if ((Walk->targetIndex == DNode) &&
				    (Walk->outif == SR->opt.hopList[0].outif) &&
				    (Walk->inif == SR->opt.hopList[fwd_max].inif)) {
				    //
				    // Found link (should match at most one).
				    //
				    curMetric = Walk->Metric;
				    break;
				}//end if link equal
			}//end for all links
		}//end if dnode > 0 

		if (DNode <= 0 || Walk == NULL || Walk->Metric == 0 || 
			(*VA->IsInfinite)(curMetric) ||
			(*VA->ConvMetric)(curMetric) >= DEFAULT_ETX_FORWARD_SELECTION_THRESH) 
			{
				break; 
			}

		fwd_max++; 
	}//end while	
	
	//in the loop above, we always advance 1 too many.
	fwd_max--;
	if (fwd_max == 0 || fwd_max == Index)
		fwd_max++; //need to have max be 1 at very least.

	//set all fwd list bitmaps
	for (i = Index + 1 ; i <= fwd_max ; i++) {
		setSRPbitmap(&(SR->opt.fwd_bitmap), i);		
	}
	ASSERT(VirtualAddressEqual(Dest, srp->sr->opt.hopList[SOURCE_ROUTE_HOPS(srp->sr->opt.optDataLen)-1].addr));
	return NDIS_STATUS_SUCCESS ;
}




//* ReceiveSRPacket
//
//  Performs receive processing for an LQSR packet,
//  after it has been parsed into an SRPacket structure.
//
//  Disposes of srp (only if it doesn't locally cache packet for local recovery).
//
void
ReceiveSRPacket(
    MiniportAdapter *VA,
    ProtocolAdapter *PA,
    SRPacket *srp)
{
    InternalAcknowledgement **PrevAck;
    InternalAcknowledgement *Ack;
    InternalInfoRequest *InfoReq;
    InternalInfoReply *InfoRep;
	EtherHeader UNALIGNED *Ether;
    int Index;
	NDIS_STATUS Status;
	boolint forwBitMapSet = 0 ; 
	boolint forwSentAck = 0 ; 
#if E2E_RATE_LIMIT_ENABLED	
	boolint isE2EAck = FALSE ;
#endif
	KIRQL OldIrql;

	uint ackCount = 0 ; 



    //
    // Deal with probes immediately.  
    //
    if ((srp->Probe != NULL) &&
        (VA->ReceiveProbe != NULL)) {
		
			uint okay = 1;
#if STOP_LINK_METRIC
			//ejr- want to stop updating the link weights
			// so do nothing here.
			if (KeQueryInterruptTime() > VA->start_link_time + (Time) (LINK_MEASURE_TIME * SECOND))
			{
				//KdPrint(("Setting okay = 0 in ReceiveSRPacket, time is up!\n"));
				okay = 0; 
			}	
#endif
        //
        // Process the probe.
        //
		//if(debug) KdPrint(("Received probe in SRPacketToPkt()\n")) ;
		if (okay)
		{
	        (*VA->ReceiveProbe)(VA, srp->Probe, (LQSRIf) PA->Index);
		}
    }

    //
    // Process a Probe Reply.
    //
    if ((srp->ProbeReply != NULL) &&
        VirtualAddressEqual(srp->ProbeReply->Opt.To, VA->Address) &&
        (srp->ProbeReply->Opt.InIf == PA->Index)) {
        //
        // Process the probe reply according to whatever metric
        // we are using. 
        // 
        switch (VA->MetricType) {
        case METRIC_TYPE_RTT:
            RttReceiveProbeReply(VA, srp->ProbeReply);
            break;
        case METRIC_TYPE_PKTPAIR:
            PktPairReceiveProbeReply(VA, srp->ProbeReply);
            break;
        case METRIC_TYPE_WCETT:
            WcettReceivePktPairProbeReply(VA, srp->ProbeReply);
            break;
        }
    } 

    //
    // Process any Information Requests.
    //
    for (InfoReq = srp->inforeq; InfoReq != NULL; InfoReq = InfoReq->next)
        MiniportSendInfo(VA, PA, srp, InfoReq);

    //
    // Process any Information options.
    //
    for (InfoRep = srp->inforep; InfoRep != NULL; InfoRep = InfoRep->next) {
        VA->MinVersionSeen = min(VA->MinVersionSeen, InfoRep->opt.version);
        VA->MaxVersionSeen = max(VA->MaxVersionSeen, InfoRep->opt.version);
    }

	//ejr- check to see if we rcv'd route reply
	if (srp->rep != NULL)
			if(debug) KdPrint(("Received route reply\n")) ;

    //
    // Send an Ack immediately, if requested.
    //
	Ether = NdisFirstBuffer(srp->Packet)->MappedSystemVa;	
    if (srp->ackreq != NULL) {
		ASSERT(srp->sr != NULL);
        Index = SOURCE_ROUTE_HOPS(srp->sr->opt.optDataLen) - srp->sr->opt.segmentsLeft;

        //
        // See if I am on the forwarding list.  If so then I
        // need to send an ACK.
		//
		if (OnForwardingList(VA, &(VA->Address), &(srp->sr->opt))) {
			if (srp->PacketLength == srp->PayloadOffset) {				
				if(debug) KdPrint(("Need to send an ACK for recvd (empty) packet\n"));	
			} else {	
				if(debug) KdPrint(("Need to send an ACK for recvd packet %d\n", srp->ackreq->opt.identification));				
			}

			//if we have to forward the data, then we need to know if
			//the bitmap was already set
			forwBitMapSet = MaintBufSendAck(VA, srp);
			forwSentAck = TRUE ;
		} 
		else
			if(debug) KdPrint(("Received ackreq for packet %d, but not on fwd list\n", srp->ackreq->opt.identification));	
    }

    //
    // Process any received Acks immediately,
    // if they are for us.
    // 
    PrevAck = &srp->ack;
    while ((Ack = *PrevAck) != NULL) {
		boolint updateAckState = FALSE ; 
		boolint higherETX = FALSE; 
		//if(debug) KdPrint(("INSIDE processACK"));
        //
        // Is this Ack for us?
        // NB: We do not check Ack->opt.outif == PA->Index!
        // It doesn't matter how the Ack got to us -
        // it might not have been via the reverse route.
        //
		
		//PRINT_ADDRESS("Ack local: ", Ack->opt.from, Ack->opt.to,  Ack->opt.identification);
				
		//
        // Remove the Ack from the packet.
        //
		ackCount++;
		KeAcquireSpinLock(&VA->RecoveryLock, &OldIrql);
		updateAckState = LocalRecoveryProcessAck(VA, Ack, &higherETX); 
		ProtocolRemoveFromPreIFQ(VA, Ack, updateAckState, higherETX);
		KeReleaseSpinLock(&VA->RecoveryLock, OldIrql);
#if E2E_RATE_LIMIT_ENABLED	
		if (Ack->opt.isE2E) {
			if(debug) KdPrint(("Holy snickies, recieved an E2E ack: %d!\n",++VA->numE2EAcksRecv));
			//if(debug) 
			KdPrint(("rlNumber Recv_E2E_ACK %d (%d) BitMap: %x %x %x %x %x %x %x %x\n", Ack->opt.identification, Ack->opt.isE2E,
				Ack->opt.bitmap.bm[0], Ack->opt.bitmap.bm[1], Ack->opt.bitmap.bm[2], Ack->opt.bitmap.bm[3], 
				Ack->opt.bitmap.bm[4], Ack->opt.bitmap.bm[5], Ack->opt.bitmap.bm[6], Ack->opt.bitmap.bm[7])); 	
#if E2E_DEBUG
			//debug
			if (Ack->next) {
				ASSERT(Ack->next->opt.optionType == LQSR_OPTION_TYPE_ACK);
				ASSERT(srp->PacketLength == srp->PayloadOffset);
				ASSERT(Ack->next == NULL);
			}
#endif
					
			ASSERT(Ack->next == NULL);
			ASSERT(ackCount == 1); 
			isE2EAck = TRUE ; 

			break; 
		} else {
			PRINT_ADDRESS("Ack flow: ", Ack->opt.flowSource, Ack->opt.flowDest,  Ack->opt.identification);
			if(debug) KdPrint(("DEBUG ZZ (%d) Receive_ACK %d BitMap: %x %x %x %x %x %x %x %x\n", ackCount, Ack->opt.identification,
						Ack->opt.bitmap.bm[0], Ack->opt.bitmap.bm[1], Ack->opt.bitmap.bm[2], Ack->opt.bitmap.bm[3], 
						Ack->opt.bitmap.bm[4], Ack->opt.bitmap.bm[5], Ack->opt.bitmap.bm[6], Ack->opt.bitmap.bm[7])); 	
			//verify the linked-list is correct.
			if (Ack->next) {
				ASSERT(Ack->next->opt.optionType == LQSR_OPTION_TYPE_ACK);
			}	
			*PrevAck = Ack->next;
			ExFreePool(Ack); //when we forward below, we assume ACK is deleted
		}
		if(debug) KdPrint(("Number of Packets in cache after process ACK: %d\n", VA->myCache.numPackets));
#else
		PRINT_ADDRESS("Ack flow: ", Ack->opt.flowSource, Ack->opt.flowDest,  Ack->opt.identification);
		if(debug) KdPrint(("DEBUG ZZ Receive_ACK %d BitMap: %x %x %x %x %x %x %x %x\n", Ack->opt.identification,
					Ack->opt.bitmap.bm[0], Ack->opt.bitmap.bm[1], Ack->opt.bitmap.bm[2], Ack->opt.bitmap.bm[3], 
					Ack->opt.bitmap.bm[4], Ack->opt.bitmap.bm[5], Ack->opt.bitmap.bm[6], Ack->opt.bitmap.bm[7])); 	
        *PrevAck = Ack->next;
		ExFreePool(Ack); //when we forward below, we assume ACK is deleted
		if(debug) KdPrint(("Number of Packets in cache after process ACK: %d\n", VA->myCache.numPackets));

#endif		
        continue;
	}


#if E2E_RATE_LIMIT_ENABLED	
	if (!isE2EAck) {
		srp->ack = NULL; //all ACKs should be deleted, so let's be defensive
	}
#else
	srp->ack = NULL; //all ACKs should be deleted, so let's be defensive
#endif
	
    if (srp->sr != NULL) {
        Index = SOURCE_ROUTE_HOPS(srp->sr->opt.optDataLen) - srp->sr->opt.segmentsLeft;

        //
        // This packet has a source route.
        // Check to make sure I'm actually the destination, since right now
        // sometimes the packets are MAC broadcast...
        //

		if (!OnForwardingList(VA, &(VA->Address), &(srp->sr->opt))) {

			//If we are not on the forwarding list, then we simply just exit.
			//We do ACK piggybacking, so don't need this method anymore.
			//if(debug) KdPrint(("Received packet, not on fwd list\n"));
			goto FreeSrpAndReturn ;		

		}

        //
        // Inspect segmentsLeft to decide if we should receive
        // the packet or forward it on its next hop.
        //
		--srp->sr->opt.segmentsLeft; 
        if (destinationShouldReceive(VA, &(srp->sr->opt))) {
			
			uint maxIndex = SOURCE_ROUTE_HOPS(srp->sr->opt.optDataLen)-1;
			maxIndex += 0 ;

			//I don't think this breaks anything, but we'll assign segmentsLeft to zero, just to be safe
			srp->sr->opt.segmentsLeft = (uchar) 0  ;
			
			//this assert can certainly fail b/c there may multiple fwding nodes that receive the packet
			// when the destination is on the fwding list
			
			if (srp->PacketLength == srp->PayloadOffset) {			
				ASSERT(VirtualAddressEqual(VA->Address, srp->sr->opt.hopList[maxIndex].addr));
			} else {
				ASSERT(VirtualAddressEqual(VA->Address, srp->sr->opt.hopList[maxIndex].addr));
			}

			if (srp->ackreq)
				if(debug) KdPrint(("Receive srpacket sending %d to MRL.\n", srp->ackreq->opt.identification)); 
			/*
			if (srp->ackreq != NULL) {
				//if(debug) KdPrint(("DEBUG Receive_Packet %d\n", srp->ackreq->opt.identification)); 
				PRINT_ADDRESS("rcvsrpkt1 (srp)", srp->Source, srp->Dest, srp->ackreq->opt.identification);
				//PRINT_ADDRESS("rcvsrpkt1 (real-srp)", srp->Source, srp->realDest, Ether->Type);
				PRINT_ADDRESS("rcvsrpkt1 (eth)", Ether->Source, Ether->Dest, srp->ackreq->opt.identification);
			} else {
				PRINT_ADDRESS("rcvsrpkt1 (srp)", srp->Source, srp->Dest, -1);
				//PRINT_ADDRESS("rcvsrpkt1 (real-srp)", srp->Source, srp->realDest, Ether->Type);
				PRINT_ADDRESS("rcvsrpkt1 (eth)", Ether->Source, Ether->Dest, -1);
			}
			*/
            MiniportReceiveLocally(VA, srp, MiniportReceiveComplete);
		}
        else {
		
			int max_fwd_slot ; //normalized slot in fwding list of highest priority forwarder
			int my_fwd_slot ; //normalized slot in fwding list of me
	
			InternalSourceRoute *old_sr;
			
			uint etxThresh ;
			uint j; 
			InternalSourceRoute *SR_;

			uint dcnt = 0  ;

			//if we already received packet, then we are done
			if (forwBitMapSet) {
				if(debug) KdPrint(("Already received packet %d, not forwarding.\n", srp->ackreq->opt.identification)); 
				goto FreeSrpAndReturn; 
			}
			
			//do some book-keeping for later--determines our forward timer	
			getFwdSlots(VA, srp, &max_fwd_slot, &my_fwd_slot); 
			ASSERT(max_fwd_slot != -1 && my_fwd_slot != -1); 
	
			if (srp->ackreq)
				if(debug) KdPrint(("Receive srpacket try forward: %d\n", srp->ackreq->opt.identification)); 
	
			old_sr = ExAllocatePool(NonPagedPool,
                            sizeof *old_sr + sizeof(SRAddr)*SOURCE_ROUTE_HOPS(srp->sr->opt.optDataLen));
			
			if (old_sr != NULL) {
				RtlCopyMemory(old_sr, srp->sr,
					      sizeof *old_sr + sizeof(SRAddr)*SOURCE_ROUTE_HOPS(srp->sr->opt.optDataLen));
				//readjust segmentsLeft
				old_sr->opt.segmentsLeft++; 
				srp->old_sr = old_sr;
			
			}


			//
            // We must forward the packet.
            //
			InterlockedIncrement((PLONG)&VA->CountXmitForwardUnicast);

			//
			// We need to update segmentsLeft for the appropiate node and get the new 
			// opt.max value. This method also preprocesses sr for LinkCacheInsertForwardingList
			//
			Status = updateForwardInfo(VA, srp); 
			if (Status != NDIS_STATUS_SUCCESS) {
				if (srp->ackreq)
					if(debug) KdPrint(("Can't updateForwardInfo b/c memory, dropping packet %d\n", srp->ackreq->opt.identification));
				else
					if(debug) KdPrint(("Can't updateForwardInfo b/c memory, dropping packet.\n"));
				goto FreeSrpAndReturn; 
			}

		
			//
			// Packets which dont' have an ACK request can just be forwarded.
			//
			if (srp->ackreq == NULL) {	
				//If this is not a data packet, then we just want one forwarder, so reset the
				//forwarding node bitmap to just the next hop.
				uint Index = SOURCE_ROUTE_HOPS(srp->sr->opt.optDataLen) - srp->sr->opt.segmentsLeft ;//next hop
				srp->sr->opt.fwd_bitmap = 0 ; 
				setSRPbitmap(&(srp->sr->opt.fwd_bitmap), Index);	

				//
				// This must be done *after* updateForwardInfo()
				//
				if (! LinkCacheUseSR(VA, srp)) {
	                InterlockedIncrement((PLONG)&VA->CountXmitForwardQueueFull);
		            goto FreeSrpAndReturn;
			    }
				MaintBufSendPacket(VA, srp, MiniportReceiveComplete);
				PRINT_ADDRESS("Forw uni pkt (srp)", srp->Source, srp->Dest, Index);
				//PRINT_ADDRESS("Forw uni pkt (real-srp)", srp->Source, srp->realDest, Ether->Type);
				//PRINT_ADDRESS("Forw uni pkt (eth)", Ether->Source, Ether->Dest, -1);
				return; 
			}

#if E2E_RATE_LIMIT_ENABLED	
			ASSERT(!isE2EAck);
#endif

			KeAcquireSpinLock(&(VA->LC->Lock), &OldIrql);	
			//get new forwarding list
			SR_ = LinkCacheFindForwarders(VA, srp, &etxThresh);
			if (SR_ != NULL) {
				uint i = 0 ; 
				for (i = 0 ; i < SOURCE_ROUTE_HOPS(SR_->opt.optDataLen); i++) {
					VirtualAddress *Address;
					Address = &(SR_->opt.hopList[i].addr) ;
					PRINT_VA("\t\t\tFFFFFFF Forwarding node: ", *Address); 
				} 	
			}


			Status = LinkCacheInsertForwardingList(VA, SR_, srp, etxThresh);
			if (Status != NDIS_STATUS_SUCCESS) {
				//no action required b/c we should still have the packet
				// prepared to send anyways.
				if(debug) KdPrint(("Unsuccessful in linkcacheinsertforwardinglist\n")); 
			}
			KeReleaseSpinLock(&(VA->LC->Lock), OldIrql);
			//for debug, can remove later.
			srp->updateTime = KeQueryInterruptTime(); 


			//debug
			Index = SOURCE_ROUTE_HOPS(srp->sr->opt.optDataLen) - srp->sr->opt.segmentsLeft ;
			if(debug) KdPrint(("%d hops on the fwd list, %d is next hop\n", SOURCE_ROUTE_HOPS(srp->sr->opt.optDataLen), Index));
			for (j = 0 ; j < SOURCE_ROUTE_HOPS(srp->sr->opt.optDataLen); j++) {
				VirtualAddress *Address;
				Address = &(srp->sr->opt.hopList[j].addr) ;
				if (checkSRPbitmap(srp->sr->opt.bitmap, j)) {
					dcnt++; 
					if (checkSRPbitmap(srp->sr->opt.fwd_bitmap, j))
						PRINT_VA("\t\t\tDFDFDFDFDF Forwarding node: ", *Address); 
					else
						PRINT_VA("\t\t\tDNDNDNDNDN Forwarding node: ", *Address); 
				} else {
					if (checkSRPbitmap(srp->sr->opt.fwd_bitmap, j))
						PRINT_VA("\t\t\tNFNFNFNFNF Forwarding node: ", *Address); 
					else
						PRINT_VA("\t\t\tNNNNNNNNNN Forwarding node: ", *Address); 
				}
			} 
			if(debug) KdPrint(("Done with %d default hops on the fwd list.\n", dcnt));


			//
			// This must be done *after* updateForwardInfo()
			//
			if (! LinkCacheUseSR(VA, srp)) {
                InterlockedIncrement((PLONG)&VA->CountXmitForwardQueueFull);
                goto FreeSrpAndReturn;
            }


			//
			// We want to piggy-back ACKs on data packets only.
			// The value 'lowerBound - 1' should point to the original
			// sender of the packet and 'lowerBound' should indicate the original
			// sender's next hop
			//
			if (srp->ackreq != NULL) {
				InternalAcknowledgement *curAck; 
				srp->ack = MaintBufFillInPiggyBackAck(VA, srp);

				//debug
				for (curAck = srp->ack ; curAck != NULL ; curAck = curAck->next) {
					PRINT_ADDRESS("forwardPacket Ack flow: ", curAck->opt.flowSource, curAck->opt.flowDest, 
						curAck->opt.identification);

#if E2E_DEBUG
					ASSERT(curAck->opt.isE2E == 0); 
#endif
				}

			}


			//If I am the highest priority node in the forwarding list, then I send immediately.
			//Otherwise, I must wait before forwarding.
			if (max_fwd_slot == my_fwd_slot) {	
				MaintBufSendPacket(VA, srp, MiniportForwardComplete);
			} else {
				//need to queue packet before sending
				KeAcquireSpinLock(&VA->RecoveryLock, &OldIrql);	
				LocalRecoveryAddForwarding(VA, srp, max_fwd_slot - my_fwd_slot); 
				KeReleaseSpinLock(&VA->RecoveryLock, OldIrql);
			}//end else priority
        }
    }
    else if (srp->req != NULL) {
        RouteRequest *RR = &srp->req->opt;
        uint Hops = ROUTE_REQUEST_HOPS(RR->optDataLen);
        uint Metric;

		//this was supposed to be done earlier, let's just do it now.
		if (srp->ackreq != NULL) {
				ExFreePool(srp->ackreq);
				srp->ackreq = NULL;
		}

        //
        // Find the metric for the link this packet just used to reach us.
        //
        Metric = LinkCacheLookupMetric(VA,
                                       RR->hopList[Hops - 1].addr,
                                       VA->Address,
                                       (LQSRIf) PA->Index,
                                       RR->hopList[Hops - 1].outif);
        if ((*VA->IsInfinite)(Metric)) {
            //
            // Obviously the link does exist, but we don't know the metric.
            //            
			Metric = 0;
        } else if ((*VA->IsBad)(Metric)) {
            //
            // We must ignore route requests that reach us over bad links.
            //
            goto FreeSrpAndReturn;
        }

        //
        // If we have already seen this Route Request before,
        // then ignore this one. This must be the last check
        // because it leaves state in the request table.
        //
        if (! ForwardRequestP(VA, srp))
            goto FreeSrpAndReturn;

        //
        // Add ourself to the Route Request Hop List.
        //
        RR->optDataLen = ROUTE_REQUEST_LEN(Hops + 1);
        RtlCopyMemory(RR->hopList[Hops].addr,
                      VA->Address, SR_ADDR_LEN);
        RR->hopList[Hops].inif = (LQSRIf) PA->Index;
        // hopList[Hops].outif will be initialized in ProtocolForwardRequest
        // and MiniportSendRouteReply.
        RR->hopList[Hops].Metric = Metric;

        if (VirtualAddressEqual(srp->req->opt.targetAddress,
                                VA->Address)) {
            //
            // This Route Request is for us, so send a Route Reply.
            //
            MiniportSendRouteReply(VA, srp);

            //
            // Then receive the packet locally.
            //
			/*
			if (srp->ackreq != NULL) {
				PRINT_ADDRESS("rcvsrpkt2 (srp)", srp->Source, srp->Dest, srp->ackreq->opt.identification);
				//PRINT_ADDRESS("rcvsrpkt2 (real-srp)", srp->Source, srp->realDest, Ether->Type);
				PRINT_ADDRESS("rcvsrpkt2 (eth)", Ether->Source, Ether->Dest, srp->ackreq->opt.identification);		
			} else {
				PRINT_ADDRESS("rcvsrpkt2 (srp)", srp->Source, srp->Dest, -1);
				//PRINT_ADDRESS("rcvsrpkt2 (real-srp)", srp->Source, srp->realDest, Ether->Type);
				PRINT_ADDRESS("rcvsrpkt2 (eth)", Ether->Source, Ether->Dest, -1);	
			}
			*/
            MiniportReceiveLocally(VA, srp, MiniportReceiveComplete);
        }
        else {
            //
            // Stall our execution briefly. This prevents undesirable
            // synchronization when multiple nodes simultaneously receive
            // a broadcast route request, and simultaneously try to forward
            // it and collide.
            //
            MiniportForwardingStall(VA);
            InterlockedIncrement((PLONG)&VA->CountXmitForwardBroadcast);
			
		            
			//
            // Broadcast the Route Request via our physical adapters.
            // Then completion function will receive the packet locally.
            //
            srp->TransmitComplete = MiniportForwardRequestComplete;
            MiniportForwardRequest(VA, srp);
        }
    }
    else {
        //
        // The packet does not have a Route Request or a Source Route.
        //
        goto FreeSrpAndReturn;
    }

    return;

  FreeSrpAndReturn:
    SRPacketFree(srp);
}

//* MiniportReceivePacket
//
//  Receive an LQSR packet from a virtual link.
//
void
MiniportReceivePacket(
    MiniportAdapter *VA,
    ProtocolAdapter *PA,
    NDIS_PACKET *Packet,
    void (*FreePacket)(ProtocolAdapter *PA, NDIS_PACKET *Packet))
{
    SRPacket *srp;
    NDIS_STATUS Status;
    EtherHeader UNALIGNED *Ether;

    InterlockedIncrement((PLONG)&VA->CountRecv);
	//KdPrint(("Got packet in MRP1\n"));


	Ether = NdisFirstBuffer(Packet)->MappedSystemVa;
	//PRINT_ADDRESS("sr bef", Ether->Source, Ether->Dest, Ether->Type) ; 
	Status = SRPacketFromPkt(VA, Packet, &srp);
	//PRINT_ADDRESS("sr aft", Ether->Source, Ether->Dest, Ether->Type) ; 
	
    if (Status != NDIS_STATUS_SUCCESS) {
		if(debug) KdPrint(("SRPacketFromPkt failed in MiniportReceive\n")); 
        (*FreePacket)(PA, Packet);
        return;
    }

    srp->PA = PA;
    srp->FreePacket = FreePacket;

    //
    // See if we should make this link behave worse than it is.
    // For testing purposes.
    //
    if (ArtificialDrop(VA, PA, srp)) {
		//if(debug) KdPrint(("Dropped packet bc of the Artificial Drop!\n")); 
        SRPacketFree(srp);
        return;
    }
	

    CacheSRPacket(VA, PA, srp);
    
	ReceiveSRPacket(VA, PA, srp);
}

//* MiniportSendRouteError
//
//  Given an SRPacket that Route Maintenance has given up on,
//  sends a Route Error.
//
//  Does not dispose of srp or srp->Packet.
//
void
MiniportSendRouteError(
    MiniportAdapter *VA,
    const SRPacket *srp)
{
    InternalRouteError *RE;
    uint Index;
    uint Metric;

    //
    // The Maintenance Buffer only works with Source Routed packets.
    //
    ASSERT(srp->sr != NULL);

    //
    // Do not send errors for statically routed
    // packets.
    //
    if (srp->sr->opt.staticRoute) 
        return;
   
    //
    // Do not send an error for an error.
    //
    if (srp->err != NULL)
        return;

    //
    // Figure out which hop in the Source Route has failed.
    //
    Index = SOURCE_ROUTE_HOPS(srp->sr->opt.optDataLen) - srp->sr->opt.segmentsLeft - 1;
    ASSERT(VirtualAddressEqual(srp->sr->opt.hopList[Index].addr, VA->Address));

    //
    // Do not send an error to ourselves.
    //
    if (Index == 0)
        return;

    //
    // Lookup the current metric for the failed link.
    // The link metric has already been penalized.
    //
    Metric = LinkCacheLookupMetric(VA,
                                   srp->sr->opt.hopList[Index].addr,
                                   srp->sr->opt.hopList[Index + 1].addr,
                                   srp->sr->opt.hopList[Index + 1].inif,
                                   srp->sr->opt.hopList[Index].outif);
    if (Metric == 0) {
        if (FindPhysicalAdapterFromIndex(VA,
                        srp->sr->opt.hopList[Index].outif) != NULL) {
            //
            // Hmm, the outgoing interface still exists.
            // Somehow the link managed to disappear from the link cache.
            //
            InterlockedIncrement((PLONG)&VA->CountXmitRouteErrorNoLink);
            return;
        }

        //
        // Report that the link is gone.
        //
        Metric = (uint)-1;
    }

    InterlockedIncrement((PLONG)&VA->CountXmitRouteError);

    //
    // Allocate the Route Error option.
    //
    RE = ExAllocatePool(NonPagedPool, sizeof *RE);
    if (RE == NULL)
        return;

    //
    // Initialize the Route Error option.
    //
    RtlZeroMemory(RE, sizeof *RE);
    RE->opt.optionType = LQSR_OPTION_TYPE_ERROR;
    RE->opt.optDataLen = ROUTE_ERROR_LENGTH;
    RtlCopyMemory(RE->opt.errorSrc, srp->sr->opt.hopList[Index].addr, SR_ADDR_LEN);
    RtlCopyMemory(RE->opt.errorDst, srp->sr->opt.hopList[0].addr, SR_ADDR_LEN);
    RtlCopyMemory(RE->opt.unreachNode, srp->sr->opt.hopList[Index + 1].addr,
                  SR_ADDR_LEN);
    RE->opt.inIf = srp->sr->opt.hopList[Index + 1].inif;
    RE->opt.outIf = srp->sr->opt.hopList[Index].outif;
    RE->opt.Metric = Metric;

    //
    // Send the Route Error option.
    //
    PbackSendOption(VA, RE->opt.errorDst,
                    (InternalOption *) RE,
                    PBACK_ROUTE_ERROR_TIMEOUT);
}

uint MiniportInfoRequestIdentifier = 0;

//* MiniportSendInfoRequest
//
//  Sends an Information Request.
//
void
MiniportSendInfoRequest(
    MiniportAdapter *VA)
{
    SRPacket *srp;
    InternalInfoRequest *IR;
    VirtualAddress Target;
    LQSRReqId Identifier;

    InterlockedIncrement((PLONG)&VA->CountXmitInfoRequest);

    //
    // Initialize a new SRPacket for the Information Request.
    //
    srp = ExAllocatePool(NonPagedPool, sizeof *srp);
    if (srp == NULL)
        return;
    RtlZeroMemory(srp, sizeof *srp);
	//srp->MAC = NULL; 

    //
    // Initialize the Information Request option.
    //
    IR = ExAllocatePool(NonPagedPool, sizeof *IR);
    if (IR == NULL) {
        SRPacketFree(srp);
        return;
    }

    srp->inforeq = IR;
    RtlZeroMemory(IR, sizeof *IR);
    IR->opt.optionType = LQSR_OPTION_TYPE_INFOREQ;
    IR->opt.optDataLen = INFO_REQUEST_LEN;
    IR->opt.identification =
        InterlockedIncrement((PLONG)&MiniportInfoRequestIdentifier);
    RtlCopyMemory(IR->opt.sourceAddress, VA->Address, SR_ADDR_LEN);

    //
    // We use a dummy Target address.
    //
    RtlZeroMemory(Target, SR_ADDR_LEN);

    Identifier = ReqTableIdentifier(VA, Target);
    MiniportGenerateRouteRequest(VA, srp, Target, Identifier,
                                 MiniportControlSendComplete);
}

//* MiniportSendInfo
//
//  Sends an Information packet.
//
static void
MiniportSendInfo(
    MiniportAdapter *VA,
    ProtocolAdapter *PA,
    const SRPacket *srp,
    const InternalInfoRequest *InfoReq)
{
    InternalInfoReply *InfoRep;

    //
    // Do not send an Info packet to ourself.
    //
    if (VirtualAddressEqual(InfoReq->opt.sourceAddress, VA->Address))
        return;

    InterlockedIncrement((PLONG)&VA->CountXmitInfoReply);

    //
    // Allocate the Info option.
    //
    InfoRep = ExAllocatePool(NonPagedPool, sizeof *InfoRep);
    if (InfoRep == NULL)
        return;

    //
    // Allocate the Info option.
    //
    RtlZeroMemory(InfoRep, sizeof *InfoRep);
    InfoRep->opt.optionType = LQSR_OPTION_TYPE_INFO;
    InfoRep->opt.optDataLen = INFO_REPLY_LEN;
    InfoRep->opt.identification = InfoReq->opt.identification;
    InfoRep->opt.version = Version;

    //
    // Send the Info option.
    //
    PbackSendOption(VA, InfoReq->opt.sourceAddress,
                    (InternalOption *) InfoRep,
                    PBACK_INFO_REPLY_TIMEOUT);

    //
    // This code is similar to MiniportSendRouteReply,
    // with two differences:
    // 1. We are not yet added to the Route Request option,
    // so we have to add ourself here to the Route Reply.
    // 2. The timeout is different.
    // 3. We do not increment CountXmitRouteReply.
    //
    if (srp->req != NULL) {
        uint Hops = ROUTE_REQUEST_HOPS(srp->req->opt.optDataLen);
        InternalRouteReply *RR;

        //
        // Allocate the Route Reply option.
        //
        RR = ExAllocatePool(NonPagedPool,
                            sizeof *RR + sizeof(SRAddr)*MAX_SR_LEN);
        if (RR == NULL)
            return;

        //
        // Initialize the Route Reply option.
        //
        RtlZeroMemory(RR, sizeof *RR + sizeof(SRAddr)*MAX_SR_LEN);

        RtlCopyMemory(RR->opt.hopList, srp->req->opt.hopList,
                      Hops * sizeof(SRAddr));

        if (Hops < MAX_SR_LEN) {
            //
            // Add ourself to the Route Reply.
            //
            RtlCopyMemory(RR->opt.hopList[Hops].addr,
                          VA->Address, SR_ADDR_LEN);
            RR->opt.hopList[Hops].inif = (LQSRIf) PA->Index;
            RR->opt.hopList[Hops].outif = 0;
            Hops++;
        }

        RR->opt.optionType = LQSR_OPTION_TYPE_REPLY;
        RR->opt.optDataLen = ROUTE_REPLY_LEN(Hops);

        //
        // Send the Route Reply option.
        //
        PbackSendOption(VA, InfoReq->opt.sourceAddress,
                        (InternalOption *) RR,
                        PBACK_INFO_REPLY_TIMEOUT);
    }
}

//* MiniportSendRouteRequest
//
//  Sends a Route Request.
//
void
MiniportSendRouteRequest(
    MiniportAdapter *VA,
    const VirtualAddress Target,
    LQSRReqId Identifier)
{
    SRPacket *srp;

    InterlockedIncrement((PLONG)&VA->CountXmitRouteRequest);

    srp = ExAllocatePool(NonPagedPool, sizeof *srp);
    if (srp == NULL)
        return;
    RtlZeroMemory(srp, sizeof *srp);
	//srp->MAC = NULL; 

    MiniportGenerateRouteRequest(VA, srp,
                                 Target, Identifier,
                                 MiniportControlSendComplete);
}

//* MiniportResetStatistics
//
//  Resets all counters and statistics gathering for the virtual adapter
//  and associated data structures.
//
void
MiniportResetStatistics(MiniportAdapter *VA)
{
    ProtocolAdapter *PA;
    KIRQL OldIrql;

    VA->CountPacketPoolFailure = 0;

    VA->CountXmit = 0;
    VA->CountXmitSmall = 0;
    VA->CountXmitLarge = 0;
    VA->CountXmitLocally = 0;
    VA->CountXmitMulticast = 0;
    VA->CountXmitRouteRequest = 0;
    VA->CountXmitRouteReply = 0;
    VA->CountXmitRouteError = 0;
    VA->CountXmitSendBuf = 0;
    VA->CountXmitMaintBuf = 0;
    VA->CountXmitForwardUnicast = 0;
    VA->CountXmitForwardBroadcast = 0;
    VA->CountXmitInfoRequest = 0;
    VA->CountXmitInfoReply = 0;
    VA->CountXmitProbe = 0;
    VA->CountXmitProbeReply = 0;
    VA->CountXmitLinkInfo = 0;
    VA->CountSalvageAttempt = 0;
    VA->CountSalvageStatic = 0;
    VA->CountSalvageOverflow = 0;
    VA->CountSalvageNoRoute = 0;
    VA->CountSalvageTransmit = 0;
    VA->CountRecv = 0;
    VA->CountRecvLocally = 0;
    VA->CountRecvLocallySalvaged = 0;
    VA->CountRecvBadMAC = 0;
    VA->CountRecvRouteRequest = 0;
    VA->CountRecvRouteReply = 0;
    VA->CountRecvRouteError = 0;
    VA->CountRecvAckRequest = 0;
    VA->CountRecvAck = 0;
    VA->CountRecvSourceRoute = 0;
	VA->CountRecvRecovery = 0 ;
    VA->CountRecvInfoRequest = 0;
    VA->CountRecvInfoReply = 0;
    VA->CountRecvDupAckReq = 0;
    VA->CountRecvProbe = 0;
    VA->CountRecvProbeReply = 0;
    VA->CountRecvLinkInfo = 0;
    VA->CountRecvRecursive = 0;
    VA->CountRecvEmpty = 0;
    VA->CountRecvSmall = 0;
    VA->CountRecvLarge = 0;
    VA->CountRecvDecryptFailure = 0;

    VA->MinVersionSeen = Version;
    VA->MaxVersionSeen = Version;

    VA->TotalForwardingStall = 0;

    KeAcquireSpinLock(&VA->Lock, &OldIrql);
    VA->ForwardMax = VA->ForwardNum;
    VA->CountForwardFast = 0;
    VA->CountForwardQueue = 0;
    VA->CountForwardDrop = 0;

    VA->TimeoutMinLoops = MAXULONG;
    VA->TimeoutMaxLoops = 0;

    for (PA = VA->PhysicalAdapters; PA != NULL; PA = PA->Next)
        ProtocolResetStatistics(PA);
    KeReleaseSpinLock(&VA->Lock, OldIrql);

    LinkCacheResetStatistics(VA);
    MaintBufResetStatistics(VA);
    SendBufResetStatistics(VA);
    ReqTableResetStatistics(VA);
    PbackResetStatistics(VA);
}

//* MiniportHalt
//
//  Called by NDIS. Cleans up and frees the virtual adapter.
//
//  BUGBUG: There is a race here. It is possible for a reference
//  to prevent the driver from unloading after MiniportHalt returns.
//  Then a new MiniportAdapter can be created and so we are running
//  but there is no ioctl access because there is no device.
//  Furthermore a subsequent MiniportHalt blue-screens.
//
void
MiniportHalt(
    IN NDIS_HANDLE MiniportAdapterContext)
{
    MiniportAdapter *VA = (MiniportAdapter *) MiniportAdapterContext;
    KIRQL OldIrql;
    NDIS_STATUS Status;
	AckCache *AC ; 
	int numOut; 


    if(debug) KdPrint(("MCL!MiniportHalt(VA %p)\n", VA));

    //
    // Remove the virtual adapter from our list.
    //
    KeAcquireSpinLock(&MiniportAdapters.Lock, &OldIrql);
    if (VA->Next != NULL)
        VA->Next->Prev = VA->Prev;
    *VA->Prev = VA->Next;
    KeReleaseSpinLock(&MiniportAdapters.Lock, OldIrql);

	KeAcquireSpinLock(&VA->RecoveryLock, &OldIrql);
	//ejr
	//We have to delete the packets from the local recovery cache
	//This may be a problem if cache packets received after this
	//method is called.
	if(debug) KdPrint(("MiniportHalt: Starting local recovery halt 1\n")); 
	LocalRecoveryHalt(VA);
	if(debug) KdPrint(("MiniportHalt: Ending local recovery halt 1\n")); 
	KeReleaseSpinLock(&VA->RecoveryLock, OldIrql);

	//have to delete our AckCache linked list.
	AC = VA->acl.head; 
	while (AC != NULL) {
		AckCache *tmp = AC->next; 
#if SLIDING_RATE_LIMIT_ENABLED	
		TimeStruct *tmpt = AC->AckTimes;
		while(tmpt) {
			TimeStruct *tmpn = tmpt->next; 
			ExFreePool(tmpt); 
			tmpt = tmpn; 
		}
		tmpt = AC->SendTimes;
		while(tmpt) {
			TimeStruct *tmpn = tmpt->next; 
			ExFreePool(tmpt); 
			tmpt = tmpn; 
		}
#endif

		ExFreePool(AC); 
		AC = tmp ; 
	}

	//delete the linked list for arb_dijkstra
	KeAcquireSpinLock(&VA->LC->Lock, &OldIrql);
	delete_arb_dijkstra_caches(VA);
	KeReleaseSpinLock(&VA->LC->Lock, OldIrql);

	/*
		ejr- Potential problem: LocalRecoveryHalt will delete all Holder
		entries.  However, if there are some outstanding packets in which
		LocalRecoveryTransmitComplete hasn't been called, then there could be a
		problem.  The reason is that after ProtocolCleanup is called, we are 
		not allowed to use any NdisXXX functions.  However, LocalRecoveryTransmitComplete
		will do this.  We need to ensure all LocalRecoveryTransmitComplete's are 
		finished beforet this point.
	*/


	if(debug) KdPrint(("MiniportHalt: Starting protocol cleanup\n")); 
	//might cause deadlock if we call this w/ lock held...
    ProtocolCleanup(VA);
	if(debug) KdPrint(("MiniportHalt: Ending protocol cleanup\n")); 

    //
    // Stop our periodic timer.
    // We need to do this after ProtocolCleanup,
    // so that MiniportTimeout can finish forwarding packets
    // so NdisDeregisterProtocol can complete.
    //
    KeCancelTimer(&VA->Timer);

	KeCancelTimer(&VA->RecoveryTimer); 

#if 0 // This API does not exist in WinXP.
    //
    // Wait for any DPCs (in particular our timer DPC) to finish.
    //
    KeFlushQueuedDpcs();
#endif

    //
    // If we are holding onto packets that we are originating,
    // we will not be halted. After NdisDeregisterProtocol,
    // we will not hold any packets that we are forwarding.
    // But we can still have internally-originated packets
    // in the list. We need to do this after stopping the timer
    // because PbackTimeout can add packets to VA->ForwardList.
    //
    MiniportForwardCleanup(VA);

	


    //ASSERT(NdisPacketPoolUsage(VA->PacketPool) == 0);
	numOut = NdisPacketPoolUsage(VA->PacketPool);
	if (numOut != 0) {
		if(debug) KdPrint(("\n\n1ERROR IN PACKET POOL USAGE\n %d AMOUNT OF PACKETS IN PACKET POOL\nERROR\n\n", numOut));

	}

    NeighborCacheCleanup(&VA->NC);
    PbackCleanup(VA);
    LinkCacheFree(VA);
    ReqTableFree(VA);
    SendBufFree(VA);
    MaintBufFree(VA);
    NdisFreePacketPool(VA->PacketPool);
    NdisFreeBufferPool(VA->BufferPool);
    ExFreePool(VA);


    //
    // If there are no more virtual adapters,
    // delete our custom device. This will allow us to unload.
    //
    if (MiniportAdapters.VirtualAdapters == NULL) {

        Status = NdisMDeregisterDevice(OurDeviceHandle);
        if(debug) KdPrint(("MCL!NdisMDeregisterDevice -> %x\n", Status));
    }
}

//* MiniportShutdown
//
//  Called by NDIS to shutdown the adapter.
//
void
MiniportShutdown(
    IN PVOID ShutdownContext)
{
    MiniportAdapter *VA = (MiniportAdapter *) ShutdownContext;

    UNREFERENCED_PARAMETER(VA);
    if(debug) KdPrint(("MCL!MiniportShutdown(VA %p)\n", VA));
}

//* MiniportIndicateStatusConnected
//
//  Indicate that our virtual adapter is now connected.
//
void
MiniportIndicateStatusConnected(MiniportAdapter *VA)
{
    if(debug) KdPrint(("MCL!MiniportIndicateStatusConnected(VA %p)\n", VA));

    NdisMIndicateStatus(VA->MiniportHandle, NDIS_STATUS_MEDIA_CONNECT,
                        NULL, 0);
    NdisMIndicateStatusComplete(VA->MiniportHandle);
}

//* MiniportIndicateStatusDisconnected
//
//  Indicate that our virtual adapter is now disconnected.
//
void
MiniportIndicateStatusDisconnected(
    MiniportAdapter *VA)
{
    if(debug) KdPrint(("MCL!MiniportIndicateStatusDisconnected(VA %p)\n", VA));

    NdisMIndicateStatus(VA->MiniportHandle, NDIS_STATUS_MEDIA_DISCONNECT,
                        NULL, 0);
    NdisMIndicateStatusComplete(VA->MiniportHandle);
}

//* MiniportSendLinkInfo
//
//  Sends a null Route Request for the purpose of disseminating
//  link information.
//
void
MiniportSendLinkInfo(
    MiniportAdapter *VA)
{
    const static VirtualAddress Dest = { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff };

    ASSERT(VA->MetricType != METRIC_TYPE_HOP);
    InterlockedIncrement((PLONG)&VA->CountXmitLinkInfo);

    MiniportSendRouteRequest(VA, Dest, ReqTableIdentifier(VA, Dest));
}

//* MiniportSendPeriodicLinkInfo
//
//  If we have not sent link information recently,
//  then do it now.
//
//  Called at DPC level from MiniportTimeout.
//
Time
MiniportSendPeriodicLinkInfo(
    MiniportAdapter *VA,
    Time Now)
{
    Time LastLinkInfo = InterlockedRead64((LONGLONG *)&VA->LastLinkInfo);

    ASSERT(VA->MetricType != METRIC_TYPE_HOP);

    //
    // If we have not sent link information recently,
    // do it now.
    //
    if (LastLinkInfo + LINKINFO_PERIOD <= Now) {
        //
        // This updates VA->LastLinkInfo if it succeeds.
        // But we use our local variable to guarantee that
        // our returned timeout is in the future.
        //
        LastLinkInfo = Now;
        MiniportSendLinkInfo(VA);
    }

    //
    // Return the time by which we should next send link information.
    //
    return LastLinkInfo + LINKINFO_PERIOD;
}

//* MiniportRescheduleTimeoutHelper
//
//  Reschedules the timer so that MiniportTimeout
//  is called no later than Timeout.
//
//  Called with the miniport lock held.
//
static void
MiniportRescheduleTimeoutHelper(
    MiniportAdapter *VA,
    Time Now,
    Time Timeout)
{
    LARGE_INTEGER Delay;

    ASSERT(Now < Timeout);
    ASSERT(Timeout < VA->Timeout);

    //
    // Switch to the new earlier time.
    //
    VA->Timeout = Timeout;

    //
    // We want to reschedule the timer.
    // Try to cancel it.
    //
    if (KeCancelTimer(&VA->Timer)) {
        //
        // We canceled the timer, so we can reschedule.
        //
        Delay.QuadPart = Now - VA->Timeout; // Negative so its relative.
        KeSetTimer(&VA->Timer, Delay, &VA->TimeoutDpc);
    }
    else {
        //
        // We failed to cancel the timer.
        // This means MiniportTimeout is already running.
        //
    }
}

//* MiniportRescheduleTimeout
//
//  Reschedules the timer so that MiniportTimeout
//  is called no later than Timeout.
//  NB: Unfortunately the granularity of the system clock
//  interrupt preventsf MiniportTimeout from waking up accurately.
//
//  Called with no locks held.
//
void
MiniportRescheduleTimeout(
    MiniportAdapter *VA,
    Time Now,
    Time Timeout)
{
    KIRQL OldIrql;

    KeAcquireSpinLock(&VA->Lock, &OldIrql);
    if (Timeout < VA->Timeout)
        MiniportRescheduleTimeoutHelper(VA, Now, Timeout);
    KeReleaseSpinLock(&VA->Lock, OldIrql);
}


//* MiniportTimeout
//
//  Called aperiodically (at least every MINIPORT_TIMEOUT milliseconds)
//  at DPC level.
//
static void
MiniportTimeout(
    PKDPC Dpc,
    void *Context,
    void *Unused1,
    void *Unused2)
{
    MiniportAdapter *VA = (MiniportAdapter *) Context;
    Time Now;
    Time Timeout;
    Time UpdateTimeout;
    LARGE_INTEGER TimerTimeout;
    uint NumLoops = 0;
	uint okay;
    
	UNREFERENCED_PARAMETER(Dpc);
    UNREFERENCED_PARAMETER(Unused1);
    UNREFERENCED_PARAMETER(Unused2);

    for (;;) {
        KeAcquireSpinLockAtDpcLevel(&VA->Lock);
        Now = KeQueryInterruptTime();

        //
        // If we have not reached the Timeout yet,
        // we should re-arm the timer and wait.
        //
        if (Now < VA->Timeout)
            break;

        NumLoops++;

        //
        // Set our next Timeout.
        // NB: KeQueryInterruptTime does not advance continuously.
        // It can and will jump forward when the system sleeps.
        // Hence VA->Timeout += (Time) (MINIPORT_TIMEOUT * 10000)
        // is not good: we will loop multiple times catching up
        // after a jump.
        //
        VA->Timeout = Now + (Time) (MINIPORT_TIMEOUT * 10000);
        KeReleaseSpinLockFromDpcLevel(&VA->Lock);

        //
        // Perform our timeout processing,
        // keeping track of a new desired timeout.
        // NB: MiniportRescheduleTimeout may also lower VA->Timeout.
        //
        Timeout = MAXTIME;

        if (VA->MetricTimeout != NULL) {
            //
            // Metric-specific procedures. 
            //
			okay = 1; 
#if STOP_LINK_METRIC
			//ejr- want to stop updating the link weights
			// so do nothing here.
			if (Now > VA->start_link_time + (Time) (LINK_MEASURE_TIME * SECOND))
			{
				okay = 0; 
				UpdateTimeout = Now + VA->MetricParams.Etx.ProbePeriod;
				ASSERT(Now < UpdateTimeout);
				if (UpdateTimeout < Timeout)
					Timeout = UpdateTimeout;
			}	
#endif
			if (okay)
			{
				UpdateTimeout = (*VA->MetricTimeout)(VA, Now);
				ASSERT(Now < UpdateTimeout);
				if (UpdateTimeout < Timeout)
					Timeout = UpdateTimeout;
			}
        }
	
		
        MiniportForwardDelayedRequests(VA, Now);

        SendBufCheck(VA);

        UpdateTimeout = MaintBufTimer(VA, Now);
        ASSERT(Now < UpdateTimeout);
        if (UpdateTimeout < Timeout)
            Timeout = UpdateTimeout;

        UpdateTimeout = PbackTimeout(VA, Now);
        ASSERT(Now < UpdateTimeout);
        if (UpdateTimeout < Timeout)
            Timeout = UpdateTimeout;


		okay = 1;
#if STOP_LINK_METRIC
		if (Now > VA->start_link_time + (Time) (LINK_MEASURE_TIME * SECOND))
		{
			Time LastLinkInfo;
			okay = 0 ;
			LastLinkInfo = InterlockedRead64((LONGLONG *)&VA->LastLinkInfo);
			//
			// If we have not sent link information recently,
			// do it now.
			//
			if (LastLinkInfo + LINKINFO_PERIOD <= Now) {
				//
				// This updates VA->LastLinkInfo if it succeeds.
				// But we use our local variable to guarantee that
				// our returned timeout is in the future.
				//
				LastLinkInfo = Now;
			}
			UpdateTimeout =  LastLinkInfo + LINKINFO_PERIOD;
			ASSERT(Now < UpdateTimeout);
			if (UpdateTimeout < Timeout)
				Timeout = UpdateTimeout;
		}
#endif
        if (okay && VA->MetricType != METRIC_TYPE_HOP) {
            UpdateTimeout = MiniportSendPeriodicLinkInfo(VA, Now);
            ASSERT(Now < UpdateTimeout);
            if (UpdateTimeout < Timeout)
                Timeout = UpdateTimeout;
        }

        //
        // Update VA->Timeout.
        //
        if (Timeout < VA->Timeout)
            VA->Timeout = Timeout;
    }

    if (NumLoops > VA->TimeoutMaxLoops)
        VA->TimeoutMaxLoops = NumLoops;
    if (NumLoops < VA->TimeoutMinLoops)
        VA->TimeoutMinLoops = NumLoops;

    //
    // Re-arm the timer so we get called again.
    //
    TimerTimeout.QuadPart = Now - VA->Timeout; // Negative so its relative.
    KeSetTimer(&VA->Timer, TimerTimeout, &VA->TimeoutDpc);
    KeReleaseSpinLockFromDpcLevel(&VA->Lock);
}


//* MiniportRescheduleRecoveryTimeoutHelper
//
//  Reschedules the timer so that MiniportRecoveryTimeout
//  is called no later than Timeout.
//
//  Called with the recovery lock held.
//
static void
MiniportRescheduleRecoveryTimeoutHelper(
    MiniportAdapter *VA,
    Time Now,
    Time Timeout)
{
    LARGE_INTEGER Delay;

    ASSERT(Now < Timeout);
    ASSERT(Timeout < VA->RecoveryTimeout);

    //
    // Switch to the new earlier time.
    //
    VA->RecoveryTimeout = Timeout;

    //
    // We want to reschedule the timer.
    // Try to cancel it.
    //
    if (KeCancelTimer(&VA->RecoveryTimer)) {
        //
        // We canceled the timer, so we can reschedule.
        //
        Delay.QuadPart = Now - VA->RecoveryTimeout; // Negative so its relative.
        KeSetTimer(&VA->RecoveryTimer, Delay, &VA->RecoveryTimeoutDpc);
    }
    else {
        //
        // We failed to cancel the timer.
        // This means MiniportRecoveryTimeout is already running.
        //
    }
}


//* MiniportRescheduleRecoveryTimeout
//
//  Reschedules the timer so that MiniportRecoveryTimeout
//  is called no later than RecoveryTimeout.
//  NB: Unfortunately the granularity of the system clock
//  interrupt prevents MiniportRecoveryTimeout from waking up accurately.
//
//  Called with recovery lock held.
//
void
MiniportRescheduleRecoveryTimeout(
    MiniportAdapter *VA,
    Time Now,
    Time Timeout)
{
    if (Timeout < VA->RecoveryTimeout)
        MiniportRescheduleRecoveryTimeoutHelper(VA, Now, Timeout);
}




//* MiniportRecoveryTimeout
//
//  Called aperiodically (at least every MINIPORT_RECOVERY_TIMEOUT milliseconds)
//  at DPC level.
//
static void
MiniportRecoveryTimeout(
    PKDPC Dpc,
    void *Context,
    void *Unused1,
    void *Unused2)
{
    MiniportAdapter *VA = (MiniportAdapter *) Context;
    Time Now;
    Time Timeout;
    Time UpdateTimeout;
    LARGE_INTEGER TimerTimeout;
    

    UNREFERENCED_PARAMETER(Dpc);
    UNREFERENCED_PARAMETER(Unused1);
    UNREFERENCED_PARAMETER(Unused2);

    KeAcquireSpinLockAtDpcLevel(&VA->RecoveryLock);
  
	Now = KeQueryInterruptTime();

    //
    // If we have not reached the Timeout yet,
    // we should re-arm the timer and wait.
    //
    if (Now < VA->RecoveryTimeout)
		goto ExitReturn;

    //
    // Set our next Timeout.
    // NB: KeQueryInterruptTime does not advance continuously.
    // It can and will jump forward when the system sleeps.
    // Hence VA->RecoveryTimeout += (Time) (MINIPORT_RECOVERY_TIMEOUT * 10000)
    // is not good: we will loop multiple times catching up
    // after a jump.
    //
    VA->RecoveryTimeout = Now + (Time) (MINIPORT_RECOVERY_TIMEOUT * 10000);
    
	//
	// Perform our timeout processing,
	// keeping track of a new desired timeout.
	// NB: MiniportRescheduleRecoveryTimeout may also lower VA->RecoveryTimeout.
	//
	Timeout = MAXTIME;

	//we will see if there are any packets in 
	//the local recovery queue which need to be sent.
	UpdateTimeout = LocalRecoveryCheckRecovery(VA) ;
	ASSERT(Now < UpdateTimeout);
	if (UpdateTimeout < Timeout)
		Timeout = UpdateTimeout;


	//
	// Update VA->RecoveryTimeout.
	//
	if (Timeout < VA->RecoveryTimeout)
		VA->RecoveryTimeout = Timeout;


	//
    // Re-arm the timer so we get called again.
    //
	Now = KeQueryInterruptTime();
ExitReturn:
    TimerTimeout.QuadPart = Now - VA->RecoveryTimeout; // Negative so its relative.
    KeSetTimer(&VA->RecoveryTimer, TimerTimeout, &VA->RecoveryTimeoutDpc);
    KeReleaseSpinLockFromDpcLevel(&VA->RecoveryLock);
}






//* MiniportUnload
//
//  Called by NDIS when the miniport is unloading.
//
void
MiniportUnload(
    PDRIVER_OBJECT DriverObject)
{
    UNREFERENCED_PARAMETER(DriverObject);
    if(debug) KdPrint(("MCL!MiniportUnload(Driver %p)\n", DriverObject));
    
    //
    // Restore the system default timer resolution.
    // NB: If other drivers have also modified the timer resolution,
    // the default is not restored until they also request it.
    //
    (void) ExSetTimerResolution(0, FALSE);

#if COUNTING_MALLOC
    DumpCountingMallocStats();
    UnloadCountingMalloc();
#endif
}

//* MiniportInit
//
//  Initializes the MCL miniport module.
//  Registers with NDIS as a miniport.
//
//  Returns FALSE to indicate failure.
//
boolint
MiniportInit(
    IN PDRIVER_OBJECT DriverObject,   // MCL driver object.
    IN PUNICODE_STRING RegistryPath)  // Path to our info in the registry.
{
    UNICODE_STRING DeviceName;
    UNICODE_STRING WinDeviceName;
    NDIS_MINIPORT_CHARACTERISTICS MPC;
    NDIS_STATUS Status;
    ULONG TimerResolution;
#if DBG
    LARGE_INTEGER PerformanceFrequency;
#endif

    //
    // Ask for 100us timer resolution and see what we get.
    // Typically we will get about 1ms.
    //
    TimerResolution = ExSetTimerResolution(1000, TRUE);
    if(debug) KdPrint(("ExSetTimerResolution -> %u\n", TimerResolution));

#if DBG
    //
    // See how often the interval clock interrupts.
    //
    if(debug) KdPrint(("KeQueryTimeIncrement -> %u\n", KeQueryTimeIncrement()));
    (void)KeQueryPerformanceCounter(&PerformanceFrequency);
    if(debug) KdPrint(("PerformanceFrequency -> %I64d\n", PerformanceFrequency));
#endif

    //
    // We maintain a list of virtual adapters.
    //
    KeInitializeSpinLock(&MiniportAdapters.Lock);
    ASSERT(MiniportAdapters.VirtualAdapters == NULL);

    //
    // Register with NDIS as a miniport.
    // As a result, NDIS will call MiniportInitialize
    // for each virtual adapter.
    //

    NdisMInitializeWrapper(&MiniportHandle, DriverObject, RegistryPath, NULL);
    if(debug) KdPrint(("MCL!NdisMInitializeWrapper -> %x\n", MiniportHandle));
    if (MiniportHandle == NULL)
        return FALSE;

    RtlZeroMemory(&MPC, sizeof MPC);
    MPC.MajorNdisVersion = NDIS_MINIPORT_MAJOR_VERSION;
    MPC.MinorNdisVersion = NDIS_MINIPORT_MINOR_VERSION;
    MPC.InitializeHandler = MiniportInitialize;
    MPC.QueryInformationHandler = MiniportQueryInformation;
    MPC.SetInformationHandler = MiniportSetInformation;
    MPC.SendPacketsHandler = MiniportTransmitPackets;
    MPC.ReturnPacketHandler = MiniportReturnPacket;
    MPC.HaltHandler = MiniportHalt;
    MPC.AdapterShutdownHandler = MiniportShutdown;

    Status = NdisMRegisterMiniport(MiniportHandle, &MPC, sizeof MPC);
    if(debug) KdPrint(("MCL!NdisMRegisterMiniport -> %x\n", Status));
    if (Status != NDIS_STATUS_SUCCESS) {
        NdisTerminateWrapper(MiniportHandle, NULL);
        return FALSE;
    }

    //
    // Create a custom device object that we can use for IOCTLs.
    //

    RtlInitUnicodeString(&DeviceName, DD_MCL_DEVICE_NAME);
    RtlInitUnicodeString(&WinDeviceName, L"\\??\\" WIN_MCL_BASE_DEVICE_NAME);

    Status = NdisMRegisterDevice(MiniportHandle,
                                 &DeviceName, &WinDeviceName,
                                 IoMajorFunctions,
                                 &OurDeviceObject,
                                 &OurDeviceHandle);
    if(debug) KdPrint(("MCL!NdisMRegisterDevice -> %x\n", Status));
	    if (Status != NDIS_STATUS_SUCCESS) {
        NdisTerminateWrapper(MiniportHandle, NULL);
        return FALSE;
    }

    //
    // Register an unload handler for cleanup.
    //
    NdisMRegisterUnloadHandler(MiniportHandle, MiniportUnload);

    return TRUE;
}
