// -*- mode: C++; tab-width: 4; indent-tabs-mode: nil -*- (for GNU Emacs)
//
// Copyright (c) Microsoft Corporation
//
// This file is part of the Microsoft Research Mesh Connectivity Layer.
// You should have received a copy of the Microsoft Research Shared Source
// license agreement (MSR-SSLA) for this software; see the file "license.txt".
// If not, please see http://research.microsoft.com/mesh/license.htm,
// or write to Microsoft Research, One Microsoft Way, Redmond, WA 98052-6399.
//

#include "headers.h"
#include "atheros.h"
#include <stdio.h>

#define debug_enabled 2

typedef struct ProtocolRequest {
    NDIS_REQUEST Request;
    KEVENT Event;
    NDIS_STATUS Status;
} ProtocolRequest;

#define PRINT_ADDRESS(f, s, d, t) KdPrint(("%s: from %02x-%02x-%02x-%02x-%02x-%02x\
	to %02x-%02x-%02x-%02x-%02x-%02x type %i", f, (s)[0], (s)[1], (s)[2], (s)[3], (s)[4], (s)[5], \
	(d)[0], (d)[1], (d)[2], (d)[3], (d)[4], (d)[5], t)) 


#define PRINT_VA(f, s)  KdPrint(("%s: from %02x-%02x-%02x-%02x-%02x-%02x\n", \
	f, (s)[0], (s)[1], (s)[2], (s)[3], (s)[4], (s)[5])) 


//* FindPhysicalAdapterFromIndex
//
//  Given an index, returns the physical adapter.
//
ProtocolAdapter *
FindPhysicalAdapterFromIndex(MiniportAdapter *VA, uint Index)
{
    ProtocolAdapter *PA;
    KIRQL OldIrql;

    KeAcquireSpinLock(&VA->Lock, &OldIrql);
    for (PA = VA->PhysicalAdapters;
         PA != NULL;
         PA = PA->Next) {
        //
        // Is this the desired physical adapter?
        //
        if (PA->Index == Index)
            break;
    }
    KeReleaseSpinLock(&VA->Lock, OldIrql);

    return PA;
}

//* FindPhysicalAdapterFromGuid
//
//  Given a guid, returns the physical adapter.
//
ProtocolAdapter *
FindPhysicalAdapterFromGuid(MiniportAdapter *VA, const GUID *Guid)
{
    ProtocolAdapter *PA;
    KIRQL OldIrql;

    KeAcquireSpinLock(&VA->Lock, &OldIrql);
    for (PA = VA->PhysicalAdapters;
         PA != NULL;
         PA = PA->Next) {
        //
        // Is this the desired physical adapter?
        //
        if (RtlEqualMemory(Guid, &PA->Guid, sizeof(GUID)))
            break;
    }
    KeReleaseSpinLock(&VA->Lock, OldIrql);

    return PA;
}

//* ProtocolQueueFull
//
//  Returns TRUE if the transmit queue of the specified
//  interface is full.
//
boolint
ProtocolQueueFull(MiniportAdapter *VA, LQSRIf OutIf)
{
    ProtocolAdapter *PA;

    PA = FindPhysicalAdapterFromIndex(VA, OutIf);
	if (PA == NULL) {
		if(debug) KdPrint(("PA == NULL in ProtocolQueueFull.\n")); 
	} else {
		if(debug) KdPrint(("Count outstanding: %d and myCache.numPackets: %d and mlq.count: %d\n", 
						PA->CountSentOutstanding, VA->myCache.numPackets, VA->mlq.count)); 
	}
    //ejr- we'll bound the number of packets based on the number of outstanding retransmissions as well.
	return (PA != NULL) && ((VA->mlq.count >= MAC_LIMIT_MAX_SIZE) || (VA->myCache.numPackets  >= PROTOCOL_MAX_QUEUE) || 
			(PA->CountSentOutstanding >= PROTOCOL_MAX_QUEUE));
}

//* ProtocolResetStatistics
//
//  Resets all counters and statistics gathering for the physical adapter.
//
//  The caller must hold a reference for the physical adapter
//  or have the virtual adapter locked.
//
void
ProtocolResetStatistics(ProtocolAdapter *PA)
{
    PA->CountXmitLarge = 0;
    PA->PacketsSent = 0;
    PA->PacketsReceived = 0;
    PA->PacketsReceivedTD = 0;
    PA->PacketsReceivedFlat = 0;

    PA->CountPacketPoolFailure = 0;

    PA->MaxRecvOutstanding = PA->CountRecvOutstanding;
    PA->MaxSentOutstanding = PA->MaxSentOutstanding;
}

//* ProtocolRequestHelper
//
//  This is a utility routine to submit a general request to an NDIS
//  driver.  The caller specifes the request code (OID), a buffer and
//  a length.  This routine allocates a request structure, fills it in,
//  and submits the request.
//
NDIS_STATUS
ProtocolRequestHelper(
    ProtocolAdapter *PA,
    NDIS_REQUEST_TYPE RT,   // Type of request to be done (Set or Query).
    NDIS_OID OID,           // Value to be set/queried.
    void *Info,             // Pointer to the buffer to be passed.
    uint Length,            // Length of data in above buffer.
    uint *Needed)           // Location to fill in with bytes needed in buffer.
{
    ProtocolRequest Request;
    NDIS_STATUS Status;

    // Now fill it in.
    Request.Request.RequestType = RT;
    if (RT == NdisRequestSetInformation) {
        Request.Request.DATA.SET_INFORMATION.Oid = OID;
        Request.Request.DATA.SET_INFORMATION.InformationBuffer = Info;
        Request.Request.DATA.SET_INFORMATION.InformationBufferLength = Length;
    } else {
        Request.Request.DATA.QUERY_INFORMATION.Oid = OID;
        Request.Request.DATA.QUERY_INFORMATION.InformationBuffer = Info;
        Request.Request.DATA.QUERY_INFORMATION.InformationBufferLength = Length;
    }

    //
    // Note that we can NOT use PA->Event and PA->Status here.
    // There may be multiple concurrent ProtocolRequestHelper calls.
    //

    KeInitializeEvent(&Request.Event, SynchronizationEvent, FALSE);

    NdisRequest(&Status, PA->Handle, &Request.Request);

    if (Status == NDIS_STATUS_PENDING) {
        (void) KeWaitForSingleObject(&Request.Event, UserRequest,
                                     KernelMode, FALSE, NULL);
        Status = Request.Status;
    }

    if (Needed != NULL)
        *Needed = Request.Request.DATA.QUERY_INFORMATION.BytesNeeded;

    return Status;
}

//* IsAtherosAdapter
//
//  Returns TRUE if the physical adapter is known to be based on the
//  Atheros chipset.
//
boolint
IsAtherosAdapter(ProtocolAdapter *PA)
{
    uchar Buffer[100];
    NDIS_STATUS Status;
    int Loop;
    static char *Descriptions[] = {
        "Atheros",
        "ORiNOCO 802.11ag ComboCard Gold",
        "NETGEAR WAG511 802.11a/b/g Dual Band Wireless PC Card",
        "WLAN AR5001X+ Wireless Network Adapter",
        NULL};

    Status = ProtocolRequestHelper(PA, NdisRequestQueryInformation,
                                   OID_GEN_VENDOR_DESCRIPTION,
                                   Buffer,
                                   sizeof Buffer,
                                   NULL);
    if (Status != NDIS_STATUS_SUCCESS) {
        if(debug) KdPrint(("MCL!ProtocolQuery: OID_GEN_VENDOR_DESCRIPTION: %x\n",
                 Status));
        return FALSE;
    }

    for (Loop = 0; Descriptions[Loop] != NULL; Loop++) {
        uint Len;

        Len = strlen(Descriptions[Loop]);
        if (RtlCompareMemory(Buffer, Descriptions[Loop], Len) == Len)
            return TRUE;
    }

    return FALSE;
}

//* ProtocolQuery
//
//  Returns information about the physical adapter.
//
void
ProtocolQuery(
    ProtocolAdapter *PA,
    MCL_INFO_PHYSICAL_ADAPTER *Info)
{
    NDIS_STATUS Status;

    RtlCopyMemory(Info->Address, PA->Address, sizeof(PhysicalAddress));
    Info->Use = PA->Use;
    Info->MaxFrameSize = PA->MaxFrameSize;
    Info->MaxTotalSize = PA->MaxTotalSize;
    Info->Promiscuous = PA->Promiscuous;
    Info->ReceiveOnly = PA->ReceiveOnly;
    Info->Channel = PA->Channel;
    Info->Bandwidth = PA->Bandwidth;

    Info->CountXmitLarge = PA->CountXmitLarge;
    Info->PacketsSent = PA->PacketsSent;
    Info->PacketsReceived = PA->PacketsReceived;
    Info->PacketsReceivedTD = PA->PacketsReceivedTD;
    Info->PacketsReceivedFlat = PA->PacketsReceivedFlat;

    Info->CountPacketPoolFailure = PA->CountPacketPoolFailure;

    Info->CountRecvOutstanding = PA->CountRecvOutstanding;
    Info->MaxRecvOutstanding = PA->MaxRecvOutstanding;
    Info->CountSentOutstanding = PA->CountSentOutstanding;
    Info->MaxSentOutstanding = PA->MaxSentOutstanding;

    Status = ProtocolRequestHelper(PA, NdisRequestQueryInformation,
                                   OID_802_11_BSSID,
                                   &Info->W.BSSID, sizeof Info->W.BSSID,
                                   NULL);
    if (Status != NDIS_STATUS_SUCCESS) {
        if(debug) KdPrint(("MCL!ProtocolQuery: OID_802_11_BSSID: %x\n", Status));
        Info->Type = MCL_PHYSICAL_ADAPTER_ETHERNET;
        //
        // Initialize the 802.11-specific parameters to something.
        //
        memset(&Info->W.BSSID, 0, sizeof Info->W.BSSID);
        Info->W.SSID.SsidLength = 0;
        Info->W.Mode = (NDIS_802_11_NETWORK_INFRASTRUCTURE) -1;
        memset(&Info->W.Radio, 0, sizeof Info->W.Radio);
        Info->W.PowerMode = (NDIS_802_11_POWER_MODE) -1;
        Info->W.TxPowerLevel = 0;
        Info->W.RTSThreshold = (NDIS_802_11_RTS_THRESHOLD) -1;
        memset(&Info->W.Statistics, 0, sizeof Info->W.Statistics);
        return;
    }

    Info->Type = MCL_PHYSICAL_ADAPTER_802_11;
    Status = ProtocolRequestHelper(PA, NdisRequestQueryInformation,
                                   OID_802_11_SSID,
                                   &Info->W.SSID, sizeof Info->W.SSID,
                                   NULL);
    if (Status != NDIS_STATUS_SUCCESS) {
        if(debug) KdPrint(("MCL!ProtocolQuery: OID_802_11_SSID: %x\n", Status));
        Info->W.SSID.SsidLength = 0;
    }

    Status = ProtocolRequestHelper(PA, NdisRequestQueryInformation,
                                   OID_802_11_INFRASTRUCTURE_MODE,
                                   &Info->W.Mode, sizeof Info->W.Mode,
                                   NULL);
    if (Status != NDIS_STATUS_SUCCESS) {
        if(debug) KdPrint(("MCL!ProtocolQuery: OID_802_11_INFRASTRUCTURE_MODE: %x\n", Status));
        Info->W.Mode = (NDIS_802_11_NETWORK_INFRASTRUCTURE) -1;
    }

    Status = ProtocolRequestHelper(PA, NdisRequestQueryInformation,
                                   OID_802_11_CONFIGURATION,
                                   &Info->W.Radio, sizeof Info->W.Radio,
                                   NULL);
    if (Status != NDIS_STATUS_SUCCESS) {
        if(debug) KdPrint(("MCL!ProtocolQuery: OID_802_11_CONFIGURATION: %x\n", Status));
        memset(&Info->W.Radio, 0, sizeof Info->W.Radio);
    } else {
        //
        // We've encounted some radios that return the frequency in
        // incorrect units.  Attempt to compensate.
        //
        if (Info->W.Radio.DSConfig < 2412000)
            Info->W.Radio.DSConfig = Info->W.Radio.DSConfig * 100;
    }

    Status = ProtocolRequestHelper(PA, NdisRequestQueryInformation,
                                   OID_802_11_POWER_MODE,
                                   &Info->W.PowerMode,
                                   sizeof Info->W.PowerMode,
                                   NULL);
    if (Status != NDIS_STATUS_SUCCESS) {
        if(debug) KdPrint(("MCL!ProtocolQuery: OID_802_11_POWER_MODE: %x\n", Status));
        Info->W.PowerMode = (NDIS_802_11_POWER_MODE) -1;
    }

    Status = ProtocolRequestHelper(PA, NdisRequestQueryInformation,
                                   OID_802_11_TX_POWER_LEVEL,
                                   &Info->W.TxPowerLevel,
                                   sizeof Info->W.TxPowerLevel,
                                   NULL);
    if (Status != NDIS_STATUS_SUCCESS) {
        if(debug) KdPrint(("MCL!ProtocolQuery: OID_802_11_TX_POWER_LEVEL: %x\n", Status));
        Info->W.TxPowerLevel = 0;
    }

    Status = ProtocolRequestHelper(PA, NdisRequestQueryInformation,
                                   OID_802_11_RTS_THRESHOLD,
                                   &Info->W.RTSThreshold,
                                   sizeof Info->W.RTSThreshold,
                                   NULL);
    if (Status != NDIS_STATUS_SUCCESS) {
        if(debug) KdPrint(("MCL!ProtocolQuery: OID_802_11_RTS_THRESHOLD: %x\n", Status));
        Info->W.RTSThreshold = (NDIS_802_11_RTS_THRESHOLD) -1;
    }

    Status = ProtocolRequestHelper(PA, NdisRequestQueryInformation,
                                   OID_802_11_STATISTICS,
                                   &Info->W.Statistics,
                                   sizeof Info->W.Statistics,
                                   NULL);
    if (Status != NDIS_STATUS_SUCCESS) {
        if(debug) KdPrint(("MCL!ProtocolQuery: OID_802_11_STATISTICS: %x\n", Status));
        memset(&Info->W.Statistics, 0, sizeof Info->W.Statistics);
    }

    Info->Atheros = IsAtherosAdapter(PA);
    if (!Info->Atheros) {
        //
        // Initialize the Atheros-specific parameters to something.
        //
        Info->W.AtherosTxPowerLevel = 0;
        Info->W.AtherosTxLinkSpeed = 0;
        return;
    }

    Status = ProtocolRequestHelper(PA, NdisRequestQueryInformation,
                                   OID_ATHEROS_TX_POWER_LEVEL,
                                   &Info->W.AtherosTxPowerLevel,
                                   sizeof Info->W.AtherosTxPowerLevel,
                                   NULL);
    if (Status != NDIS_STATUS_SUCCESS) {
        if(debug) KdPrint(("MCL!ProtocolQuery: OID_ATHEROS_TX_POWER_LEVEL: %x\n",
                 Status));
        Info->W.AtherosTxPowerLevel = 0;
    }

    Status = ProtocolRequestHelper(PA, NdisRequestQueryInformation,
                                   OID_ATHEROS_TX_LINK_SPEED,
                                   &Info->W.AtherosTxLinkSpeed,
                                   sizeof Info->W.AtherosTxLinkSpeed,
                                   NULL);
    if (Status != NDIS_STATUS_SUCCESS) {
        if(debug) KdPrint(("MCL!ProtocolQuery: OID_ATHEROS_TX_LINK_SPEED: %x\n",
                 Status));
        Info->W.AtherosTxLinkSpeed = 0;
    }
}

//* ProtocolControl
//
//  Sets protocol configuration parameters from the structure.
//
//  Return values:
//      STATUS_INVALID_PARAMETER_3      Bad Channel.
//      STATUS_INVALID_PARAMETER_4      Bad Bandwidth.
//      STATUS_SUCCESS
//
NTSTATUS
ProtocolControl(
    ProtocolAdapter *PA,
    MCL_CONTROL_PHYSICAL_ADAPTER *Control)
{
    //
    // Check parameters before any updates.
    //

    if (Control->Channel != 0) {
        if (Control->Channel >= 256)
            return STATUS_INVALID_PARAMETER_3;
    }

    if (Control->Bandwidth != 0) {
        if (((Control->Bandwidth >> 2) >= 1024) ||
            ((Control->Bandwidth >> 2) == 0))
            return STATUS_INVALID_PARAMETER_4;
    }

    //
    // Update the attributes of the physical adapter.
    // Note that the "Use" flag cannot be changed at run-time.
    //

    ExAcquireFastMutex(&PA->Mutex);

    if (Control->ReceiveOnly != (boolint) -1)
        PA->ReceiveOnly = Control->ReceiveOnly;

    if (Control->Channel != 0) {
        PA->Channel = Control->Channel;
        PA->ChannelConfigured = TRUE;
    }

    if (Control->Bandwidth != 0) {
        PA->Bandwidth = Control->Bandwidth;
        PA->BandwidthConfigured = TRUE;
    }

    ExReleaseFastMutex(&PA->Mutex);
    return STATUS_SUCCESS;
}

//* ProtocolOpenAdapterComplete
//
//  Called by NDIS when an NdisOpenAdapter completes.
//
void
ProtocolOpenAdapterComplete(
    NDIS_HANDLE Handle,       // Binding handle.
    NDIS_STATUS Status,       // Final status of command.
    NDIS_STATUS ErrorStatus)  // Extra status for some errors.
{
    ProtocolAdapter *PA = (ProtocolAdapter *)Handle;

    UNREFERENCED_PARAMETER(ErrorStatus);

    if(debug) KdPrint(("MCL!ProtocolOpenAdapterComplete(VA %p PA %p)\n",
             PA->VirtualAdapter, PA));

    //
    // Signal whoever is waiting and pass the final status.
    //
    PA->Status = Status;
    KeSetEvent(&PA->Event, 0, FALSE);
}

//* ProtocolCloseAdapterComplete
//
//  Called by NDIS when an NdisCloseAdapter completes.
//
//  At this point, NDIS guarantees that it has no other outstanding
//  calls to us.
//
void
ProtocolCloseAdapterComplete(
    NDIS_HANDLE Handle,  // Binding handle.
    NDIS_STATUS Status)  // Final status of command.
{
    ProtocolAdapter *PA = (ProtocolAdapter *)Handle;
    if(debug) KdPrint(("MCL!ProtocolCloseAdapterComplete(VA %p PA %p)\n",
             PA->VirtualAdapter, PA));

    //
    // Signal whoever is waiting and pass the final status.
    //
    PA->Status = Status;
    KeSetEvent(&PA->Event, 0, FALSE);
}


#if SLIDING_RATE_LIMIT_ENABLED

//
// Given the current rate-limiting state, should we drop a packet?
// Returns true if so, false otherwise.
//
//
boolint shouldRateLimitDrop(MiniportAdapter *VA, 
							AckCache *AC)
{
	ASSERT(AC); 
	ASSERT(VirtualAddressEqual(VA->Address, AC->source));

	//prune lists first
	pruneTimeStruct(&(AC->SendTimes), &(AC->lastSendTimes), &(AC->numSendTimes));
	pruneTimeStruct(&(AC->AckTimes), &(AC->lastAckTimes), &(AC->numAckTimes));

	if (AC->numAckTimes >= AC->numSendTimes) {
		//do nothing--we're ok to send
		if(debug) KdPrint(("NumAckTimes: %u and numSendTimes; %u\n",AC->numAckTimes, AC->numSendTimes));
		return FALSE; 
	} else {
		uint diff = AC->numSendTimes - AC->numAckTimes;
		uint fdr = LOSS_SCALE - AC->virtualLoss;  //a virtual loss of 0.05 is represented by 50, so the FDR would be 950
		if(debug) KdPrint(("NumAckTimes: %u and numSendTimes; %u\n",AC->numAckTimes, AC->numSendTimes));
		ASSERT(fdr);
		if (diff > (LOSS_SCALE*SLIDING_RATE_LIMIT_THRESH*AC->numForwarders)/fdr ) {//might need scaling, no fp
			if(debug) KdPrint(("Dropping packet due to rate limiting, diff: %u numForw: %u fdr: %u\n",diff, AC->numForwarders, fdr));
			return TRUE;
		}
	}
	return FALSE; 
}
//
// Prunes the linked lists used for rate-limiting
// 
// TimeStruct **ts - pointer to a AC->TimeStruct (head) variable
// uint *count - pointer to AC->numTimeStruct
// 
// Both ts and count are updated and returned
//
void pruneTimeStruct(TimeStruct **ts, TimeStruct **last_ts, uint *count) {
	Time Add = KeQueryInterruptTime(); //Now
	int tmpCnt = *count; 
	TimeStruct *tmp = *ts;
	TimeStruct *head = *ts;  //head of linked list
	TimeStruct *tail = *last_ts; //tail of linked list

	while (tmp != NULL) {
		TimeStruct *tmpnext = tmp->next; 
		if (tmp->timestamp + SLIDING_RATE_LIMIT_TIMEOUT < Add) {
			//remove tmp
			if (tmp->prev)
				tmp->prev->next = tmp->next; 
			else
				head = tmp->next;
			if (tmp->next)
				tmp->next->prev = tmp->prev; 
			else
				tail = tmp->prev; 

			ExFreePool(tmp); 
			tmpCnt--; 
		} else {
			break; //all times from now on are ok
		}

        tmp = tmpnext; 	
	}

	//update variables and return
	*ts = head; 
	*last_ts = tail ; 
	*count = tmpCnt; 
}


//
// Updates the linked lists used for rate-limiting
// 
// TimeStruct **ts - pointer to a AC->TimeStruct variable
// TimeStruct **last_ts - pointer to tail of AC->TimeStruct LL
// uint *count - pointer to AC->numTimeStruct
// 
// Both ts and count are updated and returned
//
void updateTimeStruct(TimeStruct **ts, TimeStruct **last_ts, uint *count) {
	Time Add = KeQueryInterruptTime(); //Now
	uint tmpCnt = *count; 
	TimeStruct *tmp ;
	TimeStruct *head = *ts;  //head of linked list
	TimeStruct *tail = *last_ts; //tail of linked list

	//perform pruning on the list
	pruneTimeStruct(&head, &tail, &tmpCnt);

	//add the current entry (always to end)
	tmp = ExAllocatePool(NonPagedPool, sizeof *tmp); //reuse tmp
	if (tmp == NULL) {
		if(debug) KdPrint(("Can't alloc memory in updateTimeStruct.\n")); 	
		goto uts_done ;
	}
    RtlZeroMemory(tmp, sizeof *tmp);
	tmp->next = NULL;
	if (tmpCnt == 0) {
		ASSERT(tail == NULL);
	}
	tmp->prev = tail; 
	if (tail)
		tail->next = tmp; 
    tmp->timestamp = Add; 
	tail = tmp; 
	if (tmpCnt == 0) {
		ASSERT(head == NULL);
		head = tmp; 
	}
	tmpCnt++; 

uts_done:
	//update variables to return
	*ts = head; 
	*last_ts = tail ; 
	*count = tmpCnt; 
}
#endif

//* ProtocolIsProbe
// 
// Returns true when the packet is a probe packet. Determines
// this based upon the PC(Packet)->TransmitComplete function
//
boolint 
ProtocolIsProbe(NDIS_PACKET *Packet)
{
	if (PC(Packet)->TransmitComplete == MiniportProbeComplete ||
		PC(Packet)->TransmitComplete == RttSendProbeComplete ||	
		PC(Packet)->TransmitComplete == PktPairSendProbeComplete ||
		PC(Packet)->TransmitComplete == UltSendUcastProbeComplete) {
			return 1; 
		}
		
	return 0; 
}


//* preIFQCanDelete
//
// Given the current Ack, can I delete srp?
//
// Conditions we check:
//  1. Higher priority on old list
//  2. Higher priority node on new list
//  3. myETX >> fwderETX
//
// This method also used by: LocalRecoveryProcessAck
//
// higherETX is condition #3 from above.
//
boolint
preIFQCanDelete(MiniportAdapter *VA,   
				InternalAcknowledgement *Ack, 
				SRPacket *srp,
				boolint higherETX) 
{
	//
	// ProtocolRemoveFromPreIFQ, the caller of this method, has
	// already verified that we are on the same flow. Now just
	// need to check the sequence numbers.
	//

	//AckCacheDeletePacket(Ack, srp->ackreq->opt.identification)
	boolint ACKokay = AckCacheDeletePacket(Ack, srp->ackreq->opt.identification);
	if (!ACKokay) {
		return 0 ; 
	}

    //
    // See if node that sent the ACK is higher priority
	// First, we check 'new' source route, then we check 'old' source route
	if (higherETX || HigherForwardPriority(VA, &(srp->sr->opt), &(Ack->opt.from))) {
		return 1 ;
	} else if (srp->old_sr && HigherForwardPriority(VA, &(srp->old_sr->opt), &(Ack->opt.from))) {
		return 1; 
	}
	return 0 ;
}


//* ProtocolRemoveFromPreIFQ
//
//  Called by when a packet in the preIFQ 
// should be removed. For instance, when an ACK
// is received for a packet we are about to send.
//
// We don't want to touch CountSentOutstanding
// or call ProtocolCheckQueue.
// 
// We have to ensure that the packet we want
// to delete is not outstanding.
//
// Assumes that caller takes the RecoveryLock
//
// Also assumes that the deletion on the RecoveryCache was
// already called.
//
//
// if 'ackAlreadyUpdated' is set, then we don't have to update
//  the ack cache
//
// 'higherETX' is calculated in LocalRecoveryProcessACK
//	--says if myETX >> fwdETX
//
/*
*  Conditions to delete:
* 1. Packet ACK'd by higher priority node
*	--All data packets have piggy-backed ACKs,
*     so we only need to consider ACK state?
*   --How to check higher priority?
*		Possible solution:
*			1. If ACK from higher priority node on pkt we're storing: del
*			2. If ACK from higher priority node on the pkt we 
*					received: del
*			3. If myETX >> fwderETX: del  <--also case were original sender sends
*			Want to prevent case where forwarders have very similar
*			ETX values (prevent flapping). But also case were higher
*			priority node sends data, but not on any of my fwding lists
*			(because some other node may have sent or I am original sender).
*			--For now, we ignore case 3.
*
*	
*/
void
ProtocolRemoveFromPreIFQ(
    MiniportAdapter *VA,   
	InternalAcknowledgement *Ack,
	boolint ackAlreadyUpdated,
	boolint higherETX) //the ACK we just received
{
	//assuming only one interface for now
	ProtocolAdapter *PA = VA->PhysicalAdapters;
	MACLimitHolder *mlh_; 


	//go thru all packets in preIFQ to see if any can be deleted.
	for (mlh_ = VA->mlq.firstHolder ; mlh_ != NULL ; ) {
		//may be able to AND on srp->ack b/c all data pkts piggy-back	
		if (!mlh_->outstanding && !protocolIsACK(VA, mlh_->Packet) &&
			PC(mlh_->Packet)->srp && PC(mlh_->Packet)->srp->ackreq &&
			PC(mlh_->Packet)->srp->sr) {
				
				PNDIS_PACKET savePacket = mlh_->Packet;
				SRPacket *srp = PC(savePacket)->srp;
				if (VirtualAddressEqual(Ack->opt.flowDest, srp->Dest) &&
					VirtualAddressEqual(Ack->opt.flowSource, srp->Source)) {

						//For now just check against (new) forwarding list
						if (preIFQCanDelete(VA, Ack, srp, higherETX)) {
							if (!ackAlreadyUpdated) {
								//update our ack cache b/c ack rcv'd from higher priority node
								AckCache *AC; 
								AC = MaintBufFindAckFlow(VA, &Ack->opt.flowSource, 
										&Ack->opt.flowDest, Ack->opt.inif, Ack->opt.outif, NULL); 
								AckCacheUpdateAckState(AC, Ack);
								ackAlreadyUpdated = TRUE; 
							}
							ASSERT(PC(mlh_->Packet)->TransmitComplete == PC(savePacket)->TransmitComplete);
							if(debug) KdPrint(("Deleting packet %u from preIFQ\n", srp->ackreq->opt.identification));
							mlh_ = mlh_->next ;
							ProtocolDequeue(PA, savePacket); 			
							srp->sendCompleteShouldDelete = (boolint) 1 ;
							//If this packet has not been sent (ie- we've never put it in the recoverycache),
							//then we need to revert the TransmitComplete methods back to ensure that the 
							//data actually gets deleted. If a packet gets put in the recoverycache, then
							// the PC(packet)->TransmitComplete function is immediately changed.
							if (PC(savePacket)->TransmitComplete == MaintBufStaticSendComplete) {
								if (srp->TransmitComplete == MiniportRetransmitSendComplete)
									srp->TransmitComplete = MiniportSendComplete ;
								else if (srp->TransmitComplete == MiniportForwardComplete) {
									srp->TransmitComplete = MiniportReceiveComplete; 
								} 								
							}

							// Assumption is that recovery cache
							// deletion is always called before this function.
							(*PC(savePacket)->TransmitComplete)(PA->VirtualAdapter, savePacket, NDIS_STATUS_SUCCESS);
							continue; 
						}
				}
			}
			mlh_ = mlh_->next ;
	}
}



#if (AIMD_RATE_LIMIT_ENABLED)
// updateCurAck
//
// Gets the number of curNumAck by looking at difference between ack bitmaps & seqnum
// seqNum: the current seqNum of our state, which we compare against AC->prevAckSeqNum
// bitmap: the current bitmap of our state, which we compare against AC->bitmap
// Returns the nubmer of ACKs. 
//
// AC->curNumAck = updateCurAck(AC, AC->curAckSeqNum, AC->curAckBitMap);
uint
updateCurAck(AckCache *AC, LQSRAckId seqNum, AckBitMap bitmap)
{
	uint numAcks = 0 ; 
	int ackBound; 
	int i ;

	//doesn't work w/ wrap around
	ASSERT(seqNum >= AC->prevAckSeqNum);

	if(debug) KdPrint(("rlNumber Previous ACK %u, bit: %x %x %x %x %x %x %x %x\n", AC->prevAckSeqNum,
		AC->prevAckBitMap.bm[0], AC->prevAckBitMap.bm[1], AC->prevAckBitMap.bm[2], AC->prevAckBitMap.bm[3], 
		AC->prevAckBitMap.bm[4], AC->prevAckBitMap.bm[5], AC->prevAckBitMap.bm[6], AC->prevAckBitMap.bm[7])); 	

	if(debug) KdPrint(("rlNumber Current ACK %u, bit: %x %x %x %x %x %x %x %x\n", seqNum,
				bitmap.bm[0], bitmap.bm[1], bitmap.bm[2], bitmap.bm[3], 
				bitmap.bm[4], bitmap.bm[5], bitmap.bm[6], bitmap.bm[7])); 	

	//if the cumulative seq num has advanced, then we have
	// to check the old array for this many entries. If it is 
	// set for any of these entries then we do not updated the 
	// numAcks. If it is, then we do.
	ackBound = seqNum - AC->prevAckSeqNum; 
	
	//
	// Check which entries <= current seqnum have been acked
	//
	if (ackBound > BITMAPLENGTH) {
		//all prev entries acked b/c we jumped over all old bitmap values
		numAcks = ackBound-1; 
		//anything in new bitmap is an acked packet
		for (i = 0 ;  i < BITMAPLENGTH ; i++ ) {
			if (AckCacheCheckBit(&(bitmap), i)) {
				numAcks++; 
			}
		}	
		//ahh... but the old bitmap could have some packets set, in which case we need
		// to decrement those.
		for (i = 0 ;  i < BITMAPLENGTH ; i++ ) {
			if (AckCacheCheckBit(&(AC->prevAckBitMap), i)) {
				numAcks--; 
			}
		}
		return numAcks; 
	} else {
		//check all entries on old bitmap that new cum. seqnum updates
		for (i = 0 ; i < ackBound ; i++) {
			if (!(AckCacheCheckBit(&(AC->prevAckBitMap), i))) { 
				//old bit not set, so this pkt has been acked
				numAcks++; 
			}
		}
	}
	
	//
	// We know that ackBound <= BITMAPLENGTH if we are here.
	//

	//
	// Check which entries > current seqnum have been acked
	//--but still w/in old bitmap length
	for (i = ackBound ; i < BITMAPLENGTH ; i++) {
		//if old bitmap was not set and the new one is
		if (!(AckCacheCheckBit(&(AC->prevAckBitMap), i)) && AckCacheCheckBit(&(bitmap), i-ackBound)) {
			numAcks++; 
		}
	}

	//
	// Check which entries > current seqnum have been acked
	//--but not w/in the old bitmap length
	ASSERT(ackBound <= BITMAPLENGTH); 
	for (i = BITMAPLENGTH-ackBound ;  i < BITMAPLENGTH ; i++ ) {
		//any bit set on the new bitmap is a new ack
		if (AckCacheCheckBit(&(bitmap), i)) {
			numAcks++; 
		}
	}
	//if(debug) KdPrint(("numAcks: %u\n", numAcks)); 
	return numAcks; 
}
#endif


#if AIMD_RATE_LIMIT_ENABLED

//
// Given the current rate-limiting state, should we drop a packet?
// Returns true if so, false otherwise.
//
//
boolint shouldAIMDRateLimitDrop(MiniportAdapter *VA, 
								AckCache *AC)
{
	uint Random; 
	ASSERT(AC); 
	ASSERT(VirtualAddressEqual(VA->Address, AC->source));

	//never drop from first interval
	if (AC->firstInterval) {
		return FALSE;
	}

	
	Random = GetRandomNumber(AIMD_PROB_SCALE);
	
	if (AC->aimd_prob < Random) {
		//aimd_prob is the probability of not dropping the packet,
		// in other words, the probability we should send the packet.
		if(debug) KdPrint(("Dropping packet due to rate-limit\n")); 
		return TRUE;
	}
					
	return FALSE;

}


//
// Updates state when we are past the deadline
//--assumes that check was already made taht we are past
//    the deadline
void 
updateAIMDdeadline(AckCache *AC)
{
	Time Now = KeQueryInterruptTime();
	ASSERT(Now >= AC->intervalDeadline); 
	if (AC->intervalDeadline != 0) {//if == 0, then this first time ever here
		AC->firstInterval = FALSE;
		AC->curNumAck = updateCurAck(AC, AC->AckNum, AC->bitmap);
		if (Now > AC->intervalDeadline + AIMD_RATE_LIMIT_INTERVAL) {
			//if we missed an interval, then we should reset 
			AC->curNumSent = 0 ;
			AC->curNumAck = 0 ;
			AC->prev_aimd_prob = AC->virtualLoss; //set to loss
		}
		AC->aimd_prob = updateAIMDProb(AC);
		AC->prev_aimd_prob = AC->aimd_prob ;
		if(debug) KdPrint(("Number of sent packets: %u\n", AC->curNumSent));
		if(debug) KdPrint(("Number of acked packets: %u\n", AC->curNumAck));
		if(debug) KdPrint(("New probability of send: %u\n", AC->aimd_prob)); 
		//update the 'previous' variables
		AC->prevAckSeqNum = AC->AckNum; 
		AC->prevAckBitMap = AC->bitmap; //can I do this on structs?
		AC->curNumAck = 0 ;
		AC->curNumSent = 0 ;
	} else {
		AC->prev_aimd_prob = AC->virtualLoss; //set to loss
		ASSERT(AC->firstInterval);
	}
	AC->intervalDeadline = Now + AIMD_RATE_LIMIT_INTERVAL; 
	if(debug) KdPrint(("Setting new interval: %u\n", AC->intervalDeadline));
}


//updateAIMDProb
//
// Updates the probabilty that we send a packet in the given interval.
//--returns the new prob.
uint 
updateAIMDProb(AckCache *AC) 
{
	uint fdr = 0;
	uint THRESH = 0;
	uint curNumAck = AC->curNumAck;  
	uint curNumSent = AC->curNumSent;
	uint curProb = AC->aimd_prob; 
	uint prevProb = AC->prev_aimd_prob ;
	
	
	if (curNumSent == 0)
		curNumSent = 1; //want to avoid a divide by zero error
				

	//get FDR on the link
	fdr = LOSS_SCALE - AC->virtualLoss;  //a virtual loss of 0.05 is represented by 50, so the FDR would be 950
	ASSERT(fdr);

	//want the thresh to be product of fdr
	THRESH = (LOSS_SCALE*AIMD_RATE_LIMIT_THRESH)/fdr;

	//
	//curProb is always between [0, AIMD_PROB_SCALE]. The actually probability is:
	// curProb/AIMD_PROB_SCALE
	//
	
	//new changes
	if 	(AIMD_PROB_SCALE - (AIMD_PROB_SCALE*curNumAck)/(curNumSent) > prevProb && curNumSent >= AIMD_PKT_SENT_THRESH) { 
	
		return (AIMD_PROB_SCALE*curProb)/(AIMD_DECAY); 
	} else { 
		uint ret = curProb + AIMD_INCREASE;
		if (ret > AIMD_PROB_SCALE)
			ret = AIMD_PROB_SCALE;
		return ret; 
	}
}

#endif







#if E2E_RATE_LIMIT_ENABLED	
//
// Updates state when we are past the deadline
// To do:
//  1. Move number of packets ACKed to curBonus
//		--not so easy, have to have the carry-over
//	2. Reset curNumSent and curNumAck
//	3. Update interval deadline
//
void 
updateE2Edeadline(AckCache *AC)
{
	
	Time Now = KeQueryInterruptTime();
	uint extraBonus = 0;
	uint allow = 0;
	uint acked = 0 ;
	ASSERT(AC->curNumAck >= AC->prevNumAck);
	acked = AC->curNumAck - AC->prevNumAck;

	ASSERT(Now >= AC->intervalDeadline);

	if (AC->intervalDeadline == 0) {
		AC->firstInterval = TRUE;
	} else if (AC->intervalDeadline <= Now - (1*SECOND)) { //5*E2E_RATE_LIMIT_INTERVAL){
		//reset the interval if we haven't done anything in a long time
		AC->firstInterval = TRUE;
	} else {
		AC->firstInterval = FALSE; 
	}

	//update the current number of ACKs based upon differences in the ACK
	// state.
	//AC->curNumAck = updateCurAck(AC, AC->curAckSeqNum, AC->curAckBitMap);
	//The above should already be updated when we receive the e2e ACK.


	if (AC->firstInterval) {
		//if(debug) 
			KdPrint(("rlNumber of sent packets*: %u and retx: %d\n", AC->curNumSent, AC->curNumReTx));
	} else {
		//if(debug) 
			KdPrint(("rlNumber of sent packets: %u and retx: %d\n", AC->curNumSent, AC->curNumReTx));
	}
	//if(debug)
		KdPrint(("rlNumber of acked packets: %u\n", acked));
	//if(debug) 
		KdPrint(("rlNumber of dropped packets: %u\n", AC->curNumDrop));
	//if(debug) 
		KdPrint(("rlNumber of PT dropped packets: %u\n", AC->curPTNumDrop));
	//if(debug) 
		KdPrint(("rlNumber of CQ dropped packets: %u\n", AC->curCQNumDrop));
	//if(debug) 
		KdPrint(("rlNumber of e2e ACKs recv'd: %u\n", AC->curNumE2ER));


	KdPrint(("rlNumber Cumulative sent %u and cumulative pkts e2e acked %u\n", AC->totalNumSent, AC->curNumAck));

	//we only allow an extra bonus if we send less packets than the maximum allowable and
	// if we get more pkt's ACKed than sent.
	//AC->curNumBonus + MIN_E2E_SENT is the maximum number of pkts allowed to be sent in the interval,
	// but don't want to carry over the MIN_E2E_SENT packets...
	if (AC->curNumSent < AC->saveNumAckPrev) {
		extraBonus = AC->saveNumAckPrev - AC->curNumSent; 
	}

	allow = acked + extraBonus + E2E_ADDITIVE_CONSTANT ;
	if (allow < MIN_E2E_SENT) {
		allow = MIN_E2E_SENT ;
	}

	AC->curAllowable = allow;
	if(debug) KdPrint(("rlNumber of allowable packets: %u\n", AC->curAllowable));

#if RATE_LIMIT_FILE
	if (AC->firstInterval) {
		//don't drop any pkts in the first interval
		ProtocolCreateRateFile(MAX_DATA_PKTS_PER_INTERVAL);		
	} else {
		ProtocolCreateRateFile(AC->curAllowable);
	}
#endif


	//update the 'previous' variables
	AC->prevNumAck = AC->curNumAck ;
	AC->saveNumAckPrev = acked;
	AC->curNumSent = 0 ;
	AC->curNumDrop = 0 ;
	AC->curNumE2ER = 0 ;
	AC->curNumReTx = 0 ;
	AC->curCQNumDrop = 0;
	AC->curPTNumDrop = 0 ;
	
	AC->intervalDeadline = Now + E2E_RATE_LIMIT_INTERVAL; 
	//if(debug) KdPrint(("Setting new interval: %u\n", AC->intervalDeadline));
}




//* ProtocolRateLimitSending
//
// Called when we send a fresh (ie- not retransmitted) pkt.
// Needs to do:
// 1. Check to see if we are in hte next interval (ie- past interval deadline)
// 2. Check to see if we can send this packet
// 3. Update curNumSent
//
// Returns:
//		TRUE if we should drop the current packet
//		FALSE if it is okay to send the current packet
//
//  Assumes lock is already held.
boolint
ProtocolRateLimitSending(
					AckCache *AC)
{
	Time Now = KeQueryInterruptTime();
	ASSERT(AC); 
	

	//update deadline if necessary
	if (Now >= AC->intervalDeadline) {
		updateE2Edeadline(AC);
	}

	//we don't want to drop any packets over the first interval
	//--probably want to change this to do if we haven't sent any data
	// in a while (maybe over 2 or 3 intervals...)
	if (AC->firstInterval) {
		goto OKRateLimitSend;
	}


//if using rate-limit by file (ttcp), then don't ever drop
#if !RATE_LIMIT_FILE
	//should we drop the packet?
	//-this may change to if the difference is bigger than some preset value...
	if (AC->curNumSent >= AC->curAllowable) {
		if(debug) KdPrint(("Dropping packet due to rate-limit\n"));
		AC->curNumDrop++;
		return TRUE;
	}
	else
	{
		if(debug)
			KdPrint(("Sending Packet\n"));
	}
#endif

OKRateLimitSend:

	//okay to send, let's update the number of packets sent this interval
	AC->curNumSent++;

	//debug
	AC->totalNumSent++; 
	//end debug

	return FALSE;
}
#endif




//* ProtocolUpdateSequenceNumber
//
//
void
ProtocolUpdateSequenceNumber(
				MiniportAdapter *VA,
				SRPacket *srp,
				PNDIS_PACKET Packet)
{
	AckCache *ACdest ;
	LQSRHeader *OrigLQSR1;
	NDIS_BUFFER *OrigBuffer1;
	void *OrigHeader1;
	uint OrigBufferLength1;
	uint OrigPacketLength1;
	EtherHeader *OrigEther1;
	LQSROption *Walk;

	ASSERT(srp->ackreq);
	ASSERT(srp->ackreq->opt.identification == INITIAL_SEQ_NO);
	
	ACdest =  MaintBufFindAckFlow(VA, &VA->Address, 
			&srp->Dest, srp->sr->opt.hopList[1].inif, 
			srp->sr->opt.hopList[1].outif, NULL); 
	// this does not update the actual seq number in the binary packet....
	srp->ackreq->opt.identification = ACdest->nextSeqNumToSend++; 

	//update packet
	NdisGetFirstBufferFromPacket(Packet, &OrigBuffer1, &OrigHeader1,
								&OrigBufferLength1, &OrigPacketLength1);

	ASSERT(OrigBufferLength1 <= OrigPacketLength1);
	ASSERT(OrigBufferLength1 >= sizeof(EtherHeader) + sizeof(LQSRHeader));

	OrigEther1 = (EtherHeader *) OrigHeader1;
	OrigLQSR1 = (LQSRHeader *) (OrigEther1 + 1);
	Walk = (LQSROption *) (OrigLQSR1 + 1);

	//update binary packet structure
	InsertFakeOptions(&Walk, srp->sr, LQSR_OPTION_TYPE_SOURCERT);
	//InsertFakeOptions(&Walk, SRP->recovery, LQSR_OPTION_TYPE_RECOVERY);
	InsertFakeOptions(&Walk, srp->req, LQSR_OPTION_TYPE_REQUEST);
	InsertFakeOptions(&Walk, srp->rep, LQSR_OPTION_TYPE_REPLY);
	InsertFakeOptions(&Walk, srp->err, LQSR_OPTION_TYPE_ERROR);
	InsertOptions(&Walk, srp->ackreq, LQSR_OPTION_TYPE_ACKREQ);
}



//* ProtocolTransmitComplete
//
//  Called by NDIS when a send completes.
//
void
ProtocolTransmitComplete(
    NDIS_HANDLE Handle,   // Binding handle.
    PNDIS_PACKET Packet,  // Packet that was sent.
    NDIS_STATUS Status)   // Final status of send.
{
    ProtocolAdapter *PA = (ProtocolAdapter *)Handle;
	KIRQL OldIrql;

	//SRPacket *srp = PC(Packet)->srp; 
	//boolint rlDrop = (srp && srp->rlDropped) ;

    ASSERT(PC(Packet)->PA == PA);
	
	//if(debug) KdPrint(("\t\t\tProtocolDone: %x\n",PC(Packet)->queueID));		

	if (!ProtocolIsProbe(Packet)) {
		KeAcquireSpinLock(&PA->	VirtualAdapter->RecoveryLock, &OldIrql);
		ProtocolDequeue(PA,Packet); 	
		KeReleaseSpinLock(&PA->VirtualAdapter->RecoveryLock, OldIrql);
	} 

    (*PC(Packet)->TransmitComplete)(PA->VirtualAdapter, Packet, Status);

    InterlockedDecrement((PLONG)&PA->CountSentOutstanding);

#if 0
    ReleasePA(PA);
#endif
	//checkqueue has its own locks
	ProtocolCheckQueue(PA);
}

//* ProtocolTransmit
//
//  Send a packet. The packet already has an Ethernet header.
//
void
ProtocolTransmit(
    ProtocolAdapter *PA,
    NDIS_PACKET *Packet)
{
    uint Length;
    EtherHeader UNALIGNED *Ether;
    NDIS_STATUS Status;
	KIRQL OldIrql;
	
	//Time UpdateTime; 
	//KIRQL OldIrql; 

    ASSERT(PA->Use);

    InterlockedIncrement((PLONG)&PA->VirtualAdapter->CountXmit);
	
	// 
    // We never want the physical adapter to loopback.
    //
    Packet->Private.Flags = NDIS_FLAGS_DONT_LOOPBACK;
	

    //
    // Our sender initializes the Ethernet destination.
    //
    Ether = (EtherHeader UNALIGNED *) NdisFirstBuffer(Packet)->MappedSystemVa;
    RtlCopyMemory(Ether->Source, PA->Address, IEEE_802_ADDR_LENGTH);
    Ether->Type = ETYPE_MSFT;

    //
    // Prepare for ProtocolTransmitComplete.
    // Our sender initializes PC(Packet)->TransmitComplete.
    // The packet holds a reference for the physical adapter.
    //
#if 0
    AddRefPA(PA);
#endif
    PC(Packet)->PA = PA;
	 

    NdisQueryPacketLength(Packet, &Length);
	
	
	if (Length > PA->MaxTotalSize) {
		SRPacket *srp = PC(Packet)->srp;  
        //
        // We must not send packets larger than the adapter can accept.
        // This can happen when we try to forward a packet
        // that was too large - presumably due to a bug or attack.
        //
		InterlockedIncrementHighWater((PLONG)&PA->CountSentOutstanding,
                                  (PLONG)&PA->MaxSentOutstanding);
        InterlockedIncrement((PLONG)&PA->CountXmitLarge);
		//if we fail, have to ensure that correct completion methods are called.
		if(debug) KdPrint(("ERROR: ProtocolTransmit: Packet too large\n")); 
		if (srp->TransmitComplete == MiniportRetransmitSendComplete)
			srp->TransmitComplete = MiniportSendComplete ;
		else if (srp->TransmitComplete == MiniportForwardComplete) {
			srp->TransmitComplete = MiniportReceiveComplete; 
		}
        Status = NDIS_STATUS_INVALID_LENGTH;
    }
    else if (PA->ReceiveOnly) {
		SRPacket *srp = PC(Packet)->srp;  
        //
        // This is adapter is configured to only receive.
        // This simplifies testing of asymmetric links.
        //
		InterlockedIncrementHighWater((PLONG)&PA->CountSentOutstanding,
                                  (PLONG)&PA->MaxSentOutstanding);
		//if we fail, have to ensure that correct completion methods are called.
		if(debug) KdPrint(("ERROR: ProtocolTransmit: Adapter receive only\n")); 
		if (srp->TransmitComplete == MiniportRetransmitSendComplete)
			srp->TransmitComplete = MiniportSendComplete ;
		else if (srp->TransmitComplete == MiniportForwardComplete) {
			srp->TransmitComplete = MiniportReceiveComplete; 
		}
        Status = NDIS_STATUS_FAILURE;
    }
    else {
		SRPacket *srp = PC(Packet)->srp;  
        //
        // Send the packet. NDIS will call ProtocolTransmitComplete
        // if the send completes asynchronously.
        //
        InterlockedIncrement((PLONG)&PA->PacketsSent);
		PC(Packet)->queueID = PA->VirtualAdapter->queueCounter++; 

#if E2E_RATE_LIMIT_ENABLED	
		//We'll just have E2E Acks bypass the preIFQ
		if ((PA->CountSentOutstanding >= MAX_MAC_OUTSTANDING && !ProtocolIsProbe(Packet)) && 
			!(srp && srp->ack && srp->ack->opt.isE2E) ) {
#else
		//We need to queue packets if the IFQ has too many packets in it.
		//However, we want to send probe packets directly.
		if (PA->CountSentOutstanding >= MAX_MAC_OUTSTANDING && !ProtocolIsProbe(Packet)) {
#endif
			KeAcquireSpinLock(&PA->VirtualAdapter->RecoveryLock, &OldIrql);
			Status = ProtocolEnqueue(PA,Packet); 	
			KeReleaseSpinLock(&PA->VirtualAdapter->RecoveryLock, OldIrql);

			//if queueing doesn't complete successfully, then we have to make sure
			//that PA->CountSentOutstanding gets incremented here (because it is 
			// decremented in ProtocolTransmitComplete)
			if (Status != NDIS_STATUS_PENDING) {
				InterlockedIncrementHighWater((PLONG)&PA->CountSentOutstanding,
			                          (PLONG)&PA->MaxSentOutstanding);
				//if protocolenqueue was not successful, then we have to change 
				// the complete function that gets called. This is especially true
				// when RecoveryCacheCheckForwarding this to forward a packet and
				// the preIFQ is full. Also true if we have an original packet that
				// we are sending.
				if (srp->TransmitComplete == MiniportRetransmitSendComplete)
					srp->TransmitComplete = MiniportSendComplete ;
				else if (srp->TransmitComplete == MiniportForwardComplete) {
					srp->TransmitComplete = MiniportReceiveComplete; 
				}		
			}
		} else {	 
			MiniportAdapter *VA = PA->VirtualAdapter; 

			InterlockedIncrementHighWater((PLONG)&PA->CountSentOutstanding,
			                          (PLONG)&PA->MaxSentOutstanding);


			// IF we are sending an ACK, then we want to cancel timers, update ackoutrequest, etc
#if E2E_RATE_LIMIT_ENABLED	
			if (srp && srp->ack && !srp->ack->opt.isE2E) {
#else
			if (srp && srp->ack) {
#endif
				InternalAcknowledgement *curAck; 
				KeAcquireSpinLock(&VA->RecoveryLock, &OldIrql);	
				//go through all acks in packet
				for (curAck = srp->ack ; curAck != NULL ; curAck = curAck->next) {
					AckCache *AC = MaintBufFindAckFlow(VA, &curAck->opt.flowSource, 
									&curAck->opt.flowDest, curAck->opt.inif, 
									curAck->opt.outif, NULL); 
					
					ASSERT(AC); 
#if PIGGY_ACKS_CANCEL
					//If we are going to send an ACK for this packet, then we can 
					// cancel the ack thining timeout
					AC->NumAckSkipped = 0 ;
					KeCancelTimer(&AC->AckTimer);	
					if(debug) KdPrint(("Piggyback ACK cancelling ACK timer.\n")); 
#endif
					//we just sent an up-to-date ACK, so there are no outstanding ACK requests
					AC->numOutAckRequest = 0 ; 
					
#if (E2E_RATE_LIMIT_ENABLED && !NEW_E2E_RATE_CONTROL)
					//want to see if we need to send an E2E rate-control ACK.
					//--this needs to be added to ProtocolRetransmitPacket and ProtocolCheckQueue
					if (VirtualAddressEqual(VA->Address, curAck->opt.flowDest)) {
						AC->NumAckSent++;
						if (AC->NumAckSent % NUM_E2E_SKIPS == 0) {
							//need to watch out for recursion!!!! --what to do w/ lock?
							KeReleaseSpinLock(&VA->RecoveryLock, OldIrql);	
							ProtocolSendE2EAck(AC, 1); 
							KeAcquireSpinLock(&VA->RecoveryLock, &OldIrql);	
						}
					}
#endif

				}
				KeReleaseSpinLock(&VA->RecoveryLock, OldIrql);	
			}

			//Need to add retransmission timer if necessary.
			//the last condition shouldn't matter b/c we only use this method for 
			// sending the first time. Retransmissions are handled through ProtocolRecoverTransmit.
			if (srp && srp->ackreq != NULL && srp->PayloadOffset != srp->PacketLength && 
				(srp->TransmitComplete == MiniportRetransmitSendComplete ||
				srp->TransmitComplete == MiniportForwardComplete) &&
				srp->RecoveryHolder == NULL) {

				//let's try some debugging...
				PC(Packet)->TransmitComplete = MaintBufRetransmitSendComplete;
				KeAcquireSpinLock(&VA->RecoveryLock, &OldIrql);	
			

#if SLIDING_RATE_LIMIT_ENABLED	
				//update sendTimes linked list	
				if (srp->sr && srp->ackreq && VirtualAddressEqual(srp->sr->opt.hopList[0].addr, VA->Address)) {
					//assuming only 1 interface for now!!
					AckCache *AC = MaintBufFindAckFlow(VA, &VA->Address, 
						&srp->Dest, srp->sr->opt.hopList[1].inif, 
						srp->sr->opt.hopList[1].outif, &VA->Address); 
					ASSERT(AC); 
					updateTimeStruct(&(AC->SendTimes), &(AC->lastSendTimes), &(AC->numSendTimes));
				}
#endif

#if AIMD_RATE_LIMIT_ENABLED

				//only for flows i am the original sender on
				if (srp->sr && srp->ackreq && VirtualAddressEqual(srp->sr->opt.hopList[0].addr, VA->Address)) {
					Time Now = KeQueryInterruptTime();
					AckCache *ACrate ;
					ACrate =  MaintBufFindAckFlow(VA, &VA->Address, 
							&srp->Dest, srp->sr->opt.hopList[1].inif, 
							srp->sr->opt.hopList[1].outif, &VA->Address); 
					ASSERT(ACrate); 
					if (Now >= ACrate->intervalDeadline) {
						updateAIMDdeadline(ACrate);
					} else {
						ACrate->curNumSent++;	
					}
				}

#endif
				//
				//let's try to update the sequence number for the packet
				//--only for new packets (not forwarded ones)
				//
				if (srp->ackreq && 
					(srp->TransmitComplete != MiniportReceiveComplete && srp->TransmitComplete != MiniportForwardComplete)) {
					
						
						//let's also update some rate-limiting stuff
#if E2E_RATE_LIMIT_ENABLED	
						//only for flows i am the original sender on
						if (srp->sr && VirtualAddressEqual(srp->sr->opt.hopList[0].addr, VA->Address)) {
							
							AckCache *ACrate = NULL;
							boolint drop = FALSE ;
							ACrate =  MaintBufFindAckFlow(VA, &VA->Address, 
									&srp->Dest, srp->sr->opt.hopList[1].inif, 
									srp->sr->opt.hopList[1].outif, NULL); 
							ASSERT(ACrate); 
							drop = ProtocolRateLimitSending(ACrate);
							if (drop) {								
								//PA->CountSentOutstanding already incremented.

								// We also have to change 
								// the complete function that gets called. This is especially true
								// when RecoveryCacheCheckForwarding this to forward a packet and
								// the preIFQ is full. Also true if we have an original packet that
								// we are sending.
								PC(Packet)->TransmitComplete = MaintBufStaticSendComplete;
								if (srp->TransmitComplete == MiniportRetransmitSendComplete)
									srp->TransmitComplete = MiniportSendComplete ;
								else if (srp->TransmitComplete == MiniportForwardComplete)
									srp->TransmitComplete = MiniportReceiveComplete; 

								//Status = NDIS_STATUS_RESOURCES;
								Status = NDIS_STATUS_SUCCESS;
								ACrate->curPTNumDrop++;
								KeReleaseSpinLock(&VA->RecoveryLock, OldIrql);	
								goto ProtocolTransmitDone; 
							}
						}

#endif
						//we're okay to send, so update seq number and go!
						ProtocolUpdateSequenceNumber(VA, srp, Packet);


				}

				//Have to be sure to add this logic on first time we try to add
				// the retransmission.
				//if not successful, then we must revert to old sendcomplete function
				if (!LocalRecoveryAddRetransmission(VA, srp, (MaintBufNode*) srp->MBNnexthop)) {
					PC(Packet)->TransmitComplete = MaintBufStaticSendComplete;
					if (srp->TransmitComplete == MiniportRetransmitSendComplete)
						srp->TransmitComplete = MiniportSendComplete ;
					else if (srp->TransmitComplete == MiniportForwardComplete) {
						srp->TransmitComplete = MiniportReceiveComplete; 
					} else {
						ASSERT(!"We should never be here\n"); 
					}
				}//end if addretransmission


				KeReleaseSpinLock(&VA->RecoveryLock, OldIrql);							
			}

			//If we don't add a retransmission, the we may have to modify the retransmission
			//timer for a packet (ie- it isn't the first time this packet has been sent). 
			//However: retransmissions always use ProtocolRecoverTransmit, so we are safe 
			// here. We still need to check when we dequeue from preIFQ.
			//KdPrint(("NdisSend sending packet\n"));
			NdisSend(&Status, PA->Handle, Packet);
		}
    }
	
#if E2E_RATE_LIMIT_ENABLED	
ProtocolTransmitDone:	
#endif
    if (Status != NDIS_STATUS_PENDING) {
        //
        // The send finished synchronously,
        // so we call the completion handler.
        //
        ProtocolTransmitComplete((NDIS_HANDLE)PA, Packet, Status);
    }

}




//* ejr
//	ProtocolRecoverTransmit
//
//  This method assumes it will never have to send an ACK. It is also used
//  for retransmissions/forwarding sometimes. We want to eventually get ride
//  of this method (if possible).
//
//  Send a packet. The packet may not have(?) an Ethernet header.
//  This is a special case of ProtocolTransmit for local recovery.
//  Called with lock held.
void
ProtocolRecoverTransmit(
    ProtocolAdapter *PA,
    NDIS_PACKET *Packet)
{
    uint Length;
    EtherHeader UNALIGNED *Ether;
    NDIS_STATUS Status;
	KIRQL OldIrql;
	
	UNREFERENCED_PARAMETER(Ether);

    ASSERT(PA->Use);

	//ejr- It would be nice to add some variables for local recovery
	//InterlockedIncrement((PLONG)&PA->VirtualAdapter->CountXmit);

	
	// 
    //
    // We never want the physical adapter to loopback.
    //
    Packet->Private.Flags = NDIS_FLAGS_DONT_LOOPBACK;
	

 	//
    // Prepare for ProtocolRecoverTransmitComplete.
    //
	if (PC(Packet)->TransmitComplete != MaintBufRetransmitSendComplete)
		PC(Packet)->TransmitComplete = ProtocolRecoverTransmitComplete;

#if 0
    AddRefPA(PA);
#endif
    PC(Packet)->PA = PA;

	NdisQueryPacketLength(Packet, &Length);
    if (Length > PA->MaxTotalSize) { 
        //
        // We must not send packets larger than the adapter can accept.
        // This can happen when we try to forward a packet
        // that was too large - presumably due to a bug or attack.
        //
		InterlockedIncrementHighWater((PLONG)&PA->CountSentOutstanding,
                                  (PLONG)&PA->MaxSentOutstanding);
        InterlockedIncrement((PLONG)&PA->CountXmitLarge);
		if(debug) KdPrint(("ERROR: Packet too large in ProtocolRecoverTransmit\n")); 
		Status = NDIS_STATUS_INVALID_LENGTH;
    }
    else if (PA->ReceiveOnly) {
        //
        // This is adapter is configured to only receive.
        // This simplifies testing of asymmetric links.
        //
		InterlockedIncrementHighWater((PLONG)&PA->CountSentOutstanding,
                                  (PLONG)&PA->MaxSentOutstanding);
		if(debug) KdPrint(("ERROR: Adapter is receive only in ProtocolRecoverTransmit\n")); 
        Status = NDIS_STATUS_FAILURE;
    }
    else {
        //
        // Send the packet. NDIS will call ProtocolTransmitComplete
        // if the send completes asynchronously.
        //
        InterlockedIncrement((PLONG)&PA->PacketsSent);
		PC(Packet)->queueID = PA->VirtualAdapter->queueCounter++; 
		
		if (PA->CountSentOutstanding >= MAX_MAC_OUTSTANDING) {
			
			KeAcquireSpinLock(&PA->VirtualAdapter->RecoveryLock, &OldIrql);
			Status = ProtocolEnqueue(PA,Packet); 			
			KeReleaseSpinLock(&PA->VirtualAdapter->RecoveryLock, OldIrql);

			
			//if protocolenqueue was not successful, then we have to change 
			// the complete function that gets called. This is especially true
			// when RecoveryCacheCheckForwarding this to forward a packet and
			// the preIFQ is full.

			//if queueing doesn't complete successfully, then we have to make sure
			//that PA->CountSentOutstanding gets incremented here (because it is 
			// decremented in ProtocolTransmitComplete)
			if (Status != NDIS_STATUS_PENDING) {
				InterlockedIncrementHighWater((PLONG)&PA->CountSentOutstanding,
			                          (PLONG)&PA->MaxSentOutstanding);
			}
		} else {	
			SRPacket *srp = PC(Packet)->srp;  
			InterlockedIncrementHighWater((PLONG)&PA->CountSentOutstanding,
			                          (PLONG)&PA->MaxSentOutstanding);

			// Need to add logic here to update retransmisison
			// timer of packet. Only retransmitted data packets should be 
			// here, so I don't think we need to do any checks.
			ASSERT(srp->sendCompleteShouldDelete == 0); //want to ensure that this is a clean pkt
			
			LocalRecoveryUpdateRetransmissionTimer(PA->VirtualAdapter, srp); 

			// If this packet is one that we are forwarding, then we need to
			// check if it is a piggy-backed ACK. If so, then cancel the ACK
			// thinning timeouts.
			if (srp->ack) {
				InternalAcknowledgement *curAck; 
				KeAcquireSpinLock(&PA->VirtualAdapter->RecoveryLock, &OldIrql);
				//go through all acks in packet
				for (curAck = srp->ack ; curAck != NULL ; curAck = curAck->next) {
					AckCache *AC = MaintBufFindAckFlow(PA->VirtualAdapter, &curAck->opt.flowSource, 
									&curAck->opt.flowDest, curAck->opt.inif, 
									curAck->opt.outif, NULL); 
					
					ASSERT(AC); 
#if PIGGY_ACKS_CANCEL
					//If we are going to send an ACK for this packet, then we can 
					// cancel the ack thining timeout
					AC->NumAckSkipped = 0 ;
					KeCancelTimer(&AC->AckTimer);	
					if(debug) KdPrint(("Piggyback ACK cancelling ACK timer.\n")); 
#endif
					//we just sent an up-to-date ACK, so there are no outstanding ACK requests
					AC->numOutAckRequest = 0 ;

//#if E2E_RATE_LIMIT_ENABLED	
#if (E2E_RATE_LIMIT_ENABLED && !NEW_E2E_RATE_CONTROL)
					//want to see if we need to send an E2E rate-control ACK.
					//--this needs to be added to ProtocolRetransmitPacket and ProtocolCheckQueue
					if (VirtualAddressEqual(PA->VirtualAdapter->Address, curAck->opt.flowDest)) {
						AC->NumAckSent++;
						if (AC->NumAckSent % NUM_E2E_SKIPS == 0) {
							//need to watch out for recursion!!!! --what to do w/ lock?
							KeReleaseSpinLock(&PA->VirtualAdapter->RecoveryLock, OldIrql);	
							ProtocolSendE2EAck(AC, 1); 
							KeAcquireSpinLock(&PA->VirtualAdapter->RecoveryLock, &OldIrql);	
						}
					}
#endif

				}
			
#if E2E_RATE_LIMIT_ENABLED
				if (1) {
					AckCache *ACrate = NULL;
					ACrate =  MaintBufFindAckFlow(PA->VirtualAdapter, &PA->VirtualAdapter->Address, 
							&srp->Dest, srp->sr->opt.hopList[1].inif, 
							srp->sr->opt.hopList[1].outif, NULL); 
					ASSERT(ACrate); 
					ACrate->curNumReTx++;	
				}
#endif
				KeReleaseSpinLock(&PA->VirtualAdapter->RecoveryLock, OldIrql);
			}

			
#if SLIDING_RATE_LIMIT_ENABLED		
			//update sendTimes linked list
			if (srp->sr && srp->ackreq && VirtualAddressEqual(srp->sr->opt.hopList[0].addr, PA->VirtualAdapter->Address)) {
				AckCache *AC;
				KeAcquireSpinLock(&PA->VirtualAdapter->RecoveryLock, &OldIrql);
				//assuming only 1 interface for now!!
				AC = MaintBufFindAckFlow(PA->VirtualAdapter, &(PA->VirtualAdapter->Address), 
					&srp->Dest, srp->sr->opt.hopList[1].inif, 
					srp->sr->opt.hopList[1].outif, &(PA->VirtualAdapter->Address)); 
				ASSERT(AC); 
				updateTimeStruct(&(AC->SendTimes), &(AC->lastSendTimes), &(AC->numSendTimes));
				KeReleaseSpinLock(&PA->VirtualAdapter->RecoveryLock, OldIrql);
			}
#endif



#if AIMD_RATE_LIMIT_ENABLED
			//only for flows i am the original sender on
			if (srp->sr && srp->ackreq && VirtualAddressEqual(srp->sr->opt.hopList[0].addr, PA->VirtualAdapter->Address)) {
				Time Now = KeQueryInterruptTime();
				AckCache *ACrate ;
				
				KeAcquireSpinLock(&PA->VirtualAdapter->RecoveryLock, &OldIrql);
				ACrate =  MaintBufFindAckFlow(PA->VirtualAdapter, &(PA->VirtualAdapter->Address), 
						&srp->Dest, srp->sr->opt.hopList[1].inif, 
						srp->sr->opt.hopList[1].outif, &(PA->VirtualAdapter->Address)); 
				ASSERT(ACrate); 
				if (Now >= ACrate->intervalDeadline) {
					updateAIMDdeadline(ACrate);
				} else {
					ACrate->curNumSent++;	
				}
				KeReleaseSpinLock(&PA->VirtualAdapter->RecoveryLock, OldIrql);
			}

#endif



			NdisSend(&Status, PA->Handle, Packet);
		}
    }

    if (Status != NDIS_STATUS_PENDING) {
        //
        // The send finished synchronously,
        // so we call the completion handler.
        //
        ProtocolTransmitComplete((NDIS_HANDLE)PA, Packet, Status);//calls pc(packet)->transmitcomplete

    }
}


//* ejr
//  ProtocolRecoverTransmitComplete
//
//  Called by NDIS when a recovery send completes.
//
void
ProtocolRecoverTransmitComplete(
    NDIS_HANDLE Handle,   // Binding handle.
    PNDIS_PACKET Packet,  // Packet that was sent.
    NDIS_STATUS Status)   // Final status of send.
{
	//ejr- do we need locking here? how to get it we dont'
	// have a handle to VA?
	SRPacket *srp ;
	SRPacketHolder *curHolder ;

	UNREFERENCED_PARAMETER(Status);
	UNREFERENCED_PARAMETER(Handle);
	
	ASSERT(PC(Packet) != NULL); 
	srp = PC(Packet)->srp; 
    NdisFreePacketClone(Packet);
	if (srp != NULL) { //this may happen? if so, just exit
		//if(debug) KdPrint(("TransmitRecoveryComplete on seqNum: %d\n", srp->ackreq->opt.identification)); 
		curHolder = srp->RecoveryHolder ;
		ASSERT(curHolder != NULL); //this may happen? if so, just exit
		curHolder->Packet = NULL ;
	}
}


#if E2E_RATE_LIMIT_ENABLED
//ejr
//ProtocolSendE2EAck
//
// Responsible for sending an E2E ack. numAcked is the number of pkts
// ACKd for this flow. Note that numAcked is now a cumulative count.
//
void ProtocolSendE2EAck (AckCache *AC, uint numAcked) {
	MiniportAdapter *VA = AC->VA; 
    InternalAcknowledgement *Ack;
    InternalSourceRoute *SR;
	SRPacket *srpack ;
    boolint OKToSend;
    uint Size;
    NDIS_STATUS Status;
	NDIS_PACKET *Packet = NULL;

	EtherHeader UNALIGNED *Ether; 
	void *Data;
    NDIS_BUFFER *Buffer;
    uint BufferLength;
	
	KIRQL OldIrql;


	KeAcquireSpinLock(&VA->RecoveryLock, &OldIrql);

	//
    // Allocate the Acknowledgement option.
    //
    Ack = ExAllocatePool(NonPagedPool, sizeof *Ack);
	if (Ack == NULL) {
		KeReleaseSpinLock(&VA->RecoveryLock, OldIrql);     
		return;
	}
	
    RtlZeroMemory(Ack, sizeof *Ack);
    RtlCopyMemory(Ack->opt.from, VA->Address, SR_ADDR_LEN);
	//set to e2e ack
	ASSERT(numAcked); 
	Ack->opt.isE2E = numAcked; //TRUE ;
	RtlCopyMemory(Ack->opt.flowSource, AC->source, SR_ADDR_LEN);
	RtlCopyMemory(Ack->opt.flowDest, AC->dest, SR_ADDR_LEN);
	ASSERT(VirtualAddressEqual(Ack->opt.flowDest, VA->Address));

	//assuming only 1 interface for now!
	Ack->opt.inif = AC->InIf;
	Ack->opt.outif = AC->OutIf;

	//fill in opt.identification and opt.bitmap
	AckCacheFillAck(AC ,&(Ack->opt.identification), &(Ack->opt.bitmap), 0); 
	
    //
    // Initialize the Acknowledgement option.
    //
    Ack->opt.optionType = LQSR_OPTION_TYPE_ACK;
    Ack->opt.optDataLen = ACKNOWLEDGEMENT_LEN;
	
	//Allocate a new packet to hold our ACK
	srpack = ExAllocatePool(NonPagedPool, sizeof *srpack);
	if (srpack == NULL) {
		ExFreePool(Ack);
		KeReleaseSpinLock(&VA->RecoveryLock, OldIrql);
        return;
	}
    RtlZeroMemory(srpack, sizeof *srpack);
	
    //
    // Initialize source & destination of this packet.
    //
	RtlCopyMemory(srpack->Dest, AC->source, SR_ADDR_LEN);
    RtlCopyMemory(srpack->Source, VA->Address, SR_ADDR_LEN);
	//
    // Initialize a source route to this destination.
    //
    SR = ExAllocatePool(NonPagedPool, sizeof *SR + sizeof(SRAddr)*MAX_SR_LEN);
	if (SR == NULL) {
		ExFreePool(Ack);
		ExFreePool(srpack);
		KeReleaseSpinLock(&VA->RecoveryLock, OldIrql);
        return;
	}
    RtlZeroMemory(SR, sizeof *SR);
	
	// For E2E ACKs, can we just reuse the LinkCacheFillSR() function and
	// then manually set the fwd bitmap to next node.
	Status = LinkCacheFillSR(VA, srpack->Dest, SR);
    if (Status != NDIS_STATUS_SUCCESS) {
		ExFreePool(SR);
        SRPacketFree(srpack);	
		ExFreePool(Ack);
		KeReleaseSpinLock(&VA->RecoveryLock, OldIrql);
        return;
	}
	SR->opt.bitmap = (ushort) -1;
	//just assign to next hop (in this case, index = 1)
	SR->opt.fwd_bitmap = 0;
	setSRPbitmap(&(SR->opt.fwd_bitmap), 1);	

	//
	// Send the packet using the source route.
    //
    srpack->sr = SR;

    Size = sizeof(LQSRHeader) + sizeof(LQSROption) + SOURCE_ROUTE_LEN(MAX_SR_LEN);
    ASSERT(Size == PbackPacketSize(srpack));
    ASSERT(Size <= PROTOCOL_MIN_FRAME_SIZE);

	//
	//ejr- Add option to the packet
	//
	srpack->ack = Ack;

	//
    // This packet only contains acknowledgements.
    // Requesting an ack for an ack is a bad idea,
    // so suppress acknowledgements.
    //
    SR->opt.staticRoute = TRUE;
		
    //
    // Send this packet using the maintenance buffer.
    //
	OKToSend = LinkCacheUseSR(VA, srpack);
    ASSERT(OKToSend);

    srpack->TransmitComplete = ACKSendComplete;

#if E2E_DEBUG
	ASSERT(srpack->ack->next == NULL);
#endif


    Status = SRPacketToPkt(VA, srpack, &Packet);
    if (Status != NDIS_STATUS_SUCCESS) {
		ACKSendComplete(VA, srpack, Status);
		KeReleaseSpinLock(&VA->RecoveryLock, OldIrql);
        return;
    }

    PC(Packet)->srp = srpack;
    PC(Packet)->TransmitComplete = MaintBufStaticSendComplete;

    if (PC(Packet)->PA == NULL) {
		    //
            // This means the packet is trying to use a physical adapter that
            // no longer exists.
            //
            MaintBufStaticSendComplete(VA, Packet, NDIS_STATUS_FAILURE);
			KeReleaseSpinLock(&VA->RecoveryLock, OldIrql);
    }
    else {

		uint Index = 0; 
		uint j = 0;
		uint dcnt = 0 ;
		VA->numE2EAcksSent++;
		PRINT_ADDRESS("sending E2E ack: ", srpack->Source, srpack->Dest, VA->numE2EAcksSent); 
		//if(debug)
		KdPrint(("DEBUG E2E Sending_ACK %u (%u) {%u} BitMap: %x %x %x %x %x %x %x %x\n", srpack->ack->opt.identification,
			numAcked, AC->totalReTxRcv,
				srpack->ack->opt.bitmap.bm[0], srpack->ack->opt.bitmap.bm[1], srpack->ack->opt.bitmap.bm[2], 
				srpack->ack->opt.bitmap.bm[3], srpack->ack->opt.bitmap.bm[4], srpack->ack->opt.bitmap.bm[5], 
				srpack->ack->opt.bitmap.bm[6], srpack->ack->opt.bitmap.bm[7])); 		
		
	
		
		//debug
		Index = SOURCE_ROUTE_HOPS(srpack->sr->opt.optDataLen) - srpack->sr->opt.segmentsLeft ;
		if(debug) KdPrint(("%d hops on the fwd list, %d is next hop\n", SOURCE_ROUTE_HOPS(srpack->sr->opt.optDataLen), Index));
		for (j = 0 ; j < SOURCE_ROUTE_HOPS(srpack->sr->opt.optDataLen); j++) {
			VirtualAddress *Address;
			Address = &(srpack->sr->opt.hopList[j].addr) ;
			if (checkSRPbitmap(srpack->sr->opt.bitmap, j)) {
				dcnt++; 
				if (checkSRPbitmap(srpack->sr->opt.fwd_bitmap, j))
					PRINT_VA("\t\t\tDFDFDFDFDF Forwarding node: ", *Address); 
				else
					PRINT_VA("\t\t\tDNDNDNDNDN Forwarding node: ", *Address); 
			} else {
				if (checkSRPbitmap(srpack->sr->opt.fwd_bitmap, j))
					PRINT_VA("\t\t\tNFNFNFNFNF Forwarding node: ", *Address); 
				else
					PRINT_VA("\t\t\tNNNNNNNNNN Forwarding node: ", *Address); 
			}
		} 
		if(debug) KdPrint(("Done with %d default hops on the fwd list.\n", dcnt));
		
		Buffer = NdisFirstBuffer(Packet);
		NdisQueryBuffer(Buffer, &Data, &BufferLength);
		Ether = (EtherHeader UNALIGNED *) Data;
		PRINT_VA("\t\t\tEthernet destination: ", (Ether->Dest)); 


		
		KeReleaseSpinLock(&VA->RecoveryLock, OldIrql);

		//want to change to NDIS send?		
        ProtocolTransmit(PC(Packet)->PA, Packet);
	}
	return ; 

}
#endif



//ejr
// ProtocolEnqueue
//
// Queues a packet in the MAC limiting queue if there are too many
// outstanding packets.  Will drop a packet if there are too many packets
// in the MAC limiting queue.
//
// returns NDIS_STATUS_RESOURCES if it cannot queue, NDIS_STATUS_PENDING
// otherwise
//
NDIS_STATUS ProtocolEnqueue(ProtocolAdapter *PA, NDIS_PACKET *Packet) {
	MiniportAdapter *VA = PA->VirtualAdapter ;
	MACLimitHolder *mlh_; 

	//check to see if MAC limiting queue is full
	if (VA->mlq.count >= MAC_LIMIT_MAX_SIZE) {
		if(debug) KdPrint(("Can't buffer packet in ProtocolEnqueue, count %d.\n", VA->mlq.count)); 	
		//dump the packet
        return NDIS_STATUS_RESOURCES;		
	}


	//Need to see if this packet is an ACK. If so, then we need to do
	//ACK peeking to allow this packet to be discarded. Be sure to update
	// all relevant variables (ie- CountOutstanding). Also, ACKs should
	// go in the front of the preIFQ if we do send them.

	//create packet holder
	mlh_ = ExAllocatePool(NonPagedPool, sizeof *mlh_);
	if (mlh_ == NULL) {
		if(debug) KdPrint(("Can't alloc memory in ProtocolEnqueue.\n")); 	
        return NDIS_STATUS_RESOURCES;
	}

    RtlZeroMemory(mlh_, sizeof *mlh_);
	mlh_->Packet = Packet; 
	mlh_->next = NULL ;
	mlh_->prev = NULL ;
	
	if(debug) KdPrint(("ProtocolEnqueue: Buffering packet %d.\n", PC(Packet)->queueID));

	//add packet holder to our queue
	if (VA->mlq.count == 0) {
		VA->mlq.firstHolder = mlh_;
		VA->mlq.lastHolder = mlh_ ;
	} else {
		//ACKs go at front of preIFQ
		if (protocolIsACK(VA, Packet)) {
			//if we can peek, then we discard packet
			if (protocolCanPeek(VA, Packet)) {
				if(debug) KdPrint(("Ack peek successful at preIFQ\n")); 
				//we want to get ride of ACK. If we return success, then it 
				// appears that if this packet was successfully transmitted, so all
				// state will be (correctly) deleted.
				ExFreePool(mlh_); 
				return NDIS_STATUS_SUCCESS ; 
			}
			if(debug) KdPrint(("Inserting ACK at front of preIFQ\n")); 
			//Add ack to front of preIFQ
			mlh_->next = VA->mlq.firstHolder ;
			VA->mlq.firstHolder->prev = mlh_; 
			VA->mlq.firstHolder = mlh_ ;
		} else {
			//everything else goes at end of preIFQ
			VA->mlq.lastHolder->next = mlh_; 
			mlh_->prev = VA->mlq.lastHolder; 
			VA->mlq.lastHolder = mlh_ ;
		}
	}
	//increase count
	VA->mlq.count++; 

	//return the default (expected) value if we would have just sent
	return NDIS_STATUS_PENDING;
}



// ejr
// ProtocolDequeue
//
// This method removes a packet from our MAC limiting linked list 
// This must be called *before* the transmit complete method.
//
// The ProtocolCheckSent method checks to see if we can send a 
// packet after dequeueing has been completed.
//
//
void ProtocolDequeue(ProtocolAdapter *PA, NDIS_PACKET *Packet) {
	MiniportAdapter *VA = PA->VirtualAdapter ;
	MACLimitHolder *mlh_; 
	int i = 0 ;
	i += 0 ;

	//find packet in our array
	for (mlh_ = VA->mlq.firstHolder ; mlh_ != NULL ; mlh_ = mlh_->next) {
		if (VA->mlq.count != 0)
			if(debug) KdPrint(("%d :: first_incom: %d second_loop: %d\n", i++, PC(Packet)->queueID, PC(mlh_->Packet)->queueID));
		if (PC(Packet)->queueID == PC(mlh_->Packet)->queueID)
			break ;
	}
	
	if (mlh_ == NULL) {
		return; 
	} else {
		if(debug) KdPrint(("ProtocolDequeue: Packet %d found in search.\n", PC(Packet)->queueID));
	}

	//delete entry in mac limit queue (update LL)
	if (mlh_->next == NULL)
		VA->mlq.lastHolder = mlh_->prev ;
	else 
		mlh_->next->prev = mlh_->prev ;
	if (mlh_->prev == NULL)
		VA->mlq.firstHolder = mlh_->next; 
	else
		mlh_->prev->next = mlh_->next ; 
	
	//free resources of holder
	ExFreePool(mlh_);	

	//decrement count
	VA->mlq.count--; 

}



//ejr
//ProtocolCheckQueue
//
// Checks to see if we can send a packet from the queue. Must be the last method
// called from the ProtocolTransmitComplete
//
void ProtocolCheckQueue(ProtocolAdapter *PA) {
	MiniportAdapter *VA = PA->VirtualAdapter ;
	MACLimitHolder *mlh_; 
	NDIS_STATUS Status;
	KIRQL OldIrql;
#if E2E_DEBUG
	InternalAcknowledgement *curAck; 
#endif

	KeAcquireSpinLock(&PA->VirtualAdapter->RecoveryLock, &OldIrql);
	//send new packet (if any)
	for (mlh_ = VA->mlq.firstHolder ; mlh_ != NULL ; mlh_ = mlh_->next) {
		if (!mlh_->outstanding) {
			break; 
		}
	}	
	
	//if we have a new packet to send, then let's do it
	if (mlh_ != NULL && PA->CountSentOutstanding < MAX_MAC_OUTSTANDING) {
		SRPacket *srp = PC(mlh_->Packet)->srp;  

		//we want to se this before the protocolIsAck is called because we
		// don't want that method to find the very packet we are sending.
		// Need to revert this back to zero if we can peek.
		mlh_->outstanding = (boolint) 1 ; 

		//look for any ACK peeking oppurtunities- an if statement is sufficient, since
		// we'll recursively call checkPacket again.
		if (protocolIsACK(VA, mlh_->Packet)) {
			//if we can peek, then we discard packet
			if (protocolCanPeek(VA, mlh_->Packet)) {
				mlh_->outstanding = (boolint) 0 ;	
				if(debug) KdPrint(("Ack peek successful at checkPacket preIFQ\n"));
				//need to increment CountSentOustanding if we call ProtocolTransmitComplete
				//directly b/c of error or if packet isn't sent
				InterlockedIncrementHighWater((PLONG)&PA->CountSentOutstanding,
				                      (PLONG)&PA->MaxSentOutstanding);	
				KeReleaseSpinLock(&PA->VirtualAdapter->RecoveryLock, OldIrql); 
				ProtocolTransmitComplete((NDIS_HANDLE)PA, mlh_->Packet, NDIS_STATUS_SUCCESS);
				return; 
			}
		}

#if SLIDING_RATE_LIMIT_ENABLED	
		//update sendTimes linked list
		if (srp && srp->sr && srp->ackreq && VirtualAddressEqual(srp->sr->opt.hopList[0].addr, VA->Address)) {
			//assuming only 1 interface for now!!
			AckCache *AC = MaintBufFindAckFlow(VA, &VA->Address, 
				&srp->Dest, srp->sr->opt.hopList[1].inif, 
				srp->sr->opt.hopList[1].outif, &VA->Address); 
			ASSERT(AC); 
			updateTimeStruct(&(AC->SendTimes), &(AC->lastSendTimes), &(AC->numSendTimes));
		}
#endif
		
#if AIMD_RATE_LIMIT_ENABLED
		//only for flows i am the original sender on
		if (srp && srp->sr && srp->ackreq && VirtualAddressEqual(srp->sr->opt.hopList[0].addr, VA->Address)) {
			Time Now = KeQueryInterruptTime();
			AckCache *ACrate ;
			ACrate =  MaintBufFindAckFlow(VA, &VA->Address, 
					&srp->Dest, srp->sr->opt.hopList[1].inif, 
					srp->sr->opt.hopList[1].outif, &VA->Address); 
			ASSERT(ACrate); 
			if (Now >= ACrate->intervalDeadline) {
				updateAIMDdeadline(ACrate);
				//shouldn't we advance curNumSent regardless?
			} else {
				ACrate->curNumSent++;	
			}
		}

#endif

		//send the packet
		if(debug) KdPrint(("ProtocolCheckPacket sending %d\n", PC(mlh_->Packet)->queueID));
		InterlockedIncrementHighWater((PLONG)&PA->CountSentOutstanding,
				                      (PLONG)&PA->MaxSentOutstanding);
		
		//
		//let's try to update the sequence number for the packet
		//--only for new packets (not forwarded ones)
		//
		if (srp->ackreq && 
			(srp->TransmitComplete != MiniportReceiveComplete && srp->TransmitComplete != MiniportForwardComplete)) {
				//don't want to call this function on retransmitted packets.
				if (srp->ackreq->opt.identification == INITIAL_SEQ_NO) {

					ASSERT(srp->RecoveryHolder == NULL);
#if E2E_RATE_LIMIT_ENABLED	
	
					//only for flows i am the original sender on
					if (srp->sr && VirtualAddressEqual(srp->sr->opt.hopList[0].addr, VA->Address)) {
						
						AckCache *ACrate = NULL;
						boolint drop = FALSE ;
						ACrate =  MaintBufFindAckFlow(VA, &VA->Address, 
								&srp->Dest, srp->sr->opt.hopList[1].inif, 
								srp->sr->opt.hopList[1].outif, NULL); 
						ASSERT(ACrate); 
						drop = ProtocolRateLimitSending(ACrate);
						//if (0) {								
						if (drop) {								
							//PA->CountSentOutstanding was already incremented.
							
							// We also have to change 
							// the complete function that gets called. This is especially true
							// when RecoveryCacheCheckForwarding this to forward a packet and
							// the preIFQ is full. Also true if we have an original packet that
							// we are sending.
							PC(mlh_->Packet)->TransmitComplete = MaintBufStaticSendComplete;
							if (srp->TransmitComplete == MiniportRetransmitSendComplete)
								srp->TransmitComplete = MiniportSendComplete ;
							else if (srp->TransmitComplete == MiniportForwardComplete)
								srp->TransmitComplete = MiniportReceiveComplete; 
	
							ACrate->curCQNumDrop++;

							//Status = NDIS_STATUS_RESOURCES;	
							Status = NDIS_STATUS_SUCCESS;
							KeReleaseSpinLock(&VA->RecoveryLock, OldIrql);	
							goto ProtocolCheckTransmitDone; 
						}
					}

#endif
					ProtocolUpdateSequenceNumber(VA, srp, mlh_->Packet);
				}
			}	
		
		//
		// Need to update the state of packet (ack cache, bitmap) if 
		// necessary!  Also have to update retransmit timers.
		//
#if E2E_RATE_LIMIT_ENABLED
		//if (!(srp && srp->ack && srp->ack->opt.isE2E))
		ASSERT(!(srp && srp->ack && srp->ack->opt.isE2E));
#endif

		//we just dequeued from the preIFQ and now we have to update the state of the 
		// packet, such as the ACKs, etc.
#if NEW_PREIFQ_UPDATE_DEQUEUE
		//try a new approach-- only update the fwding list if haven't updated it in
		// awhile.
        //mlh_->Packet = protocolUpdateAckAndFwd(VA, mlh_->Packet, OldIrql);		
		if (KeQueryInterruptTime() - srp->updateTime  >= UPDATE_FWD_LIST_TIME_DIFF) {
			mlh_->Packet = protocolUpdateAckAndFwd(VA, mlh_->Packet, OldIrql);
		} else {
			protocolUpdateAckState(VA, mlh_->Packet, OldIrql); 
		}
				
#else
		protocolUpdateAckState(VA, mlh_->Packet, OldIrql); 
#endif


		KeReleaseSpinLock(&PA->VirtualAdapter->RecoveryLock, OldIrql);

		//Need to add retransmission timer if necessary.
		if (srp && srp->ackreq != NULL && srp->PayloadOffset != srp->PacketLength && 
			(srp->TransmitComplete == MiniportRetransmitSendComplete ||
			srp->TransmitComplete == MiniportForwardComplete)) {
		
			KeAcquireSpinLock(&VA->RecoveryLock, &OldIrql);	

			//case of sending packet for first time
			if( srp->RecoveryHolder == NULL) {
				ASSERT(srp->sendCompleteShouldDelete == 0); //want to ensure that this is a clean pkt

				//let's try some debugging...
				PC(mlh_->Packet)->TransmitComplete = MaintBufRetransmitSendComplete;

				//Have to be sure to add this logic on first time we try to add
				// the retransmission.

				//if not successful, then we must revert to old sendcomplete function
				if (!LocalRecoveryAddRetransmission(VA, srp, (MaintBufNode*) srp->MBNnexthop)) {
					PC(mlh_->Packet)->TransmitComplete = MaintBufStaticSendComplete;
					if (srp->TransmitComplete == MiniportRetransmitSendComplete)
						srp->TransmitComplete = MiniportSendComplete ;
					else if (srp->TransmitComplete == MiniportForwardComplete) {
						srp->TransmitComplete = MiniportReceiveComplete; 
					} else {
						ASSERT(!"We should never be here\n"); 
					}
				}//end if addretransmission
						
		
			} else {
				//this is a retransmitted packet.
				// Need to add logic here to update retransmisison
				// timer of packet. Only retransmitted data packets should be 
				// here, so I don't think we need to do any checks.		
#if E2E_RATE_LIMIT_ENABLED
				AckCache *ACrate = NULL;
				ACrate =  MaintBufFindAckFlow(VA, &VA->Address, 
						&srp->Dest, srp->sr->opt.hopList[1].inif, 
						srp->sr->opt.hopList[1].outif, NULL); 
				ASSERT(ACrate); 
				ACrate->curNumReTx++;	
#endif
				LocalRecoveryUpdateRetransmissionTimer(PA->VirtualAdapter, PC(mlh_->Packet)->srp); 
			}			
			KeReleaseSpinLock(&VA->RecoveryLock, OldIrql);	
		}
	
#if E2E_DEBUG
		if (srp) {
			for (curAck = srp->ack ; curAck != NULL ; curAck = curAck->next) {
				ASSERT(curAck->opt.isE2E == 0);
			}
		}
#endif
		NdisSend(&Status, PA->Handle, mlh_->Packet);
		
#if E2E_RATE_LIMIT_ENABLED	
ProtocolCheckTransmitDone:
#endif
		if (Status != NDIS_STATUS_PENDING) {
			//
			// The send finished synchronously,
			// so we call the completion handler.
			//
			ProtocolTransmitComplete((NDIS_HANDLE)PA, mlh_->Packet, Status);
	    }
	} else { //end send packet
		KeReleaseSpinLock(&PA->VirtualAdapter->RecoveryLock, OldIrql);
	}
}

//* protocolCanPeek
//
// NDIS_PACKET *Packet: the ACK that we want to send
//
// Checks to see if any of the first ACK_PEEK_NUMBER
// packets in the preIFQ contain an ACK for the flow
// designated by the ACK: 'Packet'
//
// Must be called with lock held.
//
// ACKs in the preIFQ may change, so we can only peek
// if the packet has an ACK and the pkt is for the same
// flow as the packet we want to peek.
//
//
// Returns true if we don't need to send the ACK
//  ie- there is an ACK w/in ACK_PEEK_NUMBER for the same flow
// Returns false otherwise
//
boolint protocolCanPeek(
	MiniportAdapter *VA,
	NDIS_PACKET *Packet)
{
	MACLimitHolder *mlh_ = VA->mlq.firstHolder ; 
	SRPacket *SRP = PC(Packet)->srp;
    SRPacket *SRP_inq;
	uint i = 0 ;
	ASSERT(SRP);
	ASSERT(SRP->ack); 

	//for (i = 0 ; i < ACK_PEEK_NUMBER && i < VA->mlq.count ; i++)
	while(mlh_ != NULL)
	{
		ASSERT(mlh_);
	
		//packets that have already been sent should be not be considered
		if (mlh_->outstanding) {
			mlh_ = mlh_->next ;
			continue; 
		}

		//want to ensure that we don't peek too far
		if (i >= ACK_PEEK_NUMBER || i >= VA->mlq.count) {
			return 0;
		}

		SRP_inq = PC(mlh_->Packet)->srp ;

		if(SRP_inq != NULL && SRP_inq->ack != NULL && SRP_inq->sr != NULL) 
		{
			InternalAcknowledgement *curAck; 
#if NEW_ACK_PEEK
			uint Index = SOURCE_ROUTE_HOPS(SRP_inq->sr->opt.optDataLen) - 1; 
			//check only the flow of the current pkt, since that will always have an ACK of the current flow
			//
			// what if the pkt we're looking at is an ACK? Then the sr is opposite.
			//
			boolint dataSameFlow = !protocolIsACK(VA, mlh_->Packet) &&
				VirtualAddressEqual(SRP->ack->opt.flowDest, SRP_inq->sr->opt.hopList[Index].addr) && 
				VirtualAddressEqual(SRP->ack->opt.flowSource, SRP_inq->sr->opt.hopList[0].addr);

			boolint ackSameFlow = protocolIsACK(VA, mlh_->Packet) && 
				VirtualAddressEqual(SRP->ack->opt.flowDest, SRP_inq->ack->opt.flowDest) && 
				VirtualAddressEqual(SRP->ack->opt.flowSource, SRP_inq->ack->opt.flowSource); 
			/*
			PRINT_ADDRESS("peek ack examinng: ", SRP->ack->opt.flowSource, SRP->ack->opt.flowDest, 0);
			PRINT_ADDRESS("peek inq examinng: ", SRP_inq->sr->opt.hopList[0].addr, SRP_inq->sr->opt.hopList[Index].addr, 0);
			*/

			if (dataSameFlow || ackSameFlow) {
				/*	
				if (dataSameFlow)
					if(debug) KdPrint(("peek found data pkt flow\n"));
				if (ackSameFlow)
					if(debug) KdPrint(("peek found ack pkt flow\n")); 
				*/
				//let's just make sure it actually has the ACK, b/c maybe it didn't piggyback anything
				//Need to check over all flows
				for (curAck = SRP_inq->ack; curAck != NULL; curAck = curAck->next) {
					if (VirtualAddressEqual(SRP->ack->opt.flowSource, curAck->opt.flowSource) && 
						VirtualAddressEqual(SRP->ack->opt.flowDest, curAck->opt.flowDest)) {
							return 1; 
						}
				}
			}
#else
			
			//Need to check over all flows
			for (curAck = SRP_inq->ack; curAck != NULL; curAck = curAck->next) {
				if (VirtualAddressEqual(SRP->ack->opt.flowSource, curAck->opt.flowSource) && 
					VirtualAddressEqual(SRP->ack->opt.flowDest, curAck->opt.flowDest)) {
						return 1; 
					}
			}
#endif

		}

		i++;
		mlh_ = mlh_->next ;
	}
	return 0 ; 
}


//* protocolIsACK
//
// Checks the current packet to see if it is an ACK
// packet. It should be an ACK packet only, if it is piggy-backed,
// then false should be returned.
//
boolint protocolIsACK(
	MiniportAdapter *VA,
	NDIS_PACKET *Packet)
{
	SRPacket *SRP = PC(Packet)->srp;
    UNREFERENCED_PARAMETER(VA);

	if (SRP == NULL)
		return 0;

	if (SRP->TransmitComplete == ACKSendComplete) {
		ASSERT(SRP->ack); //getting a seg fault right here
		return 1; 
	}

	return 0; 
}


//protocolUpdateAckAndFwd
//
// Let's try a new method that updates the ACK state and the 
// forwarding list for packets we just dequeued from the preIFQ.
//
// General idea: delete the packet we've already created,
//				update all SRP state, create new packet and
//				then update PC(Packet) state.
//
// Returns: new packet for the mlh_ holder
//
//
NDIS_PACKET*
protocolUpdateAckAndFwd(
	MiniportAdapter *VA,
	NDIS_PACKET *Packet,
	KIRQL OldIrql)
{
	NDIS_PACKET *newPacket ; 
    NDIS_STATUS Status; 
	SRPacket *srp = NULL;
	EtherHeader UNALIGNED *Ether;	
	
	AckCache *AC ; 
	InternalAcknowledgement *curAck;
	
	uint etxThresh ;
	InternalSourceRoute *SR_;

    UNREFERENCED_PARAMETER(OldIrql);

	srp = PC(Packet)->srp; 
	
	//All the above was stole from SRPacketFromPkt().
	//We now need to update the ACK state.
	// stand-alone acks should always have a source route
	if (srp->ack == NULL) {
		if (srp->sr == NULL) {
			//we won't be able to do anything anyways...
			return Packet; 
		} else {
			goto TryUpdateFwdList; 
		}
	}


#if E2E_RATE_LIMIT_ENABLED	
	//e2e acks should never be here
	ASSERT(!(srp && srp->ack && srp->ack->opt.isE2E)); 
#endif

	//
	// Update the SRP state.
	//

	/*
	//Debug
	for (curAck = srp->ack ; curAck != NULL ; curAck = curAck->next) {
		if(debug) KdPrint(("1DEBUG1 1qSending_ACK %d BitMap: %x %x %x %x %x %x %x %x\n", curAck->opt.identification,
				curAck->opt.bitmap.bm[0], curAck->opt.bitmap.bm[1], curAck->opt.bitmap.bm[2], 
				curAck->opt.bitmap.bm[3], curAck->opt.bitmap.bm[4], curAck->opt.bitmap.bm[5], 
				curAck->opt.bitmap.bm[6], curAck->opt.bitmap.bm[7])); 	
	}
	//End Debug
	*/

	
	//
	// Need to re-add ACKs to our packet.
	//  We can do this by conditioning on srp->TransmitComplete
	//
	if (!VirtualAddressEqual(VA->Address, srp->sr->opt.hopList[0].addr) && srp->PayloadOffset != srp->PacketLength) {
		//packet we are forwarding
		SRFreeOptionList((InternalOption *)srp->ack);
		srp->ack = NULL;
		srp->ack = MaintBufFillInPiggyBackAck(VA, srp);
		if (srp->ack == NULL) {
			if(debug) KdPrint(("nullack\n"));
		}
	} 
	//
	// Packet we are sending 
	//
	//used to just condition on srp->transmitcomplete == miniportresendcomplete
	else if (VirtualAddressEqual(VA->Address, srp->sr->opt.hopList[0].addr) && srp->PayloadOffset != srp->PacketLength) {
		uint Size ; 
		//assuming 1 interface for now!
		AckCache *AC = MaintBufFindAckFlow(VA, &(VA->Address), &(srp->Dest),
			srp->sr->opt.hopList[1].inif, srp->sr->opt.hopList[1].inif, NULL); 

		ASSERT(AC);
		//packet we are originally sending
		SRFreeOptionList((InternalOption *)srp->ack);
		srp->ack = NULL;
		//fill in piggy back ACKs
		Size = PbackPacketSize(srp);
		if (Size < PROTOCOL_MIN_FRAME_SIZE) {
			Size = PROTOCOL_MIN_FRAME_SIZE - Size; //amount left to use
			srp->ack = MaintBufGetCrossFlowAcks(VA, AC, Size, 0);
		}
		if (srp->ack == NULL) {
			if(debug) KdPrint(("nullack\n"));
		}
	} 
	//
	// looks like we could have missed stand-alone acks above... they have cross-flow acks, so 
	// need to see which flows to include
	//
	else if (protocolIsACK(VA, Packet)) {
		uint Size ; 
		//assuming 1 interface for now!
		AckCache *AC = MaintBufFindAckFlow(VA, &(srp->ack->opt.flowSource), &(srp->ack->opt.flowDest),
			srp->sr->opt.hopList[1].inif, srp->sr->opt.hopList[1].inif, NULL); 

		ASSERT(AC);
		//free all the cross-flow acks
		SRFreeOptionList((InternalOption *)srp->ack->next);
		srp->ack->next = NULL;
		//fill in piggy back ACKs
		Size = PbackPacketSize(srp);
		if (Size < PROTOCOL_MIN_FRAME_SIZE) {
			Size = PROTOCOL_MIN_FRAME_SIZE - Size; //amount left to use
			srp->ack->next = MaintBufGetCrossFlowAcks(VA, AC, Size, 1);
		}
	}



	//We now need to update the ACK state.
	//go through all acks in packet
	for (curAck = srp->ack ; curAck != NULL ; curAck = curAck->next) {

		AC = MaintBufFindAckFlow(VA, &curAck->opt.flowSource, 
			&curAck->opt.flowDest, curAck->opt.inif, 
			curAck->opt.outif, NULL); 

		ASSERT(AC); 

		//If we are going to send an ACK for this packet, then we can 
		// cancel the ack thining timeout
#if PIGGY_ACKS_CANCEL
		AC->NumAckSkipped = 0 ;
		KeCancelTimer(&AC->AckTimer);
		if(debug) KdPrint(("Piggyback ACK cancelling ACK timer.\n")); 
#endif
		//fill in opt.identification and opt.bitmap
		AckCacheFillAck(AC ,&(curAck->opt.identification), &(curAck->opt.bitmap), curAck->opt.identification); 

		//we just sent an up-to-date ACK, so there are no outstanding ACK requests
		AC->numOutAckRequest = 0 ;

		//Debug
		if(debug) KdPrint(("DEBUG qSending_ACK %d BitMap: %x %x %x %x %x %x %x %x\n", curAck->opt.identification,
				curAck->opt.bitmap.bm[0], curAck->opt.bitmap.bm[1], curAck->opt.bitmap.bm[2], 
				curAck->opt.bitmap.bm[3], curAck->opt.bitmap.bm[4], curAck->opt.bitmap.bm[5], 
				curAck->opt.bitmap.bm[6], curAck->opt.bitmap.bm[7])); 	
		//End Debug

			
//#if E2E_RATE_LIMIT_ENABLED	
#if (E2E_RATE_LIMIT_ENABLED && !NEW_E2E_RATE_CONTROL)
		//want to see if we need to send an E2E rate-control ACK.
		//--this needs to be added to ProtocolRetransmitPacket and ProtocolCheckQueue
		if (VirtualAddressEqual(VA->Address, curAck->opt.flowDest)) {
			AC->NumAckSent++;
			if (AC->NumAckSent % NUM_E2E_SKIPS == 0) {
				//need to watch out for recursion!!!! 
				//at this point, the recoverylock was held when the method was called.
				KeReleaseSpinLock(&VA->RecoveryLock, OldIrql);	
				ProtocolSendE2EAck(AC, 1); 
				KeAcquireSpinLock(&VA->RecoveryLock, &OldIrql);	
			}
		}
#endif
		
#if E2E_DEBUG
		ASSERT(curAck->opt.isE2E == 0);
#endif

	}

	//
	//Update the forwarding list
	//
TryUpdateFwdList:
	if (srp->sr == NULL || srp->ackreq == NULL || srp->PayloadOffset == srp->PacketLength || 
		!(srp->TransmitComplete == MiniportRetransmitSendComplete || 
		srp->TransmitComplete == MiniportForwardComplete))
	{
		goto SendNewSRPacket;  		
	}

	//can remove later, debug
	if(debug) KdPrint(("forwifq difference in timers: %d\n",(KeQueryInterruptTime()-srp->updateTime)/10000));

	//We now need to update the forwarding list. Need to ensure this packet should actually update
	// the forwarding list (ie- don't do it on route replies, etc)
	Status = updateForwardInfo(VA, srp); 
	if (Status != NDIS_STATUS_SUCCESS) {
		goto SendNewSRPacket; 
	}
	//get new forwarding list
	SR_ = LinkCacheFindForwarders(VA, srp, &etxThresh);
	Status = LinkCacheInsertForwardingList(VA, SR_, srp, etxThresh);
	if (Status == NDIS_STATUS_SUCCESS) {
		srp->updateTime = KeQueryInterruptTime();
	}
	//don't need to worry if Status isn't successful.		

	
SendNewSRPacket:
	//convert updated SRP packet to new packet to send
	Status = SRPacketToPkt(VA, srp, &newPacket);
    if (Status != NDIS_STATUS_SUCCESS) {
		//just send the old packet, don't worry about updating
		if(debug) KdPrint(("Error in protocolUpdateAckAndFwd, SRPacketToPkt failed, old pkt sent.\n"));
		return Packet; 
    }
	
	PC(newPacket)->PA = PC(Packet)->PA;
	PC(newPacket)->queueID = PC(Packet)->queueID; 
	//PC(newPacket)->ReceiveComplete = PC(Packet)->ReceiveComplete ;
	PC(newPacket)->srp = PC(Packet)->srp; 
	PC(newPacket)->TransmitComplete = PC(Packet)->TransmitComplete; 
	//PC(newPacket)->OrigPacket = PC(Packet)->OrigPacket ;

//protocolTransmit does this....;
	
	// 
    // We never want the physical adapter to loopback.
    //
    newPacket->Private.Flags = NDIS_FLAGS_DONT_LOOPBACK;

    //
    // Our sender initializes the Ethernet destination.
    //
    Ether = (EtherHeader UNALIGNED *) NdisFirstBuffer(newPacket)->MappedSystemVa;
    RtlCopyMemory(Ether->Source, PC(newPacket)->PA->Address, IEEE_802_ADDR_LENGTH);
    Ether->Type = ETYPE_MSFT;
	//we don't have to worry about e2e acks-- they should never be here.
	IEEE_802_ADDR_SET_BROADCAST(&(Ether->Dest)) ;


	//We can only be here on a retransmission
	if (srp->RecoveryHolder) {
		srp->RecoveryHolder->Packet = newPacket; 
	}

	//delete the old packet
    NdisFreePacketClone(Packet);

	if(debug) KdPrint(("Updated the packet in updateAndFwd\n"));

	return newPacket;
}


//* protocolUpdateAckState
//
// This method will update the ACK state in a packet that has been sitting in the
// the preIFQ. We need to do this for any packet that is/has an ACK.
//
// It would be best to try to optimize this function later to only pick out
// the ACK state from the packet, instead of creating a whole SRP.
//
void
protocolUpdateAckState(
	MiniportAdapter *VA,
	NDIS_PACKET *Packet,
	KIRQL OldIrql)
{
	
    SRPacket *SRP;
    LQSRHeader *OrigLQSR;
	NDIS_BUFFER *OrigBuffer;
    void *OrigHeader;
	uint OrigBufferLength;
    uint OrigPacketLength;
    EtherHeader *OrigEther;
	AckCache *AC ; 
    //MaintBuf *MB = VA->MaintBuf;
    LQSROption *Walk;
	InternalAcknowledgement *curAck; 
	uint numAcks = 0 ; //number acks in pkt
	InternalAcknowledgement *holdAck;  //tmp holder we use to overwrite acks
	AckCache *holdAC ; 
	uint i ;
	
	
    UNREFERENCED_PARAMETER(OldIrql);

	NdisGetFirstBufferFromPacket(Packet, &OrigBuffer, &OrigHeader,
                                 &OrigBufferLength, &OrigPacketLength);

    ASSERT(OrigBufferLength <= OrigPacketLength);
    ASSERT(OrigBufferLength >= sizeof(EtherHeader) + sizeof(LQSRHeader));

    OrigEther = (EtherHeader *) OrigHeader;
    OrigLQSR = (LQSRHeader *) (OrigEther + 1);
    Walk = (LQSROption *) (OrigLQSR + 1);

	SRP = ExAllocatePool(NonPagedPool, sizeof *SRP);
    if (SRP == NULL)
        return ;
    RtlZeroMemory(SRP, sizeof *SRP);

	//
    // Parse the LQSR header options.
    //
	//
	// possible optimization: just use PC(Packet)->srp->ack
	//
    SRP = FillSRPacket(SRP, OrigLQSR);
    if (SRP == NULL) {
        //
        // Illegal packet format. FillSRPacket freed SRP.
        //
		ASSERT(0); 
		return; 
    }

	//All the above was stole from SRPacketFromPkt().
	//We now need to update the ACK state.
	if (SRP->ack == NULL || SRP->sr == NULL) {
		goto FreeSRandExit; 
	}
	
	//already have the recovery lock when calling this method....
    //KeAcquireSpinLock(&MB->Lock, &OldIrql);


	//find number of acks in packet
	for (curAck = SRP->ack ; curAck != NULL ; curAck = curAck->next) {
		numAcks++; 
	}
	ASSERT(numAcks); 

	//
	//add my flow to first entry
	//--need to do this for forwarded packets or stand-alone acks
	//first condition: forwarder   second condition: stand alone ack <-need to verify
	holdAck = SRP->ack ;
	if (!VirtualAddressEqual(VA->Address, SRP->sr->opt.hopList[0].addr) || SRP->PayloadOffset == SRP->PacketLength) {
		uint cnt =0 ; 
		InternalAcknowledgement *lastAck = SRP->ack;

		//yuck, the acks are stored in reverse order as we originally put them in. So
		//go to the last ack and add it first.
		for (cnt = 0; cnt < numAcks-1 ; cnt++) {
			lastAck = lastAck->next; //reuse holdAck for now
		}

		//assuming one interface for now!
		holdAC =  MaintBufFindAckFlow(VA, &lastAck->opt.flowSource, &lastAck->opt.flowDest, SRP->sr->opt.hopList[1].inif, 
			SRP->sr->opt.hopList[1].inif, NULL); 
		ASSERT(holdAC); 
		
		//update ACK
		holdAck->opt.bitmap = holdAC->bitmap; 
		holdAck->opt.identification = holdAC->AckNum; 
		RtlCopyMemory(&holdAck->opt.flowDest, &holdAC->dest, SR_ADDR_LEN);
		RtlCopyMemory(&holdAck->opt.flowSource, &holdAC->source, SR_ADDR_LEN);	
		RtlCopyMemory(&holdAck->opt.from, &VA->Address, SR_ADDR_LEN);	
		holdAC->numOutAckRequest = 0 ; //reset to zero and then we can use this in numAcks loop below

		/*
		if (!VirtualAddressEqual(VA->Address, SRP->sr->opt.hopList[0].addr)) {
			PRINT_ADDRESS("\ttesteric s: ", SRP->sr->opt.hopList[0].addr, 
			SRP->sr->opt.hopList[SOURCE_ROUTE_HOPS(SRP->sr->opt.optDataLen)-1].addr, holdAC->numOutAckRequest);
			
		}
		
		PRINT_ADDRESS("\tqcur Ack flow1: ", holdAC->source, holdAC->dest, 
			holdAC->numOutAckRequest);
		*/

		//update state
		numAcks--;
		holdAck = holdAck->next; 
	} 


	//find best numAcks packets
	for (i = 0 ; i < numAcks ; i++) {
		
		//go through all Ack Caches
		AckCache *curAckCache ;
		uint highestAckOut = 0 ;
		AckCache *highestAckCache = NULL; 

		
		//find best ACK
		for (curAckCache = VA->acl.head ; curAckCache != NULL ; curAckCache = curAckCache->next) {
			//if there are no outstanding ack requests, then skip (or we already added)
			if (curAckCache->numOutAckRequest == 0) {
				continue; 
			}
			if (curAckCache->numOutAckRequest >= highestAckOut) {
				highestAckOut = curAckCache->numOutAckRequest ;
				highestAckCache = curAckCache; 
			}
		}

		//want to set holdAC to the best ACK we choose
		if (highestAckCache == NULL) {
			if(debug) KdPrint(("No more ACKs to add in updateackstate\n"));
			break; 
		}
	
		holdAC =  highestAckCache; 
		ASSERT(holdAC); 

		/*
		PRINT_ADDRESS("\tqcur Ack flow: ", holdAC->source, holdAC->dest, 
			holdAC->numOutAckRequest);
		*/

		//update ACK
		holdAck->opt.bitmap = holdAC->bitmap; 
		holdAck->opt.identification = holdAC->AckNum; 
		RtlCopyMemory(&holdAck->opt.flowDest, &holdAC->dest, SR_ADDR_LEN);
		RtlCopyMemory(&holdAck->opt.flowSource, &holdAC->source, SR_ADDR_LEN);
		RtlCopyMemory(&holdAck->opt.from, &VA->Address, SR_ADDR_LEN);	
		holdAC->numOutAckRequest = 0 ; //reset to zero

		//update state
		holdAck = holdAck->next; 

	}

	//go through all acks in packet
	for (curAck = SRP->ack ; curAck != NULL ; curAck = curAck->next) {

		AC = MaintBufFindAckFlow(VA, &curAck->opt.flowSource, 
			&curAck->opt.flowDest, curAck->opt.inif, 
			curAck->opt.outif, NULL); 

		ASSERT(AC);

		//If we are going to send an ACK for this packet, then we can 
		// cancel the ack thining timeout
#if PIGGY_ACKS_CANCEL
		AC->NumAckSkipped = 0 ;
		KeCancelTimer(&AC->AckTimer);
		if(debug) KdPrint(("Piggyback ACK cancelling ACK timer.\n")); 
#endif

		//fill in opt.identification and opt.bitmap
		AckCacheFillAck(AC ,&(curAck->opt.identification), &(curAck->opt.bitmap), curAck->opt.identification); 

		//we just sent an up-to-date ACK, so there are no outstanding ACK requests
		AC->numOutAckRequest = 0 ;

		//Debug
		if(debug) KdPrint(("DEBUG qSending_ACK %d BitMap: %x %x %x %x %x %x %x %x\n", curAck->opt.identification,
				curAck->opt.bitmap.bm[0], curAck->opt.bitmap.bm[1], curAck->opt.bitmap.bm[2], 
				curAck->opt.bitmap.bm[3], curAck->opt.bitmap.bm[4], curAck->opt.bitmap.bm[5], 
				curAck->opt.bitmap.bm[6], curAck->opt.bitmap.bm[7])); 	
		//End Debug

			
//#if E2E_RATE_LIMIT_ENABLED	
#if (E2E_RATE_LIMIT_ENABLED && !NEW_E2E_RATE_CONTROL)
		//want to see if we need to send an E2E rate-control ACK.
		//--this needs to be added to ProtocolRetransmitPacket and ProtocolCheckQueue
		if (VirtualAddressEqual(VA->Address, curAck->opt.flowDest)) {
			AC->NumAckSent++;
			if (AC->NumAckSent % NUM_E2E_SKIPS == 0) {
				//need to watch out for recursion!!!! 
				//at this point, the recoverylock was held when the method was called.
				KeReleaseSpinLock(&VA->RecoveryLock, OldIrql);	
				ProtocolSendE2EAck(AC, 1); 
				KeAcquireSpinLock(&VA->RecoveryLock, &OldIrql);	
			}
		}
#endif

#if E2E_DEBUG
		ASSERT(curAck->opt.isE2E == 0);
#endif

	}
	//KeReleaseSpinLock(&MB->Lock, OldIrql);


	//
	// At this point, we've updated SRP->ack. We need to put this back into the Packet
	// and then we'll be done.
 	InsertFakeOptions(&Walk, SRP->sr, LQSR_OPTION_TYPE_SOURCERT);

    InsertFakeOptions(&Walk, SRP->req, LQSR_OPTION_TYPE_REQUEST);
    InsertFakeOptions(&Walk, SRP->rep, LQSR_OPTION_TYPE_REPLY);
    InsertFakeOptions(&Walk, SRP->err, LQSR_OPTION_TYPE_ERROR);
    InsertFakeOptions(&Walk, SRP->ackreq, LQSR_OPTION_TYPE_ACKREQ);
    InsertOptions(&Walk, SRP->ack, LQSR_OPTION_TYPE_ACK); //want to actually put ACK state in packet
	//we can skip the rest of 'InsertFakeOptions' now

	//should be all done now, just free SRPacket structure


 FreeSRandExit:
	//kill the SRPacket that we allocated.
	SRFreeOptionList((InternalOption *)SRP->req);
    SRFreeOptionList((InternalOption *)SRP->rep);
    SRFreeOptionList((InternalOption *)SRP->err);
    SRFreeOptionList((InternalOption *)SRP->ackreq);
    SRFreeOptionList((InternalOption *)SRP->ack);
    SRFreeOptionList((InternalOption *)SRP->sr);
    //SRFreeOptionList((InternalOption *)SRP->recovery);
    SRFreeOptionList((InternalOption *)SRP->inforeq);
    SRFreeOptionList((InternalOption *)SRP->inforep);
    SRFreeOptionList((InternalOption *)SRP->Probe);
    SRFreeOptionList((InternalOption *)SRP->ProbeReply);
    SRFreeOptionList((InternalOption *)SRP->LinkInfo);
    ExFreePool(SRP);
}



//* ProtocolForwardRequestComplete
//
//  Completes an individual packet transmission for ProtocolForwardRequest.
//
void
ProtocolForwardRequestComplete(
    MiniportAdapter *VA,
    NDIS_PACKET *Packet,
    NDIS_STATUS Status)
{
    SRPacket *srp = PC(Packet)->srp;

	
    //
    // Free the packet structure.
    //
    NdisFreePacketClone(Packet);
	
    //
    // Update the current cumulative status.
    // If the current status is NDIS_STATUS_SUCCESS,
    // we replace it with our new status.
    //
    InterlockedCompareExchange((PLONG)&srp->ForwardStatus,
                               Status,
                               NDIS_STATUS_SUCCESS);

    if (InterlockedDecrement((PLONG)&srp->ForwardCount) == 0) {
        //
        // Complete the ProtocolForwardRequest operation.
        //
        (*srp->TransmitComplete)(VA, srp, srp->ForwardStatus);
    }
}

//* ProtocolForwardRequest
//
//  Sends a Request packet via every physical adapter simultaneously.
//  The operation completes when the last underlying transmit completes.
//  The operation completes successfully only if every
//  underlying transmit is successful.
//
//  Our caller supplies srp->TransmitComplete, which always called
//  to consume the SRPacket.
//
//  The packet destination is presumably a broadcast/multicast
//  address, but that is not actually required.
//
void
ProtocolForwardRequest(
    MiniportAdapter *VA,
    SRPacket *srp)
{
    ProtocolAdapter *PA;
    NDIS_PACKET *PacketList = NULL;
    NDIS_PACKET *Packet;
    NDIS_STATUS Status;
    uint Hops;
    KIRQL OldIrql;
#if E2E_DEBUG
	InternalAcknowledgement *curAck; 
#endif
    //
    // We have already added ourselves to the hop list.
    //
    Hops = ROUTE_REQUEST_HOPS(srp->req->opt.optDataLen) - 1;
    ASSERT(RtlEqualMemory(srp->req->opt.hopList[Hops].addr,
                          VA->Address, SR_ADDR_LEN));

    //
    // Check for options that can be piggy-backed on this packet.
    // We can only piggy-back if we are originating the request.
    //
#if E2E_RATE_LIMIT_ENABLED
	if (Hops == 0 && !(srp && srp->ack && srp->ack->opt.isE2E))
#else
    if (Hops == 0)
#endif
        PbackSendPacket(VA, srp);


#if E2E_DEBUG
	for (curAck = srp->ack ; curAck != NULL ; curAck = curAck->next) {
		ASSERT(curAck->opt.isE2E == 0);
	}
#endif

    //
    // We start with one reference (our own) for the SRPacket.
    //
    srp->ForwardCount = 1;
    srp->ForwardStatus = NDIS_STATUS_SUCCESS;

    //
    // First we build a temporary list of packet structures
    // and corresponding physical adapters. While doing this,
    // we also initialize the transmission count.
    // It starts with one, for our own reference for OrigPacket.
    //
    KeAcquireSpinLock(&VA->Lock, &OldIrql);
    for (PA = VA->PhysicalAdapters;
         PA != NULL;
         PA = PA->Next) {

        if (!PA->Use)
            continue;

        srp->req->opt.hopList[Hops].outif = (LQSRIf) PA->Index;

        Status = SRPacketToPkt(VA, srp, &Packet);
        if (Status != NDIS_STATUS_SUCCESS) {
            srp->ForwardStatus = Status;
            break;
        }

        //
        // Remember the packet and corresponding physical adapter.
        // We temporarily use the OrigPacket field to build our list.
        //
        PC(Packet)->PA = PA;
        PC(Packet)->OrigPacket = PacketList;
        PacketList = Packet;
        srp->ForwardCount++;
    }
    KeReleaseSpinLock(&VA->Lock, OldIrql);

    //
    // Now we can transmit the packet via each physical adapter,
    // using the new packet structures.
    //
    while ((Packet = PacketList) != NULL) {
        PacketList = PC(Packet)->OrigPacket;
        PA = PC(Packet)->PA;

        //
        // Send the packet via a physical adapter.
        //
        PC(Packet)->srp = srp;
        PC(Packet)->TransmitComplete = ProtocolForwardRequestComplete;
        ProtocolTransmit(PA, Packet);
    }

    //
    // Release our reference for the original packet.
    //
    if (InterlockedDecrement((PLONG)&srp->ForwardCount) == 0) {
        //
        // Complete the ProtocolForwardRequest operation.
        //
        (*srp->TransmitComplete)(VA, srp, srp->ForwardStatus);
    }
}

//* ProtocolResetComplete
//
//  Called by NDIS when a reset completes.
//
void
ProtocolResetComplete(
    NDIS_HANDLE Handle,  // Binding handle.
    NDIS_STATUS Status)  // Final status of command.
{
    ProtocolAdapter *PA = (ProtocolAdapter *)Handle;

    UNREFERENCED_PARAMETER(Status);
    UNREFERENCED_PARAMETER(PA);

    if(debug) KdPrint(("MCL!ProtocolResetComplete(VA %p PA %p)\n",
             PA->VirtualAdapter, PA));
}

//* ProtocolRequestComplete
//
//  Called by NDIS when an NdisRequest completes.
//  We block on all requests, so we'll just wake up
//  whoever's blocked on this request.
//
void
ProtocolRequestComplete(
    NDIS_HANDLE Handle,     // Binding handle.
    PNDIS_REQUEST Context,  // Request that completed.
    NDIS_STATUS Status)     // Final status of requested command.
{
    ProtocolRequest *Request = (ProtocolRequest *) Context;

    UNREFERENCED_PARAMETER(Handle);

    //
    // Signal the completion of a generic synchronous request.
    // See ProtocolRequestHelper.
    //
    Request->Status = Status;
    KeSetEvent(&Request->Event, 0, FALSE);
}

//* ProtocolFreePacket
//
//  Frees a packet allocated by ProtocolReceive.
//
void
ProtocolFreePacket(
    ProtocolAdapter *PA,
    NDIS_PACKET *Packet)
{
    NDIS_BUFFER *Buffer;
    void *Data;

    Buffer = NdisFirstBuffer(Packet);
    ASSERT((Buffer != NULL) && (Buffer->Next == NULL));
    ASSERT(Buffer->MdlFlags & MDL_SOURCE_IS_NONPAGED_POOL);
    Data = Buffer->MappedSystemVa;

    NdisFreePacket(Packet);
    NdisFreeBuffer(Buffer);
    ExFreePool(Data);

    InterlockedDecrement((PLONG)&PA->CountRecvOutstanding);
}

//* ProtocolTransferDataComplete
//
//  Called by NDIS when a transfer data completes.
//  We now have a complete packet.
//
void
ProtocolTransferDataComplete(
    NDIS_HANDLE Handle,   // Binding handle.
    PNDIS_PACKET Packet,  // The packet used for the Transfer Data (TD).
    NDIS_STATUS Status,   // Final status of command.
    uint BytesCopied)     // Number of bytes copied.
{
    ProtocolAdapter *PA = (ProtocolAdapter *)Handle;
    NDIS_BUFFER *Buffer;

	//KIRQL OldIrql; 
	//Time UpdateTimeout; 

    UNREFERENCED_PARAMETER(BytesCopied);

    //
    // Undo the NdisAdjustBuffer in ProtocolReceive.
    //
    Buffer = NdisFirstBuffer(Packet);
    NdisAdjustBuffer(Buffer,
                     (uchar *)Buffer->MappedSystemVa - sizeof(EtherHeader),
                     Buffer->ByteCount + sizeof(EtherHeader));

    if (Status == NDIS_STATUS_SUCCESS) {


        //
        // We have the packet data so receive the packet.
        //
        MiniportReceivePacket(PA->VirtualAdapter, PA,
                              Packet, ProtocolFreePacket);
    }
    else {
        if(debug) KdPrint(("MCL!ProtocolTransferDataComplete(VA %p PA %p) -> %x\n",
                 PA->VirtualAdapter, PA, Status));

        //
        // Free the packet.
        //
        ProtocolFreePacket(PA, Packet);
    }
}

//* ProtocolReceive
//
//  Called by NDIS when data arrives.
//  Note that newer NDIS drivers are likely to call ProtocolReceivePacket to
//  indicate data arrival instead of this routine.
//
//  The status code tells NDIS whether or not we took the packet.
//
NDIS_STATUS
ProtocolReceive(
    NDIS_HANDLE ProtocolBindingContext,
    NDIS_HANDLE MacReceiveContext,
    void *HeaderBuffer,
    uint HeaderBufferSize,
    void *LookAheadBuffer,
    uint LookAheadBufferSize,
    uint PacketSize)
{
    ProtocolAdapter *PA = (ProtocolAdapter *) ProtocolBindingContext;
    EtherHeader UNALIGNED *Ether;
    NDIS_PACKET *Packet;
    NDIS_BUFFER *Buffer;
    void *Data;
    uint UNALIGNED *Code;
    NDIS_STATUS Status;

	//KIRQL OldIrql; 
	//Time UpdateTimeout;

    UNREFERENCED_PARAMETER(HeaderBufferSize);

    //
    // Because we only bind to Ethernets.
    //
    ASSERT(HeaderBufferSize == sizeof *Ether);
    Ether = (EtherHeader UNALIGNED *) HeaderBuffer;

    //
    // Check both the EtherType and the following code value
    // to ensure that we only receive LQSR packets.
    //

	if (Ether->Type != ETYPE_MSFT) {  
		return NDIS_STATUS_NOT_RECOGNIZED;
	}

	if ((PacketSize < sizeof *Code) || (LookAheadBufferSize < sizeof *Code)) {
		return NDIS_STATUS_NOT_RECOGNIZED;
	}

    Code = (uint UNALIGNED *) LookAheadBuffer;
	if (*Code != LQSR_CODE) {
		return NDIS_STATUS_NOT_RECOGNIZED;
	}
    InterlockedIncrement((PLONG)&PA->PacketsReceived);

    //
    // Allocate non-paged pool to hold the packet data.
    //
    Data = ExAllocatePool(NonPagedPool, sizeof *Ether + PacketSize);
	if (Data == NULL) {
		return NDIS_STATUS_RESOURCES;
	}

    //
    // Allocate a packet structure.
    //
    NdisAllocatePacket(&Status, &Packet, PA->PacketPool);
    if (Status != NDIS_STATUS_SUCCESS) {
        InterlockedIncrement((PLONG)&PA->CountPacketPoolFailure);
        ExFreePool(Data);
		return Status;
    }

    //
    // Allocate a packet buffer.
    //
    NdisAllocateBuffer(&Status, &Buffer, PA->BufferPool,
                       Data, sizeof *Ether + PacketSize);
    if (Status != NDIS_STATUS_SUCCESS) {
        NdisFreePacket(Packet);
        ExFreePool(Data);
        return Status;
    }

    NdisChainBufferAtFront(Packet, Buffer);

    InterlockedIncrementHighWater((PLONG)&PA->CountRecvOutstanding,
                                  (PLONG)&PA->MaxRecvOutstanding);

    //
    // Do we have the entire packet?
    //
    if (LookAheadBufferSize < PacketSize) {
        uint Transferred;

        InterlockedIncrement((PLONG)&PA->PacketsReceivedTD);

        //
        // Copy the Ethernet header.
        //
        RtlCopyMemory(Data, HeaderBuffer, sizeof *Ether);

        //
        // We must asynchronously transfer the packet data.
        // To do this we must adjust the buffer to move past
        // the Ethernet header.
        //
        NdisAdjustBuffer(Buffer, (uchar *)Data + sizeof *Ether, PacketSize);
        NdisTransferData(&Status, PA->Handle, MacReceiveContext,
                         0, PacketSize, Packet, &Transferred);

        if (Status != NDIS_STATUS_PENDING) {
            //
            // The transfer completed synchronously.
            //
            ProtocolTransferDataComplete(ProtocolBindingContext,
                                         Packet, Status, Transferred);
        }
    }
    else {
        InterlockedIncrement((PLONG)&PA->PacketsReceivedFlat);



        //
        // We already have access to the packet data.
        //
        RtlCopyMemory(Data, HeaderBuffer, sizeof *Ether);
        RtlCopyMemory((uchar *)Data + sizeof *Ether,
                      LookAheadBuffer, PacketSize);
        MiniportReceivePacket(PA->VirtualAdapter, PA,
                              Packet, ProtocolFreePacket);
    }

  
    
	return NDIS_STATUS_SUCCESS;
}

//* ProtocolReceiveComplete
//
//  Called by NDIS after some number of receives.
//  In some sense, it indicates 'idle time'.
//
void
ProtocolReceiveComplete(
    NDIS_HANDLE Handle)  // Binding handle.
{
    UNREFERENCED_PARAMETER(Handle);
}

typedef struct ProtocolMediaConnectContext {
    PIO_WORKITEM Item;
    ProtocolAdapter *PA;
} ProtocolMediaConnectContext;

//* ProtocolQueryRadioConfiguration
//
//  Uses ProtocolRequestHelper to query the channel and bandwidth
//  of the physical adapter.
//
//  Called in a thread context, not at DPC level.
//
void
ProtocolQueryRadioConfiguration(
    ProtocolAdapter *PA)
{
    NDIS_STATUS Status;

    //
    // Note that a burst of media-connect events can lead to races
    // with multiple simultaneous ProtocolQueryRadioConfiguration
    // executions.  We need the mutex to ensure that after the last
    // media-connect, the adapter configuration stabilizes with
    // correct Channel and Bandwidth values.
    //
    ExAcquireFastMutex(&PA->Mutex);

    if (! PA->BandwidthConfigured) {
        uint Speed;

        //
        // Query the maximum link bandwidth.
        //
        Status = ProtocolRequestHelper(PA, NdisRequestQueryInformation,
                                       OID_GEN_LINK_SPEED,
                                       &Speed, sizeof Speed,
                                       NULL);
        if (Status != NDIS_STATUS_SUCCESS) {
            //
            // Use 10Mbs as a default.
            //
            Speed = 100000;
        }

        //
        // The OID returns the speed value in 100 bps units.
        //
        PA->Bandwidth = WcettEncodeBandwidth(Speed * 100);
    }

    if (! PA->ChannelConfigured) {
        NDIS_802_11_CONFIGURATION Radio;

        //
        // Use Channel zero as a default.
        //
        PA->Channel = 0;

        //
        // Query the radio configuration.
        // NB: If the radio is still associating, this query will fail!
        //
        Status = ProtocolRequestHelper(PA, NdisRequestQueryInformation,
                                       OID_802_11_CONFIGURATION,
                                       &Radio, sizeof Radio,
                                       NULL);
        if (Status == NDIS_STATUS_SUCCESS) {
            //
            // The OID returns the frequency in kHz.
            // Convert to a channel number.
            //

            //
            // We've encounted some radios that return the frequency in
            // incorrect units.  Attempt to compensate.
            //
            if (Radio.DSConfig < 2412000)
                Radio.DSConfig = Radio.DSConfig * 100;

            switch (Radio.DSConfig) {

#define Case24GhzChannel(ch) \
                case ((ch * 5) + 2407) * 1000: \
                    PA->Channel = ch; \
                    break

                Case24GhzChannel(1);
                Case24GhzChannel(2);
                Case24GhzChannel(3);
                Case24GhzChannel(4);
                Case24GhzChannel(5);
                Case24GhzChannel(6);
                Case24GhzChannel(7);
                Case24GhzChannel(8);
                Case24GhzChannel(9);
                Case24GhzChannel(10);
                Case24GhzChannel(11);

#define Case5GhzChannel(ch) \
                case ((ch * 5) + 5000) * 1000: \
                    PA->Channel = ch; \
                    break

                Case5GhzChannel(36);
                Case5GhzChannel(37);
                Case5GhzChannel(38);
                Case5GhzChannel(39);
                Case5GhzChannel(40);
                Case5GhzChannel(41);
                Case5GhzChannel(42);
                Case5GhzChannel(43);
                Case5GhzChannel(44);
                Case5GhzChannel(45);
                Case5GhzChannel(46);
                Case5GhzChannel(47);
                Case5GhzChannel(48);
                Case5GhzChannel(49);
                Case5GhzChannel(50);
                Case5GhzChannel(51);
                Case5GhzChannel(52);
                Case5GhzChannel(53);
                Case5GhzChannel(54);
                Case5GhzChannel(55);
                Case5GhzChannel(56);
                Case5GhzChannel(57);
                Case5GhzChannel(58);
                Case5GhzChannel(59);
                Case5GhzChannel(60);
                Case5GhzChannel(61);
                Case5GhzChannel(62);
                Case5GhzChannel(63);
                Case5GhzChannel(64);
                Case5GhzChannel(149);
                Case5GhzChannel(150);
                Case5GhzChannel(151);
                Case5GhzChannel(152);
                Case5GhzChannel(153);
                Case5GhzChannel(154);
                Case5GhzChannel(155);
                Case5GhzChannel(156);
                Case5GhzChannel(157);
                Case5GhzChannel(158);
                Case5GhzChannel(159);
                Case5GhzChannel(160);
                Case5GhzChannel(161);
                Case5GhzChannel(162);
                Case5GhzChannel(163);
                Case5GhzChannel(164);
                Case5GhzChannel(165);
            }
        }
    }

    ExReleaseFastMutex(&PA->Mutex);
}

//* ProtocolMediaConnect
//
//  Called as a work-item on a system thread,
//  after NDIS indicates a media-connect event.
//
void
ProtocolMediaConnect(
    DEVICE_OBJECT *DeviceObject,
    void *Context)
{
    ProtocolMediaConnectContext *PMCC =
        (ProtocolMediaConnectContext *) Context;
    PIO_WORKITEM Item = PMCC->Item;
    ProtocolAdapter *PA = PMCC->PA;

    UNREFERENCED_PARAMETER(DeviceObject);

    IoFreeWorkItem(Item);
    ExFreePool(PMCC);

    ProtocolQueryRadioConfiguration(PA);
    LinkCacheDeleteInterface(PA->VirtualAdapter, (LQSRIf) PA->Index);
}

//* ProtocolStatus
//
//  Called by NDIS when some sort of status change occurs.
//  We can take action depending on the type of status.
//
//  Called at DPC level.
//
void
ProtocolStatus(
    NDIS_HANDLE Handle,   // Binding handle.
    NDIS_STATUS GStatus,  // General status type which caused the call.
    void *Status,         // Pointer to buffer of status specific info.
    uint StatusSize)      // Size of the above status buffer.
{
    ProtocolAdapter *PA = (ProtocolAdapter *)Handle;

    UNREFERENCED_PARAMETER(Status);
    UNREFERENCED_PARAMETER(StatusSize);

    if(debug) KdPrint(("MCL!ProtocolStatus(VA %p PA %p GStatus %x)\n",
             PA->VirtualAdapter, PA, GStatus));

    if (GStatus == NDIS_STATUS_MEDIA_CONNECT) {
        ProtocolMediaConnectContext *PMCC;
        PIO_WORKITEM Item;

        //
        // We can not use ProtocolRequestHelper at DPC level.
        // So we queue a work-item for a system thread.
        //

        PMCC = ExAllocatePool(NonPagedPool, sizeof *PMCC);
        if (PMCC == NULL)
            return;

        Item = IoAllocateWorkItem(PA->VirtualAdapter->DeviceObject);
        if (Item == NULL) {
            ExFreePool(PMCC);
            return;
        }

        PMCC->Item = Item;
        PMCC->PA = PA;

        IoQueueWorkItem(Item, ProtocolMediaConnect, DelayedWorkQueue, PMCC);
    }
}

//* ProtocolStatusComplete
//
//  Called by NDIS after a status event.
//
void
ProtocolStatusComplete(
    NDIS_HANDLE Handle)  // Binding handle.
{
    UNREFERENCED_PARAMETER(Handle);
}

//* ProtocolReturnPacket
//
//  Returns when we have finished processing a packet
//  from ProtocolReceivePacket.
//
void
ProtocolReturnPacket(
    ProtocolAdapter *PA,
    NDIS_PACKET *Packet)
{
    NdisReturnPackets(&Packet, 1);
    InterlockedDecrement((PLONG)&PA->CountRecvOutstanding);
}

//
//* ProtocolReceivePacket
//
//  Called by NDIS when data arrives.
//  Note that older NDIS drivers are likely to call ProtocolReceive to
//  indicate data arrival instead of this routine.
//
//  Returns the number of references we hold to Packet upon return.
//
int
ProtocolReceivePacket(
    NDIS_HANDLE Handle,   // Binding handle.
    PNDIS_PACKET Packet)  // Packet descriptor for incoming packet.
{
    ProtocolAdapter *PA = (ProtocolAdapter *)Handle;
    NDIS_BUFFER *Buffer;
    void *Address;
    uint Length;
    EtherHeader UNALIGNED *Ether;
    uint UNALIGNED *Code;

	
    //KIRQL OldIrql; 
	//Time UpdateTimeout;

    //
    // Find out about the packet we've been handed.
    //
    Buffer = NdisFirstBuffer(Packet);
    NdisQueryBuffer(Buffer, &Address, &Length);

    if (Length < sizeof *Ether + sizeof *Code) {
        //
        // Header data too small to contain an Ethernet header + Code.
        //
        if(debug) KdPrint(("MCL!ProtocolReceivePacket(VA %p PA %p): Length %u\n",
                 PA->VirtualAdapter, PA, Length));
        return 0;
    }

    Ether = (EtherHeader UNALIGNED *) Address;
    Code = (uint UNALIGNED *) (Ether + 1);

   
    InterlockedIncrement((PLONG)&PA->PacketsReceived);
    InterlockedIncrementHighWater((PLONG)&PA->CountRecvOutstanding,
                                  (PLONG)&PA->MaxRecvOutstanding);



	MiniportReceivePacket(PA->VirtualAdapter, PA,
                          Packet, ProtocolReturnPacket);

	
	
    return 1;
}


//Format the address so we can print to the screen
char *
FormatPhysicalAddress(PhysicalAddress Address)
{

    static char buffer[128];

    sprintf(buffer, "%02x-%02x-%02x-%02x-%02x-%02x",
            Address[0], Address[1], Address[2],
            Address[3], Address[4], Address[5]);

    return buffer;
}

//* ProtocolCloseAdapter
//
//  Closes our connection to NDIS and frees the adapter.
//  The adapter has already been removed
//  from the MiniportAdapter->PhysicalAdapters list.
//
//  Called from thread context, with no locks held.
//
void
ProtocolCloseAdapter(ProtocolAdapter *PA)
{
    NDIS_STATUS Status;
	uint numOut = 0;

    if(debug) KdPrint(("MCL!ProtocolCloseAdapter(VA %p PA %p)\n",
             PA->VirtualAdapter, PA));

    //
    // Initialize an event in case NdisCloseAdapter completes
    // asynchronously via ProtocolCloseAdapterComplete.
    //
    KeInitializeEvent(&PA->Event, SynchronizationEvent, FALSE);

	if(debug) KdPrint(("Initialize event done\n"));


    //
    // Close the connection to NDIS.
    //
    // NB: We must not use the Handle after we call NdisCloseAdapter.
    //
    NdisCloseAdapter(&Status, PA->Handle);

	if(debug) KdPrint(("NDIS close adapter done\n")); 
    //
    // Wait for the close to complete.
    //
    if (Status == NDIS_STATUS_PENDING) {
        (void) KeWaitForSingleObject(&PA->Event,
                                     UserRequest, KernelMode,
                                     FALSE, NULL);
        Status = PA->Status;
    }

    if(debug) KdPrint(("MCL!NdisCloseAdapter(VA %p PA %p) -> %x\n",
             PA->VirtualAdapter, PA, Status));

    //ASSERT(NdisPacketPoolUsage(PA->PacketPool) == 0);
	numOut = NdisPacketPoolUsage(PA->PacketPool);
	if (numOut != 0) {
		if(debug) KdPrint(("\n\n2ERROR IN PACKET POOL USAGE\n %d AMOUNT OF PACKETS IN PACKET POOL\nERROR\n\n", numOut));

	}

    NdisFreeBufferPool(PA->BufferPool);
    NdisFreePacketPool(PA->PacketPool);
    ExFreePool(PA);
}

//* ProtocolOpenRegKey
//
//  Given an adapter, opens the registry key that holds
//  persistent configuration information for the adapter.
//
//  Callable from thread context, not DPC context.
//
NTSTATUS
ProtocolOpenRegKey(
    ProtocolAdapter *PA,
    OpenRegKeyAction Action,
    OUT HANDLE *RegKey)
{
    MiniportAdapter *VA = PA->VirtualAdapter;
    UNICODE_STRING GuidName;
    HANDLE DeviceKey;
    HANDLE ParamsKey;
    HANDLE AdaptersKey;
    NTSTATUS Status;

    PAGED_CODE();

    //
    // Open our configuration information in the registry.
    //
    Status = IoOpenDeviceRegistryKey(VA->PhysicalDeviceObject,
                                     PLUGPLAY_REGKEY_DRIVER,
                                     GENERIC_READ,
                                     &DeviceKey);
    if (! NT_SUCCESS(Status)) {
        if(debug) KdPrint(("MCL!IoOpenDeviceRegistryKey -> %x\n", Status));
        return Status;
    }

    Status = OpenRegKey(&ParamsKey, DeviceKey,
                        L"Parameters",
                        ((Action == OpenRegKeyCreate) ?
                         OpenRegKeyCreate : OpenRegKeyRead));
    ZwClose(DeviceKey);
    if (! NT_SUCCESS(Status))
        return Status;

    Status = OpenRegKey(&AdaptersKey, ParamsKey,
                        L"Adapters",
                        ((Action == OpenRegKeyCreate) ?
                         OpenRegKeyCreate : OpenRegKeyRead));
    ZwClose(ParamsKey);
    if (! NT_SUCCESS(Status))
        return Status;

    //
    // Convert the guid to string form.
    //
    Status = RtlStringFromGUID(&PA->Guid, &GuidName);
    if (! NT_SUCCESS(Status)) {
        ZwClose(AdaptersKey);
        return Status;
    }

    //
    // It will be null-terminated.
    //
    ASSERT(GuidName.MaximumLength == GuidName.Length + sizeof(WCHAR));
    ASSERT(((WCHAR *)GuidName.Buffer)[GuidName.Length/sizeof(WCHAR)] == UNICODE_NULL);

    Status = OpenRegKey(RegKey, AdaptersKey,
                        (WCHAR *)GuidName.Buffer, Action);

    RtlFreeUnicodeString(&GuidName);
    ZwClose(AdaptersKey);
    return Status;
}

//* ProtocolPersistControl
//
//  Writes configuration parameters to the registry.
//
NTSTATUS
ProtocolPersistControl(
    ProtocolAdapter *PA,
    MCL_CONTROL_PHYSICAL_ADAPTER *Control)
{
    NTSTATUS Status;
    HANDLE Key;

    //
    // We do not need to validate the configuration parameters.
    // They will be validated after being read from the registry.
    //

    //
    // Open our configuration information in the registry.
    //
    Status = ProtocolOpenRegKey(PA, OpenRegKeyCreate, &Key);
    if (! NT_SUCCESS(Status)) {
        if(debug) KdPrint(("LQSR:ProtocolOpenRegKey: %x\n", Status));
        return Status;
    }

    if (Control->Use != (uint)-1) {
        Status = SetRegDWORDValue(Key, L"Use", Control->Use);
        if (! NT_SUCCESS(Status))
            goto CloseReturn;
    }

    if (Control->ReceiveOnly != (uint)-1) {
        Status = SetRegDWORDValue(Key, L"ReceiveOnly", Control->ReceiveOnly);
        if (! NT_SUCCESS(Status))
            goto CloseReturn;
    }

    if (Control->Channel != 0) {
        Status = SetRegDWORDValue(Key, L"Channel", Control->Channel);
        if (! NT_SUCCESS(Status))
            goto CloseReturn;
    }

    if (Control->Bandwidth != 0) {
        Status = SetRegDWORDValue(Key, L"Bandwidth", Control->Bandwidth);
        if (! NT_SUCCESS(Status))
            goto CloseReturn;
    }

    Status = STATUS_SUCCESS;

  CloseReturn:
    ZwClose(Key);
    return Status;
}

//* ProtocolInitAdapter
//
//  Initializes a physical adapter from the registry.
//
void
ProtocolInitAdapter(ProtocolAdapter *PA)
{
    MCL_CONTROL_PHYSICAL_ADAPTER Control;
    HANDLE Key;
    NTSTATUS Status;

    PAGED_CODE();

    //
    // Open our configuration information in the registry.
    //
    Status = ProtocolOpenRegKey(PA, OpenRegKeyCreate, &Key);
    if (! NT_SUCCESS(Status)) {
        if(debug) KdPrint(("LQSR:ProtocolOpenRegKey: %x\n", Status));
        return;
    }

    MCL_INIT_CONTROL_PHYSICAL_ADAPTER(&Control);

    //
    // Read configuration parameters.
    //
    (void) GetRegDWORDValue(Key, L"Use", (PULONG)&Control.Use);
    (void) GetRegDWORDValue(Key, L"ReceiveOnly", (PULONG)&Control.ReceiveOnly);
    (void) GetRegDWORDValue(Key, L"Channel", (PULONG)&Control.Channel);
    (void) GetRegDWORDValue(Key, L"Bandwidth", (PULONG)&Control.Bandwidth);
    ZwClose(Key);

    //
    // Update the adapter using the parameters.
    //
    (void) ProtocolControl(PA, &Control);
}

//* ProtocolBindAdapter
//
//  Called by NDIS to tell us about a new adapter.
//
//  Called in a thread context, not at DPC level.
//
void
ProtocolBindAdapter(
    PNDIS_STATUS RetStatus,    // Where to return status of this call.
    NDIS_HANDLE BindContext,   // Handle for calling NdisCompleteBindAdapter.
    PNDIS_STRING AdapterName,  // Name of adapter.
    PVOID SS1,                 // System specific parameter 1.
    PVOID SS2)                 // System specific parameter 2.
{
    //
    // Grovel through the BindContext to get the NDIS protocol handle,
    // and from that get our MiniportAdapter structure.
    // We do this because ProtocolBindAdapter does not give us context.
    //
    NDIS_HANDLE Protocol = NdisProtocolFromBindContext(BindContext);
    MiniportAdapter *VA = (MiniportAdapter *)
        *NdisReservedFieldInProtocol(Protocol);

    UNICODE_STRING GuidName;
    GUID Guid;
    uint BindPrefixLength;

    ProtocolAdapter *PA;
    uint MediaIndex;
    static NDIS_MEDIUM MediaArray[] = { NdisMedium802_3 };
    NDIS_STATUS Status, ErrorStatus;
    uint Filter;
    KIRQL OldIrql;

    UNREFERENCED_PARAMETER(SS1);
    UNREFERENCED_PARAMETER(SS2);

    //
    // Parse past the NDIS binding prefix:
    // We skip the first character (a backslash) and look for a backslash.
    // If we don't find a backslash, RtlGUIDFromString will fail
    // when given a zero-length string.
    //
    BindPrefixLength = 1;
    while (sizeof(WCHAR)*BindPrefixLength < AdapterName->Length) {
        if (((WCHAR *)AdapterName->Buffer)[BindPrefixLength++] == L'\\')
            break;
    }
    BindPrefixLength *= sizeof(WCHAR); // Convert characters to bytes.

    //
    // Convert the NDIS AdapterName to a Guid.
    //
    GuidName.Buffer = (PVOID)((char *)AdapterName->Buffer + BindPrefixLength);
    GuidName.Length = (ushort)(AdapterName->Length - BindPrefixLength);
    GuidName.MaximumLength = (ushort)(AdapterName->MaximumLength - BindPrefixLength);

    if (! NT_SUCCESS(RtlGUIDFromString(&GuidName, &Guid))) {

        if(debug) KdPrint(("MCL!ProtocolBindAdapter(%.*ls) - bad guid\n",
                 AdapterName->Length / sizeof(WCHAR),
                 AdapterName->Buffer));
        *RetStatus = NDIS_STATUS_FAILURE;
        return;
    }

    PA = (ProtocolAdapter *) ExAllocatePool(NonPagedPool, sizeof *PA);
    if (PA == NULL) {
        *RetStatus = NDIS_STATUS_FAILURE;
        return;
    }

    if(debug) KdPrint(("MCL!ProtocolBindAdapter(VA %p, Name %.*ls) -> PA %p\n",
             VA,
             AdapterName->Length / sizeof(WCHAR),
             AdapterName->Buffer,
             PA));

    RtlZeroMemory(PA, sizeof *PA);
    PA->VirtualAdapter = VA;
    PA->Index = InterlockedIncrement((PLONG)&VA->NextPhysicalAdapterIndex);
    PA->Guid = Guid;

    NdisAllocatePacketPool(&Status, &PA->PacketPool, PACKET_POOL_SZ, 0);
    if (Status != NDIS_STATUS_SUCCESS) {
        ExFreePool(PA);
        *RetStatus = Status;
        return;
    }

    NdisAllocateBufferPool(&Status, &PA->BufferPool, PACKET_POOL_SZ);
    if (Status != NDIS_STATUS_SUCCESS) {
        NdisFreePacketPool(PA->PacketPool);
        ExFreePool(PA);
        *RetStatus = Status;
        return;
    }

    //
    // Initialize an event in case NdisOpenAdapter completes
    // asynchronously via ProtocolOpenAdapterComplete.
    //
    KeInitializeEvent(&PA->Event, SynchronizationEvent, FALSE);

    //
    // Initialize a mutex used to synchronize attribute retrieval
    // from NDIS.
    //
    ExInitializeFastMutex(&PA->Mutex);

    NdisOpenAdapter(&Status, &ErrorStatus, &PA->Handle,
                    &MediaIndex, MediaArray, sizeof MediaArray / sizeof MediaArray[0],
                    VA->ProtocolHandle, PA, AdapterName,
                    0, NULL);

    //
    // Wait for the open to complete.
    //
    if (Status == NDIS_STATUS_PENDING) {
        (void) KeWaitForSingleObject(&PA->Event, UserRequest, KernelMode,
                                     FALSE, NULL);
        Status = PA->Status;
    }

    if (Status != NDIS_STATUS_SUCCESS) {
        NdisFreeBufferPool(PA->BufferPool);
        NdisFreePacketPool(PA->PacketPool);
        ExFreePool(PA);
        *RetStatus = Status;
        return;
    }

    //
    // Query the maximum frame size.
    // We can only use the adapter if it has a sufficiently large frame,
    // because we need to insert headers.
    //
    Status = ProtocolRequestHelper(PA, NdisRequestQueryInformation,
                                   OID_GEN_MAXIMUM_FRAME_SIZE,
                                   &PA->MaxFrameSize, sizeof PA->MaxFrameSize,
                                   NULL);
    if ((Status != NDIS_STATUS_SUCCESS) ||
        (PA->MaxFrameSize < PROTOCOL_MIN_FRAME_SIZE)) {

        ProtocolCloseAdapter(PA);
        *RetStatus = NDIS_STATUS_FAILURE;
        return;
    }

    //
    // Query the maximum total size.
    // This is the frame size plus header.
    //
    Status = ProtocolRequestHelper(PA, NdisRequestQueryInformation,
                                   OID_GEN_MAXIMUM_TOTAL_SIZE,
                                   &PA->MaxTotalSize, sizeof PA->MaxTotalSize,
                                   NULL);
    if ((Status != NDIS_STATUS_SUCCESS) ||
        (PA->MaxTotalSize < PROTOCOL_MIN_FRAME_SIZE + sizeof(EtherHeader))) {

        ProtocolCloseAdapter(PA);
        *RetStatus = NDIS_STATUS_FAILURE;
        return;
    }

    //
    // Query the adapter address.
    //
    Status = ProtocolRequestHelper(PA, NdisRequestQueryInformation,
                                   OID_802_3_CURRENT_ADDRESS,
                                   &PA->Address, sizeof PA->Address,
                                   NULL);
    if (Status != NDIS_STATUS_SUCCESS) {

        ProtocolCloseAdapter(PA);
        *RetStatus = NDIS_STATUS_FAILURE;
        return;
    }

    //
    // Initialize from the registry.
    // Adapter defaults to "use" unless registry over-rides.
    //
    PA->Use = TRUE;
    ProtocolInitAdapter(PA);

    //
    // Query the adapter bandwidth and channel,
    // if they were not configured from the registry.
    //
    ProtocolQueryRadioConfiguration(PA);

    //
    // Add the new physical adapter to our list.
    //
    KeAcquireSpinLock(&VA->Lock, &OldIrql);
    PA->Prev = &VA->PhysicalAdapters;
    PA->Next = VA->PhysicalAdapters;
    if (VA->PhysicalAdapters != NULL)
        VA->PhysicalAdapters->Prev = &PA->Next;
    VA->PhysicalAdapters = PA;
    KeReleaseSpinLock(&VA->Lock, OldIrql);

    //
    // If we've been configured to not use this adapter,
    // then don't start receiving frames on it.
    //
    if (!PA->Use) {
        *RetStatus = NDIS_STATUS_SUCCESS;
        return;
    }

    //
    // Set the receive filter so that we start receiving frames.
    // We want promiscuous mode but not all adapters support it.
    //
    Filter = NDIS_PACKET_TYPE_PROMISCUOUS;
    Status = ProtocolRequestHelper(PA, NdisRequestSetInformation,
                                   OID_GEN_CURRENT_PACKET_FILTER,
                                   &Filter, sizeof Filter, NULL);
    if (Status == NDIS_STATUS_SUCCESS) {
        //
        // We got promiscuous mode.
        //
        PA->Promiscuous = TRUE;
    }
    else {
        ASSERT(PA->Promiscuous == FALSE);

        Filter = (NDIS_PACKET_TYPE_BROADCAST |
                  NDIS_PACKET_TYPE_DIRECTED |
                  NDIS_PACKET_TYPE_MULTICAST);
        Status = ProtocolRequestHelper(PA, NdisRequestSetInformation,
                                       OID_GEN_CURRENT_PACKET_FILTER,
                                       &Filter, sizeof Filter, NULL);
        if (Status != NDIS_STATUS_SUCCESS)
            if(debug) KdPrint(("MCL!ProtocolRequestHelper(PA %p, PF) -> %x\n",
                     PA, Status));
    }

    //
    // If this is the first physical adapter, indicate that
    // that our miniport is now connected.
    //
    if (PA->Next == NULL)
        MiniportIndicateStatusConnected(VA);

    *RetStatus = NDIS_STATUS_SUCCESS;
}

//* ProtocolUnbindAdapter
//
//  Called by NDIS when we need to unbind from an adapter.
//
void
ProtocolUnbindAdapter(
    PNDIS_STATUS RetStatus,       // Where to return status from this call.
    NDIS_HANDLE Handle,           // Context we gave NDIS earlier.
    NDIS_HANDLE UnbindContext)    // Context for completing this request.
{
    ProtocolAdapter *PA = (ProtocolAdapter *)Handle;
    MiniportAdapter *VA = PA->VirtualAdapter;
    KIRQL OldIrql;

    UNREFERENCED_PARAMETER(UnbindContext);

    if(debug) KdPrint(("MCL!ProtocolUnbindAdapter(VA %p PA %p)\n", VA, PA));

    //
    // Remove the adapter from our list.
    //
    KeAcquireSpinLock(&VA->Lock, &OldIrql);
    if (PA->Next != NULL)
        PA->Next->Prev = PA->Prev;
    *PA->Prev = PA->Next;
    KeReleaseSpinLock(&VA->Lock, OldIrql);

    //
    // Remove the adapter from the link cache.
    //
    LinkCacheDeleteInterface(VA, (LQSRIf) PA->Index);

    //
    // If this was the last physical adapter, indicate that
    // our miniport is no longer connected to a physical adapter.
    //
    // NB: We have a race with ProtocolBindAdapter and could
    // indicate disconnected after it indicates connected.
    //
    if (VA->PhysicalAdapters == NULL)
        MiniportIndicateStatusDisconnected(VA);

    ProtocolCloseAdapter(PA);
    *RetStatus = NDIS_STATUS_SUCCESS;
}

//* ProtocolPnPEvent
//
//  Called by NDIS for plug'n'play and power-management events.
//
//  Called in a thread context, not at DPC level.
//
NDIS_STATUS
ProtocolPnPEvent(
    NDIS_HANDLE Handle,   // Binding handle.
    PNET_PNP_EVENT NetPnPEvent)
{
    static char *NetEvents[] = {
        "set power",
        "query power",
        "query remove",
        "cancel remove",
        "reconfigure",
        "bind list",
        "binds complete",
        "pnp capabilities",
    };
    ProtocolAdapter *PA = (ProtocolAdapter *)Handle;

    if(debug) KdPrint(("MCL!ProtocolPnPEvent(PA %p Event %s)\n", PA,
             (NetPnPEvent->NetEvent < sizeof NetEvents / sizeof NetEvents[0]) ?
             NetEvents[NetPnPEvent->NetEvent] : "unknown"));

    if (NetPnPEvent->NetEvent == NetEventSetPower) {
        NET_DEVICE_POWER_STATE PowerState;

        //
        // Get the power state of the interface.
        //
        ASSERT(NetPnPEvent->BufferLength >= sizeof PowerState);
        PowerState = * (NET_DEVICE_POWER_STATE *) NetPnPEvent->Buffer;

        if (PowerState == NetDeviceStateD0) {
            //
            // The device woke up.
            //
            ProtocolQueryRadioConfiguration(PA);
        }
        LinkCacheDeleteInterface(PA->VirtualAdapter, (LQSRIf) PA->Index);
    }

    return NDIS_STATUS_SUCCESS;
}

//* ProtocolInit
//
//  Registers an LQSR transport with NDIS.
//
NDIS_STATUS
ProtocolInit(
    IN const WCHAR *Name,
    IN MiniportAdapter *VA)
{
    NDIS_PROTOCOL_CHARACTERISTICS PC;
    NDIS_STATUS Status;

    RtlZeroMemory(&PC, sizeof PC);
    RtlInitUnicodeString(&PC.Name, Name);
    PC.MajorNdisVersion = NDIS_PROTOCOL_MAJOR_VERSION;
    PC.MinorNdisVersion = NDIS_PROTOCOL_MINOR_VERSION;
    PC.OpenAdapterCompleteHandler = ProtocolOpenAdapterComplete;
    PC.CloseAdapterCompleteHandler = ProtocolCloseAdapterComplete;
    PC.SendCompleteHandler = ProtocolTransmitComplete;
    PC.TransferDataCompleteHandler = ProtocolTransferDataComplete;
    PC.ResetCompleteHandler = ProtocolResetComplete;
    PC.RequestCompleteHandler = ProtocolRequestComplete;
    PC.ReceiveHandler = ProtocolReceive;
    PC.ReceiveCompleteHandler = ProtocolReceiveComplete;
    PC.StatusHandler = ProtocolStatus;
    PC.StatusCompleteHandler = ProtocolStatusComplete;
    PC.ReceivePacketHandler = ProtocolReceivePacket;
    PC.BindAdapterHandler = ProtocolBindAdapter;
    PC.UnbindAdapterHandler = ProtocolUnbindAdapter;
    PC.PnPEventHandler = ProtocolPnPEvent;

    NdisRegisterProtocol(&Status, &VA->ProtocolHandle, &PC, sizeof PC);
    if(debug) KdPrint(("MCL!NdisRegisterProtocol(VA %p) -> %x\n", VA, Status));
    if (Status != NDIS_STATUS_SUCCESS)
        return Status;

    //
    // Modify the NDIS protocol structure to point
    // to our MiniportAdapter structure. We need this because
    // ProtocolBindAdapter does not supply us with a context.
    //
    *NdisReservedFieldInProtocol(VA->ProtocolHandle) = VA;

    return NDIS_STATUS_SUCCESS;
}

//* ProtocolCleanup
//
//  Uninitializes an LQSR transport module.
//  Closes any remaining adapters.
//
void
ProtocolCleanup(MiniportAdapter *VA)
{
    NDIS_STATUS Status;

    //
    // NDIS will call ProtocolUnbindAdapter for each adapter,
    // so we are left with no physical adapters.
    //
	if(debug) KdPrint(("Going to deregister protocol\n")); 
    NdisDeregisterProtocol(&Status, VA->ProtocolHandle);
    if(debug) KdPrint(("MCL!NdisDeregisterProtocol(VA %p) -> %x\n", VA, Status));
    ASSERT(VA->PhysicalAdapters == NULL);
}


#if RATE_LIMIT_FILE
//* ProtocolCreateRateFile
//
//  Prints the current rate-limiting parameters for ttcp
//  to read in.
//
//  dataPkts- how many pkts we are allowed to send in the next interval
//
void
ProtocolCreateRateFile(uint dataPkts)
{
	//zwcreatefile	
	//http://msdn2.microsoft.com/en-us/library/ms804358.aspx
	HANDLE  FileHandle ;
	ACCESS_MASK  DesiredAccess = GENERIC_ALL;
	OBJECT_ATTRIBUTES  ObjectAttributes;
	IO_STATUS_BLOCK  IoStatusBlock ;
	PLARGE_INTEGER  AllocationSize = NULL ;
	ULONG  FileAttributes = FILE_ATTRIBUTE_NORMAL ;
	ULONG  ShareAccess = FILE_SHARE_READ|FILE_SHARE_WRITE|FILE_SHARE_DELETE ;
	ULONG  CreateDisposition = FILE_SUPERSEDE ;
	ULONG  CreateOptions = FILE_NON_DIRECTORY_FILE;//|FILE_SEQUENTIAL_ONLY ;
	PVOID  EaBuffer = NULL;
	ULONG  EaLength = 0 ;
	NTSTATUS fStatus;

	//initializeobjectattributes
	UNICODE_STRING  ObjectName;
	ULONG  Attributes = OBJ_CASE_INSENSITIVE | OBJ_KERNEL_HANDLE | OBJ_FORCE_ACCESS_CHECK;
	HANDLE  RootDirectory = NULL;
	PSECURITY_DESCRIPTOR  SecurityDescriptor = NULL;

	//writefile
	PIO_APC_ROUTINE  ApcRoutine = NULL;
	PVOID  ApcContext = NULL;
	LARGE_INTEGER  ByteOffset ;
	PULONG  Key  = NULL ;

	WCHAR Buffer[E2E_FILE_BUFFER_LENGTH] ; //WCHAR is 16 bit character
	//WCHAR Buffer[] = L"123465"; //WCHAR is 16 bit character
	ULONG  Length = sizeof Buffer;

	
	uint i = 0 ;
	for (i = 0 ; i < E2E_FILE_BUFFER_LENGTH ; i++) {
		Buffer[i] = 0 ;
	}
	
	if(debug) KdPrint(("size of buffer: %d\n",Length));
    UNREFERENCED_PARAMETER(dataPkts);

	RtlZeroMemory(&ByteOffset, sizeof(ByteOffset)) ;
	
	//first interval, don't drop any pkts
	if (dataPkts >= MAX_DATA_PKTS_PER_INTERVAL) {
		sprintf((char*)Buffer, "1.0"); 	
	} else {
		//uint ratio = dataPkts*100/MAX_DATA_PKTS_PER_INTERVAL; 
		uint ratio = dataPkts*1000/MAX_DATA_PKTS_PER_INTERVAL; 
		if (ratio == 0) ratio = 1; 
		sprintf((char*)Buffer, "0.%.3d", ratio); 
	}


	//setup ObjectName
	//http://msdn2.microsoft.com/en-us/library/ms795642.aspx
	//
	RtlInitUnicodeString(&ObjectName, L"\\??\\C:\\e2e-rl.txt");

	//
	// set up the object attributes
	//
	InitializeObjectAttributes(&ObjectAttributes, &ObjectName, Attributes, RootDirectory, SecurityDescriptor);

	//
	// get a file handle
	//
	fStatus = ZwCreateFile(&FileHandle, DesiredAccess, &ObjectAttributes, &IoStatusBlock, 
						AllocationSize, FileAttributes, ShareAccess, CreateDisposition,
						CreateOptions, EaBuffer, EaLength);

	if (!NT_SUCCESS(fStatus)) {
		//check ntstatus.h for error codes
		if(debug) KdPrint(("--ERROR-- In opening file fStatus: %x and IoStatus: %x\n", (fStatus), IoStatusBlock));
		goto FileDone;
	}

	//
	// write to the file
	//
	fStatus = ZwWriteFile(FileHandle, NULL, ApcRoutine, ApcContext, &IoStatusBlock, 
						Buffer, Length, &ByteOffset, Key);

	if (!NT_SUCCESS(fStatus)) {
		if(debug) KdPrint(("--ERROR-- In writing file fStatus: %x and IoStatus: %x\n", (fStatus), IoStatusBlock));
	}

	//
	// close the file handle
	//
	ZwClose(FileHandle);
FileDone:
	return;
}
#endif
