// *- mode: C++; tab-width: 4; indent-tabs-mode: nil -*- (for GNU Emacs)
//
// Copyright (c) Microsoft Corporation
//
// This file is part of the Microsoft Research Mesh Connectivity Layer.
// You should have received a copy of the Microsoft Research Shared Source
// license agreement (MSR-SSLA) for this software; see the file "license.txt".
// If not, please see http://research.microsoft.com/mesh/license.htm,
// or write to Microsoft Research, One Microsoft Way, Redmond, WA 98052-6399.
//
// File created at University of Texas-Austin research.
//


#include "headers.h"

#define PRINT_ADDRESS(f, s, d, t) KdPrint(("%s: from %02x-%02x-%02x-%02x-%02x-%02x\
	to %02x-%02x-%02x-%02x-%02x-%02x seqNum %i\n", f, (s)[0], (s)[1], (s)[2], (s)[3], (s)[4], (s)[5], \
	(d)[0], (d)[1], (d)[2], (d)[3], (d)[4], (d)[5], t)) 

#define PRINT_VA(f, s) KdPrint(("%s: from %02x-%02x-%02x-%02x-%02x-%02x\n", \
	f, (s)[0], (s)[1], (s)[2], (s)[3], (s)[4], (s)[5])) 


//
// Definitions for SRPacketHolder.Type
//
#define PRIORITYFWD		2
#define RETRANSMISSION	3


void LocalRecoveryInitialize(MiniportAdapter *VA) {
	VA->myCache.numPackets = 0 ;
	VA->myCache.firstHolder = NULL; 
	VA->myCache.lastHolder = NULL; 
	
	VA->timerCache.firstHolder = NULL ; 
	VA->timerCache.lastHolder = NULL; 
	VA->timerCache.numEvents = 0 ; 
}



//We have to delete all of the packets that we are holding onto
//when it is time to disable the miniport adapter, or else
//we will have memory leaks.  Also have to free the SRPacketHolder
//structures.  Must be called w/ lock held.
void LocalRecoveryHalt(MiniportAdapter *VA) {
	SRPacketHolder *curHolder; 
	
	curHolder = VA->myCache.firstHolder ;
	while (curHolder != NULL) {
		SRPacketHolder *tmp = curHolder->next; 
		deletePacket(curHolder, VA); 
		curHolder = tmp; 
	}
	//cleanup
	VA->myCache.firstHolder = NULL ;
	VA->myCache.lastHolder = NULL ; 
	VA->myCache.numPackets = 0 ; 
}




//Goes through packets cached to see if any need to be resent.
//Need to ensure we do this in chronological order to avoid
//reordering.  Must be called w/ lock held.
Time LocalRecoveryCheckRecovery(MiniportAdapter *VA) {
	TimerHolder		*tHolder; 
	//TimerHolder		*ttmp; 
	SRPacketHolder *curHolder;
	//NDIS_STATUS Status ;
	Time Now; 
	Time retTimer = MAXTIME; 
	Time helperTimer; 
		
	tHolder = VA->timerCache.firstHolder; 
	Now = KeQueryInterruptTime();
	while (tHolder != NULL && tHolder->timeout <= Now) {
		curHolder = tHolder->myPacketHolder; 
		ASSERT(curHolder != NULL) ;

		//deal with other types differently
		if (curHolder->Type == PRIORITYFWD || curHolder->Type == RETRANSMISSION) {
			TimerHolder *tmp = tHolder->next; 		
			//we'll have the correct method deal with this holder.
			helperTimer = (tHolder->TimeoutCaller) (VA, tHolder); 	
			if (helperTimer < retTimer)
					retTimer = helperTimer; 

			tHolder = tmp; 
			Now = KeQueryInterruptTime();
			continue; 
		}

		//should never be here anymore--clean this up later...
		ASSERT(0); 

	}
	if (tHolder != NULL && tHolder->timeout < retTimer)
		return tHolder->timeout ;
	return retTimer; 
}


//
// Calculates the size of the forwarding list.
//
// Note: This WILL NOT work for AddForwarding(), but should work for
//	retransmissions and caching.
uint forwardListSize(SourceRoute *SR) {
	int ret = 0;
	uint i ; 
	for (i = 0 ; i < SOURCE_ROUTE_HOPS(SR->optDataLen) ; i++) {
		if (checkSRPbitmap(SR->fwd_bitmap, i))
			ret++; 
	}
	ASSERT(ret > 0); 
	return (uint) ret; 

}


//must be called w/ the lock held, adds a timer to our
//linked list of timing events.  We want to add our timers such that the link list
//is sorted.  The most recent timeouts will be at the beginning of the list.
void addTimer (TimerHolder *newth, MiniportAdapter *VA) {
	TimerHolder *curHolder ; //we'll iterate through the list to see were we put our new timer

	curHolder = VA->timerCache.lastHolder ;

	//empty list
	if (curHolder == NULL) {
		ASSERT(VA->timerCache.numEvents == 0); 
		VA->timerCache.firstHolder = newth ;
		VA->timerCache.lastHolder = newth; 
		VA->timerCache.numEvents++;
		return; 
	}


	while (curHolder != NULL && newth->timeout <= curHolder->timeout) {
		curHolder = curHolder->prev;
	}

	//if curHolder is NULL, we put at beginning of list, otherwise, we add newth
	//behind the curHolder.
	if (curHolder == NULL) {
		TimerHolder *head ;
		head = VA->timerCache.firstHolder ;
		head->prev = newth; 
		newth->next = head ;
		VA->timerCache.firstHolder = newth; 
		VA->timerCache.numEvents++; 
		return ;
	}

	newth->prev = curHolder; 
	newth->next = curHolder->next ;
	if (curHolder->next != NULL) {
		curHolder->next->prev = newth; 
	} else {
		VA->timerCache.lastHolder = newth; 
	}
	curHolder->next = newth; 
	VA->timerCache.numEvents++; 
}


//must be called w/ the lock held, adds a recovery packet
//to our linked list. 
void addPacket(SRPacketHolder *newph, MiniportAdapter *VA) {
	SRPacketHolder *tail  ; //the last Holder in our LL

	//we'll add to end of list
	tail = VA->myCache.lastHolder ;

	//check to see if LL is empty, act accordingly
	if (tail == NULL) {
		ASSERT(VA->myCache.numPackets == 0); 
		VA->myCache.firstHolder = newph ;
		VA->myCache.lastHolder = newph; 
		VA->myCache.numPackets++;
		return; 
	}

	//we have at least one entry in LL, add our new entry to tail
	tail->next = newph; 
	newph->prev = tail ; 
	VA->myCache.lastHolder = newph; 
	VA->myCache.numPackets++; 	
}


//deletes a packet from our LL or SRPacketHolders, needs to be called
//with the lock held.  If curHolder->Packet needs to be deleted, then the
//caller of this function should do so (and then also set curHolder->Packet
//to NULL).
void deletePacket (SRPacketHolder *curHolder, MiniportAdapter *VA) {
	
	ASSERT(curHolder->srp != NULL) ; 
	ASSERT(curHolder->srp->FreePacket != NULL  || curHolder->srp->FreePacket2 != NULL); 
	
	
	//first delete the timer if there is one
	if (curHolder->timerHolder != NULL) {
		deleteTimer(curHolder->timerHolder, VA); 
		curHolder->timerHolder = NULL; 
	}
	
	//if(debug) KdPrint(("Inside deletePacket\n")); 

	if (curHolder->Type == RETRANSMISSION && !curHolder->srp->sendCompleteDone) {
		//need to delete the holder from the LL (and update LL)
		if (curHolder->next == NULL)
			VA->myCache.lastHolder = curHolder->prev ;
		else 
			curHolder->next->prev = curHolder->prev ;
		if (curHolder->prev == NULL)
			VA->myCache.firstHolder = curHolder->next; 
		else
			curHolder->prev->next = curHolder->next ; 

		curHolder->srp->sendCompleteShouldDelete = 1 ;
		
		//decrement number of packets in the cache
		VA->myCache.numPackets--;
		//if(debug) KdPrint(("deletePacket called, but need to have sendComplete finish\n")); 
		return; 
	}

	
	if (OLD_NDIS_M_COMPLETE) {

	//send complete has already finished, so we need to free resources.
	if (curHolder->Type == RETRANSMISSION && curHolder->srp->TransmitComplete == MiniportRetransmitSendComplete) {
		//this was supposed to be done in MiniportSendComplete...
		NdisMSendComplete(VA->MiniportHandle,
			              PC(curHolder->srp->Packet)->OrigPacket,
				          curHolder->srp->ForwardStatus);
	}
	
	
	}

	//free the srp from the holder
	SRPacketFree(curHolder->srp);
	
	curHolder->srp = NULL; 
	//this prevents errors in ProtocolRecoverTransmitComplete
	if (curHolder->Packet != NULL && PC(curHolder->Packet) != NULL) {
		PC(curHolder->Packet)->srp = NULL ;
	}

	/* Recovery:
	*  If curHolder->Packet isn't null, that means we've sent the packet, but the 
	*	ProtocolRecoverTransmitComplete method hasn't been called yet. We don't want
	*	to delete the packet here, because we are not allowed to so until the driver
	*	calls the ProtocolRecoverTransmitComplete method.  Thus, we need to ensure
	*	that the curHolder->Packet pointer and its curHolder structure are not removed
	*	until ProtocolRecoverTransmitComplete method is called.
	*	-Is it safe to remove the curHolder structure? NdisSend was given a pointer
	*	 to the packet.  I will just delete the curHolder structure for now.
	*/
	
	//need to delete the holder from the LL (and update LL)
	if (curHolder->next == NULL)
		VA->myCache.lastHolder = curHolder->prev ;
	else 
		curHolder->next->prev = curHolder->prev ;
	if (curHolder->prev == NULL)
		VA->myCache.firstHolder = curHolder->next; 
	else
		curHolder->prev->next = curHolder->next ; 

	//need to free the holder
	ExFreePool(curHolder); 

	//decrement number of packets in the cache
	VA->myCache.numPackets--;
	
	//defensive- should be taken care of above.
	if (VA->myCache.numPackets == 0) {
		ASSERT(VA->myCache.firstHolder == NULL) ;
		ASSERT(VA->myCache.lastHolder == NULL) ;
	}

}

//deletes the timer from the timing LL... needs to be called 
//with the lock held.
void deleteTimer (TimerHolder *curHolder, MiniportAdapter *VA) {
	if (curHolder->myPacketHolder != NULL) {
		curHolder->myPacketHolder->timerHolder = NULL ;
	}

	//need to delete the holder from the LL (and update LL)
	if (curHolder->next == NULL)
		VA->timerCache.lastHolder = curHolder->prev ;
	else 
		curHolder->next->prev = curHolder->prev ;
	if (curHolder->prev == NULL)
		VA->timerCache.firstHolder = curHolder->next; 
	else
		curHolder->prev->next = curHolder->next ; 

	//need to free the holder
	ExFreePool(curHolder); 

	//decrement number of packets in the cache
	VA->timerCache.numEvents--;
	
	//defensive- should be taken care of above.
	if (VA->timerCache.numEvents == 0) {
		ASSERT(VA->timerCache.firstHolder == NULL) ;
		ASSERT(VA->timerCache.lastHolder == NULL) ;
	}

}


//updates the curHolder's timeout value to a new timeout value
//must be called with lock held.  Has to keep the LL sorted,
//so it needs to reorder.
void updateTimer(TimerHolder *curHolder, Time NewTime, MiniportAdapter *VA) {
	//start debug
	Time Now ; 
	Now = KeQueryInterruptTime();
	ASSERT(NewTime >= Now);
	//ASSERT(NewTime >= curHolder->timeout); //this may not hold under new timers!!!!!!!!!!!
    //end debug

	curHolder->timeout = NewTime; 

	//need to delete the holder from the LL
	if (curHolder->next == NULL)
		VA->timerCache.lastHolder = curHolder->prev ;
	else 
		curHolder->next->prev = curHolder->prev ;
	if (curHolder->prev == NULL)
		VA->timerCache.firstHolder = curHolder->next; 
	else
		curHolder->prev->next = curHolder->next ;

	//addTimer updates the number of timers.
	VA->timerCache.numEvents--; 

	curHolder->next = NULL;
	curHolder->prev = NULL; 
	//list have been updated, readd the entry
	addTimer(curHolder,VA);
}



// Checks to see if the node "first" is higher priority than
// me in the *forwarding* list.
//
// Have to be careful because this can be called on the new sr or
// the old sr. If it is the new sr, then the following conditions hold:
//   -Since it is a new sr, I am the one who sent the packet. All we do
//    is see if the other node is on the fwd list. After that, we cannot 
//    make any guarenteees.
//
// Otherwise, if old sr, I did not send the packet. Then we look at all
// forwarders in that packet and check their ordering.
//
// ASSUMPTIONS: this method is only  called from preIFQCanDelete(). Other methods
// must check for compatibility before using this function.
//
// returns true if so, false otherwise
//
boolint HigherForwardPriority(MiniportAdapter *VA, SourceRoute *sr, VirtualAddress *first) {
	int firstIndex = -1 ;
	int myIndex = -1 ;
	uint i ; 
	uint Index = SOURCE_ROUTE_HOPS(sr->optDataLen) - sr->segmentsLeft ;

	ASSERT(sr); 
	ASSERT(Index > 0); 

	if (VirtualAddressEqual(sr->hopList[Index - 1].addr, VA->Address)) {
		//this is the new sr because I was the sender

		//all we need to do is see if 'first' is on the forward list
		for (i = 0 ; i < SOURCE_ROUTE_HOPS(sr->optDataLen) ; i++) {
			if (VirtualAddressEqual(*first, sr->hopList[i].addr) ) {
				if (checkSRPbitmap(sr->fwd_bitmap, i))
					return TRUE ;
				else
					return FALSE ; 
			}
		}
		return FALSE ;
	} else {
		//this is the old sr because I am not the sender

		//so then we need to find our index on the fwding list. Then we 
		// need to find the index of 'first'. If 'first' as a higher index, then
		// it is higher priority.
		for (i = 0 ; i < SOURCE_ROUTE_HOPS(sr->optDataLen) ; i++) {
			if (VirtualAddressEqual(*first, sr->hopList[i].addr)) {
				if (checkSRPbitmap(sr->fwd_bitmap, i))
					firstIndex = i ;
				else
					return FALSE ;
			}
			if (VirtualAddressEqual(VA->Address, sr->hopList[i].addr)) {
				if (checkSRPbitmap(sr->fwd_bitmap, i))
					myIndex = i ;
				else
					return FALSE ;
			}
		}

		if (firstIndex == -1 || myIndex == -1)
			return FALSE; 

		//on source route, the higher the index, the higher the priority
		if (firstIndex > myIndex)
			return TRUE; 

	}
	return FALSE; 
}

//
// Returns true if:
// ETX(Fwder, Dest) < 1.0 + ETX(Me, Dest)
// --or special case where Fwder == Dest
//
// Called by LocalRecoveryProcessAck
//
// Fwder is the node that we recieved the ACK from
// Dest is the destination of the flow
boolint HigherETXPriority(MiniportAdapter *VA, VirtualAddress *Fwder, VirtualAddress *Dest) {
	uint myETX  = (uint) -1;
	uint fwdETX = (uint) -1; 
	uint myret ;
	uint fwdret; 
	Time Now = KeQueryInterruptTime();

	//special case
	if (VirtualAddressEqual(*Fwder, *Dest)) {
		if(debug) KdPrint(("HigherETX, dest was forwarder\n"));
		return TRUE; 
	}

	myret = arb_dijkstra(VA, VA->Address, *Dest, &myETX, Now);
	fwdret = arb_dijkstra(VA, *Fwder, *Dest, &fwdETX, Now);

	if (!myret || !fwdret || myETX == -1 || fwdETX == -1) {
		if(debug) KdPrint(("HigherETX, arb dijkstra failed\n"));
		return FALSE; 
	}


	if(debug) KdPrint(("HigherETX, fwdETX: %u and myETX: %u\n", fwdETX, myETX));
	ASSERT(4096 + fwdETX > fwdETX); //overflow
	if (4096 + fwdETX < myETX) {
		return TRUE; 
	} else {	
		return FALSE; 
	}
}




//We have to check if packet is in our cache.
//If so, we remove packet from cache and cancel any timers we have.
//ReceiveSRPacket will free the incoming Ack, so we don't have to worry about that.
//But we still need to free the holder data structure.  Needs to be called with
//lock held.
//
//Update: ACKs are now cumulative with bitmap, need to have additional checks.
//
// returns- true if we updated our ack cache from this ACK, false otherwise
//	*higherETX (returned value) -- if myETX >> fwdETX
//
//
boolint LocalRecoveryProcessAck(MiniportAdapter *VA, InternalAcknowledgement *Ack, boolint *higherETX) {
	SRPacketHolder *curHolder ;
	boolint updateAckState = FALSE ; 
	AckCache *AC; 
	SRPacketHolder *tmpnext; 
	ASSERT(Ack != NULL); 
	AC = MaintBufFindAckFlow(VA, &Ack->opt.flowSource, &Ack->opt.flowDest, Ack->opt.inif, Ack->opt.outif, NULL); 
	ASSERT(AC); 
	
	*higherETX = HigherETXPriority(VA, &(Ack->opt.from), &(Ack->opt.flowDest));
	if (*higherETX) {
		//update the ACK state b/c higher priority node send an ACK
		AckCacheUpdateAckState(AC, Ack);
		updateAckState = TRUE; 
	}
	
	//we assume an ACK was for a recently heard packet, so we start from
	//the tail of the LL and progress backwards.
	curHolder = VA->myCache.firstHolder; 
	while (curHolder != NULL) {
		tmpnext = curHolder->next; 
	
		ASSERT(curHolder->srp != NULL); 
		ASSERT(curHolder->srp->sr != NULL);
		ASSERT(curHolder->srp->sr->opt.optDataLen != 0); //weird crash below(?)
		ASSERT(curHolder->srp->sr->opt.segmentsLeft >= 0); 

		if (VirtualAddressEqual(Ack->opt.flowDest, curHolder->srp->Dest) &&
			VirtualAddressEqual(Ack->opt.flowSource, curHolder->srp->Source) &&
			Ack != NULL && curHolder->srp->ackreq != NULL)
		{
			KeAcquireSpinLockAtDpcLevel(&VA->MaintBuf->Lock);
			if (preIFQCanDelete(VA, Ack, curHolder->srp, *higherETX)) { //don't let the name fool you...
				if(debug) KdPrint(("ZZ Process ACK deleting %d\n", curHolder->srp->ackreq->opt.identification)); 
				if (curHolder->Type == RETRANSMISSION && curHolder->count == 1) {
					//this is a candidate for srtt calculation
					Time now = KeQueryInterruptTime();
					Time rtt = now - curHolder->sendTime ;
					if(debug) KdPrint(("rtt value is: %u\n", rtt)); 
					if (AC->firstRTT) {
						AC->srtt = rtt ;
						AC->rttvar = rtt / 2 ;
						AC->firstRTT = FALSE; 
					} else {
						Time atmp ; 
						if (AC->srtt > rtt)
							atmp = AC->srtt - rtt; 
						else
							atmp = rtt - AC->srtt; 
						AC->rttvar = ((1000 - RTT_BETA)*AC->rttvar + RTT_BETA*atmp)/1000 ;
						AC->srtt = ((1000 - RTT_ALPHA)*AC->srtt + RTT_ALPHA*rtt)/1000 ;
					}
					if(debug) KdPrint(("rttvar value is: %u\n", AC->rttvar));
					if(debug) KdPrint(("srtt value is: %u\n", AC->srtt)); 
				}
				deletePacket(curHolder, VA);

#if SLIDING_RATE_LIMIT_ENABLED	
				//rate-limit
				//update linked list for times we received an ack
				if (VirtualAddressEqual(AC->source, VA->Address)) {
					updateTimeStruct(&(AC->AckTimes), &(AC->lastAckTimes), &(AC->numAckTimes));
				}
#endif

				//only need to update ack state once based upon this ACK
				if (!updateAckState) {
					AckCacheUpdateAckState(AC, Ack);
					updateAckState = TRUE;
				}
			}
			KeReleaseSpinLockFromDpcLevel(&VA->MaintBuf->Lock);
		}

		
		curHolder = tmpnext; 
	}


#if AIMD_RATE_LIMIT_ENABLED
	//only for flows i am the original sender on 
	//look at this regardless if i update ACK state or not...
	if (VirtualAddressEqual(AC->source, VA->Address)) {
		Time Now = KeQueryInterruptTime();
		
		if (Now >= AC->intervalDeadline) {
			updateAIMDdeadline(AC);
		} 
	}

#endif


#if E2E_RATE_LIMIT_ENABLED
	//only for flows i am the original sender on (and only if received E2E ack).
	//look at this regardless if i update ACK state or not...
	if (Ack->opt.isE2E && VirtualAddressEqual(AC->source, VA->Address)) {
		Time Now = KeQueryInterruptTime();					
		//check interval deadline
		if (Now >= AC->intervalDeadline) {
			updateE2Edeadline(AC);
		} 
		//update most current E2E ACK state
		//AC->curAckSeqNum = Ack->opt.identification; 
		//AC->curAckBitMap = Ack->opt.bitmap;
		if (Ack->opt.isE2E > AC->curNumAck) { //have to worry about reordering
			AC->curNumAck = Ack->opt.isE2E; 
		}
		AC->curNumE2ER++;
	}

#endif

	return updateAckState; 
}



// LocalRecoveryAddRetransmission
//
// Adds the given srp to the LL.  MBN is that of the destination, it is unused for now.
//
// Needs to be called w/ the lock held.  Have to ensure that srp isn't deleted by the 
// callers of this method!
boolint LocalRecoveryAddRetransmission(MiniportAdapter *VA, SRPacket *srp, MaintBufNode *MBN) {
	SRPacketHolder *newph ;
	TimerHolder	   *newth ; 
	SRPacketHolder *tail  ; //the last Holder in our LL
	Time Now ; 
		
	//VirtualAddress *nextHop ;
	Time rexmitTimer; 
	//uint myETX;
	//uint Index; 
	
	AckCache *AC; 

	//assume only 1 interface for now!
	AC = MaintBufFindAckFlow(VA, &(srp->Source), &(srp->Dest), srp->sr->opt.hopList[1].inif, 
							srp->sr->opt.hopList[1].outif, NULL); 

	UNREFERENCED_PARAMETER(MBN); 

	ASSERT(srp->PacketLength != srp->PayloadOffset && srp->ackreq != NULL); 

	ASSERT(srp->sr != NULL);
	ASSERT(!srp->sr->opt.staticRoute) ;
#if E2E_RATE_LIMIT_ENABLED	
	ASSERT(!(srp->ack && srp->ack->opt.isE2E));
#endif

	if(debug) KdPrint(("ZZ Adding packet %d to retransmission.\n", srp->ackreq->opt.identification)); 

	//we'll add to end of list
	tail = VA->myCache.lastHolder ;
  
	newph = ExAllocatePool(NonPagedPool, sizeof *newph);
	if (newph == NULL) {
		if(debug) KdPrint(("newph null in LocalRecoveryAddRetransmission\n"));
		//SRPacketFree(srp);	
        return (boolint) FALSE;
	}
    RtlZeroMemory(newph, sizeof *newph);
	newph->Packet = NULL; 
	newph->count = 1 ; 
	newph->Type = RETRANSMISSION;

	//
	// Preprocessing on SRPacket structure.  We don't want to send any probe info that may have been
	// piggybacked to this packet.  It will cause our link metrics to be inaccurate.
	//
	ASSERT(srp->Probe == NULL) ;
	ASSERT(srp->ProbeReply == NULL) ;

	//set the pointer to our packet
	srp->RecoveryHolder = newph;
	newph->srp = srp ;

	//defensive
	srp->sendCompleteDone = 0 ;
	srp->sendCompleteShouldDelete = 0 ;

	//
	// Set up the timer structure
	//
	newth = ExAllocatePool(NonPagedPool, sizeof *newth);
	if (newth == NULL) {
		if(debug) KdPrint(("newth null in LocalRecoveryAddRetransmission\n"));
		//SRPacketFree(srp); 
		ExFreePool(newph); 
		srp->RecoveryHolder = NULL; 
        return (boolint) FALSE;
	}
    RtlZeroMemory(newth, sizeof *newth);
	//we want to set the retransmission timer for the new packet
	//this will be a function of ETX to next hop
	newth->myPacketHolder = newph; 
	newth->TimeoutCaller = LocalRecoveryCheckRetransmission ; 
	
	
	//need to set the rexmitTimer using the new RTO calculations
	// If we haven't had any RTT measurements, then we use the old
	// timeout value.
	if (AC->firstRTT && AC->rttvar == 0 && AC->srtt == 0) {
		//default value
		rexmitTimer = (Time) (INITIAL_RTO * MILLISECOND); 
		//if(debug) KdPrint(("default rto used\n")); 
	} else {
		Time max = RTO_K*AC->rttvar; 
		//if(debug) KdPrint(("g: %u\n", RTO_G)); 
		//if(debug) KdPrint(("max: %u\n", max)); 
		if (RTO_G > max)
			max = RTO_G; 
		//if(debug) KdPrint(("max: %u\n", max));
		rexmitTimer = (Time) ((3*AC->srtt)/2 + max);  
	}


	if(debug) KdPrint(("ReTransmit rxmit timer in AddRetransmission: %u\n", rexmitTimer)); 
	newph->timerHolder = newth; 
	Now = KeQueryInterruptTime() ;
	newph->sendTime = Now; //needed for srtt
	newth->timeout = Now + rexmitTimer; 
	MiniportRescheduleRecoveryTimeout(VA, Now, newth->timeout); 
				
	addTimer(newth, VA); 

	addPacket(newph, VA); 

	return (boolint) TRUE ;
	
}


//Will handle the Retransmission for the SRPacketHolder pointed to by
// tHolder.
//
// Needs to be called w/ lock held. Returns the next timer the retransmission is
// due.
Time LocalRecoveryCheckRetransmission (MiniportAdapter *VA, TimerHolder *tHolder) {
	SRPacketHolder *curHolder ;
	Time Now; 
	Time retTime = MAXTIME; 
	NDIS_STATUS Status ;
	
	NDIS_BUFFER *Buffer;
    void *Data;
    uint BufferLength;
    EtherHeader UNALIGNED *Ether;	
	uint save_count = 0; 

	//uint myETX  ; 
	//Time rexmitTimer ;
	//VirtualAddress *nextHop ;
	uint Index; 
	MaintBufNode *MBN;
    MaintBuf *MB = VA->MaintBuf;
	//uint numRec = 0 ;
	//InternalAcknowledgement *tack ; //we'll use to get new piggyback ack state

#if E2E_DEBUG
	InternalAcknowledgement *curAck; 
#endif

	curHolder = tHolder->myPacketHolder ;
	ASSERT(curHolder->timerHolder == tHolder); 

	//
	// Need to:
	// 1. Transmit packet
	// 2. Update timer for the next retransmission
	// 3. Update any counter variables for number of rounds
	//

	
	//send the packet, we need to reconstruct the packet (along the lines of SRPacketToPkt)
	//This probably shouldn't happen, just a safeguard for now.
	ASSERT(curHolder->srp != NULL); 
	ASSERT(curHolder->srp->Packet != NULL) ;

	//we dont' want to have a situation where sendCompleteDone = 1 and sendCompleteShouldDelete = 1
	ASSERT(!(curHolder->srp->sendCompleteDone && curHolder->srp->sendCompleteShouldDelete)); 


	if (!curHolder->srp->sendCompleteDone && curHolder->count != 0) {
		//This isn't necessarily the case b/c local recovery or overhearing
		//packets can increase count.

		//ASSERT(curHolder->Packet != NULL); 
	} 
	if (curHolder->srp->sendCompleteDone) {
		ASSERT(curHolder->Packet == NULL); 
	}

	//We always want this to be null to avoid memory leaks.  The only time this should be NULL
	//is if ProtocolRecoverTransmitComplete hasn't been called since we last sent (probably
	// shouldn't hapen)-- actually did occur... hmmm.  I am just going to delete this entry
	//and just hope that the packet completion occurs.
	//
	//We can have curHolder->srp->sendCompleteDone == 1 and curHolder->Packet == NULL when
	//the original packet hasn't finished sending yet.
	if (curHolder->Packet != NULL || !curHolder->srp->sendCompleteDone) {
		if(debug) KdPrint(("MLR: rexmit curHolder->Packet wasn't null, deleting entry %d\n", 
					curHolder->srp->ackreq->opt.identification)); 
		ASSERT(!curHolder->srp->sendCompleteDone);
		deletePacket(curHolder,VA); 
		return retTime ;
	}

	Now = KeQueryInterruptTime();
	//debug to test accuracy of our timers.		
	if(debug) KdPrint(("Round: %d Difference in timers: %d\n", curHolder->count, Now - curHolder->timerHolder->timeout)); 

	//
	// Let's update the fwd'ing list
	//
	if (Now - curHolder->srp->updateTime >= UPDATE_FWD_LIST_TIME_DIFF) {
		Status = updateForwardInfo(VA, curHolder->srp); 
		if (Status == NDIS_STATUS_SUCCESS) {
			uint etxThresh ;
			InternalSourceRoute *SR_;

			//get new forwarding list
			SR_ = LinkCacheFindForwarders(VA, curHolder->srp, &etxThresh);
		
			Status = LinkCacheInsertForwardingList(VA, SR_, curHolder->srp, etxThresh);
			if (Status == NDIS_STATUS_SUCCESS) {
			
				curHolder->srp->updateTime = KeQueryInterruptTime();			
				
			}

		}
	}

	
	if (! LinkCacheUseSR(VA, curHolder->srp)) {
        InterlockedIncrement((PLONG)&VA->CountXmitForwardQueueFull);
		if(debug) KdPrint(("rexmit dropping pkt b/c link cache full\n")); 
		curHolder->Packet = NULL; 
		deletePacket(curHolder, VA); 
		return retTime; 
	}

	//
	// Need to update ACK state here b/c we may not have to 
	// use the preIFQ
	//

	//update our piggy-back ACK state-- before SRPacketToPkt
	//--if i am the original sender, then we do not want this behavior
	if (curHolder->srp->ack) {
		SRFreeOptionList((InternalOption *)curHolder->srp->ack);
		curHolder->srp->ack = NULL; 
		if (curHolder->srp->TransmitComplete == MiniportForwardComplete) {
			curHolder->srp->ack = MaintBufFillInPiggyBackAck(VA, curHolder->srp);
			
#if E2E_DEBUG
			for (curAck = curHolder->srp->ack ; curAck != NULL ; curAck = curAck->next) {
				ASSERT(curAck->opt.isE2E == 0);
			}
#endif
		} else if (curHolder->srp->TransmitComplete == MiniportRetransmitSendComplete) {	
			uint Size ; 
			//assuming 1 interface for now!
			AckCache *AC = MaintBufFindAckFlow(VA, &(VA->Address), &(curHolder->srp->Dest),
				curHolder->srp->sr->opt.hopList[1].inif, curHolder->srp->sr->opt.hopList[1].inif, NULL); 

			//fill in piggy back ACKs
			Size = PbackPacketSize(curHolder->srp);
			if (Size < PROTOCOL_MIN_FRAME_SIZE && AC) {
				Size = PROTOCOL_MIN_FRAME_SIZE - Size; //amount left to use
				curHolder->srp->ack = MaintBufGetCrossFlowAcks(VA, AC, Size, 0);
			}			
#if E2E_DEBUG
			for (curAck = curHolder->srp->ack ; curAck != NULL ; curAck = curAck->next) {
				ASSERT(curAck->opt.isE2E == 0);
			}
#endif
		} else {
			//shouldn't be here.
			ASSERT(0);
		}		
	}



	//create our packet to send
	Status = SRPacketToPkt(VA, curHolder->srp, &curHolder->Packet);

	if (Status != NDIS_STATUS_SUCCESS) {
		if(debug) KdPrint(("MLR:LocalRecovery rxmit: SRPacketToPkt Failed!\n")); 
		//we'll just delete the entry for now
		curHolder->Packet = NULL; 
		deletePacket(curHolder, VA); 
		return retTime; 
	}

	if (PC(curHolder->Packet) == NULL) {
		//delete entry....
		if(debug) KdPrint(("MLR:LocalRecovery rxmit: PC(packet) is null!\n")); 
		NdisFreePacketClone(curHolder->Packet);
		curHolder->Packet = NULL; 
		deletePacket(curHolder,VA); 
		return retTime; 
	}
			
	PC(curHolder->Packet)->srp = curHolder->srp;
	PC(curHolder->Packet)->TransmitComplete = MaintBufRetransmitSendComplete ;

	if (PC(curHolder->Packet)->PA == NULL) {
		if(debug) KdPrint(("MLR:LocalRecovery rexmit: Protocol Adaptor is null!\n")); 
		//delete entry....
		NdisFreePacketClone(curHolder->Packet);
		curHolder->Packet = NULL; 
		deletePacket(curHolder, VA); 
		return retTime; 
	}



	Buffer = NdisFirstBuffer(curHolder->Packet);
	NdisQueryBuffer(Buffer, &Data, &BufferLength);
	Ether = (EtherHeader UNALIGNED *) Data;
	IEEE_802_ADDR_SET_BROADCAST(&(Ether->Dest)) ;

	//changing this so we don't muck up the neighbor cache entries....
    RtlCopyMemory(Ether->Source, PC(curHolder->Packet)->PA->Address, IEEE_802_ADDR_LENGTH);
    Ether->Type = ETYPE_MSFT;
	//RtlCopyMemory(Ether->Source, VA->Address, SR_ADDR_LEN);	


	//segmentsLeft has already been adjusted, so we are okay.
	Index = SOURCE_ROUTE_HOPS(curHolder->srp->sr->opt.optDataLen) - curHolder->srp->sr->opt.segmentsLeft; 
	ASSERT(Index > 0) ;	

	//Need to update some variables for MBN.
    KeAcquireSpinLockAtDpcLevel(&MB->Lock);
	MBN = MaintBufFindNode(VA, MB, curHolder->srp->sr->opt.hopList[Index].addr,
						   curHolder->srp->sr->opt.hopList[Index].inif,
                           curHolder->srp->sr->opt.hopList[Index-1].outif);
	if (MBN == NULL) {
		if(debug) KdPrint(("rxmit: Not resending bc can't find bufnode\n")); 
		NdisFreePacketClone(curHolder->Packet);
		curHolder->Packet = NULL; 
		deletePacket(curHolder, VA); 
		KeReleaseSpinLockFromDpcLevel(&MB->Lock);
		return retTime; 
	}

	/*
	*  Need to go thru the MaintBufTimer function and mimic it here.  Eventually,
	* we'd like to totally replace that function.  For now, we'll assume no route
	* errors in the topology.
	*
	*/
	MBN->NumAckReqs++;
    MBN->LastAckReq = Now;


	//need to release spin lock!!
    KeReleaseSpinLockFromDpcLevel(&MB->Lock);

	
	//let's just see what we would send
	PRINT_ADDRESS("MLR:Going to rexmit: ", curHolder->srp->Source, curHolder->srp->Dest, 
					curHolder->srp->ackreq->opt.identification); 
	

	
	//need to reset the variables for sendComplete, since we are retransmitting.
	curHolder->srp->sendCompleteDone = 0 ; 
	curHolder->srp->sendCompleteShouldDelete = 0 ; 

	
	//the LL is sorted by time. We can't leave the old timeout in here, so let's reset to
	// the max. The timer will be set to the correct value by either ProtocolRecoverTransmit or
	// when dequeuing from preIFQ.
	updateTimer(curHolder->timerHolder, MAXTIME, VA); 

	save_count = curHolder->count; 
	//ProtocolRecoverTransmit reassigns our transmitComplete....
	ProtocolRecoverTransmit(PC(curHolder->Packet)->PA, curHolder->Packet);	
	//curHolder->count = curHolder->count + 1 ; //let's move this to localrecoveryupdateretransmittimer

	// Let's delete the packet if the number of rounds
	// has exceeded.
	//--Want move this after we call protocolrecovertransmit
	// and update the count b/c we don't want this sitting
	// in the cache for so long.
	if (save_count >= MAX_NUM_ROUNDS) {
		if(debug) KdPrint(("MLR: Retransmit count exceeded for packet: %u\n", curHolder->srp->ackreq->opt.identification)); 
		deletePacket(curHolder, VA); 
		return retTime; 
	}


	return retTime; 

}


//
// Updates the retransmission for the packet when we actually send the packet.
//
// Have to call this method with the lock held.
// 
void LocalRecoveryUpdateRetransmissionTimer(MiniportAdapter *VA, SRPacket *srp) {
	SRPacketHolder *curHolder ;
	Time Now; 
	Time rexmitTimer ;
	//uint exponent = 0 ;
	
	AckCache *AC; 

	//for exponential backoff
	uint numerator = 1;
	uint denominator = 1; 
	uint i = 0 ;

	ASSERT(srp); 
	curHolder = srp->RecoveryHolder;
	ASSERT(curHolder); 

	//---this can happen when the packet is retransmitted the max number of times... want to exit
	//--------can also happen if packet gets an error in CheckRetransmission() and needs to be deleted
	if(!curHolder->timerHolder) {
		if(debug) KdPrint(("Not rexmitting packet %u because tHolder is null\n", srp->ackreq->opt.identification));
		return; 
	}

	//assume only 1 interface for now!
	AC = MaintBufFindAckFlow(VA, &(srp->Source), &(srp->Dest), srp->sr->opt.hopList[1].inif, 
							srp->sr->opt.hopList[1].outif, NULL); 

	ASSERT(curHolder->timerHolder); //this assert fails when we don't delete from the preIFQ.
	ASSERT(curHolder->srp != NULL); 
	ASSERT(curHolder->srp->sr != NULL);
	ASSERT(curHolder->srp->ackreq != NULL); 
	ASSERT(srp->PacketLength != srp->PayloadOffset && srp->ackreq != NULL); 
	ASSERT(srp->sr != NULL);
	ASSERT(!srp->sr->opt.staticRoute) ;

	ASSERT(curHolder->srp == srp); 


	//need to set the rexmitTimer using the new RTO calculations
	// If we haven't had any RTT measurements, then we use the old
	// timeout value.
	if (AC->firstRTT && AC->rttvar == 0 && AC->srtt == 0) {
		//default value
		rexmitTimer = (Time) (INITIAL_RTO * MILLISECOND); 
		//if(debug) KdPrint(("default rto used\n")); 
	} else {
		Time max = RTO_K*AC->rttvar; 
		//if(debug) KdPrint(("g: %u\n", RTO_G)); 
		//if(debug) KdPrint(("max: %u\n", max)); 
		if (RTO_G > max)
			max = RTO_G; 
		//if(debug) KdPrint(("max: %u\n", max));
		rexmitTimer = (Time) ((3*AC->srtt)/2 + max);  
	}
	
	for (i = 0 ; i < curHolder->count; i++) {
		numerator *= RESEND_EXP_FACTOR_NUM;
		denominator *= RESEND_EXP_FACTOR_DEM; 	
	}
	rexmitTimer = (rexmitTimer * numerator) / denominator ;

	if(debug) KdPrint(("ReTransmit rxmit timer in LocalRecoveryUpdateRetransmissionTimer: %u\n", rexmitTimer)); 
	if(debug) KdPrint(("ejr rxmit seqnum: %u and count: %u\n", srp->ackreq->opt.identification, curHolder->count));
	
	//update the number of times this packet has been transmitted.
	curHolder->count = curHolder->count + 1 ; 
	

	Now = KeQueryInterruptTime() ;
	updateTimer(curHolder->timerHolder, Now + rexmitTimer, VA); 

	//we have the new timer in the curHolder->timeHolder structure. Need to 
	// make sure that the timeout is called if necessary.
	MiniportRescheduleRecoveryTimeout(VA, Now, curHolder->timerHolder->timeout); 
}


//
// Adds a packet to be forwarded to our linked list.  
//  If the addition is unsuccessful, then we need to delete srp.
//	Have to call this method with the lock held.
//
// my_slot gives my priority in the fwding list (0 is highest priority).
//
void LocalRecoveryAddForwarding(MiniportAdapter *VA, SRPacket *srp, int my_slot) {
	SRPacketHolder *newph ;
	TimerHolder	   *newth ; 
	SRPacketHolder *tail  ; //the last Holder in our LL
	Time Now ; 
	Time rexmitTimer; 


	ASSERT(srp->PacketLength != srp->PayloadOffset && srp->ackreq != NULL); 
	ASSERT(srp->sr != NULL);
	ASSERT(!srp->sr->opt.staticRoute) ;
#if E2E_RATE_LIMIT_ENABLED	
	ASSERT(!(srp->ack && srp->ack->opt.isE2E));
#endif

	if(debug) KdPrint(("ZZ Adding packet %d to forwarding timer.\n", srp->ackreq->opt.identification)); 

	//we'll add to end of list
	tail = VA->myCache.lastHolder ;
 
	newph = ExAllocatePool(NonPagedPool, sizeof *newph);
	if (newph == NULL) {
		SRPacketFree(srp);	
        return ;
	}
    RtlZeroMemory(newph, sizeof *newph);
	newph->Packet = NULL; 
	newph->count = 0 ; 
	newph->Type = PRIORITYFWD;

	//set the pointer to our packet
	srp->RecoveryHolder = newph;
	newph->srp = srp ;

	//defensive
	srp->sendCompleteDone = 0 ;
	srp->sendCompleteShouldDelete = 0 ;

	//
	// Set up the timer structure
	//
	newth = ExAllocatePool(NonPagedPool, sizeof *newth);
	if (newth == NULL) {
		SRPacketFree(srp); 
		ExFreePool(newph); 
        return ;
	}

    RtlZeroMemory(newth, sizeof *newth);
	newth->myPacketHolder = newph; 
	newth->TimeoutCaller = LocalRecoveryCheckForwarding ; 


	ASSERT(my_slot > 0);
	rexmitTimer = (Time) ((my_slot)*DATA_TX_TIME); 

	if(debug) KdPrint(("Forwarding timer in AddForwarding: %u\n", rexmitTimer)); 
	newph->timerHolder = newth; 
	Now = KeQueryInterruptTime() ;
	newth->timeout = Now + rexmitTimer; 
	MiniportRescheduleRecoveryTimeout(VA, Now, newth->timeout); 
				
	addTimer(newth, VA); 

	addPacket(newph, VA); 

	return  ;
}

//
// Called when it is time to forward a packet.  Needs to be called with the lock 
// held.  We have to ensure that calling maintbufsendpacket does not cause a 
// problem here. (ie- we will call addretransmission, which may cause fits).
//
// Returns the Time for when the packet will have to be retransmitted (not).
//
Time LocalRecoveryCheckForwarding(MiniportAdapter *VA, TimerHolder *tHolder) {
	SRPacketHolder *curHolder = tHolder->myPacketHolder; 
	SRPacket *savesrp = curHolder->srp;
	NDIS_STATUS Status ;
	
#if E2E_DEBUG
	InternalAcknowledgement *curAck; 
#endif
	curHolder->srp->RecoveryHolder = NULL; 

	//
	//we want to delete the timerHolder and curHolder for the forwarding entry,
	// but we DO NOT want to delete the SRPacket (srp) because the retransmissions will
	// need it.
	//

	//first delete the timer
	deleteTimer(tHolder, VA); 

	if (curHolder->next == NULL)
		VA->myCache.lastHolder = curHolder->prev ;
	else 
		curHolder->next->prev = curHolder->prev ;
	if (curHolder->prev == NULL)
		VA->myCache.firstHolder = curHolder->next; 
	else
		curHolder->prev->next = curHolder->next ; 

	//need to free the holder
	ExFreePool(curHolder); 
	
	//decrement number of packets in the cache
	VA->myCache.numPackets--;

	//
	//try to update the forwarding list if we've been waiting too long
	//
	if (KeQueryInterruptTime() - savesrp->updateTime >= UPDATE_FWD_LIST_TIME_DIFF) {
		Status = updateForwardInfo(VA, savesrp); 
		if (Status == NDIS_STATUS_SUCCESS) {
			uint etxThresh ;
			InternalSourceRoute *SR_;

			//get new forwarding list
			SR_ = LinkCacheFindForwarders(VA, savesrp, &etxThresh);
			
			Status = LinkCacheInsertForwardingList(VA, SR_, savesrp, etxThresh);
			if (Status == NDIS_STATUS_SUCCESS) {
				savesrp->updateTime = KeQueryInterruptTime();
			}
		}
	}


	//have to ensure it is okay to send the packet
	if (! LinkCacheUseSR(VA, savesrp)) {
		InterlockedIncrement((PLONG)&VA->CountXmitForwardQueueFull);
        goto FwdFreeSrpAndReturn;
	}


	//we want to update the ACK state for this packet.	
	if (savesrp->ack) {
		SRFreeOptionList((InternalOption *)savesrp->ack);
		savesrp->ack = NULL;
		savesrp->ack = MaintBufFillInPiggyBackAck(VA, savesrp);
		#if E2E_DEBUG
			for (curAck = savesrp->ack ; curAck != NULL ; curAck = curAck->next) {
				ASSERT(curAck->opt.isE2E== 0);
			}
		#endif
	}
	
	MaintBufSendPacket(VA, savesrp, MiniportForwardComplete);
	
	return MAXTIME; //maintbufsendpacket will call addretransmission, which will update our timer appropiately.
FwdFreeSrpAndReturn:
	
    SRPacketFree(savesrp);
	return MAXTIME;
}





