// -*- mode: C++; tab-width: 4; indent-tabs-mode: nil -*- (for GNU Emacs)
//
// Copyright (c) Microsoft Corporation
//
// This file is part of the Microsoft Research Mesh Connectivity Layer.
// You should have received a copy of the Microsoft Research Shared Source
// license agreement (MSR-SSLA) for this software; see the file "license.txt".
// If not, please see http://research.microsoft.com/mesh/license.htm,
// or write to Microsoft Research, One Microsoft Way, Redmond, WA 98052-6399.
//

#include "headers.h"

//
// We send unicast probes to each neighbor this frequently.
//
#define DEFAULT_ULT_UCAST_PROBE_PERIOD          (10 * SECOND)

//
// Loss probabilities less than this indicates that the link is broken.
//
#define DEFAULT_ULT_UCAST_LOW_THRESHOLD 60      // 60%

//
// Loss probabilities greater than this indicates that the link is working.
//
#define DEFAULT_ULT_UCAST_HIGH_THRESHOLD        80      // 80%

//
// How much to lower the perceived quality when not receiving a probe.
//
#define DEFAULT_ULT_UCAST_LOSS_PENALTY          10      // 10%

//
// We send broadcast probes this frequently.
//
#define DEFAULT_ULT_BCAST_PROBE_PERIOD          (1 * SECOND)

//
// How many broadcast probes must be lost before we stop unicast probes.
//
#define DEFAULT_ULT_BCAST_LOSS_THRESHOLD        4

//* UltInitLinkMetric
//
// Init metric information for a new link. 
//
// Called with LinkCache locked.
//
void 
UltInitLinkMetric(
    MiniportAdapter *VA,
    int SNode,
    Link *Link, 
    Time Now)
{
    uint i;

    UNREFERENCED_PARAMETER(VA);
    UNREFERENCED_PARAMETER(SNode);
    UNREFERENCED_PARAMETER(Now);

    //
    // Start by assuming the link is down.
    // We use (uint)-2 so that MiniportIsInfinite returns false,
    // preventing the link from being reinitialized.
    //
    Link->Metric = (uint)-2;

    Link->MetricInfo.Ult.UcastProbeSeq = 0;
    Link->MetricInfo.Ult.SentUcastProbes = 0;
    Link->MetricInfo.Ult.RcvdUcastProbes = 0;
    Link->MetricInfo.Ult.RcvdQuality = 0;
    Link->MetricInfo.Ult.SentQuality = 0;
    Link->MetricInfo.Ult.PcvdQuality = 100;
    Link->MetricInfo.Ult.Quality = 0;

    Link->MetricInfo.Ult.PHNext = 0;
    for (i = 0; i < MAX_ULT_PROBE_HISTORY; i++)
        Link->MetricInfo.Ult.PH[i] = 0;

    Link->MetricInfo.Ult.LostBcastProbes = 0;
}

//* UltIsBad
//
//  Returns TRUE if the link metric indicates
//  that the link is bad and should never be used.
//
boolint
UltIsBad(uint Metric)
{
    return Metric == (uint)-2;
}

//* UltConvMetric
//
//  Converts a link metric to a path metric.
//
uint
UltConvMetric(uint LinkMetric)
{
    //
    // Because we use (uint)-2 to indicate a down link,
    // we must convert that value to infinity.
    //
    if (LinkMetric == (uint)-2)
        return (uint)-1;
    else
        return LinkMetric;
}

//* UltSendUcastProbeComplete
//
//  Called after sending a probe. Cleans up after UltCreateProbePacket.
//
//static void
void
UltSendUcastProbeComplete(
    MiniportAdapter *VA,
    NDIS_PACKET *Packet,
    NDIS_STATUS Status)
{
    SRPacket *srp = PC(Packet)->srp;

    UNREFERENCED_PARAMETER(VA);
    UNREFERENCED_PARAMETER(Status);

    NdisFreePacketClone(Packet);
    SRPacketFree(srp);
}

//* UltSendBcastProbeComplete
//
//  Called after sending a probe. Cleans up after UltCreateProbePacket.
//
static void
UltSendBcastProbeComplete(
    MiniportAdapter *VA,
    SRPacket *srp,
    NDIS_STATUS Status)
{
    UNREFERENCED_PARAMETER(VA);
    UNREFERENCED_PARAMETER(Status);

    SRPacketFree(srp);
}

//* UltCreateProbePacket
//
//  Creates a packet to send a probe.
//
//  For unicast probes (indicated by non-null Adjacent link),
//  increments the link's probe sequence number.
//
//  Called with LinkCache locked.
//
NDIS_STATUS
UltCreateProbePacket(
    MiniportAdapter *VA,
    Link *Adjacent,
    SRPacket **pSRP)
{
    SRPacket *SRP;

    InterlockedIncrement((PLONG)&VA->CountXmitProbe);

    //
    // Initialize an SRPacket for the Probe.
    // The Probe carries no data so it does not need an IV.
    //
    SRP = ExAllocatePool(NonPagedPool, sizeof *SRP);
    if (SRP == NULL)
        return NDIS_STATUS_RESOURCES;
    RtlZeroMemory(SRP, sizeof *SRP);

    //
    // Initialize the Probe option.
    //
    SRP->Probe = ExAllocatePool(NonPagedPool, sizeof *SRP->Probe);
    if (SRP->Probe == NULL) {
        SRPacketFree(SRP);
        return NDIS_STATUS_RESOURCES;
    }
    RtlZeroMemory(SRP->Probe, sizeof *SRP->Probe);
    SRP->Probe->Opt.OptionType = LQSR_OPTION_TYPE_PROBE;

    //
    // For ULT, we do not have any special data. 
    // So probe length is equal to PROBE_LEN. 
    //
    SRP->Probe->Opt.OptDataLen = PROBE_LEN;
    SRP->Probe->Opt.ProbeType = METRIC_TYPE_ULT;
    RtlCopyMemory(SRP->Probe->Opt.From, VA->Address, sizeof(VirtualAddress));

    if (Adjacent != NULL) {
        //
        // This is a unicast probe.
        //
        SRP->Probe->Opt.Seq = Adjacent->MetricInfo.Ult.UcastProbeSeq++;
        RtlCopyMemory(SRP->Probe->Opt.To,
                      VA->LC->nodes[Adjacent->targetIndex].address,
                      sizeof(VirtualAddress));
        SRP->Probe->Opt.OutIf = Adjacent->outif;
        SRP->Probe->Opt.InIf = Adjacent->inif;
        if (Adjacent->MetricInfo.Ult.RcvdQuality < Adjacent->MetricInfo.Ult.PcvdQuality)
            SRP->Probe->Opt.Metric = Adjacent->MetricInfo.Ult.RcvdQuality;
        else
            SRP->Probe->Opt.Metric = Adjacent->MetricInfo.Ult.PcvdQuality;
    }
    else {
        const static VirtualAddress BcastDest = { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff };

        //
        // This is a broadcast probe.
        //
        SRP->Probe->Opt.Seq = 0;
        RtlCopyMemory(SRP->Probe->Opt.To, BcastDest, sizeof(VirtualAddress));
        SRP->Probe->Opt.OutIf = 0;
        SRP->Probe->Opt.InIf = 0;
        SRP->Probe->Opt.Metric = 0;
    }

    //
    // Initialize the source & destination of this packet.
    //
    RtlCopyMemory(SRP->Dest, SRP->Probe->Opt.To, sizeof(VirtualAddress));
    RtlCopyMemory(SRP->Source, VA->Address, sizeof(VirtualAddress));

    //
    // Check for other options that can be piggy-backed on this packet.
    //
    PbackSendPacket(VA, SRP);

    *pSRP = SRP;
    return NDIS_STATUS_SUCCESS;
}

//* UltUpdateQuality
//
//  Updates the quality of the link, and consequently its metric.
//
//  Called with LinkCache locked.
//
void
UltUpdateQuality(
    MiniportAdapter *VA,
    Link *Link)
{
    //
    // Take the minimum of the three qualities.
    //
    if (Link->MetricInfo.Ult.RcvdQuality < Link->MetricInfo.Ult.SentQuality) {
        if (Link->MetricInfo.Ult.RcvdQuality < Link->MetricInfo.Ult.PcvdQuality)
            Link->MetricInfo.Ult.Quality = Link->MetricInfo.Ult.RcvdQuality;
        else
            Link->MetricInfo.Ult.Quality = Link->MetricInfo.Ult.PcvdQuality;
    }
    else {
        if (Link->MetricInfo.Ult.SentQuality < Link->MetricInfo.Ult.PcvdQuality)
            Link->MetricInfo.Ult.Quality = Link->MetricInfo.Ult.SentQuality;
        else
            Link->MetricInfo.Ult.Quality = Link->MetricInfo.Ult.PcvdQuality;
    }

    //
    // Use the thresholds to see if the metric changes.
    //
    if (Link->MetricInfo.Ult.Quality >= VA->MetricParams.Ult.UcastHighThreshold) {
        //
        // We consider the link to be up.
        //
        Link->Metric = 1;
    }
    else if (Link->MetricInfo.Ult.Quality <= VA->MetricParams.Ult.UcastLowThreshold) {
        //
        // We consider the link to be down.
        //
        Link->Metric = (uint)-2;
    }
}

//* UltAddProbe
//
//  Adds a sequence number to the probe history of this link.
//  Recalculates the RcvdQuality.
//
//  Called with LinkCache locked.
//
void
UltAddProbe(Link *Link, uint Seq)
{
    Ult *Ult = &(Link->MetricInfo.Ult);
    uint MUPH = MAX_ULT_PROBE_HISTORY; 
    uint Found;
    uint Next;
    uint i;

    //
    // Add the sequence number to the history.
    //
    Next = Ult->PHNext;
    Ult->PH[Next] = Seq;
    Ult->PHNext = (Next + 1) % MUPH;

    //
    // Calculate RcvdQuality, by looking backwards in the history
    // for numbers from Seq to Seq - MUPH + 1.
    // Note that Found is at least one, because we just added Seq.
    //
    Found = 0;
    for (i = 0; i < MUPH; i++) {
        if (Ult->PH[Next] == Seq - i) {
            //
            // We found the sequence number, so step back to previous probe.
            //
            Found++;
            Next = (Next + MUPH - 1) %  MUPH;
        }
    }
    Link->MetricInfo.Ult.RcvdQuality = (Found * MAXPERCENT) / MUPH;
}

//* UltSendUcastProbes
//
//  Sends probes to adjacent nodes as needed. Called from MiniportTimeout.
//
Time
UltSendUcastProbes(
    MiniportAdapter *VA,
    Time Now)
{
    LinkCache *LC = VA->LC;
    NDIS_PACKET *ProbePackets = NULL;
    NDIS_PACKET *Packet;
    Link *Adjacent;
    KIRQL OldIrql;
    NDIS_STATUS Status;

    if (VA->MetricParams.Ult.UcastProbeTimeout <= Now) {
        //
        // Loop through outgoing links, creating a probe packet for each.
        // Store the probe packets temporarily on a list since
        // we can't send them while holding the lock on the link cache.
        //
        KeAcquireSpinLock(&LC->Lock, &OldIrql);
        for (Adjacent = LC->nodes[0].AdjOut;
             Adjacent != NULL;
             Adjacent = Adjacent->NextOut) {
            SRPacket *SRP;

            //
            // If the link is dead (according to the broadcast probes),
            // then we do not probe it.
            //
            if (Adjacent->Metric == (uint)-1)
                continue;

            //
            // Perceived quality may have changed due to last timeout -
            // time to incorporate it into quality.
            //
            UltUpdateQuality(VA, Adjacent);

            Status = UltCreateProbePacket(VA, Adjacent, &SRP);
            if (Status == NDIS_STATUS_SUCCESS) {
                //
                // Convert the SRPacket to an NDIS Packet.
                //
                Status = SRPacketToPkt(VA, SRP, &Packet);
                if (Status == NDIS_STATUS_SUCCESS) {
                    PC(Packet)->srp = SRP;
                    PC(Packet)->TransmitComplete = UltSendUcastProbeComplete;

                    //
                    // Queue the packet using the OrigPacket link.
                    //
                    PC(Packet)->OrigPacket = ProbePackets;
                    ProbePackets = Packet;
                    Adjacent->MetricInfo.Ult.SentUcastProbes++;
                }
                else {
                    SRPacketFree(SRP);
                }
            }

            //
            // Decrement the perceived quality, but do not update
            // quality until the next timeout. This gives
            // UltReceiveProbe a chance to reset perceived quality.
            //
            if (Adjacent->MetricInfo.Ult.PcvdQuality > VA->MetricParams.Ult.UcastLossPenalty)
                Adjacent->MetricInfo.Ult.PcvdQuality -= VA->MetricParams.Ult.UcastLossPenalty;
            else
                Adjacent->MetricInfo.Ult.PcvdQuality = 0;
        }
        KeReleaseSpinLock(&LC->Lock, OldIrql);

        //
        // Loop through queued packets, sending out each probe.
        //
        while ((Packet = ProbePackets) != NULL) {

            ProbePackets = PC(Packet)->OrigPacket;
            PC(Packet)->OrigPacket = NULL;

            if (PC(Packet)->PA == NULL) {
                //
                // This means the packet is trying to use a physical
                // adapter that no longer exists.
                // LinkCacheDeleteInterface has been called.
                //
                ASSERT(PC(Packet)->TransmitComplete == UltSendUcastProbeComplete);
                UltSendUcastProbeComplete(VA, Packet, NDIS_STATUS_FAILURE);
            }
            else {
                //
                // Transmit the packet.
                //
                ProtocolTransmit(PC(Packet)->PA, Packet);
            }
        }

        VA->MetricParams.Ult.UcastProbeTimeout = Now + VA->MetricParams.Ult.UcastProbePeriod;
    }

    return VA->MetricParams.Ult.UcastProbeTimeout;
}

//* UltSendBcastProbes
//
//  Sends broadcast probes as needed. Called from MiniportTimeout.
//
Time
UltSendBcastProbes(
    MiniportAdapter *VA,
    Time Now)
{
    if (VA->MetricParams.Ult.BcastProbeTimeout <= Now) {
        LinkCache *LC = VA->LC;
        Link *Adjacent;
        SRPacket *SRP;
        KIRQL OldIrql;
        NDIS_STATUS Status;

        //
        // Check the number of missed broadcast probes for each adjacent link,
        // and decide if the link is dead (as oppposed to down).
        //
        KeAcquireSpinLock(&LC->Lock, &OldIrql);
        for (Adjacent = LC->nodes[0].AdjOut;
             Adjacent != NULL;
             Adjacent = Adjacent->NextOut) {

            if (++Adjacent->MetricInfo.Ult.LostBcastProbes > VA->MetricParams.Ult.BcastLossThreshold) {
                //
                // The link is dead; we will stop unicast probes.
                //
                Adjacent->Metric = (uint)-1;
            }
        }
        KeReleaseSpinLock(&LC->Lock, OldIrql);

        //
        // Create a probe packet, and send it. It is a broadcast packet,
        // so no need to loop through links individually. It will go out
        // on all links.
        // 
        Status = UltCreateProbePacket(VA, NULL, &SRP);
        if (Status == NDIS_STATUS_SUCCESS) {
            //
            // Send the packet.
            //
            SRP->TransmitComplete = UltSendBcastProbeComplete;
            MiniportForwardProbe(VA, SRP);
        }

        //
        // Calculate next probe timeout.
        //
        VA->MetricParams.Ult.BcastProbeTimeout =
            Now + VA->MetricParams.Ult.BcastProbePeriod;
    }

    return VA->MetricParams.Ult.BcastProbeTimeout;  
}

//* UltSendProbes
//
//  Sends probes as needed. Called from MiniportTimeout.
//
Time
UltSendProbes(
    MiniportAdapter *VA,
    Time Now)
{
    Time Ucast, Bcast;

    Bcast = UltSendBcastProbes(VA, Now);
    Ucast = UltSendUcastProbes(VA, Now);

    //
    // Return whichever time is sooner.
    //
    if (Bcast < Ucast)
        return Bcast;
    else
        return Ucast;
}

//* UltReceiveUcastProbe
//
//  Processes a unicast probe.
//
void
UltReceiveUcastProbe(
    MiniportAdapter *VA,
    InternalProbe *Probe)
{
    LinkCache *LC = VA->LC; 
    Link *Adj;
    KIRQL OldIrql;

    //
    // Find the link over which the probe arrived.
    // Actually this is the reverse link -
    // the ULT metric assumes bidirectionality so it should exist.
    //
    KeAcquireSpinLock(&LC->Lock, &OldIrql);
    for (Adj = LC->nodes[0].AdjOut;
         Adj != NULL;
         Adj = Adj->NextOut) {

        if (VirtualAddressEqual(Probe->Opt.From,
                                LC->nodes[Adj->targetIndex].address) &&
            (Probe->Opt.OutIf == Adj->inif) &&
            (Probe->Opt.InIf == Adj->outif)) {

            Adj->MetricInfo.Ult.RcvdUcastProbes++;

            //
            // Save the quality sent by our neighbor.
            // NB: We must sanity-check all data received from the wire.
            //
            if (Probe->Opt.Metric <= MAXPERCENT)
                Adj->MetricInfo.Ult.SentQuality = Probe->Opt.Metric;

            //
            // Add the sequence number to our probe history,
            // recalculating RcvdQuality.
            //
            UltAddProbe(Adj, Probe->Opt.Seq);

            //
            // Reset the perceived quality of this link.
            //
            Adj->MetricInfo.Ult.PcvdQuality = MAXPERCENT;

            //
            // Update the quality of the link.
            //
            UltUpdateQuality(VA, Adj);
            break;
        }
    }
    KeReleaseSpinLock(&LC->Lock, OldIrql);
}

//* UltReceiveBcastProbe
//
//  Processes a broadcast probe.
//
void
UltReceiveBcastProbe(
    MiniportAdapter *VA,
    InternalProbe *Probe,
    LQSRIf InIf)
{
    LinkCache *LC = VA->LC; 
    Link *Adj;
    KIRQL OldIrql;

    //
    // Find the link over which the probe arrived.
    // Actually this is the reverse link -
    // the ULT metric assumes bidirectionality so it should exist.
    //
    KeAcquireSpinLock(&LC->Lock, &OldIrql);
    for (Adj = LC->nodes[0].AdjOut;
         Adj != NULL;
         Adj = Adj->NextOut) {

        if (VirtualAddressEqual(Probe->Opt.From,
                                LC->nodes[Adj->targetIndex].address) &&
            (Probe->Opt.OutIf == Adj->inif) &&
            (InIf == Adj->outif)) {
            //
            // Reset our count of the number of lost broadcast probes.
            //
            Adj->MetricInfo.Ult.LostBcastProbes = 0;
            break;
        }
    }
    KeReleaseSpinLock(&LC->Lock, OldIrql);
}

//* UltReceiveProbe
//
//  Receives a probe. Called from ReceiveSRPacket.
//  We have already verified that this is indeed a
//  ULT probe.
//
void
UltReceiveProbe(
    MiniportAdapter *VA,
    InternalProbe *Probe,
    LQSRIf InIf)
{
    const static VirtualAddress BcastDest = { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff };

    if (! VirtualAddressEqual(Probe->Opt.From, VA->Address) &&
        (Probe->Opt.ProbeType == METRIC_TYPE_ULT) &&
        (VirtualAddressEqual(Probe->Opt.To, BcastDest) ||
         (Probe->Opt.InIf == InIf))) {

        //
        // We know there is a link to us.
        //
		//if(debug) if(debug) KdPrint(("EEE Calling LinkCacheAddLink from UltReceiveProbe1 \n")); 
        LinkCacheAddLink(VA,
                         Probe->Opt.From,
                         VA->Address,
                         InIf,
                         Probe->Opt.OutIf,
                         0,
                         LINK_STATE_CHANGE_ADD_PROBE);

        //
        // ULT needs bidirectional links for its algorithm.
        //
		//if(debug) if(debug) KdPrint(("EEE Calling LinkCacheAddLink from UltReceiveProbe2 \n"));
        LinkCacheAddLink(VA,
                         VA->Address,
                         Probe->Opt.From,
                         Probe->Opt.OutIf,
                         InIf,
                         0,
                         LINK_STATE_CHANGE_ADD_PROBE);

        if (VirtualAddressEqual(Probe->Opt.To, BcastDest))
            UltReceiveBcastProbe(VA, Probe, InIf);
        else
            UltReceiveUcastProbe(VA, Probe);
    }
}

//* UltInit
//
//  Called by MiniportInitialize.
//
void UltInit(
    MiniportAdapter *VA)
{
    Time Now = KeQueryInterruptTime();

    VA->IsInfinite = MiniportIsInfinite;
    VA->IsBad = UltIsBad;
    VA->ConvMetric = UltConvMetric;
    VA->InitLinkMetric = UltInitLinkMetric;
    VA->PathMetric = MiniportPathMetric;
    VA->ReceiveProbe = UltReceiveProbe;
    VA->MetricTimeout = UltSendProbes;

    VA->MetricParams.Ult.UcastProbeTimeout = Now;
    VA->MetricParams.Ult.UcastProbePeriod = DEFAULT_ULT_UCAST_PROBE_PERIOD;
    VA->MetricParams.Ult.UcastLowThreshold = DEFAULT_ULT_UCAST_LOW_THRESHOLD;
    VA->MetricParams.Ult.UcastHighThreshold = DEFAULT_ULT_UCAST_HIGH_THRESHOLD;
    VA->MetricParams.Ult.UcastLossPenalty = DEFAULT_ULT_UCAST_LOSS_PENALTY;

    VA->MetricParams.Ult.BcastProbeTimeout = Now;
    VA->MetricParams.Ult.BcastProbePeriod = DEFAULT_ULT_BCAST_PROBE_PERIOD;
    VA->MetricParams.Ult.BcastLossThreshold = DEFAULT_ULT_BCAST_LOSS_THRESHOLD;
}
